/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "adbc_driver_manager/_blocking_impl.h",
            "adbc_driver_manager/arrow-adbc/adbc.h",
            "adbc_driver_manager/arrow-adbc/adbc_driver_manager.h"
        ],
        "extra_compile_args": [
            "-std=c++17"
        ],
        "include_dirs": [
            "adbc_driver_manager",
            "/build/python-arrow-adbc/src/python-arrow-adbc/python/adbc_driver_manager/adbc_driver_manager"
        ],
        "language": "c++",
        "name": "adbc_driver_manager._lib",
        "sources": [
            "adbc_driver_manager/_lib.pyx",
            "adbc_driver_manager/_blocking_impl.cc",
            "adbc_driver_manager/adbc_driver_manager.cc",
            "adbc_driver_manager/adbc_driver_manager_api.cc",
            "adbc_driver_manager/adbc_driver_manager_driver_loading.cc",
            "adbc_driver_manager/adbc_driver_manager_profiles.cc"
        ]
    },
    "module_name": "adbc_driver_manager._lib"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CppInitCode */
#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__adbc_driver_manager___lib
#define __PYX_HAVE_API__adbc_driver_manager___lib
/* Early includes */
#include <stdint.h>
#include "arrow-adbc/adbc.h"
#include "arrow-adbc/adbc_driver_manager.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string_view>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <string>
#include <vector>
#include "_blocking_impl.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "adbc_driver_manager/_lib.pyx",
  "string.to_py",
  "cpython/type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle;
struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle;
struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle;
struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle;
struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase;
struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection;
struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement;
struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error;
struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr;

/* "adbc_driver_manager/_lib.pyx":303
 * 
 * 
 * cdef class _AdbcHandle:             # <<<<<<<<<<<<<<
 *     """
 *     Base class for ADBC handles, which are context managers.
*/
struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle {
  PyObject_HEAD
  struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_vtab;
  size_t _open_children;
  PyObject *_lock;
  PyObject *_child_type;
};


/* "adbc_driver_manager/_lib.pyx":366
 * 
 * 
 * cdef class ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *     """
 *     A wrapper for an allocated ArrowSchema.
*/
struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle {
  PyObject_HEAD
  struct ArrowSchema schema;
};


/* "adbc_driver_manager/_lib.pyx":413
 * 
 * 
 * cdef class ArrowArrayHandle:             # <<<<<<<<<<<<<<
 *     """
 *     A wrapper for an allocated ArrowArray.
*/
struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle {
  PyObject_HEAD
  struct ArrowArray array;
};


/* "adbc_driver_manager/_lib.pyx":449
 * 
 * 
 * cdef class ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *     """
 *     A wrapper for an allocated ArrowArrayStream.
*/
struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle {
  PyObject_HEAD
  struct ArrowArrayStream stream;
};


/* "adbc_driver_manager/_lib.pyx":537
 * 
 * 
 * cdef class AdbcDatabase(_AdbcHandle):             # <<<<<<<<<<<<<<
 *     """
 *     An instance of a database.
*/
struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase {
  struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle __pyx_base;
  struct AdbcDatabase database;
};


/* "adbc_driver_manager/_lib.pyx":841
 * 
 * 
 * cdef class AdbcConnection(_AdbcHandle):             # <<<<<<<<<<<<<<
 *     """
 *     An active database connection.
*/
struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection {
  struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle __pyx_base;
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *database;
  struct AdbcConnection connection;
};


/* "adbc_driver_manager/_lib.pyx":1411
 * 
 * 
 * cdef class AdbcStatement(_AdbcHandle):             # <<<<<<<<<<<<<<
 *     """
 *     A database statement.
*/
struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement {
  struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle __pyx_base;
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *connection;
  struct AdbcStatement statement;
};


/* "adbc_driver_manager/_lib.pyx":180
 * 
 * 
 * cdef object convert_error(CAdbcStatusCode status, CAdbcError* error):             # <<<<<<<<<<<<<<
 *     cdef CAdbcErrorDetail c_detail
 * 
*/
struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error {
  PyObject_HEAD
  struct AdbcError *__pyx_v_error;
};


/* "adbc_driver_manager/_lib.pyx":202
 *             message += f". Vendor code: {vendor_code}"
 *         if error.sqlstate[0] != 0:
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))             # <<<<<<<<<<<<<<
 *             sqlstate = sqlstate.decode("ascii", "replace")
 *             message += f". SQLSTATE: {sqlstate}"
*/
struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error *__pyx_outer_scope;
  long __pyx_v_i;
  long __pyx_t_0;
};



/* "adbc_driver_manager/_lib.pyx":303
 * 
 * 
 * cdef class _AdbcHandle:             # <<<<<<<<<<<<<<
 *     """
 *     Base class for ADBC handles, which are context managers.
*/

struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle {
  PyObject *(*_open_child)(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *);
  PyObject *(*_close_child)(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *);
  PyObject *(*_check_open_children)(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *);
};
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_vtabptr_19adbc_driver_manager_4_lib__AdbcHandle;


/* "adbc_driver_manager/_lib.pyx":537
 * 
 * 
 * cdef class AdbcDatabase(_AdbcHandle):             # <<<<<<<<<<<<<<
 *     """
 *     An instance of a database.
*/

struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcDatabase {
  struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle __pyx_base;
};
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcDatabase;


/* "adbc_driver_manager/_lib.pyx":841
 * 
 * 
 * cdef class AdbcConnection(_AdbcHandle):             # <<<<<<<<<<<<<<
 *     """
 *     An active database connection.
*/

struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcConnection {
  struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle __pyx_base;
};
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcConnection;


/* "adbc_driver_manager/_lib.pyx":1411
 * 
 * 
 * cdef class AdbcStatement(_AdbcHandle):             # <<<<<<<<<<<<<<
 *     """
 *     A database statement.
*/

struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcStatement {
  struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle __pyx_base;
};
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcStatement;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseKeywordRequired.proto */
static void __Pyx_RaiseKeywordRequired(const char* func_name, PyObject* kw_name);

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyObjectVectorCallKwBuilder.proto (used by PyObjectVectorCallMethodKwBuilder) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* GetException.proto (used by pep479) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* decode_c_string_utf16.proto (used by decode_c_string) */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* decode_c_bytes.proto (used by decode_bytes) */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyBytes_AS_STRING(string);
    size = PyBytes_GET_SIZE(string);
#else
    if (PyBytes_AsStringAndSize(string, &as_c_string, &size) < 0) {
        return NULL;
    }
#endif
    return __Pyx_decode_c_bytes(
        as_c_string, size,
        start, stop, encoding, errors, decode_func);
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyValueError_Check.proto */
#define __Pyx_PyExc_ValueError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ValueError)

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared)
    #endif
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_OwnStrongReference)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_SharedReference)
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right))

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyRuntimeError_Check.proto */
#define __Pyx_PyExc_RuntimeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_RuntimeError)

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* BuildPyUnicode.proto (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* COrdinalToPyUnicode.proto (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value);
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t width, char padding_char);

/* GCCDiagnostics.proto (used by CIntToPyUnicode) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* IncludeStdlibH.proto (used by CIntToPyUnicode) */
#include <stdlib.h>

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_int(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_int(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_int(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* PyNotImplementedError_Check.proto */
#define __Pyx_PyExc_NotImplementedError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_NotImplementedError)

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(const char* function_name, PyObject *kw);

/* IterFinish.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto (used by UnpackTuple2) */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto (used by UnpackItemEndCheck) */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto (used by UnpackTuple2) */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto (used by UnpackTupleError) */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
    PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple);
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyObject_Unicode.proto */
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* pyint_simplify.proto */
static CYTHON_INLINE int __Pyx_PyInt_FromNumber(PyObject **number_var, const char *argname, int accept_none);

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* HasAttr.proto */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyRuntimeWarning_Check.proto */
#define __Pyx_PyExc_RuntimeWarning_Check(obj)  __Pyx_TypeCheck(obj, PyExc_RuntimeWarning)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyKeyboardInterrupt_Check.proto */
#define __Pyx_PyExc_KeyboardInterrupt_Check(obj)  __Pyx_TypeCheck(obj, PyExc_KeyboardInterrupt)

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* DefaultPlacementNew.proto */
#include <new>
template<typename T>
void __Pyx_default_placement_construct(T* x) {
    new (static_cast<void*>(x)) T();
}

/* CheckTypeForFreelists.proto */
#if CYTHON_USE_FREELISTS
#if CYTHON_USE_TYPE_SPECS
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t) == (expected_tp)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  Py_TPFLAGS_IS_ABSTRACT
#else
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t)->tp_basicsize == (expected_size)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)
#endif
#define __PYX_CHECK_TYPE_FOR_FREELISTS(t, expected_tp, expected_size)\
    (__PYX_CHECK_FINAL_TYPE_FOR_FREELISTS((t), (expected_tp), (expected_size)) &\
     (int) (!__Pyx_PyType_HasFeature((t), __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS)))
#endif

/* FunctionExport.proto */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig);

/* GetApiDict.proto */
static PyObject *__Pyx_ApiExport_GetApiDict(void);

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* PyObjectCall2Args.proto (used by Py3ClassCreate) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* PyUserWarning_Check.proto */
#define __Pyx_PyExc_UserWarning_Check(obj)  __Pyx_TypeCheck(obj, PyExc_UserWarning)

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CyFunctionClassCell.proto */
static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CheckUnpickleChecksum.proto */
static CYTHON_INLINE int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members);

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint8_t(uint8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int32_t(int32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_char(char value);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyLong_As_int32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyLong_As_char(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* UpdateUnpickledDict.proto */
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* IterNextPlain.proto (used by CoroutineBase) */
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator);
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void);
#endif

/* ReturnWithStopIteration.proto (used by CoroutineBase) */
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext);

/* CoroutineBase.proto (used by Generator) */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *gi_weakreflist;
#endif
    PyObject *classobj;
    PyObject *yieldfrom;
    __Pyx_pyiter_sendfunc yieldfrom_am_send;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
#if CYTHON_USE_SYS_MONITORING && (CYTHON_PROFILE || CYTHON_TRACE)
    PyMonitoringState __pyx_pymonitoring_state[__Pyx_MonitoringEventTypes_CyGen_count];
    uint64_t __pyx_pymonitoring_version;
#endif
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static __Pyx_PySendResult __Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static __Pyx_PySendResult __Pyx_Coroutine_Close(PyObject *self, PyObject **retval);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen);
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen);
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen);
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure);
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send);
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_mstate_global->__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);
static CYTHON_INLINE PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static PyObject *__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__open_child(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__close_child(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__check_open_children(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self); /* proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "cython" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "libc.stdlib" */

/* Module declarations from "libcpp.string_view" */

/* Module declarations from "libcpp.string" */

/* Module declarations from "libcpp.vector" */

/* Module declarations from "adbc_driver_manager._lib" */
static void __pyx_f_19adbc_driver_manager_4_lib_check_error(AdbcStatusCode, struct AdbcError *); /*proto*/
static PyObject *__pyx_f_19adbc_driver_manager_4_lib_convert_error(AdbcStatusCode, struct AdbcError *); /*proto*/
static struct AdbcError __pyx_f_19adbc_driver_manager_4_lib_empty_error(void); /*proto*/
static PyObject *__pyx_f_19adbc_driver_manager_4_lib__to_bytes(PyObject *, PyObject *); /*proto*/
static void __pyx_f_19adbc_driver_manager_4_lib_pycapsule_schema_deleter(PyObject *); /*proto*/
static void __pyx_f_19adbc_driver_manager_4_lib_pycapsule_stream_deleter(PyObject *); /*proto*/
static void __pyx_f_19adbc_driver_manager_4_lib__handle_blocking_call(void *); /*proto*/
static PyObject *__pyx_f_19adbc_driver_manager_4_lib___pyx_unpickle__AdbcHandle__set_state(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "adbc_driver_manager._lib"
extern int __pyx_module_is_main_adbc_driver_manager___lib;
int __pyx_module_is_main_adbc_driver_manager___lib = 0;

/* Implementation of "adbc_driver_manager._lib" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_arrow_array[] = "arrow_array";
static const char __pyx_k_arrow_schema[] = "arrow_schema";
static const char __pyx_k_Low_level_ADBC_API[] = "Low-level ADBC API.";
static const char __pyx_k_arrow_array_stream[] = "arrow_array_stream";
static const char __pyx_k_additional_manifest_search_path[] = "additional_manifest_search_path_list";
static const char __pyx_k_child_type__lock__open_children[] = "_child_type, _lock, _open_children";
static const char __pyx_k_additional_profile_search_path_l[] = "additional_profile_search_path_list";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_5Error___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_status_code, PyObject *__pyx_v_vendor_code, PyObject *__pyx_v_sqlstate, PyObject *__pyx_v_details); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17NotSupportedError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_vendor_code, PyObject *__pyx_v_sqlstate, PyObject *__pyx_v_details); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13convert_error_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib__test_error(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_status_code, PyObject *__pyx_v_message, PyObject *__pyx_v_vendor_code, PyObject *__pyx_v_sqlstate); /* proto */
static int __pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle___init__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self, PyObject *__pyx_v_child_type); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_2__enter__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_4__exit__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc_type, CYTHON_UNUSED PyObject *__pyx_v_exc_val, CYTHON_UNUSED PyObject *__pyx_v_exc_tb); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_6__reduce_cython__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_8__setstate_cython__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_2is_pycapsule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_name); /* proto */
static void __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle___del__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7address___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8is_valid___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_2release(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_4__arrow_c_schema__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle___del__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_8is_valid___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_2release(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7address___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle___del__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7address___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8is_valid___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_2release(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_4__arrow_c_schema__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_6__arrow_c_stream__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self, PyObject *__pyx_v_requested_schema); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase___init__(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_2close(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_4get_option(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_6get_option_bytes(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_8get_option_float(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_10get_option_int(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_12set_options(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection___init__(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_database, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_2cancel(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_4commit(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_6get_info(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_info_codes); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_8get_objects(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_depth, PyObject *__pyx_v_catalog, PyObject *__pyx_v_db_schema, PyObject *__pyx_v_table_name, CYTHON_UNUSED PyObject *__pyx_v_table_types, PyObject *__pyx_v_column_name); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_10get_option(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_12get_option_bytes(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_14get_option_float(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_16get_option_int(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_18get_statistics(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_catalog, PyObject *__pyx_v_db_schema, PyObject *__pyx_v_table_name, PyObject *__pyx_v_approximate); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_20get_statistic_names(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_22get_table_schema(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_catalog, PyObject *__pyx_v_db_schema, PyObject *__pyx_v_table_name); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_24get_table_types(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_26read_partition(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_partition); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_28rollback(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_30set_autocommit(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, int __pyx_v_enabled); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_32set_options(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_34close(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement___init__(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_2bind(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_4bind_stream(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_stream); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_6cancel(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_8close(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_10execute_query(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_12execute_partitions(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_14execute_schema(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_16execute_update(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_18get_option(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_20get_option_bytes(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_22get_option_float(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_24get_option_int(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_26get_parameter_schema(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_28prepare(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_30set_options(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_32set_sql_query(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_query); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_34set_substrait_plan(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_plan); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_4_init_blocking_call(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_6_blocking_call_impl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, PyObject *__pyx_v_cancel); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_8_blocking_call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_cancel); /* proto */
static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_10__pyx_unpickle__AdbcHandle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib__AdbcHandle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowSchemaHandle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayHandle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcDatabase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcConnection(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcStatement(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib__AdbcHandle;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib_AdbcConnection;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib_AdbcStatement;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error;
  PyObject *__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error;
  PyTypeObject *__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  PyObject *__pyx_tuple[4];
  PyObject *__pyx_codeobj_tab[71];
  PyObject *__pyx_string_tab[460];
  PyObject *__pyx_number_tab[3];
/* #### Code section: module_state_contents ### */

#if CYTHON_USE_FREELISTS
struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error *__pyx_freelist_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error[8];
int __pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *__pyx_freelist_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr[8];
int __pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr;
#endif
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* IterNextPlain.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
PyObject *__Pyx_GetBuiltinNext_LimitedAPI_cache;
#endif

/* Generator.module_state_decls */
PyTypeObject *__pyx_GeneratorType;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_A_status_code_indicating_the_ty __pyx_string_tab[1]
#define __pyx_kp_u_An_operation_or_some_functionali __pyx_string_tab[2]
#define __pyx_kp_u_ArrowArrayStreamHandle_already_c __pyx_string_tab[3]
#define __pyx_kp_u_Can_not_provide_a_schema_when_pa __pyx_string_tab[4]
#define __pyx_kp_u_Cannot_close __pyx_string_tab[5]
#define __pyx_kp_u_Could_not_read_option __pyx_string_tab[6]
#define __pyx_kp_u_Errors_related_to_database_inter __pyx_string_tab[7]
#define __pyx_kp_u_Errors_related_to_database_opera __pyx_string_tab[8]
#define __pyx_kp_u_Errors_related_to_processed_data __pyx_string_tab[9]
#define __pyx_kp_u_Errors_related_to_relational_int __pyx_string_tab[10]
#define __pyx_kp_u_Errors_related_to_the_database __pyx_string_tab[11]
#define __pyx_kp_u_Errors_related_to_the_database_i __pyx_string_tab[12]
#define __pyx_kp_u_Errors_related_to_user_errors __pyx_string_tab[13]
#define __pyx_kp_u_Failed_to_get_schema __pyx_string_tab[14]
#define __pyx_kp_u_Failed_to_initialize_KeyboardInt __pyx_string_tab[15]
#define __pyx_kp_u_Failed_to_restore_SIGINT_handler __pyx_string_tab[16]
#define __pyx_kp_u_Failed_to_set_SIGINT_handler __pyx_string_tab[17]
#define __pyx_kp_u_How_much_data_to_fetch_for_adbc __pyx_string_tab[18]
#define __pyx_kp_u_None __pyx_string_tab[19]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[20]
#define __pyx_kp_u_PEP_249_compliant_base_exceptio __pyx_string_tab[21]
#define __pyx_kp_u_PEP_249_compliant_base_warning __pyx_string_tab[22]
#define __pyx_kp_u_SQLSTATE __pyx_string_tab[23]
#define __pyx_kp_u_Self_2 __pyx_string_tab[24]
#define __pyx_kp_u_Tuple_ArrowArrayStreamHandle_int __pyx_string_tab[25]
#define __pyx_kp_u_Tuple_List_bytes_Optional_ArrowS __pyx_string_tab[26]
#define __pyx_kp_u_Underflow_in_closing_this __pyx_string_tab[27]
#define __pyx_kp_u_Unsupported_type __pyx_string_tab[28]
#define __pyx_kp_u_Vendor_code __pyx_string_tab[29]
#define __pyx_kp_u__3 __pyx_string_tab[30]
#define __pyx_kp_u__4 __pyx_string_tab[31]
#define __pyx_kp_u__5 __pyx_string_tab[32]
#define __pyx_kp_u__6 __pyx_string_tab[33]
#define __pyx_kp_u_adbc_driver_manager__lib_pyx __pyx_string_tab[34]
#define __pyx_kp_u_add_note __pyx_string_tab[35]
#define __pyx_kp_u_bytes __pyx_string_tab[36]
#define __pyx_kp_u_cannot_be_None __pyx_string_tab[37]
#define __pyx_kp_u_data_must_be_Arrow_compatible_da __pyx_string_tab[38]
#define __pyx_kp_u_data_must_be_a_PyCapsule_int_or __pyx_string_tab[39]
#define __pyx_kp_u_disable __pyx_string_tab[40]
#define __pyx_kp_u_enable __pyx_string_tab[41]
#define __pyx_kp_u_etc_adbc_drivers __pyx_string_tab[42]
#define __pyx_kp_u_etc_adbc_profiles __pyx_string_tab[43]
#define __pyx_kp_u_for_value __pyx_string_tab[44]
#define __pyx_kp_u_gc __pyx_string_tab[45]
#define __pyx_kp_u_isenabled __pyx_string_tab[46]
#define __pyx_kp_u_key_cannot_be_None __pyx_string_tab[47]
#define __pyx_kp_u_must_be_str_or_bytes __pyx_string_tab[48]
#define __pyx_kp_u_no_child_type __pyx_string_tab[49]
#define __pyx_kp_u_of_option __pyx_string_tab[50]
#define __pyx_kp_u_option_key __pyx_string_tab[51]
#define __pyx_kp_u_option_value __pyx_string_tab[52]
#define __pyx_kp_u_schema_must_be_a_PyCapsule_int_o __pyx_string_tab[53]
#define __pyx_kp_u_self_array_cannot_be_converted_t __pyx_string_tab[54]
#define __pyx_kp_u_self_connection_cannot_be_conver __pyx_string_tab[55]
#define __pyx_kp_u_self_database_cannot_be_converte __pyx_string_tab[56]
#define __pyx_kp_u_self_schema_cannot_be_converted __pyx_string_tab[57]
#define __pyx_kp_u_self_statement_cannot_be_convert __pyx_string_tab[58]
#define __pyx_kp_u_self_stream_cannot_be_converted __pyx_string_tab[59]
#define __pyx_kp_u_str_bytes __pyx_string_tab[60]
#define __pyx_kp_u_sys_prefix __pyx_string_tab[61]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[62]
#define __pyx_kp_u_utf_8 __pyx_string_tab[63]
#define __pyx_kp_u_value_for_key __pyx_string_tab[64]
#define __pyx_kp_u_with_open __pyx_string_tab[65]
#define __pyx_kp_u_would_need_more_than __pyx_string_tab[66]
#define __pyx_n_u_ALL __pyx_string_tab[67]
#define __pyx_n_u_ALREADY_EXISTS __pyx_string_tab[68]
#define __pyx_n_u_AdbcConnection __pyx_string_tab[69]
#define __pyx_n_u_AdbcConnection___reduce_cython __pyx_string_tab[70]
#define __pyx_n_u_AdbcConnection___setstate_cython __pyx_string_tab[71]
#define __pyx_n_u_AdbcConnection_cancel __pyx_string_tab[72]
#define __pyx_n_u_AdbcConnection_close __pyx_string_tab[73]
#define __pyx_n_u_AdbcConnection_commit __pyx_string_tab[74]
#define __pyx_n_u_AdbcConnection_get_info __pyx_string_tab[75]
#define __pyx_n_u_AdbcConnection_get_objects __pyx_string_tab[76]
#define __pyx_n_u_AdbcConnection_get_option __pyx_string_tab[77]
#define __pyx_n_u_AdbcConnection_get_option_bytes __pyx_string_tab[78]
#define __pyx_n_u_AdbcConnection_get_option_float __pyx_string_tab[79]
#define __pyx_n_u_AdbcConnection_get_option_int __pyx_string_tab[80]
#define __pyx_n_u_AdbcConnection_get_statistic_nam __pyx_string_tab[81]
#define __pyx_n_u_AdbcConnection_get_statistics __pyx_string_tab[82]
#define __pyx_n_u_AdbcConnection_get_table_schema __pyx_string_tab[83]
#define __pyx_n_u_AdbcConnection_get_table_types __pyx_string_tab[84]
#define __pyx_n_u_AdbcConnection_read_partition __pyx_string_tab[85]
#define __pyx_n_u_AdbcConnection_rollback __pyx_string_tab[86]
#define __pyx_n_u_AdbcConnection_set_autocommit __pyx_string_tab[87]
#define __pyx_n_u_AdbcConnection_set_options __pyx_string_tab[88]
#define __pyx_n_u_AdbcDatabase __pyx_string_tab[89]
#define __pyx_n_u_AdbcDatabase___reduce_cython __pyx_string_tab[90]
#define __pyx_n_u_AdbcDatabase___setstate_cython __pyx_string_tab[91]
#define __pyx_n_u_AdbcDatabase_close __pyx_string_tab[92]
#define __pyx_n_u_AdbcDatabase_get_option __pyx_string_tab[93]
#define __pyx_n_u_AdbcDatabase_get_option_bytes __pyx_string_tab[94]
#define __pyx_n_u_AdbcDatabase_get_option_float __pyx_string_tab[95]
#define __pyx_n_u_AdbcDatabase_get_option_int __pyx_string_tab[96]
#define __pyx_n_u_AdbcDatabase_set_options __pyx_string_tab[97]
#define __pyx_n_u_AdbcHandle __pyx_string_tab[98]
#define __pyx_n_u_AdbcHandle___enter __pyx_string_tab[99]
#define __pyx_n_u_AdbcHandle___exit __pyx_string_tab[100]
#define __pyx_n_u_AdbcHandle___reduce_cython __pyx_string_tab[101]
#define __pyx_n_u_AdbcHandle___setstate_cython __pyx_string_tab[102]
#define __pyx_n_u_AdbcInfoCode __pyx_string_tab[103]
#define __pyx_n_u_AdbcStatement __pyx_string_tab[104]
#define __pyx_n_u_AdbcStatement___reduce_cython __pyx_string_tab[105]
#define __pyx_n_u_AdbcStatement___setstate_cython __pyx_string_tab[106]
#define __pyx_n_u_AdbcStatement_bind __pyx_string_tab[107]
#define __pyx_n_u_AdbcStatement_bind_stream __pyx_string_tab[108]
#define __pyx_n_u_AdbcStatement_cancel __pyx_string_tab[109]
#define __pyx_n_u_AdbcStatement_close __pyx_string_tab[110]
#define __pyx_n_u_AdbcStatement_execute_partitions __pyx_string_tab[111]
#define __pyx_n_u_AdbcStatement_execute_query __pyx_string_tab[112]
#define __pyx_n_u_AdbcStatement_execute_schema __pyx_string_tab[113]
#define __pyx_n_u_AdbcStatement_execute_update __pyx_string_tab[114]
#define __pyx_n_u_AdbcStatement_get_option __pyx_string_tab[115]
#define __pyx_n_u_AdbcStatement_get_option_bytes __pyx_string_tab[116]
#define __pyx_n_u_AdbcStatement_get_option_float __pyx_string_tab[117]
#define __pyx_n_u_AdbcStatement_get_option_int __pyx_string_tab[118]
#define __pyx_n_u_AdbcStatement_get_parameter_sche __pyx_string_tab[119]
#define __pyx_n_u_AdbcStatement_prepare __pyx_string_tab[120]
#define __pyx_n_u_AdbcStatement_set_options __pyx_string_tab[121]
#define __pyx_n_u_AdbcStatement_set_sql_query __pyx_string_tab[122]
#define __pyx_n_u_AdbcStatement_set_substrait_plan __pyx_string_tab[123]
#define __pyx_n_u_AdbcStatusCode __pyx_string_tab[124]
#define __pyx_n_u_ArrowArrayHandle __pyx_string_tab[125]
#define __pyx_n_u_ArrowArrayHandle___reduce_cython __pyx_string_tab[126]
#define __pyx_n_u_ArrowArrayHandle___setstate_cyth __pyx_string_tab[127]
#define __pyx_n_u_ArrowArrayHandle_release __pyx_string_tab[128]
#define __pyx_n_u_ArrowArrayStreamHandle __pyx_string_tab[129]
#define __pyx_n_u_ArrowArrayStreamHandle___arrow_c __pyx_string_tab[130]
#define __pyx_n_u_ArrowArrayStreamHandle___arrow_c_2 __pyx_string_tab[131]
#define __pyx_n_u_ArrowArrayStreamHandle___reduce __pyx_string_tab[132]
#define __pyx_n_u_ArrowArrayStreamHandle___setstat __pyx_string_tab[133]
#define __pyx_n_u_ArrowArrayStreamHandle_release __pyx_string_tab[134]
#define __pyx_n_u_ArrowSchemaHandle __pyx_string_tab[135]
#define __pyx_n_u_ArrowSchemaHandle___arrow_c_sche __pyx_string_tab[136]
#define __pyx_n_u_ArrowSchemaHandle___reduce_cytho __pyx_string_tab[137]
#define __pyx_n_u_ArrowSchemaHandle___setstate_cyt __pyx_string_tab[138]
#define __pyx_n_u_ArrowSchemaHandle_release __pyx_string_tab[139]
#define __pyx_n_u_CANCELLED __pyx_string_tab[140]
#define __pyx_n_u_CATALOGS __pyx_string_tab[141]
#define __pyx_n_u_COLUMNS __pyx_string_tab[142]
#define __pyx_n_u_DB_SCHEMAS __pyx_string_tab[143]
#define __pyx_n_u_DRIVER_ARROW_VERSION __pyx_string_tab[144]
#define __pyx_n_u_DRIVER_NAME __pyx_string_tab[145]
#define __pyx_n_u_DRIVER_VERSION __pyx_string_tab[146]
#define __pyx_n_u_DataError __pyx_string_tab[147]
#define __pyx_n_u_DatabaseError __pyx_string_tab[148]
#define __pyx_n_u_Enum __pyx_string_tab[149]
#define __pyx_n_u_Error __pyx_string_tab[150]
#define __pyx_n_u_Error___init __pyx_string_tab[151]
#define __pyx_n_u_GetObjectsDepth __pyx_string_tab[152]
#define __pyx_n_u_INGEST_OPTION_MODE __pyx_string_tab[153]
#define __pyx_n_u_INGEST_OPTION_MODE_APPEND __pyx_string_tab[154]
#define __pyx_n_u_INGEST_OPTION_MODE_CREATE __pyx_string_tab[155]
#define __pyx_n_u_INGEST_OPTION_MODE_CREATE_APPEND __pyx_string_tab[156]
#define __pyx_n_u_INGEST_OPTION_MODE_REPLACE __pyx_string_tab[157]
#define __pyx_n_u_INGEST_OPTION_TARGET_TABLE __pyx_string_tab[158]
#define __pyx_n_u_INTEGRITY __pyx_string_tab[159]
#define __pyx_n_u_INTERNAL __pyx_string_tab[160]
#define __pyx_n_u_INVALID_ARGUMENT __pyx_string_tab[161]
#define __pyx_n_u_INVALID_DATA __pyx_string_tab[162]
#define __pyx_n_u_INVALID_STATE __pyx_string_tab[163]
#define __pyx_n_u_IO __pyx_string_tab[164]
#define __pyx_n_u_IntEnum __pyx_string_tab[165]
#define __pyx_n_u_IntegrityError __pyx_string_tab[166]
#define __pyx_n_u_InterfaceError __pyx_string_tab[167]
#define __pyx_n_u_InternalError __pyx_string_tab[168]
#define __pyx_n_u_List __pyx_string_tab[169]
#define __pyx_n_u_Lock __pyx_string_tab[170]
#define __pyx_n_u_MAX_OPTION_SIZE __pyx_string_tab[171]
#define __pyx_n_u_NOT_FOUND __pyx_string_tab[172]
#define __pyx_n_u_NOT_IMPLEMENTED __pyx_string_tab[173]
#define __pyx_n_u_None __pyx_string_tab[174]
#define __pyx_n_u_NotSupportedError __pyx_string_tab[175]
#define __pyx_n_u_NotSupportedError___init __pyx_string_tab[176]
#define __pyx_n_u_OK __pyx_string_tab[177]
#define __pyx_n_u_OperationalError __pyx_string_tab[178]
#define __pyx_n_u_Optional __pyx_string_tab[179]
#define __pyx_n_u_Path __pyx_string_tab[180]
#define __pyx_n_u_ProgrammingError __pyx_string_tab[181]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[182]
#define __pyx_n_u_Self __pyx_string_tab[183]
#define __pyx_n_u_TABLES __pyx_string_tab[184]
#define __pyx_n_u_TIMEOUT __pyx_string_tab[185]
#define __pyx_n_u_TYPE_CHECKING __pyx_string_tab[186]
#define __pyx_n_u_Tuple __pyx_string_tab[187]
#define __pyx_n_u_UNAUTHENTICATED __pyx_string_tab[188]
#define __pyx_n_u_UNAUTHORIZED __pyx_string_tab[189]
#define __pyx_n_u_UNKNOWN __pyx_string_tab[190]
#define __pyx_n_u_VENDOR_ARROW_VERSION __pyx_string_tab[191]
#define __pyx_n_u_VENDOR_NAME __pyx_string_tab[192]
#define __pyx_n_u_VENDOR_VERSION __pyx_string_tab[193]
#define __pyx_n_u_Warning __pyx_string_tab[194]
#define __pyx_n_u_adbc_driver_manager __pyx_string_tab[195]
#define __pyx_n_u_adbc_driver_manager__lib __pyx_string_tab[196]
#define __pyx_n_u_additional_manifest_search_path __pyx_string_tab[197]
#define __pyx_n_u_additional_profile_search_path_l __pyx_string_tab[198]
#define __pyx_n_u_allocated __pyx_string_tab[199]
#define __pyx_n_u_approximate __pyx_string_tab[200]
#define __pyx_n_u_args __pyx_string_tab[201]
#define __pyx_n_u_arrow_c_array __pyx_string_tab[202]
#define __pyx_n_u_arrow_c_schema __pyx_string_tab[203]
#define __pyx_n_u_arrow_c_stream __pyx_string_tab[204]
#define __pyx_n_u_ascii __pyx_string_tab[205]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[206]
#define __pyx_n_u_base_prefix __pyx_string_tab[207]
#define __pyx_n_u_bind __pyx_string_tab[208]
#define __pyx_n_u_bind_stream __pyx_string_tab[209]
#define __pyx_n_u_blocking_call __pyx_string_tab[210]
#define __pyx_n_u_blocking_call_impl __pyx_string_tab[211]
#define __pyx_n_u_blocking_exc __pyx_string_tab[212]
#define __pyx_n_u_blocking_lock __pyx_string_tab[213]
#define __pyx_n_u_bool __pyx_string_tab[214]
#define __pyx_n_u_buf __pyx_string_tab[215]
#define __pyx_n_u_bytes_2 __pyx_string_tab[216]
#define __pyx_n_u_c_approximate __pyx_string_tab[217]
#define __pyx_n_u_c_array __pyx_string_tab[218]
#define __pyx_n_u_c_buf __pyx_string_tab[219]
#define __pyx_n_u_c_catalog __pyx_string_tab[220]
#define __pyx_n_u_c_column_name __pyx_string_tab[221]
#define __pyx_n_u_c_db_schema __pyx_string_tab[222]
#define __pyx_n_u_c_depth __pyx_string_tab[223]
#define __pyx_n_u_c_double __pyx_string_tab[224]
#define __pyx_n_u_c_error __pyx_string_tab[225]
#define __pyx_n_u_c_info_codes __pyx_string_tab[226]
#define __pyx_n_u_c_int __pyx_string_tab[227]
#define __pyx_n_u_c_key __pyx_string_tab[228]
#define __pyx_n_u_c_len __pyx_string_tab[229]
#define __pyx_n_u_c_partitions __pyx_string_tab[230]
#define __pyx_n_u_c_plan __pyx_string_tab[231]
#define __pyx_n_u_c_query __pyx_string_tab[232]
#define __pyx_n_u_c_schema __pyx_string_tab[233]
#define __pyx_n_u_c_stream __pyx_string_tab[234]
#define __pyx_n_u_c_table_name __pyx_string_tab[235]
#define __pyx_n_u_c_value __pyx_string_tab[236]
#define __pyx_n_u_c_value_len __pyx_string_tab[237]
#define __pyx_n_u_cancel __pyx_string_tab[238]
#define __pyx_n_u_capsule __pyx_string_tab[239]
#define __pyx_n_u_catalog __pyx_string_tab[240]
#define __pyx_n_u_child_type __pyx_string_tab[241]
#define __pyx_n_u_class __pyx_string_tab[242]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[243]
#define __pyx_n_u_close __pyx_string_tab[244]
#define __pyx_n_u_column_name __pyx_string_tab[245]
#define __pyx_n_u_commit __pyx_string_tab[246]
#define __pyx_n_u_connection __pyx_string_tab[247]
#define __pyx_n_u_convert_error_locals_genexpr __pyx_string_tab[248]
#define __pyx_n_u_current_thread __pyx_string_tab[249]
#define __pyx_n_u_data __pyx_string_tab[250]
#define __pyx_n_u_database __pyx_string_tab[251]
#define __pyx_n_u_db_schema __pyx_string_tab[252]
#define __pyx_n_u_decode __pyx_string_tab[253]
#define __pyx_n_u_depth __pyx_string_tab[254]
#define __pyx_n_u_details __pyx_string_tab[255]
#define __pyx_n_u_dict __pyx_string_tab[256]
#define __pyx_n_u_dict_2 __pyx_string_tab[257]
#define __pyx_n_u_doc __pyx_string_tab[258]
#define __pyx_n_u_e __pyx_string_tab[259]
#define __pyx_n_u_enabled __pyx_string_tab[260]
#define __pyx_n_u_encode __pyx_string_tab[261]
#define __pyx_n_u_encoding __pyx_string_tab[262]
#define __pyx_n_u_enter __pyx_string_tab[263]
#define __pyx_n_u_enum __pyx_string_tab[264]
#define __pyx_n_u_err __pyx_string_tab[265]
#define __pyx_n_u_error __pyx_string_tab[266]
#define __pyx_n_u_errors __pyx_string_tab[267]
#define __pyx_n_u_exc __pyx_string_tab[268]
#define __pyx_n_u_exc_info __pyx_string_tab[269]
#define __pyx_n_u_exc_tb __pyx_string_tab[270]
#define __pyx_n_u_exc_type __pyx_string_tab[271]
#define __pyx_n_u_exc_val __pyx_string_tab[272]
#define __pyx_n_u_execute_partitions __pyx_string_tab[273]
#define __pyx_n_u_execute_query __pyx_string_tab[274]
#define __pyx_n_u_execute_schema __pyx_string_tab[275]
#define __pyx_n_u_execute_update __pyx_string_tab[276]
#define __pyx_n_u_exit __pyx_string_tab[277]
#define __pyx_n_u_float __pyx_string_tab[278]
#define __pyx_n_u_func __pyx_string_tab[279]
#define __pyx_n_u_func_2 __pyx_string_tab[280]
#define __pyx_n_u_functools __pyx_string_tab[281]
#define __pyx_n_u_genexpr __pyx_string_tab[282]
#define __pyx_n_u_get_info __pyx_string_tab[283]
#define __pyx_n_u_get_objects __pyx_string_tab[284]
#define __pyx_n_u_get_option __pyx_string_tab[285]
#define __pyx_n_u_get_option_bytes __pyx_string_tab[286]
#define __pyx_n_u_get_option_float __pyx_string_tab[287]
#define __pyx_n_u_get_option_int __pyx_string_tab[288]
#define __pyx_n_u_get_parameter_schema __pyx_string_tab[289]
#define __pyx_n_u_get_statistic_names __pyx_string_tab[290]
#define __pyx_n_u_get_statistics __pyx_string_tab[291]
#define __pyx_n_u_get_table_schema __pyx_string_tab[292]
#define __pyx_n_u_get_table_types __pyx_string_tab[293]
#define __pyx_n_u_getstate __pyx_string_tab[294]
#define __pyx_n_u_handle __pyx_string_tab[295]
#define __pyx_n_u_i __pyx_string_tab[296]
#define __pyx_n_u_info_code __pyx_string_tab[297]
#define __pyx_n_u_info_codes __pyx_string_tab[298]
#define __pyx_n_u_init __pyx_string_tab[299]
#define __pyx_n_u_init_blocking_call __pyx_string_tab[300]
#define __pyx_n_u_init_func __pyx_string_tab[301]
#define __pyx_n_u_int __pyx_string_tab[302]
#define __pyx_n_u_is_coroutine __pyx_string_tab[303]
#define __pyx_n_u_is_pycapsule __pyx_string_tab[304]
#define __pyx_n_u_is_valid __pyx_string_tab[305]
#define __pyx_n_u_items __pyx_string_tab[306]
#define __pyx_n_u_join __pyx_string_tab[307]
#define __pyx_n_u_key __pyx_string_tab[308]
#define __pyx_n_u_key_bytes __pyx_string_tab[309]
#define __pyx_n_u_kwargs __pyx_string_tab[310]
#define __pyx_n_u_length __pyx_string_tab[311]
#define __pyx_n_u_load_flags __pyx_string_tab[312]
#define __pyx_n_u_lru_cache __pyx_string_tab[313]
#define __pyx_n_u_main __pyx_string_tab[314]
#define __pyx_n_u_main_thread __pyx_string_tab[315]
#define __pyx_n_u_message __pyx_string_tab[316]
#define __pyx_n_u_metaclass __pyx_string_tab[317]
#define __pyx_n_u_module __pyx_string_tab[318]
#define __pyx_n_u_mro_entries __pyx_string_tab[319]
#define __pyx_n_u_name __pyx_string_tab[320]
#define __pyx_n_u_name_2 __pyx_string_tab[321]
#define __pyx_n_u_new __pyx_string_tab[322]
#define __pyx_n_u_new_len __pyx_string_tab[323]
#define __pyx_n_u_next __pyx_string_tab[324]
#define __pyx_n_u_nt __pyx_string_tab[325]
#define __pyx_n_u_num_partitions __pyx_string_tab[326]
#define __pyx_n_u_obj __pyx_string_tab[327]
#define __pyx_n_u_object __pyx_string_tab[328]
#define __pyx_n_u_os __pyx_string_tab[329]
#define __pyx_n_u_partition __pyx_string_tab[330]
#define __pyx_n_u_partition_lengths __pyx_string_tab[331]
#define __pyx_n_u_partitions __pyx_string_tab[332]
#define __pyx_n_u_path __pyx_string_tab[333]
#define __pyx_n_u_pathlib __pyx_string_tab[334]
#define __pyx_n_u_plan __pyx_string_tab[335]
#define __pyx_n_u_pop __pyx_string_tab[336]
#define __pyx_n_u_prefix __pyx_string_tab[337]
#define __pyx_n_u_prepare __pyx_string_tab[338]
#define __pyx_n_u_prepare_2 __pyx_string_tab[339]
#define __pyx_n_u_private_data __pyx_string_tab[340]
#define __pyx_n_u_pyx_capi __pyx_string_tab[341]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[342]
#define __pyx_n_u_pyx_result __pyx_string_tab[343]
#define __pyx_n_u_pyx_state __pyx_string_tab[344]
#define __pyx_n_u_pyx_type __pyx_string_tab[345]
#define __pyx_n_u_pyx_unpickle__AdbcHandle __pyx_string_tab[346]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[347]
#define __pyx_n_u_qualname __pyx_string_tab[348]
#define __pyx_n_u_query __pyx_string_tab[349]
#define __pyx_n_u_query_data __pyx_string_tab[350]
#define __pyx_n_u_rc __pyx_string_tab[351]
#define __pyx_n_u_read_partition __pyx_string_tab[352]
#define __pyx_n_u_reduce __pyx_string_tab[353]
#define __pyx_n_u_reduce_cython __pyx_string_tab[354]
#define __pyx_n_u_reduce_ex __pyx_string_tab[355]
#define __pyx_n_u_release __pyx_string_tab[356]
#define __pyx_n_u_requested_schema __pyx_string_tab[357]
#define __pyx_n_u_return __pyx_string_tab[358]
#define __pyx_n_u_rollback __pyx_string_tab[359]
#define __pyx_n_u_rows_affected __pyx_string_tab[360]
#define __pyx_n_u_s __pyx_string_tab[361]
#define __pyx_n_u_schema __pyx_string_tab[362]
#define __pyx_n_u_self __pyx_string_tab[363]
#define __pyx_n_u_send __pyx_string_tab[364]
#define __pyx_n_u_set_autocommit __pyx_string_tab[365]
#define __pyx_n_u_set_name __pyx_string_tab[366]
#define __pyx_n_u_set_options __pyx_string_tab[367]
#define __pyx_n_u_set_sql_query __pyx_string_tab[368]
#define __pyx_n_u_set_substrait_plan __pyx_string_tab[369]
#define __pyx_n_u_setdefault __pyx_string_tab[370]
#define __pyx_n_u_setstate __pyx_string_tab[371]
#define __pyx_n_u_setstate_cython __pyx_string_tab[372]
#define __pyx_n_u_sqlstate __pyx_string_tab[373]
#define __pyx_n_u_state __pyx_string_tab[374]
#define __pyx_n_u_status __pyx_string_tab[375]
#define __pyx_n_u_status_code __pyx_string_tab[376]
#define __pyx_n_u_str __pyx_string_tab[377]
#define __pyx_n_u_stream __pyx_string_tab[378]
#define __pyx_n_u_strict __pyx_string_tab[379]
#define __pyx_n_u_super __pyx_string_tab[380]
#define __pyx_n_u_sys __pyx_string_tab[381]
#define __pyx_n_u_table_name __pyx_string_tab[382]
#define __pyx_n_u_table_types __pyx_string_tab[383]
#define __pyx_n_u_test __pyx_string_tab[384]
#define __pyx_n_u_test_error __pyx_string_tab[385]
#define __pyx_n_u_threading __pyx_string_tab[386]
#define __pyx_n_u_throw __pyx_string_tab[387]
#define __pyx_n_u_typing __pyx_string_tab[388]
#define __pyx_n_u_update __pyx_string_tab[389]
#define __pyx_n_u_use_setstate __pyx_string_tab[390]
#define __pyx_n_u_value __pyx_string_tab[391]
#define __pyx_n_u_values __pyx_string_tab[392]
#define __pyx_n_u_vendor_code __pyx_string_tab[393]
#define __pyx_n_u_warn __pyx_string_tab[394]
#define __pyx_n_u_warnings __pyx_string_tab[395]
#define __pyx_n_u_with_traceback __pyx_string_tab[396]
#define __pyx_kp_b_PyObject_AdbcStatusCode_struct_A __pyx_string_tab[397]
#define __pyx_kp_b__2 __pyx_string_tab[398]
#define __pyx_kp_b_iso88591_1 __pyx_string_tab[399]
#define __pyx_kp_b_iso88591_1_Q_C1_1_Q_Qk_1_9A_a_T_Q_Q_V1_Q __pyx_string_tab[400]
#define __pyx_kp_b_iso88591_1_q __pyx_string_tab[401]
#define __pyx_kp_b_iso88591_2_1_G1_Qa_q_A_q_4AQ_a_4y_G_a_q __pyx_string_tab[402]
#define __pyx_kp_b_iso88591_31_Q_WAQ_1_Q_1_1HAQ __pyx_string_tab[403]
#define __pyx_kp_b_iso88591_5Q_F __pyx_string_tab[404]
#define __pyx_kp_b_iso88591_5_A_Ry_a_Q_A __pyx_string_tab[405]
#define __pyx_kp_b_iso88591_99J_2_1_Q_C1_6_q_87_iq_WA_Q_gQ __pyx_string_tab[406]
#define __pyx_kp_b_iso88591_9_Q_6a_S_2_1HAQ __pyx_string_tab[407]
#define __pyx_kp_b_iso88591_A_4vYc_haq_A_k __pyx_string_tab[408]
#define __pyx_kp_b_iso88591_A_4wis_xq_Q __pyx_string_tab[409]
#define __pyx_kp_b_iso88591_A_6_Q_C1_q_87_iq_WA_Q_gQ_q_1_8_0 __pyx_string_tab[410]
#define __pyx_kp_b_iso88591_A_7vQa_A_q_aq_a_4y_G_a_q __pyx_string_tab[411]
#define __pyx_kp_b_iso88591_A_Q_9_1_vV6_Q_31_Q_q_1HAQ_Q_E_a __pyx_string_tab[412]
#define __pyx_kp_b_iso88591_A_Q_IQe1_1_A_1_iq_Cq_A_0_T_gWAWA __pyx_string_tab[413]
#define __pyx_kp_b_iso88591_A_Q_IQe1_1_A_1_iq_Cq_A_a_T_G7_7 __pyx_string_tab[414]
#define __pyx_kp_b_iso88591_A_Q_IQe1_1_A_1_iq_Cq_A_q_T_WG1G1 __pyx_string_tab[415]
#define __pyx_kp_b_iso88591_A_Ry_O_O1_L_Kxs __pyx_string_tab[416]
#define __pyx_kp_b_iso88591_A_iq_A_q__A__A_q_87_1_1_U_1_Yau __pyx_string_tab[417]
#define __pyx_kp_b_iso88591_A_q_4t1_AQ_7vQa_A_q_aq_T_1AT_3c __pyx_string_tab[418]
#define __pyx_kp_b_iso88591_A_t2XQ __pyx_string_tab[419]
#define __pyx_kp_b_iso88591_B_Q_9_Ya_1_q_87_iq_WA_Q_1_Q_q_1 __pyx_string_tab[420]
#define __pyx_kp_b_iso88591_E_s_q_Q_q __pyx_string_tab[421]
#define __pyx_kp_b_iso88591_HA_Q_IQe1_1_A_a_iq_Cq_A_31_T_G7 __pyx_string_tab[422]
#define __pyx_kp_b_iso88591_HA_Q_IQe1_1_A_a_iq_Cq_A_4A_T_WG __pyx_string_tab[423]
#define __pyx_kp_b_iso88591_HA_Q_IQe1_1_A_a_iq_Cq_A_5Q_T_gW __pyx_string_tab[424]
#define __pyx_kp_b_iso88591_HA_Q_IQe1_1_a_0_q_1HAQ_q __pyx_string_tab[425]
#define __pyx_kp_b_iso88591_HA_Q_IQe1_1_a_1_1_1HAQ_q __pyx_string_tab[426]
#define __pyx_kp_b_iso88591_HA_Q_IQe1_1_a_2_1A_1HAQ_q __pyx_string_tab[427]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[428]
#define __pyx_kp_b_iso88591_Q_1AT_aq_1HAQ __pyx_string_tab[429]
#define __pyx_kp_b_iso88591_Q_4_1A_1HAQ __pyx_string_tab[430]
#define __pyx_kp_b_iso88591_Q_4_AQ_1HAQ __pyx_string_tab[431]
#define __pyx_kp_b_iso88591_Q_9_q_Q_q_1HAQ_q __pyx_string_tab[432]
#define __pyx_kp_b_iso88591_Q_AU_1 __pyx_string_tab[433]
#define __pyx_kp_b_iso88591_Q_C1_0_6_1_1HAQ_q __pyx_string_tab[434]
#define __pyx_kp_b_iso88591_Q_C1_4A_S_0_Q_q_1HAQ_q __pyx_string_tab[435]
#define __pyx_kp_b_iso88591_Q_G1F_q_JavQ_wgQ_j_G4_1_q_a_o_9 __pyx_string_tab[436]
#define __pyx_kp_b_iso88591_Q_IQe1_1_q_Q_q_1HAQ_q __pyx_string_tab[437]
#define __pyx_kp_b_iso88591_Q_IQe1_1_q_a_1_1HAQ_q __pyx_string_tab[438]
#define __pyx_kp_b_iso88591_Q_IQe1_1_q_q_1A_1HAQ_q __pyx_string_tab[439]
#define __pyx_kp_b_iso88591_Q_Q_1HA_Jaxq_V_a_q_q_r_2_q_A_1A __pyx_string_tab[440]
#define __pyx_kp_b_iso88591_Q_Q_C1_4A_Q_q_1HAQ_q __pyx_string_tab[441]
#define __pyx_kp_b_iso88591_Q_Q_a_Q_1HAQ_q __pyx_string_tab[442]
#define __pyx_kp_b_iso88591_Q_l_1_1HAQ __pyx_string_tab[443]
#define __pyx_kp_b_iso88591_T_t84q_G1F_a_vWE_Q_q_t_uCt7_q_D __pyx_string_tab[444]
#define __pyx_kp_b_iso88591__7 __pyx_string_tab[445]
#define __pyx_kp_b_iso88591_a_Q_1_A_A_A_Q_1HAQ __pyx_string_tab[446]
#define __pyx_kp_b_iso88591_a_Q_9_4A_1HAQ_q __pyx_string_tab[447]
#define __pyx_kp_b_iso88591_q_0_kQR_haq_7_QnN_1 __pyx_string_tab[448]
#define __pyx_kp_b_iso88591_q_Q_1_Q_E_a_1E_A_vS_31_1G1_31_1 __pyx_string_tab[449]
#define __pyx_kp_b_iso88591_q_Q_1_Q_E_a_1E_A_vS_4A_1_1G1_4A __pyx_string_tab[450]
#define __pyx_kp_b_iso88591_q_Q_1_Q_E_a_1E_A_vS_a_T_G9AQ_1G __pyx_string_tab[451]
#define __pyx_kp_b_iso88591_q_Q_C1_Q_a_Q_q_1HAQ __pyx_string_tab[452]
#define __pyx_kp_b_iso88591_q_Q_K_A_t_3a_Qat_q_q __pyx_string_tab[453]
#define __pyx_kp_b_iso88591_q_Q_t9N_Q_AQd_Qa_q __pyx_string_tab[454]
#define __pyx_kp_b_iso88591_q_Q_t_nCq_aq_M_q_a __pyx_string_tab[455]
#define __pyx_kp_b_iso88591_y_s_a_t2XQ_q_1_A_E_AQ_5WA_AQ_q __pyx_string_tab[456]
#define __pyx_n_b_arrow_array __pyx_string_tab[457]
#define __pyx_n_b_arrow_array_stream __pyx_string_tab[458]
#define __pyx_n_b_arrow_schema __pyx_string_tab[459]
#define __pyx_int_1024 __pyx_number_tab[0]
#define __pyx_int_86393024 __pyx_number_tab[1]
#define __pyx_int_134217728 __pyx_number_tab[2]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib__AdbcHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib_AdbcConnection);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib_AdbcStatement);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error);
  Py_CLEAR(clear_module_state->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr);
  for (int i=0; i<4; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<71; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<460; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<3; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* Generator.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib__AdbcHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib_AdbcConnection);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib_AdbcStatement);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error);
  Py_VISIT(traverse_module_state->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr);
  for (int i=0; i<4; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<71; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<460; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<3; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* Generator.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "string.to_py":31
 *     cdef object __Pyx_PyObject_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string", 0);

  /* "string.to_py":33
 * @cname("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  __pyx_t_1 = (__pyx_v_s.size() > ((size_t)PY_SSIZE_T_MAX));
  if (unlikely(__pyx_t_1)) {

    /* "string.to_py":34
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
 * cdef extern from *:
*/
    PyErr_NoMemory(); __PYX_ERR(1, 34, __pyx_L1_error)

    /* "string.to_py":33
 * @cname("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  }

  /* "string.to_py":35
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), <Py_ssize_t> s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyUnicode_FromStringAndSize(const char*, size_t)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_FromStringAndSize(__pyx_v_s.data(), ((Py_ssize_t)__pyx_v_s.size())); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "string.to_py":31
 *     cdef object __Pyx_PyObject_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":39
 *     cdef object __Pyx_PyUnicode_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string", 0);

  /* "string.to_py":41
 * @cname("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  __pyx_t_1 = (__pyx_v_s.size() > ((size_t)PY_SSIZE_T_MAX));
  if (unlikely(__pyx_t_1)) {

    /* "string.to_py":42
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
 * cdef extern from *:
*/
    PyErr_NoMemory(); __PYX_ERR(1, 42, __pyx_L1_error)

    /* "string.to_py":41
 * @cname("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  }

  /* "string.to_py":43
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), <Py_ssize_t> s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyBytes_FromStringAndSize(const char*, size_t)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyUnicode_FromStringAndSize(__pyx_v_s.data(), ((Py_ssize_t)__pyx_v_s.size())); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "string.to_py":39
 *     cdef object __Pyx_PyUnicode_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":47
 *     cdef object __Pyx_PyBytes_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string", 0);

  /* "string.to_py":49
 * @cname("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  __pyx_t_1 = (__pyx_v_s.size() > ((size_t)PY_SSIZE_T_MAX));
  if (unlikely(__pyx_t_1)) {

    /* "string.to_py":50
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
 * cdef extern from *:
*/
    PyErr_NoMemory(); __PYX_ERR(1, 50, __pyx_L1_error)

    /* "string.to_py":49
 * @cname("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  }

  /* "string.to_py":51
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), <Py_ssize_t> s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyByteArray_FromStringAndSize(const char*, size_t)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s.data(), ((Py_ssize_t)__pyx_v_s.size())); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "string.to_py":47
 *     cdef object __Pyx_PyBytes_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":55
 *     cdef object __Pyx_PyByteArray_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string", 0);

  /* "string.to_py":57
 * @cname("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  __pyx_t_1 = (__pyx_v_s.size() > ((size_t)PY_SSIZE_T_MAX));
  if (unlikely(__pyx_t_1)) {

    /* "string.to_py":58
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
    PyErr_NoMemory(); __PYX_ERR(1, 58, __pyx_L1_error)

    /* "string.to_py":57
 * @cname("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), <Py_ssize_t> s.size())
*/
  }

  /* "string.to_py":59
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), <Py_ssize_t> s.size())             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyByteArray_FromStringAndSize(__pyx_v_s.data(), ((Py_ssize_t)__pyx_v_s.size())); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "string.to_py":55
 *     cdef object __Pyx_PyByteArray_FromStringAndSize(const char*, size_t)
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     if s.size() > <size_t> PY_SSIZE_T_MAX:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":101
 *     """
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         message,
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_5Error_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_5Error_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_5Error_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_5Error_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_status_code = 0;
  PyObject *__pyx_v_vendor_code = 0;
  PyObject *__pyx_v_sqlstate = 0;
  PyObject *__pyx_v_details = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_message,&__pyx_mstate_global->__pyx_n_u_status_code,&__pyx_mstate_global->__pyx_n_u_vendor_code,&__pyx_mstate_global->__pyx_n_u_sqlstate,&__pyx_mstate_global->__pyx_n_u_details,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 101, __pyx_L3_error)
    if (likely(__pyx_kwds_len > 0)) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 101, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 101, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 101, __pyx_L3_error)

      /* "adbc_driver_manager/_lib.pyx":106
 *         *,
 *         status_code,
 *         vendor_code=None,             # <<<<<<<<<<<<<<
 *         sqlstate=None,
 *         details=None,
*/
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":107
 *         status_code,
 *         vendor_code=None,
 *         sqlstate=None,             # <<<<<<<<<<<<<<
 *         details=None,
 *     ):
*/
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":108
 *         vendor_code=None,
 *         sqlstate=None,
 *         details=None,             # <<<<<<<<<<<<<<
 *     ):
 *         super().__init__(message)
*/
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 101, __pyx_L3_error) }
      }
      for (Py_ssize_t i = 2; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseKeywordRequired("__init__", *(__pyx_pyargnames[i - 0])); __PYX_ERR(0, 101, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      __Pyx_RaiseKeywordRequired("__init__", __pyx_mstate_global->__pyx_n_u_status_code); __PYX_ERR(0, 101, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_message = values[1];
    __pyx_v_status_code = values[2];
    __pyx_v_vendor_code = values[3];
    __pyx_v_sqlstate = values[4];
    __pyx_v_details = values[5];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 101, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.Error.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_5Error___init__(__pyx_self, __pyx_v_self, __pyx_v_message, __pyx_v_status_code, __pyx_v_vendor_code, __pyx_v_sqlstate, __pyx_v_details);

  /* "adbc_driver_manager/_lib.pyx":101
 *     """
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         message,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_5Error___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_status_code, PyObject *__pyx_v_vendor_code, PyObject *__pyx_v_sqlstate, PyObject *__pyx_v_details) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "adbc_driver_manager/_lib.pyx":110
 *         details=None,
 *     ):
 *         super().__init__(message)             # <<<<<<<<<<<<<<
 *         self.status_code = AdbcStatusCode(status_code)
 *         self.vendor_code = vendor_code
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_5) { PyErr_SetString(PyExc_RuntimeError, "super(): empty __class__ cell"); __PYX_ERR(0, 110, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_5, __pyx_v_self};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_message};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":111
 *     ):
 *         super().__init__(message)
 *         self.status_code = AdbcStatusCode(status_code)             # <<<<<<<<<<<<<<
 *         self.vendor_code = vendor_code
 *         self.sqlstate = sqlstate
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_AdbcStatusCode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_status_code};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_status_code, __pyx_t_1) < (0)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":112
 *         super().__init__(message)
 *         self.status_code = AdbcStatusCode(status_code)
 *         self.vendor_code = vendor_code             # <<<<<<<<<<<<<<
 *         self.sqlstate = sqlstate
 *         self.details = details or []
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_vendor_code, __pyx_v_vendor_code) < (0)) __PYX_ERR(0, 112, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":113
 *         self.status_code = AdbcStatusCode(status_code)
 *         self.vendor_code = vendor_code
 *         self.sqlstate = sqlstate             # <<<<<<<<<<<<<<
 *         self.details = details or []
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_sqlstate, __pyx_v_sqlstate) < (0)) __PYX_ERR(0, 113, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":114
 *         self.vendor_code = vendor_code
 *         self.sqlstate = sqlstate
 *         self.details = details or []             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_details); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 114, __pyx_L1_error)
  if (!__pyx_t_7) {
  } else {
    __Pyx_INCREF(__pyx_v_details);
    __pyx_t_1 = __pyx_v_details;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_details, __pyx_t_1) < (0)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":101
 *     """
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         message,
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("adbc_driver_manager._lib.Error.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":148
 *     """An operation or some functionality is not supported."""
 * 
 *     def __init__(self, message, *, vendor_code=None, sqlstate=None, details=None):             # <<<<<<<<<<<<<<
 *         super().__init__(
 *             message,
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17NotSupportedError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_17NotSupportedError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17NotSupportedError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17NotSupportedError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_vendor_code = 0;
  PyObject *__pyx_v_sqlstate = 0;
  PyObject *__pyx_v_details = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_message,&__pyx_mstate_global->__pyx_n_u_vendor_code,&__pyx_mstate_global->__pyx_n_u_sqlstate,&__pyx_mstate_global->__pyx_n_u_details,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 148, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 148, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 148, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 148, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 148, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 148, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 148, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_self = values[0];
    __pyx_v_message = values[1];
    __pyx_v_vendor_code = values[2];
    __pyx_v_sqlstate = values[3];
    __pyx_v_details = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 148, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.NotSupportedError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_17NotSupportedError___init__(__pyx_self, __pyx_v_self, __pyx_v_message, __pyx_v_vendor_code, __pyx_v_sqlstate, __pyx_v_details);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17NotSupportedError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_vendor_code, PyObject *__pyx_v_sqlstate, PyObject *__pyx_v_details) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "adbc_driver_manager/_lib.pyx":149
 * 
 *     def __init__(self, message, *, vendor_code=None, sqlstate=None, details=None):
 *         super().__init__(             # <<<<<<<<<<<<<<
 *             message,
 *             status_code=AdbcStatusCode.NOT_IMPLEMENTED,
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_5) { PyErr_SetString(PyExc_RuntimeError, "super(): empty __class__ cell"); __PYX_ERR(0, 149, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_5, __pyx_v_self};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);

  /* "adbc_driver_manager/_lib.pyx":151
 *         super().__init__(
 *             message,
 *             status_code=AdbcStatusCode.NOT_IMPLEMENTED,             # <<<<<<<<<<<<<<
 *             vendor_code=vendor_code,
 *             sqlstate=sqlstate,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_AdbcStatusCode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_NOT_IMPLEMENTED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "adbc_driver_manager/_lib.pyx":154
 *             vendor_code=vendor_code,
 *             sqlstate=sqlstate,
 *             details=details,             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 4 : 0)] = {__pyx_t_2, __pyx_v_message};
    __pyx_t_5 = __Pyx_MakeVectorcallBuilderKwds(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_status_code, __pyx_t_4, __pyx_t_5, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 149, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_vendor_code, __pyx_v_vendor_code, __pyx_t_5, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 149, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sqlstate, __pyx_v_sqlstate, __pyx_t_5, __pyx_callargs+2, 2) < (0)) __PYX_ERR(0, 149, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_details, __pyx_v_details, __pyx_t_5, __pyx_callargs+2, 3) < (0)) __PYX_ERR(0, 149, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":148
 *     """An operation or some functionality is not supported."""
 * 
 *     def __init__(self, message, *, vendor_code=None, sqlstate=None, details=None):             # <<<<<<<<<<<<<<
 *         super().__init__(
 *             message,
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("adbc_driver_manager._lib.NotSupportedError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_19adbc_driver_manager_4_lib_13convert_error_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "adbc_driver_manager/_lib.pyx":202
 *             message += f". Vendor code: {vendor_code}"
 *         if error.sqlstate[0] != 0:
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))             # <<<<<<<<<<<<<<
 *             sqlstate = sqlstate.decode("ascii", "replace")
 *             message += f". SQLSTATE: {sqlstate}"
*/

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13convert_error_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *)__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 202, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_19adbc_driver_manager_4_lib_13convert_error_2generator, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_genexpr, __pyx_mstate_global->__pyx_n_u_convert_error_locals_genexpr, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib); if (unlikely(!gen)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.convert_error.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_19adbc_driver_manager_4_lib_13convert_error_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 202, __pyx_L1_error)
  }
  for (__pyx_t_1 = 0; __pyx_t_1 < 5; __pyx_t_1+=1) {
    __pyx_cur_scope->__pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_PyLong_From_char((__pyx_cur_scope->__pyx_outer_scope->__pyx_v_error->sqlstate[__pyx_cur_scope->__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 202, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":180
 * 
 * 
 * cdef object convert_error(CAdbcStatusCode status, CAdbcError* error):             # <<<<<<<<<<<<<<
 *     cdef CAdbcErrorDetail c_detail
 * 
*/

static PyObject *__pyx_f_19adbc_driver_manager_4_lib_convert_error(AdbcStatusCode __pyx_v_status, struct AdbcError *__pyx_v_error) {
  struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error *__pyx_cur_scope;
  struct AdbcErrorDetail __pyx_v_c_detail;
  PyObject *__pyx_v_message = NULL;
  PyObject *__pyx_v_vendor_code = NULL;
  PyObject *__pyx_v_sqlstate = NULL;
  PyObject *__pyx_v_details = NULL;
  int __pyx_v_num_details;
  int __pyx_v_index;
  PyObject *__pyx_v_klass = NULL;
  PyObject *__pyx_gb_19adbc_driver_manager_4_lib_13convert_error_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char const *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  AdbcStatusCode __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_error", 0);
  __pyx_cur_scope = (struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error *)__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 180, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_error = __pyx_v_error;

  /* "adbc_driver_manager/_lib.pyx":183
 *     cdef CAdbcErrorDetail c_detail
 * 
 *     if status == ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *         return None
 * 
*/
  __pyx_t_1 = (__pyx_v_status == ADBC_STATUS_OK);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":184
 * 
 *     if status == ADBC_STATUS_OK:
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     message = CAdbcStatusCodeMessage(status).decode("utf-8")
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "adbc_driver_manager/_lib.pyx":183
 *     cdef CAdbcErrorDetail c_detail
 * 
 *     if status == ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *         return None
 * 
*/
  }

  /* "adbc_driver_manager/_lib.pyx":186
 *         return None
 * 
 *     message = CAdbcStatusCodeMessage(status).decode("utf-8")             # <<<<<<<<<<<<<<
 *     vendor_code = None
 *     sqlstate = None
*/
  __pyx_t_2 = AdbcStatusCodeMessage(__pyx_v_status);
  __pyx_t_3 = __Pyx_ssize_strlen(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(__pyx_t_2, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_message = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":187
 * 
 *     message = CAdbcStatusCodeMessage(status).decode("utf-8")
 *     vendor_code = None             # <<<<<<<<<<<<<<
 *     sqlstate = None
 *     details = []
*/
  __Pyx_INCREF(Py_None);
  __pyx_v_vendor_code = Py_None;

  /* "adbc_driver_manager/_lib.pyx":188
 *     message = CAdbcStatusCodeMessage(status).decode("utf-8")
 *     vendor_code = None
 *     sqlstate = None             # <<<<<<<<<<<<<<
 *     details = []
 * 
*/
  __Pyx_INCREF(Py_None);
  __pyx_v_sqlstate = Py_None;

  /* "adbc_driver_manager/_lib.pyx":189
 *     vendor_code = None
 *     sqlstate = None
 *     details = []             # <<<<<<<<<<<<<<
 * 
 *     if error != NULL:
*/
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_details = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":191
 *     details = []
 * 
 *     if error != NULL:             # <<<<<<<<<<<<<<
 *         if error.message != NULL:
 *             message += ": "
*/
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_error != NULL);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":192
 * 
 *     if error != NULL:
 *         if error.message != NULL:             # <<<<<<<<<<<<<<
 *             message += ": "
 *             message += error.message.decode("utf-8", "replace")
*/
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_error->message != NULL);
    if (__pyx_t_1) {

      /* "adbc_driver_manager/_lib.pyx":193
 *     if error != NULL:
 *         if error.message != NULL:
 *             message += ": "             # <<<<<<<<<<<<<<
 *             message += error.message.decode("utf-8", "replace")
 *         if (
*/
      __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_message, __pyx_mstate_global->__pyx_kp_u_); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_message, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "adbc_driver_manager/_lib.pyx":194
 *         if error.message != NULL:
 *             message += ": "
 *             message += error.message.decode("utf-8", "replace")             # <<<<<<<<<<<<<<
 *         if (
 *                 error.vendor_code and
*/
      __pyx_t_5 = __pyx_cur_scope->__pyx_v_error->message;
      __pyx_t_3 = __Pyx_ssize_strlen(__pyx_t_5); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 194, __pyx_L1_error)
      __pyx_t_4 = __Pyx_decode_c_string(__pyx_t_5, 0, __pyx_t_3, NULL, __pyx_k_replace, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 194, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_message, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 194, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_message, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "adbc_driver_manager/_lib.pyx":192
 * 
 *     if error != NULL:
 *         if error.message != NULL:             # <<<<<<<<<<<<<<
 *             message += ": "
 *             message += error.message.decode("utf-8", "replace")
*/
    }

    /* "adbc_driver_manager/_lib.pyx":196
 *             message += error.message.decode("utf-8", "replace")
 *         if (
 *                 error.vendor_code and             # <<<<<<<<<<<<<<
 *                 error.vendor_code != ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA
 *         ):
*/
    __pyx_t_7 = (__pyx_cur_scope->__pyx_v_error->vendor_code != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_1 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }

    /* "adbc_driver_manager/_lib.pyx":197
 *         if (
 *                 error.vendor_code and
 *                 error.vendor_code != ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA             # <<<<<<<<<<<<<<
 *         ):
 *             vendor_code = error.vendor_code
*/
    __pyx_t_7 = (__pyx_cur_scope->__pyx_v_error->vendor_code != ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA);
    __pyx_t_1 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "adbc_driver_manager/_lib.pyx":195
 *             message += ": "
 *             message += error.message.decode("utf-8", "replace")
 *         if (             # <<<<<<<<<<<<<<
 *                 error.vendor_code and
 *                 error.vendor_code != ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA
*/
    if (__pyx_t_1) {

      /* "adbc_driver_manager/_lib.pyx":199
 *                 error.vendor_code != ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA
 *         ):
 *             vendor_code = error.vendor_code             # <<<<<<<<<<<<<<
 *             message += f". Vendor code: {vendor_code}"
 *         if error.sqlstate[0] != 0:
*/
      __pyx_t_6 = __Pyx_PyLong_From_int32_t(__pyx_cur_scope->__pyx_v_error->vendor_code); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_vendor_code, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "adbc_driver_manager/_lib.pyx":200
 *         ):
 *             vendor_code = error.vendor_code
 *             message += f". Vendor code: {vendor_code}"             # <<<<<<<<<<<<<<
 *         if error.sqlstate[0] != 0:
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))
*/
      __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_vendor_code, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Vendor_code, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_message, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_message, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "adbc_driver_manager/_lib.pyx":195
 *             message += ": "
 *             message += error.message.decode("utf-8", "replace")
 *         if (             # <<<<<<<<<<<<<<
 *                 error.vendor_code and
 *                 error.vendor_code != ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA
*/
    }

    /* "adbc_driver_manager/_lib.pyx":201
 *             vendor_code = error.vendor_code
 *             message += f". Vendor code: {vendor_code}"
 *         if error.sqlstate[0] != 0:             # <<<<<<<<<<<<<<
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))
 *             sqlstate = sqlstate.decode("ascii", "replace")
*/
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_error->sqlstate[0]) != 0);
    if (__pyx_t_1) {

      /* "adbc_driver_manager/_lib.pyx":202
 *             message += f". Vendor code: {vendor_code}"
 *         if error.sqlstate[0] != 0:
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))             # <<<<<<<<<<<<<<
 *             sqlstate = sqlstate.decode("ascii", "replace")
 *             message += f". SQLSTATE: {sqlstate}"
*/
      __pyx_t_4 = NULL;
      __pyx_t_8 = __pyx_pf_19adbc_driver_manager_4_lib_13convert_error_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_8};
        __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_DECREF_SET(__pyx_v_sqlstate, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "adbc_driver_manager/_lib.pyx":203
 *         if error.sqlstate[0] != 0:
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))
 *             sqlstate = sqlstate.decode("ascii", "replace")             # <<<<<<<<<<<<<<
 *             message += f". SQLSTATE: {sqlstate}"
 * 
*/
      __pyx_t_6 = __Pyx_decode_bytes(__pyx_v_sqlstate, 0, PY_SSIZE_T_MAX, NULL, __pyx_k_replace, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_sqlstate, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "adbc_driver_manager/_lib.pyx":204
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))
 *             sqlstate = sqlstate.decode("ascii", "replace")
 *             message += f". SQLSTATE: {sqlstate}"             # <<<<<<<<<<<<<<
 * 
 *         num_details = AdbcErrorGetDetailCount(error)
*/
      __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_sqlstate, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_SQLSTATE, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_message, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF_SET(__pyx_v_message, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "adbc_driver_manager/_lib.pyx":201
 *             vendor_code = error.vendor_code
 *             message += f". Vendor code: {vendor_code}"
 *         if error.sqlstate[0] != 0:             # <<<<<<<<<<<<<<
 *             sqlstate = bytes(error.sqlstate[i] for i in range(5))
 *             sqlstate = sqlstate.decode("ascii", "replace")
*/
    }

    /* "adbc_driver_manager/_lib.pyx":206
 *             message += f". SQLSTATE: {sqlstate}"
 * 
 *         num_details = AdbcErrorGetDetailCount(error)             # <<<<<<<<<<<<<<
 *         for index in range(num_details):
 *             c_detail = AdbcErrorGetDetail(error, index)
*/
    __pyx_v_num_details = AdbcErrorGetDetailCount(__pyx_cur_scope->__pyx_v_error);

    /* "adbc_driver_manager/_lib.pyx":207
 * 
 *         num_details = AdbcErrorGetDetailCount(error)
 *         for index in range(num_details):             # <<<<<<<<<<<<<<
 *             c_detail = AdbcErrorGetDetail(error, index)
 *             if c_detail.key == NULL or c_detail.value == NULL:
*/
    __pyx_t_10 = __pyx_v_num_details;
    __pyx_t_11 = __pyx_t_10;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
      __pyx_v_index = __pyx_t_12;

      /* "adbc_driver_manager/_lib.pyx":208
 *         num_details = AdbcErrorGetDetailCount(error)
 *         for index in range(num_details):
 *             c_detail = AdbcErrorGetDetail(error, index)             # <<<<<<<<<<<<<<
 *             if c_detail.key == NULL or c_detail.value == NULL:
 *                 # Shouldn't happen...
*/
      __pyx_v_c_detail = AdbcErrorGetDetail(__pyx_cur_scope->__pyx_v_error, __pyx_v_index);

      /* "adbc_driver_manager/_lib.pyx":209
 *         for index in range(num_details):
 *             c_detail = AdbcErrorGetDetail(error, index)
 *             if c_detail.key == NULL or c_detail.value == NULL:             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen...
 *                 break
*/
      __pyx_t_7 = (__pyx_v_c_detail.key == NULL);
      if (!__pyx_t_7) {
      } else {
        __pyx_t_1 = __pyx_t_7;
        goto __pyx_L13_bool_binop_done;
      }
      __pyx_t_7 = (__pyx_v_c_detail.value == NULL);
      __pyx_t_1 = __pyx_t_7;
      __pyx_L13_bool_binop_done:;
      if (__pyx_t_1) {

        /* "adbc_driver_manager/_lib.pyx":211
 *             if c_detail.key == NULL or c_detail.value == NULL:
 *                 # Shouldn't happen...
 *                 break             # <<<<<<<<<<<<<<
 *             details.append(
 *                 (c_detail.key,
*/
        goto __pyx_L11_break;

        /* "adbc_driver_manager/_lib.pyx":209
 *         for index in range(num_details):
 *             c_detail = AdbcErrorGetDetail(error, index)
 *             if c_detail.key == NULL or c_detail.value == NULL:             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen...
 *                 break
*/
      }

      /* "adbc_driver_manager/_lib.pyx":213
 *                 break
 *             details.append(
 *                 (c_detail.key,             # <<<<<<<<<<<<<<
 *                  PyBytes_FromStringAndSize(
 *                      <const char*> c_detail.value,
*/
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_c_detail.key); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "adbc_driver_manager/_lib.pyx":214
 *             details.append(
 *                 (c_detail.key,
 *                  PyBytes_FromStringAndSize(             # <<<<<<<<<<<<<<
 *                      <const char*> c_detail.value,
 *                      c_detail.value_length)))
*/
      __pyx_t_8 = PyBytes_FromStringAndSize(((char const *)__pyx_v_c_detail.value), __pyx_v_c_detail.value_length); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "adbc_driver_manager/_lib.pyx":213
 *                 break
 *             details.append(
 *                 (c_detail.key,             # <<<<<<<<<<<<<<
 *                  PyBytes_FromStringAndSize(
 *                      <const char*> c_detail.value,
*/
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6) != (0)) __PYX_ERR(0, 213, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8) != (0)) __PYX_ERR(0, 213, __pyx_L1_error);
      __pyx_t_6 = 0;
      __pyx_t_8 = 0;

      /* "adbc_driver_manager/_lib.pyx":212
 *                 # Shouldn't happen...
 *                 break
 *             details.append(             # <<<<<<<<<<<<<<
 *                 (c_detail.key,
 *                  PyBytes_FromStringAndSize(
*/
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_details, __pyx_t_4); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L11_break:;

    /* "adbc_driver_manager/_lib.pyx":218
 *                      c_detail.value_length)))
 * 
 *         if error.release:             # <<<<<<<<<<<<<<
 *             error.release(error)
 * 
*/
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_error->release != 0);
    if (__pyx_t_1) {

      /* "adbc_driver_manager/_lib.pyx":219
 * 
 *         if error.release:
 *             error.release(error)             # <<<<<<<<<<<<<<
 * 
 *     klass = Error
*/
      __pyx_cur_scope->__pyx_v_error->release(__pyx_cur_scope->__pyx_v_error);

      /* "adbc_driver_manager/_lib.pyx":218
 *                      c_detail.value_length)))
 * 
 *         if error.release:             # <<<<<<<<<<<<<<
 *             error.release(error)
 * 
*/
    }

    /* "adbc_driver_manager/_lib.pyx":191
 *     details = []
 * 
 *     if error != NULL:             # <<<<<<<<<<<<<<
 *         if error.message != NULL:
 *             message += ": "
*/
  }

  /* "adbc_driver_manager/_lib.pyx":221
 *             error.release(error)
 * 
 *     klass = Error             # <<<<<<<<<<<<<<
 *     if status in (ADBC_STATUS_INVALID_DATA,):
 *         klass = DataError
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_klass = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":222
 * 
 *     klass = Error
 *     if status in (ADBC_STATUS_INVALID_DATA,):             # <<<<<<<<<<<<<<
 *         klass = DataError
 *     elif status in (
*/
  __pyx_t_14 = __pyx_v_status;
  __pyx_t_1 = (__pyx_t_14 == ADBC_STATUS_INVALID_DATA);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":223
 *     klass = Error
 *     if status in (ADBC_STATUS_INVALID_DATA,):
 *         klass = DataError             # <<<<<<<<<<<<<<
 *     elif status in (
 *         ADBC_STATUS_IO,
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DataError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_klass, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":222
 * 
 *     klass = Error
 *     if status in (ADBC_STATUS_INVALID_DATA,):             # <<<<<<<<<<<<<<
 *         klass = DataError
 *     elif status in (
*/
    goto __pyx_L16;
  }

  /* "adbc_driver_manager/_lib.pyx":224
 *     if status in (ADBC_STATUS_INVALID_DATA,):
 *         klass = DataError
 *     elif status in (             # <<<<<<<<<<<<<<
 *         ADBC_STATUS_IO,
 *         ADBC_STATUS_CANCELLED,
*/
  __pyx_t_14 = __pyx_v_status;

  /* "adbc_driver_manager/_lib.pyx":225
 *         klass = DataError
 *     elif status in (
 *         ADBC_STATUS_IO,             # <<<<<<<<<<<<<<
 *         ADBC_STATUS_CANCELLED,
 *         ADBC_STATUS_TIMEOUT,
*/
  __pyx_t_7 = (__pyx_t_14 == ADBC_STATUS_IO);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L17_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":224
 *     if status in (ADBC_STATUS_INVALID_DATA,):
 *         klass = DataError
 *     elif status in (             # <<<<<<<<<<<<<<
 *         ADBC_STATUS_IO,
 *         ADBC_STATUS_CANCELLED,
*/
  __pyx_t_7 = (__pyx_t_14 == ADBC_STATUS_CANCELLED);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L17_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":227
 *         ADBC_STATUS_IO,
 *         ADBC_STATUS_CANCELLED,
 *         ADBC_STATUS_TIMEOUT,             # <<<<<<<<<<<<<<
 *         ADBC_STATUS_UNKNOWN,
 *     ):
*/
  __pyx_t_7 = (__pyx_t_14 == ADBC_STATUS_TIMEOUT);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L17_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":224
 *     if status in (ADBC_STATUS_INVALID_DATA,):
 *         klass = DataError
 *     elif status in (             # <<<<<<<<<<<<<<
 *         ADBC_STATUS_IO,
 *         ADBC_STATUS_CANCELLED,
*/
  __pyx_t_7 = (__pyx_t_14 == ADBC_STATUS_UNKNOWN);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L17_bool_binop_done:;
  __pyx_t_7 = __pyx_t_1;
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":230
 *         ADBC_STATUS_UNKNOWN,
 *     ):
 *         klass = OperationalError             # <<<<<<<<<<<<<<
 *     elif status in (ADBC_STATUS_INTEGRITY,):
 *         klass = IntegrityError
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_OperationalError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_klass, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":224
 *     if status in (ADBC_STATUS_INVALID_DATA,):
 *         klass = DataError
 *     elif status in (             # <<<<<<<<<<<<<<
 *         ADBC_STATUS_IO,
 *         ADBC_STATUS_CANCELLED,
*/
    goto __pyx_L16;
  }

  /* "adbc_driver_manager/_lib.pyx":231
 *     ):
 *         klass = OperationalError
 *     elif status in (ADBC_STATUS_INTEGRITY,):             # <<<<<<<<<<<<<<
 *         klass = IntegrityError
 *     elif status in (ADBC_STATUS_INTERNAL,):
*/
  __pyx_t_14 = __pyx_v_status;
  __pyx_t_7 = (__pyx_t_14 == ADBC_STATUS_INTEGRITY);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":232
 *         klass = OperationalError
 *     elif status in (ADBC_STATUS_INTEGRITY,):
 *         klass = IntegrityError             # <<<<<<<<<<<<<<
 *     elif status in (ADBC_STATUS_INTERNAL,):
 *         klass = InternalError
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_IntegrityError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_klass, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":231
 *     ):
 *         klass = OperationalError
 *     elif status in (ADBC_STATUS_INTEGRITY,):             # <<<<<<<<<<<<<<
 *         klass = IntegrityError
 *     elif status in (ADBC_STATUS_INTERNAL,):
*/
    goto __pyx_L16;
  }

  /* "adbc_driver_manager/_lib.pyx":233
 *     elif status in (ADBC_STATUS_INTEGRITY,):
 *         klass = IntegrityError
 *     elif status in (ADBC_STATUS_INTERNAL,):             # <<<<<<<<<<<<<<
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,
*/
  __pyx_t_14 = __pyx_v_status;
  __pyx_t_7 = (__pyx_t_14 == ADBC_STATUS_INTERNAL);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":234
 *         klass = IntegrityError
 *     elif status in (ADBC_STATUS_INTERNAL,):
 *         klass = InternalError             # <<<<<<<<<<<<<<
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,
 *                     ADBC_STATUS_INVALID_ARGUMENT,
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_InternalError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_klass, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":233
 *     elif status in (ADBC_STATUS_INTEGRITY,):
 *         klass = IntegrityError
 *     elif status in (ADBC_STATUS_INTERNAL,):             # <<<<<<<<<<<<<<
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,
*/
    goto __pyx_L16;
  }

  /* "adbc_driver_manager/_lib.pyx":235
 *     elif status in (ADBC_STATUS_INTERNAL,):
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,             # <<<<<<<<<<<<<<
 *                     ADBC_STATUS_INVALID_ARGUMENT,
 *                     ADBC_STATUS_INVALID_STATE,
*/
  __pyx_t_14 = __pyx_v_status;
  __pyx_t_1 = (__pyx_t_14 == ADBC_STATUS_ALREADY_EXISTS);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_7 = __pyx_t_1;
    goto __pyx_L21_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":236
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,
 *                     ADBC_STATUS_INVALID_ARGUMENT,             # <<<<<<<<<<<<<<
 *                     ADBC_STATUS_INVALID_STATE,
 *                     ADBC_STATUS_NOT_FOUND,
*/
  __pyx_t_1 = (__pyx_t_14 == ADBC_STATUS_INVALID_ARGUMENT);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_7 = __pyx_t_1;
    goto __pyx_L21_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":235
 *     elif status in (ADBC_STATUS_INTERNAL,):
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,             # <<<<<<<<<<<<<<
 *                     ADBC_STATUS_INVALID_ARGUMENT,
 *                     ADBC_STATUS_INVALID_STATE,
*/
  __pyx_t_1 = (__pyx_t_14 == ADBC_STATUS_INVALID_STATE);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_7 = __pyx_t_1;
    goto __pyx_L21_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":238
 *                     ADBC_STATUS_INVALID_ARGUMENT,
 *                     ADBC_STATUS_INVALID_STATE,
 *                     ADBC_STATUS_NOT_FOUND,             # <<<<<<<<<<<<<<
 *                     ADBC_STATUS_UNAUTHENTICATED,
 *                     ADBC_STATUS_UNAUTHORIZED):
*/
  __pyx_t_1 = (__pyx_t_14 == ADBC_STATUS_NOT_FOUND);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_7 = __pyx_t_1;
    goto __pyx_L21_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":235
 *     elif status in (ADBC_STATUS_INTERNAL,):
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,             # <<<<<<<<<<<<<<
 *                     ADBC_STATUS_INVALID_ARGUMENT,
 *                     ADBC_STATUS_INVALID_STATE,
*/
  __pyx_t_1 = (__pyx_t_14 == ADBC_STATUS_UNAUTHENTICATED);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_7 = __pyx_t_1;
    goto __pyx_L21_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":240
 *                     ADBC_STATUS_NOT_FOUND,
 *                     ADBC_STATUS_UNAUTHENTICATED,
 *                     ADBC_STATUS_UNAUTHORIZED):             # <<<<<<<<<<<<<<
 *         klass = ProgrammingError
 *     elif status == ADBC_STATUS_NOT_IMPLEMENTED:
*/
  __pyx_t_1 = (__pyx_t_14 == ADBC_STATUS_UNAUTHORIZED);
  __pyx_t_7 = __pyx_t_1;
  __pyx_L21_bool_binop_done:;

  /* "adbc_driver_manager/_lib.pyx":235
 *     elif status in (ADBC_STATUS_INTERNAL,):
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,             # <<<<<<<<<<<<<<
 *                     ADBC_STATUS_INVALID_ARGUMENT,
 *                     ADBC_STATUS_INVALID_STATE,
*/
  __pyx_t_1 = __pyx_t_7;
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":241
 *                     ADBC_STATUS_UNAUTHENTICATED,
 *                     ADBC_STATUS_UNAUTHORIZED):
 *         klass = ProgrammingError             # <<<<<<<<<<<<<<
 *     elif status == ADBC_STATUS_NOT_IMPLEMENTED:
 *         return NotSupportedError(
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ProgrammingError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_klass, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":235
 *     elif status in (ADBC_STATUS_INTERNAL,):
 *         klass = InternalError
 *     elif status in (ADBC_STATUS_ALREADY_EXISTS,             # <<<<<<<<<<<<<<
 *                     ADBC_STATUS_INVALID_ARGUMENT,
 *                     ADBC_STATUS_INVALID_STATE,
*/
    goto __pyx_L16;
  }

  /* "adbc_driver_manager/_lib.pyx":242
 *                     ADBC_STATUS_UNAUTHORIZED):
 *         klass = ProgrammingError
 *     elif status == ADBC_STATUS_NOT_IMPLEMENTED:             # <<<<<<<<<<<<<<
 *         return NotSupportedError(
 *             message,
*/
  __pyx_t_1 = (__pyx_v_status == ADBC_STATUS_NOT_IMPLEMENTED);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":243
 *         klass = ProgrammingError
 *     elif status == ADBC_STATUS_NOT_IMPLEMENTED:
 *         return NotSupportedError(             # <<<<<<<<<<<<<<
 *             message,
 *             vendor_code=vendor_code,
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_NotSupportedError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "adbc_driver_manager/_lib.pyx":247
 *             vendor_code=vendor_code,
 *             sqlstate=sqlstate,
 *             details=details,             # <<<<<<<<<<<<<<
 *         )
 *     return klass(
*/
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 3 : 0)] = {__pyx_t_8, __pyx_v_message};
      __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_vendor_code, __pyx_v_vendor_code, __pyx_t_15, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 243, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sqlstate, __pyx_v_sqlstate, __pyx_t_15, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 243, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_details, __pyx_v_details, __pyx_t_15, __pyx_callargs+2, 2) < (0)) __PYX_ERR(0, 243, __pyx_L1_error)
      __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "adbc_driver_manager/_lib.pyx":242
 *                     ADBC_STATUS_UNAUTHORIZED):
 *         klass = ProgrammingError
 *     elif status == ADBC_STATUS_NOT_IMPLEMENTED:             # <<<<<<<<<<<<<<
 *         return NotSupportedError(
 *             message,
*/
  }
  __pyx_L16:;

  /* "adbc_driver_manager/_lib.pyx":249
 *             details=details,
 *         )
 *     return klass(             # <<<<<<<<<<<<<<
 *         message,
 *         status_code=status,
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = NULL;
  __Pyx_INCREF(__pyx_v_klass);
  __pyx_t_15 = __pyx_v_klass; 

  /* "adbc_driver_manager/_lib.pyx":251
 *     return klass(
 *         message,
 *         status_code=status,             # <<<<<<<<<<<<<<
 *         vendor_code=vendor_code,
 *         sqlstate=sqlstate,
*/
  __pyx_t_8 = __Pyx_PyLong_From_uint8_t(__pyx_v_status); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "adbc_driver_manager/_lib.pyx":254
 *         vendor_code=vendor_code,
 *         sqlstate=sqlstate,
 *         details=details,             # <<<<<<<<<<<<<<
 *     )
 * 
*/
  __pyx_t_9 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_15);
    assert(__pyx_t_6);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_15);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_15, __pyx__function);
    __pyx_t_9 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 4 : 0)] = {__pyx_t_6, __pyx_v_message};
    __pyx_t_16 = __Pyx_MakeVectorcallBuilderKwds(4); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_status_code, __pyx_t_8, __pyx_t_16, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_vendor_code, __pyx_v_vendor_code, __pyx_t_16, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sqlstate, __pyx_v_sqlstate, __pyx_t_16, __pyx_callargs+2, 2) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_details, __pyx_v_details, __pyx_t_16, __pyx_callargs+2, 3) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_16);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":180
 * 
 * 
 * cdef object convert_error(CAdbcStatusCode status, CAdbcError* error):             # <<<<<<<<<<<<<<
 *     cdef CAdbcErrorDetail c_detail
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("adbc_driver_manager._lib.convert_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message);
  __Pyx_XDECREF(__pyx_v_vendor_code);
  __Pyx_XDECREF(__pyx_v_sqlstate);
  __Pyx_XDECREF(__pyx_v_details);
  __Pyx_XDECREF(__pyx_v_klass);
  __Pyx_XDECREF(__pyx_gb_19adbc_driver_manager_4_lib_13convert_error_2generator);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":258
 * 
 * 
 * cdef void check_error(CAdbcStatusCode status, CAdbcError* error) except *:             # <<<<<<<<<<<<<<
 *     if status == ADBC_STATUS_OK:
 *         return
*/

static void __pyx_f_19adbc_driver_manager_4_lib_check_error(AdbcStatusCode __pyx_v_status, struct AdbcError *__pyx_v_error) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_error", 0);

  /* "adbc_driver_manager/_lib.pyx":259
 * 
 * cdef void check_error(CAdbcStatusCode status, CAdbcError* error) except *:
 *     if status == ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_1 = (__pyx_v_status == ADBC_STATUS_OK);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":260
 * cdef void check_error(CAdbcStatusCode status, CAdbcError* error) except *:
 *     if status == ADBC_STATUS_OK:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     raise convert_error(status, error)
*/
    goto __pyx_L0;

    /* "adbc_driver_manager/_lib.pyx":259
 * 
 * cdef void check_error(CAdbcStatusCode status, CAdbcError* error) except *:
 *     if status == ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "adbc_driver_manager/_lib.pyx":262
 *         return
 * 
 *     raise convert_error(status, error)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib_convert_error(__pyx_v_status, __pyx_v_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 262, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":258
 * 
 * 
 * cdef void check_error(CAdbcStatusCode status, CAdbcError* error) except *:             # <<<<<<<<<<<<<<
 *     if status == ADBC_STATUS_OK:
 *         return
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.check_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "adbc_driver_manager/_lib.pyx":265
 * 
 * 
 * cdef CAdbcError empty_error():             # <<<<<<<<<<<<<<
 *     cdef CAdbcError error
 *     memset(&error, 0, cython.sizeof(error))
*/

static struct AdbcError __pyx_f_19adbc_driver_manager_4_lib_empty_error(void) {
  struct AdbcError __pyx_v_error;
  struct AdbcError __pyx_r;

  /* "adbc_driver_manager/_lib.pyx":267
 * cdef CAdbcError empty_error():
 *     cdef CAdbcError error
 *     memset(&error, 0, cython.sizeof(error))             # <<<<<<<<<<<<<<
 *     # We always want the extended error info
 *     error.vendor_code = ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA
*/
  (void)(memset((&__pyx_v_error), 0, (sizeof(__pyx_v_error))));

  /* "adbc_driver_manager/_lib.pyx":269
 *     memset(&error, 0, cython.sizeof(error))
 *     # We always want the extended error info
 *     error.vendor_code = ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA             # <<<<<<<<<<<<<<
 *     return error
 * 
*/
  __pyx_v_error.vendor_code = ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA;

  /* "adbc_driver_manager/_lib.pyx":270
 *     # We always want the extended error info
 *     error.vendor_code = ADBC_ERROR_VENDOR_CODE_PRIVATE_DATA
 *     return error             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = __pyx_v_error;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":265
 * 
 * 
 * cdef CAdbcError empty_error():             # <<<<<<<<<<<<<<
 *     cdef CAdbcError error
 *     memset(&error, 0, cython.sizeof(error))
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":273
 * 
 * 
 * cdef bytes _to_bytes(obj, str name):             # <<<<<<<<<<<<<<
 *     if isinstance(obj, bytes):
 *         return obj
*/

static PyObject *__pyx_f_19adbc_driver_manager_4_lib__to_bytes(PyObject *__pyx_v_obj, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_to_bytes", 0);

  /* "adbc_driver_manager/_lib.pyx":274
 * 
 * cdef bytes _to_bytes(obj, str name):
 *     if isinstance(obj, bytes):             # <<<<<<<<<<<<<<
 *         return obj
 *     elif isinstance(obj, str):
*/
  __pyx_t_1 = PyBytes_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":275
 * cdef bytes _to_bytes(obj, str name):
 *     if isinstance(obj, bytes):
 *         return obj             # <<<<<<<<<<<<<<
 *     elif isinstance(obj, str):
 *         return obj.encode("utf-8")
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_obj;
    __Pyx_INCREF(__pyx_t_2);
    if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 275, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "adbc_driver_manager/_lib.pyx":274
 * 
 * cdef bytes _to_bytes(obj, str name):
 *     if isinstance(obj, bytes):             # <<<<<<<<<<<<<<
 *         return obj
 *     elif isinstance(obj, str):
*/
  }

  /* "adbc_driver_manager/_lib.pyx":276
 *     if isinstance(obj, bytes):
 *         return obj
 *     elif isinstance(obj, str):             # <<<<<<<<<<<<<<
 *         return obj.encode("utf-8")
 *     raise ValueError(f"{name} must be str or bytes")
*/
  __pyx_t_1 = PyUnicode_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":277
 *         return obj
 *     elif isinstance(obj, str):
 *         return obj.encode("utf-8")             # <<<<<<<<<<<<<<
 *     raise ValueError(f"{name} must be str or bytes")
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_v_obj;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_utf_8};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 277, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "adbc_driver_manager/_lib.pyx":276
 *     if isinstance(obj, bytes):
 *         return obj
 *     elif isinstance(obj, str):             # <<<<<<<<<<<<<<
 *         return obj.encode("utf-8")
 *     raise ValueError(f"{name} must be str or bytes")
*/
  }

  /* "adbc_driver_manager/_lib.pyx":278
 *     elif isinstance(obj, str):
 *         return obj.encode("utf-8")
 *     raise ValueError(f"{name} must be str or bytes")             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = NULL;
  __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_must_be_str_or_bytes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 278, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":273
 * 
 * 
 * cdef bytes _to_bytes(obj, str name):             # <<<<<<<<<<<<<<
 *     if isinstance(obj, bytes):
 *         return obj
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("adbc_driver_manager._lib._to_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":281
 * 
 * 
 * def _test_error(status_code, message, vendor_code, sqlstate) -> Error:             # <<<<<<<<<<<<<<
 *     cdef CAdbcError error
 *     error.release = NULL
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_1_test_error(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_1_test_error = {"_test_error", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_1_test_error, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_1_test_error(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_status_code = 0;
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_vendor_code = 0;
  PyObject *__pyx_v_sqlstate = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_error (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_status_code,&__pyx_mstate_global->__pyx_n_u_message,&__pyx_mstate_global->__pyx_n_u_vendor_code,&__pyx_mstate_global->__pyx_n_u_sqlstate,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 281, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_test_error", 0) < (0)) __PYX_ERR(0, 281, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_test_error", 1, 4, 4, i); __PYX_ERR(0, 281, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 281, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 281, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 281, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 281, __pyx_L3_error)
    }
    __pyx_v_status_code = values[0];
    __pyx_v_message = values[1];
    __pyx_v_vendor_code = values[2];
    __pyx_v_sqlstate = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_test_error", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 281, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib._test_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib__test_error(__pyx_self, __pyx_v_status_code, __pyx_v_message, __pyx_v_vendor_code, __pyx_v_sqlstate);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib__test_error(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_status_code, PyObject *__pyx_v_message, PyObject *__pyx_v_vendor_code, PyObject *__pyx_v_sqlstate) {
  struct AdbcError __pyx_v_error;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  int32_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  long __pyx_t_7;
  char __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  AdbcStatusCode __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_error", 0);
  __Pyx_INCREF(__pyx_v_message);
  __Pyx_INCREF(__pyx_v_sqlstate);

  /* "adbc_driver_manager/_lib.pyx":283
 * def _test_error(status_code, message, vendor_code, sqlstate) -> Error:
 *     cdef CAdbcError error
 *     error.release = NULL             # <<<<<<<<<<<<<<
 * 
 *     message = _to_bytes(message, "message")
*/
  __pyx_v_error.release = NULL;

  /* "adbc_driver_manager/_lib.pyx":285
 *     error.release = NULL
 * 
 *     message = _to_bytes(message, "message")             # <<<<<<<<<<<<<<
 *     error.message = message
 * 
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_message, __pyx_mstate_global->__pyx_n_u_message); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_message, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":286
 * 
 *     message = _to_bytes(message, "message")
 *     error.message = message             # <<<<<<<<<<<<<<
 * 
 *     if vendor_code:
*/
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_v_message); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_v_error.message = __pyx_t_2;

  /* "adbc_driver_manager/_lib.pyx":288
 *     error.message = message
 * 
 *     if vendor_code:             # <<<<<<<<<<<<<<
 *         error.vendor_code = vendor_code
 *     else:
*/
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_vendor_code); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 288, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "adbc_driver_manager/_lib.pyx":289
 * 
 *     if vendor_code:
 *         error.vendor_code = vendor_code             # <<<<<<<<<<<<<<
 *     else:
 *         error.vendor_code = 0
*/
    __pyx_t_4 = __Pyx_PyLong_As_int32_t(__pyx_v_vendor_code); if (unlikely((__pyx_t_4 == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L1_error)
    __pyx_v_error.vendor_code = __pyx_t_4;

    /* "adbc_driver_manager/_lib.pyx":288
 *     error.message = message
 * 
 *     if vendor_code:             # <<<<<<<<<<<<<<
 *         error.vendor_code = vendor_code
 *     else:
*/
    goto __pyx_L3;
  }

  /* "adbc_driver_manager/_lib.pyx":291
 *         error.vendor_code = vendor_code
 *     else:
 *         error.vendor_code = 0             # <<<<<<<<<<<<<<
 * 
 *     if sqlstate:
*/
  /*else*/ {
    __pyx_v_error.vendor_code = 0;
  }
  __pyx_L3:;

  /* "adbc_driver_manager/_lib.pyx":293
 *         error.vendor_code = 0
 * 
 *     if sqlstate:             # <<<<<<<<<<<<<<
 *         sqlstate = sqlstate.encode("ascii")
 *     else:
*/
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_sqlstate); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 293, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "adbc_driver_manager/_lib.pyx":294
 * 
 *     if sqlstate:
 *         sqlstate = sqlstate.encode("ascii")             # <<<<<<<<<<<<<<
 *     else:
 *         sqlstate = b"\0\0\0\0\0"
*/
    __pyx_t_5 = __pyx_v_sqlstate;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ascii};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF_SET(__pyx_v_sqlstate, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "adbc_driver_manager/_lib.pyx":293
 *         error.vendor_code = 0
 * 
 *     if sqlstate:             # <<<<<<<<<<<<<<
 *         sqlstate = sqlstate.encode("ascii")
 *     else:
*/
    goto __pyx_L4;
  }

  /* "adbc_driver_manager/_lib.pyx":296
 *         sqlstate = sqlstate.encode("ascii")
 *     else:
 *         sqlstate = b"\0\0\0\0\0"             # <<<<<<<<<<<<<<
 *     for i in range(5):
 *         error.sqlstate[i] = sqlstate[i]
*/
  /*else*/ {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__2);
    __Pyx_DECREF_SET(__pyx_v_sqlstate, __pyx_mstate_global->__pyx_kp_b__2);
  }
  __pyx_L4:;

  /* "adbc_driver_manager/_lib.pyx":297
 *     else:
 *         sqlstate = b"\0\0\0\0\0"
 *     for i in range(5):             # <<<<<<<<<<<<<<
 *         error.sqlstate[i] = sqlstate[i]
 * 
*/
  for (__pyx_t_7 = 0; __pyx_t_7 < 5; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "adbc_driver_manager/_lib.pyx":298
 *         sqlstate = b"\0\0\0\0\0"
 *     for i in range(5):
 *         error.sqlstate[i] = sqlstate[i]             # <<<<<<<<<<<<<<
 * 
 *     return check_error(AdbcStatusCode(status_code), &error)
*/
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_sqlstate, __pyx_v_i, long, 1, __Pyx_PyLong_From_long, 0, 1, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyLong_As_char(__pyx_t_1); if (unlikely((__pyx_t_8 == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_error.sqlstate[__pyx_v_i]) = __pyx_t_8;
  }

  /* "adbc_driver_manager/_lib.pyx":300
 *         error.sqlstate[i] = sqlstate[i]
 * 
 *     return check_error(AdbcStatusCode(status_code), &error)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_AdbcStatusCode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
    assert(__pyx_t_5);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_status_code};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_9, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_10 = __Pyx_PyLong_As_uint8_t(__pyx_t_1); if (unlikely((__pyx_t_10 == ((AdbcStatusCode)-1)) && PyErr_Occurred())) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_t_10, (&__pyx_v_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 300, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":281
 * 
 * 
 * def _test_error(status_code, message, vendor_code, sqlstate) -> Error:             # <<<<<<<<<<<<<<
 *     cdef CAdbcError error
 *     error.release = NULL
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("adbc_driver_manager._lib._test_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message);
  __Pyx_XDECREF(__pyx_v_sqlstate);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":313
 *         str _child_type
 * 
 *     def __init__(self, str child_type):             # <<<<<<<<<<<<<<
 *         self._lock = threading.Lock()
 *         self._child_type = child_type
*/

/* Python wrapper */
static int __pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_child_type = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_child_type,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 313, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 313, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 313, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 313, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 313, __pyx_L3_error)
    }
    __pyx_v_child_type = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 313, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_child_type), (&PyUnicode_Type), 1, "child_type", 1))) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle___init__(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self), __pyx_v_child_type);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle___init__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self, PyObject *__pyx_v_child_type) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "adbc_driver_manager/_lib.pyx":314
 * 
 *     def __init__(self, str child_type):
 *         self._lock = threading.Lock()             # <<<<<<<<<<<<<<
 *         self._child_type = child_type
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Lock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_lock);
  __Pyx_DECREF(__pyx_v_self->_lock);
  __pyx_v_self->_lock = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":315
 *     def __init__(self, str child_type):
 *         self._lock = threading.Lock()
 *         self._child_type = child_type             # <<<<<<<<<<<<<<
 * 
 *     def __enter__(self) -> "Self":
*/
  __Pyx_INCREF(__pyx_v_child_type);
  __Pyx_GIVEREF(__pyx_v_child_type);
  __Pyx_GOTREF(__pyx_v_self->_child_type);
  __Pyx_DECREF(__pyx_v_self->_child_type);
  __pyx_v_self->_child_type = __pyx_v_child_type;

  /* "adbc_driver_manager/_lib.pyx":313
 *         str _child_type
 * 
 *     def __init__(self, str child_type):             # <<<<<<<<<<<<<<
 *         self._lock = threading.Lock()
 *         self._child_type = child_type
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":317
 *         self._child_type = child_type
 * 
 *     def __enter__(self) -> "Self":             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_3__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_3__enter__ = {"__enter__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_3__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_3__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__enter__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__enter__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_2__enter__(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_2__enter__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__", 0);

  /* "adbc_driver_manager/_lib.pyx":318
 * 
 *     def __enter__(self) -> "Self":
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb) -> None:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":317
 *         self._child_type = child_type
 * 
 *     def __enter__(self) -> "Self":             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":320
 *         return self
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb) -> None:             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_5__exit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_5__exit__ = {"__exit__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_5__exit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_5__exit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_exc_type = 0;
  CYTHON_UNUSED PyObject *__pyx_v_exc_val = 0;
  CYTHON_UNUSED PyObject *__pyx_v_exc_tb = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_exc_type,&__pyx_mstate_global->__pyx_n_u_exc_val,&__pyx_mstate_global->__pyx_n_u_exc_tb,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 320, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__exit__", 0) < (0)) __PYX_ERR(0, 320, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, i); __PYX_ERR(0, 320, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 320, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 320, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 320, __pyx_L3_error)
    }
    __pyx_v_exc_type = values[0];
    __pyx_v_exc_val = values[1];
    __pyx_v_exc_tb = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 320, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_4__exit__(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self), __pyx_v_exc_type, __pyx_v_exc_val, __pyx_v_exc_tb);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_4__exit__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc_type, CYTHON_UNUSED PyObject *__pyx_v_exc_val, CYTHON_UNUSED PyObject *__pyx_v_exc_tb) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__exit__", 0);

  /* "adbc_driver_manager/_lib.pyx":321
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb) -> None:
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cdef _open_child(self):
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_close, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":320
 *         return self
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb) -> None:             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":323
 *         self.close()
 * 
 *     cdef _open_child(self):             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             self._open_children += 1
*/

static PyObject *__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__open_child(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open_child", 0);

  /* "adbc_driver_manager/_lib.pyx":324
 * 
 *     cdef _open_child(self):
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             self._open_children += 1
 * 
*/
  /*with:*/ {
    __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->_lock, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->_lock, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 324, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        (void)__pyx_t_6; (void)__pyx_t_7; (void)__pyx_t_8; /* mark used */
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":325
 *     cdef _open_child(self):
 *         with self._lock:
 *             self._open_children += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef _close_child(self):
*/
          __pyx_v_self->_open_children = (__pyx_v_self->_open_children + 1);

          /* "adbc_driver_manager/_lib.pyx":324
 * 
 *     cdef _open_child(self):
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             self._open_children += 1
 * 
*/
        }
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_1) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 324, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L13;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L1_error;
    __pyx_L13:;
  }

  /* "adbc_driver_manager/_lib.pyx":323
 *         self.close()
 * 
 *     cdef _open_child(self):             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             self._open_children += 1
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle._open_child", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":327
 *             self._open_children += 1
 * 
 *     cdef _close_child(self):             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             if self._open_children == 0:
*/

static PyObject *__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__close_child(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close_child", 0);

  /* "adbc_driver_manager/_lib.pyx":328
 * 
 *     cdef _close_child(self):
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self._open_children == 0:
 *                 raise RuntimeError(
*/
  /*with:*/ {
    __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->_lock, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->_lock, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":329
 *     cdef _close_child(self):
 *         with self._lock:
 *             if self._open_children == 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Underflow in closing this {self._child_type}")
*/
          __pyx_t_9 = (__pyx_v_self->_open_children == 0);
          if (unlikely(__pyx_t_9)) {

            /* "adbc_driver_manager/_lib.pyx":330
 *         with self._lock:
 *             if self._open_children == 0:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f"Underflow in closing this {self._child_type}")
 *             self._open_children -= 1
*/
            __pyx_t_4 = NULL;

            /* "adbc_driver_manager/_lib.pyx":331
 *             if self._open_children == 0:
 *                 raise RuntimeError(
 *                     f"Underflow in closing this {self._child_type}")             # <<<<<<<<<<<<<<
 *             self._open_children -= 1
 * 
*/
            __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_self->_child_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_10 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Underflow_in_closing_this, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 331, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_5 = 1;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_10};
              __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
            }
            __Pyx_Raise(__pyx_t_2, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_ERR(0, 330, __pyx_L7_error)

            /* "adbc_driver_manager/_lib.pyx":329
 *     cdef _close_child(self):
 *         with self._lock:
 *             if self._open_children == 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Underflow in closing this {self._child_type}")
*/
          }

          /* "adbc_driver_manager/_lib.pyx":332
 *                 raise RuntimeError(
 *                     f"Underflow in closing this {self._child_type}")
 *             self._open_children -= 1             # <<<<<<<<<<<<<<
 * 
 *     cdef _check_open_children(self):
*/
          __pyx_v_self->_open_children = (__pyx_v_self->_open_children - 1);

          /* "adbc_driver_manager/_lib.pyx":328
 * 
 *     cdef _close_child(self):
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self._open_children == 0:
 *                 raise RuntimeError(
*/
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle._close_child", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_10, &__pyx_t_4) < 0) __PYX_ERR(0, 328, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_4);
          __pyx_t_3 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 328, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (__pyx_t_9 < (0)) __PYX_ERR(0, 328, __pyx_L9_except_error)
          __pyx_t_12 = (!__pyx_t_9);
          if (unlikely(__pyx_t_12)) {
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_4);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_10, __pyx_t_4);
            __pyx_t_2 = 0;  __pyx_t_10 = 0;  __pyx_t_4 = 0; 
            __PYX_ERR(0, 328, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_1) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 328, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L17;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L1_error;
    __pyx_L17:;
  }

  /* "adbc_driver_manager/_lib.pyx":327
 *             self._open_children += 1
 * 
 *     cdef _close_child(self):             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             if self._open_children == 0:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle._close_child", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":334
 *             self._open_children -= 1
 * 
 *     cdef _check_open_children(self):             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             if self._open_children != 0:
*/

static PyObject *__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__check_open_children(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11[4];
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_open_children", 0);

  /* "adbc_driver_manager/_lib.pyx":335
 * 
 *     cdef _check_open_children(self):
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self._open_children != 0:
 *                 raise RuntimeError(
*/
  /*with:*/ {
    __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->_lock, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->_lock, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":336
 *     cdef _check_open_children(self):
 *         with self._lock:
 *             if self._open_children != 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Cannot close {self.__class__.__name__} "
*/
          __pyx_t_9 = (__pyx_v_self->_open_children != 0);
          if (unlikely(__pyx_t_9)) {

            /* "adbc_driver_manager/_lib.pyx":337
 *         with self._lock:
 *             if self._open_children != 0:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f"Cannot close {self.__class__.__name__} "
 *                     f"with open {self._child_type}")
*/
            __pyx_t_4 = NULL;

            /* "adbc_driver_manager/_lib.pyx":338
 *             if self._open_children != 0:
 *                 raise RuntimeError(
 *                     f"Cannot close {self.__class__.__name__} "             # <<<<<<<<<<<<<<
 *                     f"with open {self._child_type}")
 * 
*/
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 338, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 338, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_10, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 338, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "adbc_driver_manager/_lib.pyx":339
 *                 raise RuntimeError(
 *                     f"Cannot close {self.__class__.__name__} "
 *                     f"with open {self._child_type}")             # <<<<<<<<<<<<<<
 * 
 * 
*/
            __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_v_self->_child_type); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 339, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_11[0] = __pyx_mstate_global->__pyx_kp_u_Cannot_close;
            __pyx_t_11[1] = __pyx_t_3;
            __pyx_t_11[2] = __pyx_mstate_global->__pyx_kp_u_with_open;
            __pyx_t_11[3] = __pyx_t_10;

            /* "adbc_driver_manager/_lib.pyx":338
 *             if self._open_children != 0:
 *                 raise RuntimeError(
 *                     f"Cannot close {self.__class__.__name__} "             # <<<<<<<<<<<<<<
 *                     f"with open {self._child_type}")
 * 
*/
            __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_11, 4, 13 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10));
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 338, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_5 = 1;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_12};
              __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
            }
            __Pyx_Raise(__pyx_t_2, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_ERR(0, 337, __pyx_L7_error)

            /* "adbc_driver_manager/_lib.pyx":336
 *     cdef _check_open_children(self):
 *         with self._lock:
 *             if self._open_children != 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Cannot close {self.__class__.__name__} "
*/
          }

          /* "adbc_driver_manager/_lib.pyx":335
 * 
 *     cdef _check_open_children(self):
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self._open_children != 0:
 *                 raise RuntimeError(
*/
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle._check_open_children", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_12, &__pyx_t_4) < 0) __PYX_ERR(0, 335, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_4);
          __pyx_t_10 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_12, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 335, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 335, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (__pyx_t_9 < (0)) __PYX_ERR(0, 335, __pyx_L9_except_error)
          __pyx_t_14 = (!__pyx_t_9);
          if (unlikely(__pyx_t_14)) {
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_4);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_12, __pyx_t_4);
            __pyx_t_2 = 0;  __pyx_t_12 = 0;  __pyx_t_4 = 0; 
            __PYX_ERR(0, 335, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_1) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 335, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L17;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L1_error;
    __pyx_L17:;
  }

  /* "adbc_driver_manager/_lib.pyx":334
 *             self._open_children -= 1
 * 
 *     cdef _check_open_children(self):             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             if self._open_children != 0:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle._check_open_children", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_6__reduce_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_6__reduce_cython__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._child_type, self._lock, self._open_children)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(__pyx_v_self->_open_children); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_child_type);
  __Pyx_GIVEREF(__pyx_v_self->_child_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_child_type) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_lock);
  __Pyx_GIVEREF(__pyx_v_self->_lock);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->_lock) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._child_type, self._lock, self._open_children)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._child_type, self._lock, self._open_children)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._child_type is not None or self._lock is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._child_type, self._lock, self._open_children)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._child_type is not None or self._lock is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, None), state
*/
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->_child_type != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->_lock != Py_None);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._child_type is not None or self._lock is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._child_type is not None or self._lock is not None
 *     if use_setstate:
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__AdbcHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_86393024);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_86393024);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_86393024) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._child_type is not None or self._lock is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, None), state
 *     else:
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__AdbcHandle__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__AdbcHandle); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_86393024);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_86393024);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_86393024) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__AdbcHandle__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_8__setstate_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_11_AdbcHandle_8__setstate_cython__(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__AdbcHandle__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib___pyx_unpickle__AdbcHandle__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__AdbcHandle__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib._AdbcHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":342
 * 
 * 
 * def is_pycapsule(obj, bytes name) -> bool:             # <<<<<<<<<<<<<<
 *     """Check if an object is a PyCapsule of a specific type."""
 *     # Taken from nanoarrow
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_3is_pycapsule(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_2is_pycapsule, "Check if an object is a PyCapsule of a specific type.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_3is_pycapsule = {"is_pycapsule", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_3is_pycapsule, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_2is_pycapsule};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_3is_pycapsule(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_pycapsule (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_obj,&__pyx_mstate_global->__pyx_n_u_name_2,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 342, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 342, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 342, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "is_pycapsule", 0) < (0)) __PYX_ERR(0, 342, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("is_pycapsule", 1, 2, 2, i); __PYX_ERR(0, 342, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 342, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 342, __pyx_L3_error)
    }
    __pyx_v_obj = values[0];
    __pyx_v_name = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_pycapsule", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 342, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.is_pycapsule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyBytes_Type), 1, "name", 1))) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_2is_pycapsule(__pyx_self, __pyx_v_obj, __pyx_v_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_2is_pycapsule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_pycapsule", 0);

  /* "adbc_driver_manager/_lib.pyx":345
 *     """Check if an object is a PyCapsule of a specific type."""
 *     # Taken from nanoarrow
 *     return PyCapsule_IsValid(obj, name) == 1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 345, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_AsString(__pyx_v_name); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong((PyCapsule_IsValid(__pyx_v_obj, __pyx_t_1) == 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":342
 * 
 * 
 * def is_pycapsule(obj, bytes name) -> bool:             # <<<<<<<<<<<<<<
 *     """Check if an object is a PyCapsule of a specific type."""
 *     # Taken from nanoarrow
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.is_pycapsule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":348
 * 
 * 
 * cdef void pycapsule_schema_deleter(object capsule) noexcept:             # <<<<<<<<<<<<<<
 *     cdef CArrowSchema* allocated = <CArrowSchema*>PyCapsule_GetPointer(
 *         capsule, "arrow_schema"
*/

static void __pyx_f_19adbc_driver_manager_4_lib_pycapsule_schema_deleter(PyObject *__pyx_v_capsule) {
  struct ArrowSchema *__pyx_v_allocated;
  void *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "adbc_driver_manager/_lib.pyx":349
 * 
 * cdef void pycapsule_schema_deleter(object capsule) noexcept:
 *     cdef CArrowSchema* allocated = <CArrowSchema*>PyCapsule_GetPointer(             # <<<<<<<<<<<<<<
 *         capsule, "arrow_schema"
 *     )
*/
  __pyx_t_1 = PyCapsule_GetPointer(__pyx_v_capsule, __pyx_k_arrow_schema); if (unlikely(__pyx_t_1 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 349, __pyx_L1_error)
  __pyx_v_allocated = ((struct ArrowSchema *)__pyx_t_1);

  /* "adbc_driver_manager/_lib.pyx":352
 *         capsule, "arrow_schema"
 *     )
 *     if allocated.release != NULL:             # <<<<<<<<<<<<<<
 *         allocated.release(allocated)
 *     free(allocated)
*/
  __pyx_t_2 = (__pyx_v_allocated->release != NULL);
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":353
 *     )
 *     if allocated.release != NULL:
 *         allocated.release(allocated)             # <<<<<<<<<<<<<<
 *     free(allocated)
 * 
*/
    __pyx_v_allocated->release(__pyx_v_allocated);

    /* "adbc_driver_manager/_lib.pyx":352
 *         capsule, "arrow_schema"
 *     )
 *     if allocated.release != NULL:             # <<<<<<<<<<<<<<
 *         allocated.release(allocated)
 *     free(allocated)
*/
  }

  /* "adbc_driver_manager/_lib.pyx":354
 *     if allocated.release != NULL:
 *         allocated.release(allocated)
 *     free(allocated)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  free(__pyx_v_allocated);

  /* "adbc_driver_manager/_lib.pyx":348
 * 
 * 
 * cdef void pycapsule_schema_deleter(object capsule) noexcept:             # <<<<<<<<<<<<<<
 *     cdef CArrowSchema* allocated = <CArrowSchema*>PyCapsule_GetPointer(
 *         capsule, "arrow_schema"
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("adbc_driver_manager._lib.pycapsule_schema_deleter", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
}

/* "adbc_driver_manager/_lib.pyx":357
 * 
 * 
 * cdef void pycapsule_stream_deleter(object capsule) noexcept:             # <<<<<<<<<<<<<<
 *     cdef CArrowArrayStream* allocated = <CArrowArrayStream*> PyCapsule_GetPointer(
 *         capsule, "arrow_array_stream"
*/

static void __pyx_f_19adbc_driver_manager_4_lib_pycapsule_stream_deleter(PyObject *__pyx_v_capsule) {
  struct ArrowArrayStream *__pyx_v_allocated;
  void *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "adbc_driver_manager/_lib.pyx":358
 * 
 * cdef void pycapsule_stream_deleter(object capsule) noexcept:
 *     cdef CArrowArrayStream* allocated = <CArrowArrayStream*> PyCapsule_GetPointer(             # <<<<<<<<<<<<<<
 *         capsule, "arrow_array_stream"
 *     )
*/
  __pyx_t_1 = PyCapsule_GetPointer(__pyx_v_capsule, __pyx_k_arrow_array_stream); if (unlikely(__pyx_t_1 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
  __pyx_v_allocated = ((struct ArrowArrayStream *)__pyx_t_1);

  /* "adbc_driver_manager/_lib.pyx":361
 *         capsule, "arrow_array_stream"
 *     )
 *     if allocated.release != NULL:             # <<<<<<<<<<<<<<
 *         allocated.release(allocated)
 *     free(allocated)
*/
  __pyx_t_2 = (__pyx_v_allocated->release != NULL);
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":362
 *     )
 *     if allocated.release != NULL:
 *         allocated.release(allocated)             # <<<<<<<<<<<<<<
 *     free(allocated)
 * 
*/
    __pyx_v_allocated->release(__pyx_v_allocated);

    /* "adbc_driver_manager/_lib.pyx":361
 *         capsule, "arrow_array_stream"
 *     )
 *     if allocated.release != NULL:             # <<<<<<<<<<<<<<
 *         allocated.release(allocated)
 *     free(allocated)
*/
  }

  /* "adbc_driver_manager/_lib.pyx":363
 *     if allocated.release != NULL:
 *         allocated.release(allocated)
 *     free(allocated)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  free(__pyx_v_allocated);

  /* "adbc_driver_manager/_lib.pyx":357
 * 
 * 
 * cdef void pycapsule_stream_deleter(object capsule) noexcept:             # <<<<<<<<<<<<<<
 *     cdef CArrowArrayStream* allocated = <CArrowArrayStream*> PyCapsule_GetPointer(
 *         capsule, "arrow_array_stream"
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("adbc_driver_manager._lib.pycapsule_stream_deleter", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
}

/* "adbc_driver_manager/_lib.pyx":375
 *         CArrowSchema schema
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         self.release()
 * 
*/

/* Python wrapper */
static void __pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_1__del__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_1__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle___del__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle___del__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__del__", 0);

  /* "adbc_driver_manager/_lib.pyx":376
 * 
 *     def __del__(self):
 *         self.release()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_release, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":375
 *         CArrowSchema schema
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         self.release()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("adbc_driver_manager._lib.ArrowSchemaHandle.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "adbc_driver_manager/_lib.pyx":378
 *         self.release()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def address(self) -> int:
 *         """The address of the ArrowSchema."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7address_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7address_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7address___get__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7address___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "adbc_driver_manager/_lib.pyx":381
 *     def address(self) -> int:
 *         """The address of the ArrowSchema."""
 *         return <uintptr_t> &self.schema             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(((uintptr_t)(&__pyx_v_self->schema))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":378
 *         self.release()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def address(self) -> int:
 *         """The address of the ArrowSchema."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowSchemaHandle.address.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":383
 *         return <uintptr_t> &self.schema
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8is_valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8is_valid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8is_valid___get__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8is_valid___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "adbc_driver_manager/_lib.pyx":386
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
 *         return self.schema.release != NULL             # <<<<<<<<<<<<<<
 * 
 *     def release(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->schema.release != NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":383
 *         return <uintptr_t> &self.schema
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowSchemaHandle.is_valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":388
 *         return self.schema.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this schema without having to import it.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_3release(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_2release, "Release this schema without having to import it.\n\n        No-op if already released (if ``not self.is_valid``).\n\n        Postcondition: ``not self.is_valid``.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_3release = {"release", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_3release, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_2release};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_3release(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("release (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("release", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("release", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_2release(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_2release(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("release", 0);

  /* "adbc_driver_manager/_lib.pyx":395
 *         Postcondition: ``not self.is_valid``.
 *         """
 *         if self.schema.release != NULL:             # <<<<<<<<<<<<<<
 *             self.schema.release(&self.schema)
 *             self.schema.release = NULL
*/
  __pyx_t_1 = (__pyx_v_self->schema.release != NULL);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":396
 *         """
 *         if self.schema.release != NULL:
 *             self.schema.release(&self.schema)             # <<<<<<<<<<<<<<
 *             self.schema.release = NULL
 * 
*/
    __pyx_v_self->schema.release((&__pyx_v_self->schema));

    /* "adbc_driver_manager/_lib.pyx":397
 *         if self.schema.release != NULL:
 *             self.schema.release(&self.schema)
 *             self.schema.release = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __arrow_c_schema__(self) -> object:
*/
    __pyx_v_self->schema.release = NULL;

    /* "adbc_driver_manager/_lib.pyx":395
 *         Postcondition: ``not self.is_valid``.
 *         """
 *         if self.schema.release != NULL:             # <<<<<<<<<<<<<<
 *             self.schema.release(&self.schema)
 *             self.schema.release = NULL
*/
  }

  /* "adbc_driver_manager/_lib.pyx":388
 *         return self.schema.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this schema without having to import it.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":399
 *             self.schema.release = NULL
 * 
 *     def __arrow_c_schema__(self) -> object:             # <<<<<<<<<<<<<<
 *         """Consume this object to get a PyCapsule."""
 *         # Reference:
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_5__arrow_c_schema__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_4__arrow_c_schema__, "Consume this object to get a PyCapsule.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_5__arrow_c_schema__ = {"__arrow_c_schema__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_5__arrow_c_schema__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_4__arrow_c_schema__};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_5__arrow_c_schema__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__arrow_c_schema__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__arrow_c_schema__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__arrow_c_schema__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_4__arrow_c_schema__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_4__arrow_c_schema__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self) {
  struct ArrowSchema *__pyx_v_allocated;
  PyObject *__pyx_v_capsule = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__arrow_c_schema__", 0);

  /* "adbc_driver_manager/_lib.pyx":403
 *         # Reference:
 *         # https://arrow.apache.org/docs/dev/format/CDataInterface/PyCapsuleInterface.html#create-a-pycapsule
 *         cdef CArrowSchema* allocated = <CArrowSchema*> malloc(sizeof(CArrowSchema))             # <<<<<<<<<<<<<<
 *         allocated.release = NULL
 *         capsule = PyCapsule_New(
*/
  __pyx_v_allocated = ((struct ArrowSchema *)malloc((sizeof(struct ArrowSchema))));

  /* "adbc_driver_manager/_lib.pyx":404
 *         # https://arrow.apache.org/docs/dev/format/CDataInterface/PyCapsuleInterface.html#create-a-pycapsule
 *         cdef CArrowSchema* allocated = <CArrowSchema*> malloc(sizeof(CArrowSchema))
 *         allocated.release = NULL             # <<<<<<<<<<<<<<
 *         capsule = PyCapsule_New(
 *             <void*>allocated, "arrow_schema", &pycapsule_schema_deleter,
*/
  __pyx_v_allocated->release = NULL;

  /* "adbc_driver_manager/_lib.pyx":405
 *         cdef CArrowSchema* allocated = <CArrowSchema*> malloc(sizeof(CArrowSchema))
 *         allocated.release = NULL
 *         capsule = PyCapsule_New(             # <<<<<<<<<<<<<<
 *             <void*>allocated, "arrow_schema", &pycapsule_schema_deleter,
 *         )
*/
  __pyx_t_1 = PyCapsule_New(((void *)__pyx_v_allocated), __pyx_k_arrow_schema, (&__pyx_f_19adbc_driver_manager_4_lib_pycapsule_schema_deleter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_capsule = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":408
 *             <void*>allocated, "arrow_schema", &pycapsule_schema_deleter,
 *         )
 *         memcpy(allocated, &self.schema, sizeof(CArrowSchema))             # <<<<<<<<<<<<<<
 *         self.schema.release = NULL
 *         return capsule
*/
  (void)(memcpy(__pyx_v_allocated, (&__pyx_v_self->schema), (sizeof(struct ArrowSchema))));

  /* "adbc_driver_manager/_lib.pyx":409
 *         )
 *         memcpy(allocated, &self.schema, sizeof(CArrowSchema))
 *         self.schema.release = NULL             # <<<<<<<<<<<<<<
 *         return capsule
 * 
*/
  __pyx_v_self->schema.release = NULL;

  /* "adbc_driver_manager/_lib.pyx":410
 *         memcpy(allocated, &self.schema, sizeof(CArrowSchema))
 *         self.schema.release = NULL
 *         return capsule             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_capsule);
  __pyx_r = __pyx_v_capsule;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":399
 *             self.schema.release = NULL
 * 
 *     def __arrow_c_schema__(self) -> object:             # <<<<<<<<<<<<<<
 *         """Consume this object to get a PyCapsule."""
 *         # Reference:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowSchemaHandle.__arrow_c_schema__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_capsule);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_6__reduce_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_schema_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowSchemaHandle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowSchemaHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8__setstate_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_schema_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowSchemaHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":422
 *         CArrowArray array
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         self.release()
 * 
*/

/* Python wrapper */
static void __pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_1__del__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_1__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle___del__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle___del__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__del__", 0);

  /* "adbc_driver_manager/_lib.pyx":423
 * 
 *     def __del__(self):
 *         self.release()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_release, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":422
 *         CArrowArray array
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         self.release()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("adbc_driver_manager._lib.ArrowArrayHandle.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "adbc_driver_manager/_lib.pyx":425
 *         self.release()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_8is_valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_8is_valid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_8is_valid___get__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_8is_valid___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "adbc_driver_manager/_lib.pyx":428
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
 *         return self.array.release != NULL             # <<<<<<<<<<<<<<
 * 
 *     def release(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->array.release != NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":425
 *         self.release()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayHandle.is_valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":430
 *         return self.array.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this array without having to import it.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_3release(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_16ArrowArrayHandle_2release, "Release this array without having to import it.\n\n        No-op if already released (if ``not self.is_valid``).\n\n        Postcondition: ``not self.is_valid``.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_16ArrowArrayHandle_3release = {"release", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_3release, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_16ArrowArrayHandle_2release};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_3release(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("release (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("release", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("release", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_2release(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_2release(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("release", 0);

  /* "adbc_driver_manager/_lib.pyx":437
 *         Postcondition: ``not self.is_valid``.
 *         """
 *         if self.array.release != NULL:             # <<<<<<<<<<<<<<
 *             self.array.release(&self.array)
 *             self.array.release = NULL
*/
  __pyx_t_1 = (__pyx_v_self->array.release != NULL);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":438
 *         """
 *         if self.array.release != NULL:
 *             self.array.release(&self.array)             # <<<<<<<<<<<<<<
 *             self.array.release = NULL
 * 
*/
    __pyx_v_self->array.release((&__pyx_v_self->array));

    /* "adbc_driver_manager/_lib.pyx":439
 *         if self.array.release != NULL:
 *             self.array.release(&self.array)
 *             self.array.release = NULL             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
    __pyx_v_self->array.release = NULL;

    /* "adbc_driver_manager/_lib.pyx":437
 *         Postcondition: ``not self.is_valid``.
 *         """
 *         if self.array.release != NULL:             # <<<<<<<<<<<<<<
 *             self.array.release(&self.array)
 *             self.array.release = NULL
*/
  }

  /* "adbc_driver_manager/_lib.pyx":430
 *         return self.array.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this array without having to import it.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":441
 *             self.array.release = NULL
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def address(self) -> int:
 *         """
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7address_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7address_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7address___get__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7address___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "adbc_driver_manager/_lib.pyx":446
 *         The address of the ArrowArray.
 *         """
 *         return <uintptr_t> &self.array             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(((uintptr_t)(&__pyx_v_self->array))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":441
 *             self.array.release = NULL
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def address(self) -> int:
 *         """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayHandle.address.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_16ArrowArrayHandle_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_4__reduce_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_array_cannot_be_converted_t, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayHandle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_6__setstate_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_16ArrowArrayHandle_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_array_cannot_be_converted_t, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":458
 *         CArrowArrayStream stream
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         self.release()
 * 
*/

/* Python wrapper */
static void __pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_1__del__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_1__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle___del__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle___del__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__del__", 0);

  /* "adbc_driver_manager/_lib.pyx":459
 * 
 *     def __del__(self):
 *         self.release()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_release, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":458
 *         CArrowArrayStream stream
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         self.release()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("adbc_driver_manager._lib.ArrowArrayStreamHandle.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "adbc_driver_manager/_lib.pyx":461
 *         self.release()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def address(self) -> int:
 *         """The address of the ArrowArrayStream."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7address_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7address_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7address___get__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7address___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "adbc_driver_manager/_lib.pyx":464
 *     def address(self) -> int:
 *         """The address of the ArrowArrayStream."""
 *         return <uintptr_t> &self.stream             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(((uintptr_t)(&__pyx_v_self->stream))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":461
 *         self.release()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def address(self) -> int:
 *         """The address of the ArrowArrayStream."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.address.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":466
 *         return <uintptr_t> &self.stream
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8is_valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8is_valid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8is_valid___get__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8is_valid___get__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "adbc_driver_manager/_lib.pyx":469
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
 *         return self.stream.release != NULL             # <<<<<<<<<<<<<<
 * 
 *     def release(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->stream.release != NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":466
 *         return <uintptr_t> &self.stream
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_valid(self) -> bool:
 *         """Check validility (object has non-NULL release pointer)."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.is_valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":471
 *         return self.stream.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this stream without having to import it.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_3release(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_2release, "Release this stream without having to import it.\n\n        No-op if already released (if ``not self.is_valid``).\n\n        Postcondition: ``not self.is_valid``.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_3release = {"release", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_3release, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_2release};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_3release(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("release (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("release", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("release", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_2release(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_2release(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("release", 0);

  /* "adbc_driver_manager/_lib.pyx":478
 *         Postcondition: ``not self.is_valid``.
 *         """
 *         if self.stream.release != NULL:             # <<<<<<<<<<<<<<
 *             self.stream.release(&self.stream)
 *             self.stream.release = NULL
*/
  __pyx_t_1 = (__pyx_v_self->stream.release != NULL);
  if (__pyx_t_1) {

    /* "adbc_driver_manager/_lib.pyx":479
 *         """
 *         if self.stream.release != NULL:
 *             self.stream.release(&self.stream)             # <<<<<<<<<<<<<<
 *             self.stream.release = NULL
 * 
*/
    __pyx_v_self->stream.release((&__pyx_v_self->stream));

    /* "adbc_driver_manager/_lib.pyx":480
 *         if self.stream.release != NULL:
 *             self.stream.release(&self.stream)
 *             self.stream.release = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __arrow_c_schema__(self) -> object:
*/
    __pyx_v_self->stream.release = NULL;

    /* "adbc_driver_manager/_lib.pyx":478
 *         Postcondition: ``not self.is_valid``.
 *         """
 *         if self.stream.release != NULL:             # <<<<<<<<<<<<<<
 *             self.stream.release(&self.stream)
 *             self.stream.release = NULL
*/
  }

  /* "adbc_driver_manager/_lib.pyx":471
 *         return self.stream.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this stream without having to import it.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":482
 *             self.stream.release = NULL
 * 
 *     def __arrow_c_schema__(self) -> object:             # <<<<<<<<<<<<<<
 *         """Get a PyCapsule without consuming this object."""
 *         cdef const char* err = NULL
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_5__arrow_c_schema__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_4__arrow_c_schema__, "Get a PyCapsule without consuming this object.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_5__arrow_c_schema__ = {"__arrow_c_schema__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_5__arrow_c_schema__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_4__arrow_c_schema__};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_5__arrow_c_schema__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__arrow_c_schema__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__arrow_c_schema__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__arrow_c_schema__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_4__arrow_c_schema__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_4__arrow_c_schema__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self) {
  char const *__pyx_v_err;
  struct ArrowSchema *__pyx_v_allocated;
  PyObject *__pyx_v_capsule = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7[3];
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10[4];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__arrow_c_schema__", 0);

  /* "adbc_driver_manager/_lib.pyx":484
 *     def __arrow_c_schema__(self) -> object:
 *         """Get a PyCapsule without consuming this object."""
 *         cdef const char* err = NULL             # <<<<<<<<<<<<<<
 * 
 *         if not self.is_valid:
*/
  __pyx_v_err = NULL;

  /* "adbc_driver_manager/_lib.pyx":486
 *         cdef const char* err = NULL
 * 
 *         if not self.is_valid:             # <<<<<<<<<<<<<<
 *             raise ValueError("ArrowArrayStreamHandle already consumed")
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_is_valid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (unlikely(__pyx_t_3)) {

    /* "adbc_driver_manager/_lib.pyx":487
 * 
 *         if not self.is_valid:
 *             raise ValueError("ArrowArrayStreamHandle already consumed")             # <<<<<<<<<<<<<<
 * 
 *         cdef CArrowSchema* allocated = <CArrowSchema*> malloc(sizeof(CArrowSchema))
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_ArrowArrayStreamHandle_already_c};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 487, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":486
 *         cdef const char* err = NULL
 * 
 *         if not self.is_valid:             # <<<<<<<<<<<<<<
 *             raise ValueError("ArrowArrayStreamHandle already consumed")
 * 
*/
  }

  /* "adbc_driver_manager/_lib.pyx":489
 *             raise ValueError("ArrowArrayStreamHandle already consumed")
 * 
 *         cdef CArrowSchema* allocated = <CArrowSchema*> malloc(sizeof(CArrowSchema))             # <<<<<<<<<<<<<<
 *         allocated.release = NULL
 *         capsule = PyCapsule_New(
*/
  __pyx_v_allocated = ((struct ArrowSchema *)malloc((sizeof(struct ArrowSchema))));

  /* "adbc_driver_manager/_lib.pyx":490
 * 
 *         cdef CArrowSchema* allocated = <CArrowSchema*> malloc(sizeof(CArrowSchema))
 *         allocated.release = NULL             # <<<<<<<<<<<<<<
 *         capsule = PyCapsule_New(
 *             <void*>allocated, "arrow_schema", &pycapsule_schema_deleter,
*/
  __pyx_v_allocated->release = NULL;

  /* "adbc_driver_manager/_lib.pyx":491
 *         cdef CArrowSchema* allocated = <CArrowSchema*> malloc(sizeof(CArrowSchema))
 *         allocated.release = NULL
 *         capsule = PyCapsule_New(             # <<<<<<<<<<<<<<
 *             <void*>allocated, "arrow_schema", &pycapsule_schema_deleter,
 *         )
*/
  __pyx_t_1 = PyCapsule_New(((void *)__pyx_v_allocated), __pyx_k_arrow_schema, (&__pyx_f_19adbc_driver_manager_4_lib_pycapsule_schema_deleter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_capsule = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":494
 *             <void*>allocated, "arrow_schema", &pycapsule_schema_deleter,
 *         )
 *         rc = self.stream.get_schema(&self.stream, allocated)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             err = self.stream.get_last_error(&self.stream)
*/
  __pyx_v_rc = __pyx_v_self->stream.get_schema((&__pyx_v_self->stream), __pyx_v_allocated);

  /* "adbc_driver_manager/_lib.pyx":495
 *         )
 *         rc = self.stream.get_schema(&self.stream, allocated)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             err = self.stream.get_last_error(&self.stream)
 *             if err == NULL:
*/
  __pyx_t_3 = (__pyx_v_rc != 0);
  if (__pyx_t_3) {

    /* "adbc_driver_manager/_lib.pyx":496
 *         rc = self.stream.get_schema(&self.stream, allocated)
 *         if rc != 0:
 *             err = self.stream.get_last_error(&self.stream)             # <<<<<<<<<<<<<<
 *             if err == NULL:
 *                 raise RuntimeError(f"Failed to get schema: ({rc})")
*/
    __pyx_v_err = __pyx_v_self->stream.get_last_error((&__pyx_v_self->stream));

    /* "adbc_driver_manager/_lib.pyx":497
 *         if rc != 0:
 *             err = self.stream.get_last_error(&self.stream)
 *             if err == NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(f"Failed to get schema: ({rc})")
 *             else:
*/
    __pyx_t_3 = (__pyx_v_err == NULL);
    if (unlikely(__pyx_t_3)) {

      /* "adbc_driver_manager/_lib.pyx":498
 *             err = self.stream.get_last_error(&self.stream)
 *             if err == NULL:
 *                 raise RuntimeError(f"Failed to get schema: ({rc})")             # <<<<<<<<<<<<<<
 *             else:
 *                 s = err.decode()
*/
      __pyx_t_4 = NULL;
      __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_rc, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7[0] = __pyx_mstate_global->__pyx_kp_u_Failed_to_get_schema;
      __pyx_t_7[1] = __pyx_t_6;
      __pyx_t_7[2] = __pyx_mstate_global->__pyx_kp_u__3;
      __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_7, 3, 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 1, 127);
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 498, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":497
 *         if rc != 0:
 *             err = self.stream.get_last_error(&self.stream)
 *             if err == NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(f"Failed to get schema: ({rc})")
 *             else:
*/
    }

    /* "adbc_driver_manager/_lib.pyx":500
 *                 raise RuntimeError(f"Failed to get schema: ({rc})")
 *             else:
 *                 s = err.decode()             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(f"Failed to get schema: ({rc}) {s}")
 *         return capsule
*/
    /*else*/ {
      __pyx_t_9 = __Pyx_ssize_strlen(__pyx_v_err); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 500, __pyx_L1_error)
      __pyx_t_1 = __Pyx_decode_c_string(__pyx_v_err, 0, __pyx_t_9, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_s = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "adbc_driver_manager/_lib.pyx":501
 *             else:
 *                 s = err.decode()
 *                 raise RuntimeError(f"Failed to get schema: ({rc}) {s}")             # <<<<<<<<<<<<<<
 *         return capsule
 * 
*/
      __pyx_t_8 = NULL;
      __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_rc, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10[0] = __pyx_mstate_global->__pyx_kp_u_Failed_to_get_schema;
      __pyx_t_10[1] = __pyx_t_4;
      __pyx_t_10[2] = __pyx_mstate_global->__pyx_kp_u__4;
      __pyx_t_10[3] = __pyx_t_6;
      __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 4, 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6));
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 501, __pyx_L1_error)
    }

    /* "adbc_driver_manager/_lib.pyx":495
 *         )
 *         rc = self.stream.get_schema(&self.stream, allocated)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             err = self.stream.get_last_error(&self.stream)
 *             if err == NULL:
*/
  }

  /* "adbc_driver_manager/_lib.pyx":502
 *                 s = err.decode()
 *                 raise RuntimeError(f"Failed to get schema: ({rc}) {s}")
 *         return capsule             # <<<<<<<<<<<<<<
 * 
 *     def __arrow_c_stream__(self, requested_schema=None) -> object:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_capsule);
  __pyx_r = __pyx_v_capsule;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":482
 *             self.stream.release = NULL
 * 
 *     def __arrow_c_schema__(self) -> object:             # <<<<<<<<<<<<<<
 *         """Get a PyCapsule without consuming this object."""
 *         cdef const char* err = NULL
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.__arrow_c_schema__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_capsule);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":504
 *         return capsule
 * 
 *     def __arrow_c_stream__(self, requested_schema=None) -> object:             # <<<<<<<<<<<<<<
 *         """Consume this object to get a PyCapsule."""
 *         if requested_schema is not None:
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7__arrow_c_stream__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_6__arrow_c_stream__, "Consume this object to get a PyCapsule.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7__arrow_c_stream__ = {"__arrow_c_stream__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7__arrow_c_stream__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_6__arrow_c_stream__};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7__arrow_c_stream__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_requested_schema = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__arrow_c_stream__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_requested_schema,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 504, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 504, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__arrow_c_stream__", 0) < (0)) __PYX_ERR(0, 504, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 504, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_requested_schema = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__arrow_c_stream__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 504, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.__arrow_c_stream__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_6__arrow_c_stream__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self), __pyx_v_requested_schema);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_6__arrow_c_stream__(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self, PyObject *__pyx_v_requested_schema) {
  struct ArrowArrayStream *__pyx_v_allocated;
  PyObject *__pyx_v_capsule = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__arrow_c_stream__", 0);

  /* "adbc_driver_manager/_lib.pyx":506
 *     def __arrow_c_stream__(self, requested_schema=None) -> object:
 *         """Consume this object to get a PyCapsule."""
 *         if requested_schema is not None:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError("requested_schema")
 * 
*/
  __pyx_t_1 = (__pyx_v_requested_schema != Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "adbc_driver_manager/_lib.pyx":507
 *         """Consume this object to get a PyCapsule."""
 *         if requested_schema is not None:
 *             raise NotImplementedError("requested_schema")             # <<<<<<<<<<<<<<
 * 
 *         cdef CArrowArrayStream* allocated = \
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_n_u_requested_schema};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_NotImplementedError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 507, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 507, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":506
 *     def __arrow_c_stream__(self, requested_schema=None) -> object:
 *         """Consume this object to get a PyCapsule."""
 *         if requested_schema is not None:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError("requested_schema")
 * 
*/
  }

  /* "adbc_driver_manager/_lib.pyx":510
 * 
 *         cdef CArrowArrayStream* allocated = \
 *             <CArrowArrayStream*> malloc(sizeof(CArrowArrayStream))             # <<<<<<<<<<<<<<
 *         allocated.release = NULL
 *         capsule = PyCapsule_New(
*/
  __pyx_v_allocated = ((struct ArrowArrayStream *)malloc((sizeof(struct ArrowArrayStream))));

  /* "adbc_driver_manager/_lib.pyx":511
 *         cdef CArrowArrayStream* allocated = \
 *             <CArrowArrayStream*> malloc(sizeof(CArrowArrayStream))
 *         allocated.release = NULL             # <<<<<<<<<<<<<<
 *         capsule = PyCapsule_New(
 *             <void*>allocated, "arrow_array_stream", &pycapsule_stream_deleter,
*/
  __pyx_v_allocated->release = NULL;

  /* "adbc_driver_manager/_lib.pyx":512
 *             <CArrowArrayStream*> malloc(sizeof(CArrowArrayStream))
 *         allocated.release = NULL
 *         capsule = PyCapsule_New(             # <<<<<<<<<<<<<<
 *             <void*>allocated, "arrow_array_stream", &pycapsule_stream_deleter,
 *         )
*/
  __pyx_t_2 = PyCapsule_New(((void *)__pyx_v_allocated), __pyx_k_arrow_array_stream, (&__pyx_f_19adbc_driver_manager_4_lib_pycapsule_stream_deleter)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_capsule = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":515
 *             <void*>allocated, "arrow_array_stream", &pycapsule_stream_deleter,
 *         )
 *         memcpy(allocated, &self.stream, sizeof(CArrowArrayStream))             # <<<<<<<<<<<<<<
 *         self.stream.release = NULL
 *         return capsule
*/
  (void)(memcpy(__pyx_v_allocated, (&__pyx_v_self->stream), (sizeof(struct ArrowArrayStream))));

  /* "adbc_driver_manager/_lib.pyx":516
 *         )
 *         memcpy(allocated, &self.stream, sizeof(CArrowArrayStream))
 *         self.stream.release = NULL             # <<<<<<<<<<<<<<
 *         return capsule
 * 
*/
  __pyx_v_self->stream.release = NULL;

  /* "adbc_driver_manager/_lib.pyx":517
 *         memcpy(allocated, &self.stream, sizeof(CArrowArrayStream))
 *         self.stream.release = NULL
 *         return capsule             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_capsule);
  __pyx_r = __pyx_v_capsule;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":504
 *         return capsule
 * 
 *     def __arrow_c_stream__(self, requested_schema=None) -> object:             # <<<<<<<<<<<<<<
 *         """Consume this object to get a PyCapsule."""
 *         if requested_schema is not None:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.__arrow_c_stream__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_capsule);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8__reduce_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_stream_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_10__setstate_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_stream_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.ArrowArrayStreamHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":551
 *         CAdbcDatabase database
 * 
 *     def __init__(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         super().__init__("AdbcConnection")
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static int __pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("__init__", __pyx_kwds) == -1)) return -1;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return -1;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return -1;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase___init__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase___init__(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  char const *__pyx_v_c_key;
  char const *__pyx_v_c_value;
  int64_t __pyx_v_c_value_int;
  double __pyx_v_c_value_double;
  size_t __pyx_v_c_value_len;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_manifest_path = NULL;
  PyObject *__pyx_v_profile_path = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  struct AdbcError __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  uintptr_t __pyx_t_12;
  AdbcLoadFlags __pyx_t_13;
  PyObject *__pyx_t_14[3];
  PyObject *__pyx_t_15 = NULL;
  char const *__pyx_t_16;
  char const *__pyx_t_17;
  int64_t __pyx_t_18;
  double __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "adbc_driver_manager/_lib.pyx":552
 * 
 *     def __init__(self, **kwargs) -> None:
 *         super().__init__("AdbcConnection")             # <<<<<<<<<<<<<<
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_AdbcConnection};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":553
 *     def __init__(self, **kwargs) -> None:
 *         super().__init__("AdbcConnection")
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef const char* c_key
*/
  __pyx_t_6 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 553, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_6;

  /* "adbc_driver_manager/_lib.pyx":559
 *         cdef int64_t c_value_int
 *         cdef double c_value_double
 *         cdef size_t c_value_len = 0             # <<<<<<<<<<<<<<
 *         memset(&self.database, 0, cython.sizeof(CAdbcDatabase))
 * 
*/
  __pyx_v_c_value_len = 0;

  /* "adbc_driver_manager/_lib.pyx":560
 *         cdef double c_value_double
 *         cdef size_t c_value_len = 0
 *         memset(&self.database, 0, cython.sizeof(CAdbcDatabase))             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  (void)(memset((&__pyx_v_self->database), 0, (sizeof(struct AdbcDatabase))));

  /* "adbc_driver_manager/_lib.pyx":562
 *         memset(&self.database, 0, cython.sizeof(CAdbcDatabase))
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseNew(&self.database, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":563
 * 
 *         with nogil:
 *             status = AdbcDatabaseNew(&self.database, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcDatabaseNew((&__pyx_v_self->database), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":562
 *         memset(&self.database, 0, cython.sizeof(CAdbcDatabase))
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseNew(&self.database, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":564
 *         with nogil:
 *             status = AdbcDatabaseNew(&self.database, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         # N.B. these calls are handled by the driver manager and do not do
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":571
 * 
 *         # by default, allow searching for manifests and relative paths
 *         status = AdbcDriverManagerDatabaseSetLoadFlags(             # <<<<<<<<<<<<<<
 *             &self.database, CAdbcLoadFlagDefault, &c_error)
 *         check_error(status, &c_error)
*/
  __pyx_v_status = AdbcDriverManagerDatabaseSetLoadFlags((&__pyx_v_self->database), ADBC_LOAD_FLAG_DEFAULT, (&__pyx_v_c_error));

  /* "adbc_driver_manager/_lib.pyx":573
 *         status = AdbcDriverManagerDatabaseSetLoadFlags(
 *             &self.database, CAdbcLoadFlagDefault, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         for key, value in kwargs.items():
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 573, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":575
 *         check_error(status, &c_error)
 * 
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             if key == "init_func":
 *                 status = AdbcDriverManagerDatabaseSetInitFunc(
*/
  __pyx_t_7 = 0;
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_8), (&__pyx_t_9)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_8, &__pyx_t_7, &__pyx_t_3, &__pyx_t_2, NULL, __pyx_t_9);
    if (unlikely(__pyx_t_10 == 0)) break;
    if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":576
 * 
 *         for key, value in kwargs.items():
 *             if key == "init_func":             # <<<<<<<<<<<<<<
 *                 status = AdbcDriverManagerDatabaseSetInitFunc(
 *                     &self.database, <CAdbcDriverInitFunc> (<uintptr_t> value), &c_error)
*/
    __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_init_func, Py_EQ)); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 576, __pyx_L1_error)
    if (__pyx_t_11) {

      /* "adbc_driver_manager/_lib.pyx":578
 *             if key == "init_func":
 *                 status = AdbcDriverManagerDatabaseSetInitFunc(
 *                     &self.database, <CAdbcDriverInitFunc> (<uintptr_t> value), &c_error)             # <<<<<<<<<<<<<<
 *             elif key == "load_flags":
 *                 status = AdbcDriverManagerDatabaseSetLoadFlags(
*/
      __pyx_t_12 = __Pyx_PyLong_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_12 == ((uintptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 578, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":577
 *         for key, value in kwargs.items():
 *             if key == "init_func":
 *                 status = AdbcDriverManagerDatabaseSetInitFunc(             # <<<<<<<<<<<<<<
 *                     &self.database, <CAdbcDriverInitFunc> (<uintptr_t> value), &c_error)
 *             elif key == "load_flags":
*/
      __pyx_v_status = AdbcDriverManagerDatabaseSetInitFunc((&__pyx_v_self->database), ((AdbcDriverInitFunc)((uintptr_t)__pyx_t_12)), (&__pyx_v_c_error));

      /* "adbc_driver_manager/_lib.pyx":576
 * 
 *         for key, value in kwargs.items():
 *             if key == "init_func":             # <<<<<<<<<<<<<<
 *                 status = AdbcDriverManagerDatabaseSetInitFunc(
 *                     &self.database, <CAdbcDriverInitFunc> (<uintptr_t> value), &c_error)
*/
      goto __pyx_L8;
    }

    /* "adbc_driver_manager/_lib.pyx":579
 *                 status = AdbcDriverManagerDatabaseSetInitFunc(
 *                     &self.database, <CAdbcDriverInitFunc> (<uintptr_t> value), &c_error)
 *             elif key == "load_flags":             # <<<<<<<<<<<<<<
 *                 status = AdbcDriverManagerDatabaseSetLoadFlags(
 *                     &self.database, <CAdbcLoadFlags> (value), &c_error)
*/
    __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_load_flags, Py_EQ)); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 579, __pyx_L1_error)
    if (__pyx_t_11) {

      /* "adbc_driver_manager/_lib.pyx":581
 *             elif key == "load_flags":
 *                 status = AdbcDriverManagerDatabaseSetLoadFlags(
 *                     &self.database, <CAdbcLoadFlags> (value), &c_error)             # <<<<<<<<<<<<<<
 *             elif key is None:
 *                 raise ValueError("key cannot be None")
*/
      __pyx_t_13 = __Pyx_PyLong_As_uint32_t(__pyx_v_value); if (unlikely((__pyx_t_13 == ((AdbcLoadFlags)-1)) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":580
 *                     &self.database, <CAdbcDriverInitFunc> (<uintptr_t> value), &c_error)
 *             elif key == "load_flags":
 *                 status = AdbcDriverManagerDatabaseSetLoadFlags(             # <<<<<<<<<<<<<<
 *                     &self.database, <CAdbcLoadFlags> (value), &c_error)
 *             elif key is None:
*/
      __pyx_v_status = AdbcDriverManagerDatabaseSetLoadFlags((&__pyx_v_self->database), ((AdbcLoadFlags)__pyx_t_13), (&__pyx_v_c_error));

      /* "adbc_driver_manager/_lib.pyx":579
 *                 status = AdbcDriverManagerDatabaseSetInitFunc(
 *                     &self.database, <CAdbcDriverInitFunc> (<uintptr_t> value), &c_error)
 *             elif key == "load_flags":             # <<<<<<<<<<<<<<
 *                 status = AdbcDriverManagerDatabaseSetLoadFlags(
 *                     &self.database, <CAdbcLoadFlags> (value), &c_error)
*/
      goto __pyx_L8;
    }

    /* "adbc_driver_manager/_lib.pyx":582
 *                 status = AdbcDriverManagerDatabaseSetLoadFlags(
 *                     &self.database, <CAdbcLoadFlags> (value), &c_error)
 *             elif key is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError("key cannot be None")
 *             elif value is None:
*/
    __pyx_t_11 = (__pyx_v_key == Py_None);
    if (unlikely(__pyx_t_11)) {

      /* "adbc_driver_manager/_lib.pyx":583
 *                     &self.database, <CAdbcLoadFlags> (value), &c_error)
 *             elif key is None:
 *                 raise ValueError("key cannot be None")             # <<<<<<<<<<<<<<
 *             elif value is None:
 *                 raise ValueError(f"value for key '{key}' cannot be None")
*/
      __pyx_t_3 = NULL;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_key_cannot_be_None};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 583, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":582
 *                 status = AdbcDriverManagerDatabaseSetLoadFlags(
 *                     &self.database, <CAdbcLoadFlags> (value), &c_error)
 *             elif key is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError("key cannot be None")
 *             elif value is None:
*/
    }

    /* "adbc_driver_manager/_lib.pyx":584
 *             elif key is None:
 *                 raise ValueError("key cannot be None")
 *             elif value is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"value for key '{key}' cannot be None")
 *             else:
*/
    __pyx_t_11 = (__pyx_v_value == Py_None);
    if (unlikely(__pyx_t_11)) {

      /* "adbc_driver_manager/_lib.pyx":585
 *                 raise ValueError("key cannot be None")
 *             elif value is None:
 *                 raise ValueError(f"value for key '{key}' cannot be None")             # <<<<<<<<<<<<<<
 *             else:
 *                 key = _to_bytes(key, "key")
*/
      __pyx_t_3 = NULL;
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14[0] = __pyx_mstate_global->__pyx_kp_u_value_for_key;
      __pyx_t_14[1] = __pyx_t_4;
      __pyx_t_14[2] = __pyx_mstate_global->__pyx_kp_u_cannot_be_None;
      __pyx_t_15 = __Pyx_PyUnicode_Join(__pyx_t_14, 3, 15 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 16, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4));
      if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_15};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 585, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":584
 *             elif key is None:
 *                 raise ValueError("key cannot be None")
 *             elif value is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"value for key '{key}' cannot be None")
 *             else:
*/
    }

    /* "adbc_driver_manager/_lib.pyx":587
 *                 raise ValueError(f"value for key '{key}' cannot be None")
 *             else:
 *                 key = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *                 c_key = key
 * 
*/
    /*else*/ {
      __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "adbc_driver_manager/_lib.pyx":588
 *             else:
 *                 key = _to_bytes(key, "key")
 *                 c_key = key             # <<<<<<<<<<<<<<
 * 
 *                 if isinstance(value, pathlib.Path):
*/
      __pyx_t_16 = __Pyx_PyObject_AsString(__pyx_v_key); if (unlikely((!__pyx_t_16) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L1_error)
      __pyx_v_c_key = __pyx_t_16;

      /* "adbc_driver_manager/_lib.pyx":590
 *                 c_key = key
 * 
 *                 if isinstance(value, pathlib.Path):             # <<<<<<<<<<<<<<
 *                     value = str(value)
 *                 elif isinstance(value, enum.Enum):
*/
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pathlib); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Path); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = PyObject_IsInstance(__pyx_v_value, __pyx_t_15); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (__pyx_t_11) {

        /* "adbc_driver_manager/_lib.pyx":591
 * 
 *                 if isinstance(value, pathlib.Path):
 *                     value = str(value)             # <<<<<<<<<<<<<<
 *                 elif isinstance(value, enum.Enum):
 *                     value = value.value
*/
        __pyx_t_15 = __Pyx_PyObject_Unicode(__pyx_v_value); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_15);
        __pyx_t_15 = 0;

        /* "adbc_driver_manager/_lib.pyx":590
 *                 c_key = key
 * 
 *                 if isinstance(value, pathlib.Path):             # <<<<<<<<<<<<<<
 *                     value = str(value)
 *                 elif isinstance(value, enum.Enum):
*/
        goto __pyx_L9;
      }

      /* "adbc_driver_manager/_lib.pyx":592
 *                 if isinstance(value, pathlib.Path):
 *                     value = str(value)
 *                 elif isinstance(value, enum.Enum):             # <<<<<<<<<<<<<<
 *                     value = value.value
 * 
*/
      __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_enum); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_Enum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_11 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_11) {

        /* "adbc_driver_manager/_lib.pyx":593
 *                     value = str(value)
 *                 elif isinstance(value, enum.Enum):
 *                     value = value.value             # <<<<<<<<<<<<<<
 * 
 *                 if isinstance(value, bool):
*/
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "adbc_driver_manager/_lib.pyx":592
 *                 if isinstance(value, pathlib.Path):
 *                     value = str(value)
 *                 elif isinstance(value, enum.Enum):             # <<<<<<<<<<<<<<
 *                     value = value.value
 * 
*/
      }
      __pyx_L9:;

      /* "adbc_driver_manager/_lib.pyx":595
 *                     value = value.value
 * 
 *                 if isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                     if value:
 *                         value = ADBC_OPTION_VALUE_ENABLED
*/
      __pyx_t_11 = PyBool_Check(__pyx_v_value); 
      if (__pyx_t_11) {

        /* "adbc_driver_manager/_lib.pyx":596
 * 
 *                 if isinstance(value, bool):
 *                     if value:             # <<<<<<<<<<<<<<
 *                         value = ADBC_OPTION_VALUE_ENABLED
 *                     else:
*/
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 596, __pyx_L1_error)
        if (__pyx_t_11) {

          /* "adbc_driver_manager/_lib.pyx":597
 *                 if isinstance(value, bool):
 *                     if value:
 *                         value = ADBC_OPTION_VALUE_ENABLED             # <<<<<<<<<<<<<<
 *                     else:
 *                         value = ADBC_OPTION_VALUE_DISABLED
*/
          __pyx_t_2 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_ENABLED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "adbc_driver_manager/_lib.pyx":596
 * 
 *                 if isinstance(value, bool):
 *                     if value:             # <<<<<<<<<<<<<<
 *                         value = ADBC_OPTION_VALUE_ENABLED
 *                     else:
*/
          goto __pyx_L11;
        }

        /* "adbc_driver_manager/_lib.pyx":599
 *                         value = ADBC_OPTION_VALUE_ENABLED
 *                     else:
 *                         value = ADBC_OPTION_VALUE_DISABLED             # <<<<<<<<<<<<<<
 *                     value = _to_bytes(value, "option value")
 *                     c_value = value
*/
        /*else*/ {
          __pyx_t_2 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_DISABLED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
          __pyx_t_2 = 0;
        }
        __pyx_L11:;

        /* "adbc_driver_manager/_lib.pyx":600
 *                     else:
 *                         value = ADBC_OPTION_VALUE_DISABLED
 *                     value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                     c_value = value
 *                     status = AdbcDatabaseSetOption(
*/
        __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "adbc_driver_manager/_lib.pyx":601
 *                         value = ADBC_OPTION_VALUE_DISABLED
 *                     value = _to_bytes(value, "option value")
 *                     c_value = value             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOption(
 *                         &self.database, c_key, c_value, &c_error)
*/
        __pyx_t_17 = __Pyx_PyObject_AsString(__pyx_v_value); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 601, __pyx_L1_error)
        __pyx_v_c_value = __pyx_t_17;

        /* "adbc_driver_manager/_lib.pyx":602
 *                     value = _to_bytes(value, "option value")
 *                     c_value = value
 *                     status = AdbcDatabaseSetOption(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, c_value, &c_error)
 *                 elif isinstance(value, int):
*/
        __pyx_v_status = AdbcDatabaseSetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

        /* "adbc_driver_manager/_lib.pyx":595
 *                     value = value.value
 * 
 *                 if isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                     if value:
 *                         value = ADBC_OPTION_VALUE_ENABLED
*/
        goto __pyx_L10;
      }

      /* "adbc_driver_manager/_lib.pyx":604
 *                     status = AdbcDatabaseSetOption(
 *                         &self.database, c_key, c_value, &c_error)
 *                 elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                     c_value_int = value
 *                     status = AdbcDatabaseSetOptionInt(
*/
      __pyx_t_11 = PyLong_Check(__pyx_v_value); 
      if (__pyx_t_11) {

        /* "adbc_driver_manager/_lib.pyx":605
 *                         &self.database, c_key, c_value, &c_error)
 *                 elif isinstance(value, int):
 *                     c_value_int = value             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionInt(
 *                         &self.database, c_key, c_value_int, &c_error)
*/
        __pyx_t_18 = __Pyx_PyLong_As_int64_t(__pyx_v_value); if (unlikely((__pyx_t_18 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 605, __pyx_L1_error)
        __pyx_v_c_value_int = __pyx_t_18;

        /* "adbc_driver_manager/_lib.pyx":606
 *                 elif isinstance(value, int):
 *                     c_value_int = value
 *                     status = AdbcDatabaseSetOptionInt(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, c_value_int, &c_error)
 *                 elif isinstance(value, float):
*/
        __pyx_v_status = AdbcDatabaseSetOptionInt((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value_int, (&__pyx_v_c_error));

        /* "adbc_driver_manager/_lib.pyx":604
 *                     status = AdbcDatabaseSetOption(
 *                         &self.database, c_key, c_value, &c_error)
 *                 elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                     c_value_int = value
 *                     status = AdbcDatabaseSetOptionInt(
*/
        goto __pyx_L10;
      }

      /* "adbc_driver_manager/_lib.pyx":608
 *                     status = AdbcDatabaseSetOptionInt(
 *                         &self.database, c_key, c_value_int, &c_error)
 *                 elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                     c_value_double = value
 *                     status = AdbcDatabaseSetOptionDouble(
*/
      __pyx_t_11 = PyFloat_Check(__pyx_v_value); 
      if (__pyx_t_11) {

        /* "adbc_driver_manager/_lib.pyx":609
 *                         &self.database, c_key, c_value_int, &c_error)
 *                 elif isinstance(value, float):
 *                     c_value_double = value             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionDouble(
 *                         &self.database, c_key, c_value_double, &c_error)
*/
        __pyx_t_19 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 609, __pyx_L1_error)
        __pyx_v_c_value_double = __pyx_t_19;

        /* "adbc_driver_manager/_lib.pyx":610
 *                 elif isinstance(value, float):
 *                     c_value_double = value
 *                     status = AdbcDatabaseSetOptionDouble(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, c_value_double, &c_error)
 *                 elif isinstance(value, bytes):
*/
        __pyx_v_status = AdbcDatabaseSetOptionDouble((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value_double, (&__pyx_v_c_error));

        /* "adbc_driver_manager/_lib.pyx":608
 *                     status = AdbcDatabaseSetOptionInt(
 *                         &self.database, c_key, c_value_int, &c_error)
 *                 elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                     c_value_double = value
 *                     status = AdbcDatabaseSetOptionDouble(
*/
        goto __pyx_L10;
      }

      /* "adbc_driver_manager/_lib.pyx":612
 *                     status = AdbcDatabaseSetOptionDouble(
 *                         &self.database, c_key, c_value_double, &c_error)
 *                 elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                     c_value = value
 *                     c_value_len = len(value)
*/
      __pyx_t_11 = PyBytes_Check(__pyx_v_value); 
      if (__pyx_t_11) {

        /* "adbc_driver_manager/_lib.pyx":613
 *                         &self.database, c_key, c_value_double, &c_error)
 *                 elif isinstance(value, bytes):
 *                     c_value = value             # <<<<<<<<<<<<<<
 *                     c_value_len = len(value)
 *                     status = AdbcDatabaseSetOptionBytes(
*/
        __pyx_t_17 = __Pyx_PyObject_AsString(__pyx_v_value); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 613, __pyx_L1_error)
        __pyx_v_c_value = __pyx_t_17;

        /* "adbc_driver_manager/_lib.pyx":614
 *                 elif isinstance(value, bytes):
 *                     c_value = value
 *                     c_value_len = len(value)             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionBytes(
 *                         &self.database, c_key, <const uint8_t*> c_value,
*/
        __pyx_t_20 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_20 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)
        __pyx_v_c_value_len = __pyx_t_20;

        /* "adbc_driver_manager/_lib.pyx":615
 *                     c_value = value
 *                     c_value_len = len(value)
 *                     status = AdbcDatabaseSetOptionBytes(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
*/
        __pyx_v_status = AdbcDatabaseSetOptionBytes((&__pyx_v_self->database), __pyx_v_c_key, ((uint8_t const *)__pyx_v_c_value), __pyx_v_c_value_len, (&__pyx_v_c_error));

        /* "adbc_driver_manager/_lib.pyx":612
 *                     status = AdbcDatabaseSetOptionDouble(
 *                         &self.database, c_key, c_value_double, &c_error)
 *                 elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                     c_value = value
 *                     c_value_len = len(value)
*/
        goto __pyx_L10;
      }

      /* "adbc_driver_manager/_lib.pyx":619
 *                         c_value_len, &c_error)
 *                 else:
 *                     value = _to_bytes(value, "value")             # <<<<<<<<<<<<<<
 *                     c_value = value
 *                     status = AdbcDatabaseSetOption(
*/
      /*else*/ {
        __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "adbc_driver_manager/_lib.pyx":620
 *                 else:
 *                     value = _to_bytes(value, "value")
 *                     c_value = value             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOption(
 *                         &self.database, c_key, c_value, &c_error)
*/
        __pyx_t_17 = __Pyx_PyObject_AsString(__pyx_v_value); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 620, __pyx_L1_error)
        __pyx_v_c_value = __pyx_t_17;

        /* "adbc_driver_manager/_lib.pyx":621
 *                     value = _to_bytes(value, "value")
 *                     c_value = value
 *                     status = AdbcDatabaseSetOption(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, c_value, &c_error)
 *             check_error(status, &c_error)
*/
        __pyx_v_status = AdbcDatabaseSetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));
      }
      __pyx_L10:;
    }
    __pyx_L8:;

    /* "adbc_driver_manager/_lib.pyx":623
 *                     status = AdbcDatabaseSetOption(
 *                         &self.database, c_key, c_value, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         # check if we're running in a venv
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 623, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":626
 * 
 *         # check if we're running in a venv
 *         if sys.prefix != sys.base_prefix:             # <<<<<<<<<<<<<<
 *             # if we're in a venv, add the venv prefix to the search path list
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_prefix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_base_prefix); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_15, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {

    /* "adbc_driver_manager/_lib.pyx":629
 *             # if we're in a venv, add the venv prefix to the search path list
 * 
 *             key = "additional_manifest_search_path_list"             # <<<<<<<<<<<<<<
 *             if key not in kwargs:
 *                 manifest_path = _to_bytes(
*/
    __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_additional_manifest_search_path);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_additional_manifest_search_path);

    /* "adbc_driver_manager/_lib.pyx":630
 * 
 *             key = "additional_manifest_search_path_list"
 *             if key not in kwargs:             # <<<<<<<<<<<<<<
 *                 manifest_path = _to_bytes(
 *                     os.path.join(sys.prefix, "etc/adbc/drivers"),
*/
    __pyx_t_11 = (__Pyx_PyDict_ContainsTF(__pyx_v_key, __pyx_v_kwargs, Py_NE)); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 630, __pyx_L1_error)
    if (__pyx_t_11) {

      /* "adbc_driver_manager/_lib.pyx":632
 *             if key not in kwargs:
 *                 manifest_path = _to_bytes(
 *                     os.path.join(sys.prefix, "etc/adbc/drivers"),             # <<<<<<<<<<<<<<
 *                     "sys.prefix")
 *                 c_key = b"additional_manifest_search_path_list"
*/
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_15 = __pyx_t_3;
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_prefix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 632, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_15, __pyx_t_4, __pyx_mstate_global->__pyx_kp_u_etc_adbc_drivers};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_join, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }

      /* "adbc_driver_manager/_lib.pyx":631
 *             key = "additional_manifest_search_path_list"
 *             if key not in kwargs:
 *                 manifest_path = _to_bytes(             # <<<<<<<<<<<<<<
 *                     os.path.join(sys.prefix, "etc/adbc/drivers"),
 *                     "sys.prefix")
*/
      __pyx_t_3 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_sys_prefix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_manifest_path = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "adbc_driver_manager/_lib.pyx":634
 *                     os.path.join(sys.prefix, "etc/adbc/drivers"),
 *                     "sys.prefix")
 *                 c_key = b"additional_manifest_search_path_list"             # <<<<<<<<<<<<<<
 *                 c_value = manifest_path
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)
*/
      __pyx_v_c_key = __pyx_k_additional_manifest_search_path;

      /* "adbc_driver_manager/_lib.pyx":635
 *                     "sys.prefix")
 *                 c_key = b"additional_manifest_search_path_list"
 *                 c_value = manifest_path             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)
 *                 check_error(status, &c_error)
*/
      if (unlikely(__pyx_v_manifest_path == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
        __PYX_ERR(0, 635, __pyx_L1_error)
      }
      __pyx_t_17 = __Pyx_PyBytes_AsString(__pyx_v_manifest_path); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_17;

      /* "adbc_driver_manager/_lib.pyx":636
 *                 c_key = b"additional_manifest_search_path_list"
 *                 c_value = manifest_path
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)             # <<<<<<<<<<<<<<
 *                 check_error(status, &c_error)
 * 
*/
      __pyx_v_status = AdbcDatabaseSetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

      /* "adbc_driver_manager/_lib.pyx":637
 *                 c_value = manifest_path
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)
 *                 check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *             key = "additional_profile_search_path_list"
*/
      __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 637, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":630
 * 
 *             key = "additional_manifest_search_path_list"
 *             if key not in kwargs:             # <<<<<<<<<<<<<<
 *                 manifest_path = _to_bytes(
 *                     os.path.join(sys.prefix, "etc/adbc/drivers"),
*/
    }

    /* "adbc_driver_manager/_lib.pyx":639
 *                 check_error(status, &c_error)
 * 
 *             key = "additional_profile_search_path_list"             # <<<<<<<<<<<<<<
 *             if key not in kwargs:
 *                 profile_path = _to_bytes(
*/
    __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_additional_profile_search_path_l);
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_additional_profile_search_path_l);

    /* "adbc_driver_manager/_lib.pyx":640
 * 
 *             key = "additional_profile_search_path_list"
 *             if key not in kwargs:             # <<<<<<<<<<<<<<
 *                 profile_path = _to_bytes(
 *                     os.path.join(sys.prefix, "etc/adbc/profiles"),
*/
    __pyx_t_11 = (__Pyx_PyDict_ContainsTF(__pyx_v_key, __pyx_v_kwargs, Py_NE)); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 640, __pyx_L1_error)
    if (__pyx_t_11) {

      /* "adbc_driver_manager/_lib.pyx":642
 *             if key not in kwargs:
 *                 profile_path = _to_bytes(
 *                     os.path.join(sys.prefix, "etc/adbc/profiles"),             # <<<<<<<<<<<<<<
 *                     "sys.prefix")
 *                 c_key = b"additional_profile_search_path_list"
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_path); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_1 = __pyx_t_15;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_prefix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_kp_u_etc_adbc_profiles};
        __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_join, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 642, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }

      /* "adbc_driver_manager/_lib.pyx":641
 *             key = "additional_profile_search_path_list"
 *             if key not in kwargs:
 *                 profile_path = _to_bytes(             # <<<<<<<<<<<<<<
 *                     os.path.join(sys.prefix, "etc/adbc/profiles"),
 *                     "sys.prefix")
*/
      __pyx_t_15 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_sys_prefix); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_profile_path = ((PyObject*)__pyx_t_15);
      __pyx_t_15 = 0;

      /* "adbc_driver_manager/_lib.pyx":644
 *                     os.path.join(sys.prefix, "etc/adbc/profiles"),
 *                     "sys.prefix")
 *                 c_key = b"additional_profile_search_path_list"             # <<<<<<<<<<<<<<
 *                 c_value = profile_path
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)
*/
      __pyx_v_c_key = __pyx_k_additional_profile_search_path_l;

      /* "adbc_driver_manager/_lib.pyx":645
 *                     "sys.prefix")
 *                 c_key = b"additional_profile_search_path_list"
 *                 c_value = profile_path             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)
 *                 check_error(status, &c_error)
*/
      if (unlikely(__pyx_v_profile_path == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
        __PYX_ERR(0, 645, __pyx_L1_error)
      }
      __pyx_t_17 = __Pyx_PyBytes_AsString(__pyx_v_profile_path); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_17;

      /* "adbc_driver_manager/_lib.pyx":646
 *                 c_key = b"additional_profile_search_path_list"
 *                 c_value = profile_path
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)             # <<<<<<<<<<<<<<
 *                 check_error(status, &c_error)
 * 
*/
      __pyx_v_status = AdbcDatabaseSetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

      /* "adbc_driver_manager/_lib.pyx":647
 *                 c_value = profile_path
 *                 status = AdbcDatabaseSetOption(&self.database, c_key, c_value, &c_error)
 *                 check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
      __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 647, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":640
 * 
 *             key = "additional_profile_search_path_list"
 *             if key not in kwargs:             # <<<<<<<<<<<<<<
 *                 profile_path = _to_bytes(
 *                     os.path.join(sys.prefix, "etc/adbc/profiles"),
*/
    }

    /* "adbc_driver_manager/_lib.pyx":626
 * 
 *         # check if we're running in a venv
 *         if sys.prefix != sys.base_prefix:             # <<<<<<<<<<<<<<
 *             # if we're in a venv, add the venv prefix to the search path list
 * 
*/
  }

  /* "adbc_driver_manager/_lib.pyx":649
 *                 check_error(status, &c_error)
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseInit(&self.database, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":650
 * 
 *         with nogil:
 *             status = AdbcDatabaseInit(&self.database, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcDatabaseInit((&__pyx_v_self->database), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":649
 *                 check_error(status, &c_error)
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseInit(&self.database, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L17;
        }
        __pyx_L17:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":651
 *         with nogil:
 *             status = AdbcDatabaseInit(&self.database, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def close(self) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 651, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":551
 *         CAdbcDatabase database
 * 
 *     def __init__(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         super().__init__("AdbcConnection")
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_manifest_path);
  __Pyx_XDECREF(__pyx_v_profile_path);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":653
 *         check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the database."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_3close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_2close, "Release the handle to the database.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_3close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_3close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_2close};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_3close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_2close(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_2close(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "adbc_driver_manager/_lib.pyx":655
 *     def close(self) -> None:
 *         """Release the handle to the database."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 655, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":657
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             if self.database.private_data == NULL:
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._check_open_children(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":658
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self.database.private_data == NULL:
 *                 return
*/
  /*with:*/ {
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->__pyx_base._lock, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 658, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->__pyx_base._lock, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 658, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 658, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":659
 *         self._check_open_children()
 *         with self._lock:
 *             if self.database.private_data == NULL:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          __pyx_t_10 = (__pyx_v_self->database.private_data == NULL);
          if (__pyx_t_10) {

            /* "adbc_driver_manager/_lib.pyx":660
 *         with self._lock:
 *             if self.database.private_data == NULL:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             with nogil:
*/
            __Pyx_XDECREF(__pyx_r);
            __pyx_r = Py_None; __Pyx_INCREF(Py_None);
            goto __pyx_L11_try_return;

            /* "adbc_driver_manager/_lib.pyx":659
 *         self._check_open_children()
 *         with self._lock:
 *             if self.database.private_data == NULL:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          }

          /* "adbc_driver_manager/_lib.pyx":662
 *                 return
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseRelease(&self.database, &c_error)
 *             check_error(status, &c_error)
*/
          {
              PyThreadState * _save;
              _save = PyEval_SaveThread();
              __Pyx_FastGIL_Remember();
              /*try:*/ {

                /* "adbc_driver_manager/_lib.pyx":663
 * 
 *             with nogil:
 *                 status = AdbcDatabaseRelease(&self.database, &c_error)             # <<<<<<<<<<<<<<
 *             check_error(status, &c_error)
 * 
*/
                __pyx_v_status = AdbcDatabaseRelease((&__pyx_v_self->database), (&__pyx_v_c_error));
              }

              /* "adbc_driver_manager/_lib.pyx":662
 *                 return
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseRelease(&self.database, &c_error)
 *             check_error(status, &c_error)
*/
              /*finally:*/ {
                /*normal exit:*/{
                  __Pyx_FastGIL_Forget();
                  PyEval_RestoreThread(_save);
                  goto __pyx_L16;
                }
                __pyx_L16:;
              }
          }

          /* "adbc_driver_manager/_lib.pyx":664
 *             with nogil:
 *                 status = AdbcDatabaseRelease(&self.database, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def get_option(
*/
          __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 664, __pyx_L7_error)

          /* "adbc_driver_manager/_lib.pyx":658
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self.database.private_data == NULL:
 *                 return
*/
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 658, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_4);
          __pyx_t_11 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 658, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 658, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_10 < (0)) __PYX_ERR(0, 658, __pyx_L9_except_error)
          __pyx_t_13 = (!__pyx_t_10);
          if (unlikely(__pyx_t_13)) {
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_XGIVEREF(__pyx_t_4);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_5, __pyx_t_4);
            __pyx_t_2 = 0;  __pyx_t_5 = 0;  __pyx_t_4 = 0; 
            __PYX_ERR(0, 658, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L11_try_return:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L4_return;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 658, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L4_return: {
        __pyx_t_9 = __pyx_r;
        __pyx_r = 0;
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 658, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_r = __pyx_t_9;
        __pyx_t_9 = 0;
        goto __pyx_L0;
      }
      __pyx_L6:;
    }
    goto __pyx_L20;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L20:;
  }

  /* "adbc_driver_manager/_lib.pyx":653
 *         check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the database."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":666
 *             check_error(status, &c_error)
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_5get_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_4get_option, "\n        Get the value of a string option.\n\n        Parameters\n        ----------\n        key : str or bytes\n            The option to get.\n        encoding : str\n            The encoding of the option value.  This should almost\n            always be UTF-8.\n        errors : str\n            What to do about errors when decoding the option value\n            (see bytes.decode).\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_5get_option = {"get_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_5get_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_4get_option};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_5get_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_encoding = 0;
  PyObject *__pyx_v_errors = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,&__pyx_mstate_global->__pyx_n_u_encoding,&__pyx_mstate_global->__pyx_n_u_errors,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 666, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 666, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option", 0) < (0)) __PYX_ERR(0, 666, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_kp_u_utf_8));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_strict));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option", 1, 1, 1, i); __PYX_ERR(0, 666, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 666, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_kp_u_utf_8));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_strict));
    }
    __pyx_v_key = values[0];
    __pyx_v_encoding = values[1];
    __pyx_v_errors = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 666, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_4get_option(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self), __pyx_v_key, __pyx_v_encoding, __pyx_v_errors);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_4get_option(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  size_t __pyx_v_c_len;
  char *__pyx_v_c_buf;
  PyObject *__pyx_v_buf = NULL;
  Py_ssize_t __pyx_v_new_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10[5];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option", 0);

  /* "adbc_driver_manager/_lib.pyx":687
 *             (see bytes.decode).
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":689
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":690
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef size_t c_len = 0
 *         cdef char* c_buf = NULL
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 690, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":691
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0             # <<<<<<<<<<<<<<
 *         cdef char* c_buf = NULL
 * 
*/
  __pyx_v_c_len = 0;

  /* "adbc_driver_manager/_lib.pyx":692
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
 *         cdef char* c_buf = NULL             # <<<<<<<<<<<<<<
 * 
 *         buf = bytearray(1024)
*/
  __pyx_v_c_buf = NULL;

  /* "adbc_driver_manager/_lib.pyx":694
 *         cdef char* c_buf = NULL
 * 
 *         buf = bytearray(1024)             # <<<<<<<<<<<<<<
 *         while True:
 *             c_len = len(buf)
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_int_1024};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_buf = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":695
 * 
 *         buf = bytearray(1024)
 *         while True:             # <<<<<<<<<<<<<<
 *             c_len = len(buf)
 *             c_buf = buf
*/
  while (1) {

    /* "adbc_driver_manager/_lib.pyx":696
 *         buf = bytearray(1024)
 *         while True:
 *             c_len = len(buf)             # <<<<<<<<<<<<<<
 *             c_buf = buf
 *             with nogil:
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 696, __pyx_L1_error)
    __pyx_v_c_len = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":697
 *         while True:
 *             c_len = len(buf)
 *             c_buf = buf             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 status = AdbcDatabaseGetOption(
*/
    __pyx_t_3 = __Pyx_PyObject_AsWritableString(__pyx_v_buf); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 697, __pyx_L1_error)
    __pyx_v_c_buf = __pyx_t_3;

    /* "adbc_driver_manager/_lib.pyx":698
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseGetOption(
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
*/
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":699
 *             c_buf = buf
 *             with nogil:
 *                 status = AdbcDatabaseGetOption(             # <<<<<<<<<<<<<<
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
*/
          __pyx_v_status = AdbcDatabaseGetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_buf, (&__pyx_v_c_len), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":698
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseGetOption(
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "adbc_driver_manager/_lib.pyx":701
 *                 status = AdbcDatabaseGetOption(
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *             if c_len <= len(buf):
 *                 # Entire value read
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 701, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":702
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 702, __pyx_L1_error)
    __pyx_t_7 = (__pyx_v_c_len <= __pyx_t_6);
    if (__pyx_t_7) {

      /* "adbc_driver_manager/_lib.pyx":704
 *             if c_len <= len(buf):
 *                 # Entire value read
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 # Buffer too small
*/
      goto __pyx_L4_break;

      /* "adbc_driver_manager/_lib.pyx":702
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    }

    /* "adbc_driver_manager/_lib.pyx":707
 *             else:
 *                 # Buffer too small
 *                 new_len = len(buf) * 2             # <<<<<<<<<<<<<<
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
*/
    /*else*/ {
      __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 707, __pyx_L1_error)
      __pyx_v_new_len = (__pyx_t_6 * 2);

      /* "adbc_driver_manager/_lib.pyx":708
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      __pyx_t_2 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 708, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MAX_OPTION_SIZE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 708, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 708, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__pyx_t_7)) {

        /* "adbc_driver_manager/_lib.pyx":709
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
*/
        __pyx_t_4 = NULL;

        /* "adbc_driver_manager/_lib.pyx":710
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 710, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "adbc_driver_manager/_lib.pyx":711
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")             # <<<<<<<<<<<<<<
 *                 buf = bytearray(new_len)
 * 
*/
        __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 711, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 711, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_read_option;
        __pyx_t_10[1] = __pyx_t_2;
        __pyx_t_10[2] = __pyx_mstate_global->__pyx_kp_u_would_need_more_than;
        __pyx_t_10[3] = __pyx_t_9;
        __pyx_t_10[4] = __pyx_mstate_global->__pyx_kp_u_bytes;

        /* "adbc_driver_manager/_lib.pyx":710
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 5, 22 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 6, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 710, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
          __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_Raise(__pyx_t_8, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 709, __pyx_L1_error)

        /* "adbc_driver_manager/_lib.pyx":708
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      }

      /* "adbc_driver_manager/_lib.pyx":712
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)             # <<<<<<<<<<<<<<
 * 
 *         # Remove trailing null terminator
*/
      __pyx_t_11 = NULL;
      __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_4};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_DECREF_SET(__pyx_v_buf, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;
    }
  }
  __pyx_L4_break:;

  /* "adbc_driver_manager/_lib.pyx":715
 * 
 *         # Remove trailing null terminator
 *         if c_len > 0:             # <<<<<<<<<<<<<<
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)
*/
  __pyx_t_7 = (__pyx_v_c_len > 0);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":716
 *         # Remove trailing null terminator
 *         if c_len > 0:
 *             c_len -= 1             # <<<<<<<<<<<<<<
 *         return buf[:c_len].decode(encoding, errors)
 * 
*/
    __pyx_v_c_len = (__pyx_v_c_len - 1);

    /* "adbc_driver_manager/_lib.pyx":715
 * 
 *         # Remove trailing null terminator
 *         if c_len > 0:             # <<<<<<<<<<<<<<
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)
*/
  }

  /* "adbc_driver_manager/_lib.pyx":717
 *         if c_len > 0:
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)             # <<<<<<<<<<<<<<
 * 
 *     def get_option_bytes(self, key: str) -> bytes:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_11 = PySequence_GetSlice(__pyx_v_buf, 0, __pyx_v_c_len); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = __pyx_t_11;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_encoding, __pyx_v_errors};
    __pyx_t_8 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_decode, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
  }
  __pyx_r = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":666
 *             check_error(status, &c_error)
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":719
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_7get_option_bytes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_6get_option_bytes, "Get the value of a binary option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_7get_option_bytes = {"get_option_bytes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_7get_option_bytes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_6get_option_bytes};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_7get_option_bytes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_bytes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 719, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 719, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_bytes", 0) < (0)) __PYX_ERR(0, 719, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_bytes", 1, 1, 1, i); __PYX_ERR(0, 719, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 719, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_bytes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 719, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 719, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_6get_option_bytes(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_6get_option_bytes(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  size_t __pyx_v_c_len;
  uint8_t *__pyx_v_c_buf;
  PyObject *__pyx_v_buf = NULL;
  Py_ssize_t __pyx_v_new_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  uint8_t *__pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10[5];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_bytes", 0);

  /* "adbc_driver_manager/_lib.pyx":721
 *     def get_option_bytes(self, key: str) -> bytes:
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":723
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":724
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef size_t c_len = 0
 *         cdef uint8_t* c_buf = NULL
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 724, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 724, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":725
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0             # <<<<<<<<<<<<<<
 *         cdef uint8_t* c_buf = NULL
 * 
*/
  __pyx_v_c_len = 0;

  /* "adbc_driver_manager/_lib.pyx":726
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
 *         cdef uint8_t* c_buf = NULL             # <<<<<<<<<<<<<<
 * 
 *         buf = bytearray(1024)
*/
  __pyx_v_c_buf = NULL;

  /* "adbc_driver_manager/_lib.pyx":728
 *         cdef uint8_t* c_buf = NULL
 * 
 *         buf = bytearray(1024)             # <<<<<<<<<<<<<<
 *         while True:
 *             c_len = len(buf)
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_int_1024};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_buf = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":729
 * 
 *         buf = bytearray(1024)
 *         while True:             # <<<<<<<<<<<<<<
 *             c_len = len(buf)
 *             c_buf = buf
*/
  while (1) {

    /* "adbc_driver_manager/_lib.pyx":730
 *         buf = bytearray(1024)
 *         while True:
 *             c_len = len(buf)             # <<<<<<<<<<<<<<
 *             c_buf = buf
 *             with nogil:
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    __pyx_v_c_len = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":731
 *         while True:
 *             c_len = len(buf)
 *             c_buf = buf             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 status = AdbcDatabaseGetOptionBytes(
*/
    __pyx_t_7 = __Pyx_PyObject_AsWritableUString(__pyx_v_buf); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 731, __pyx_L1_error)
    __pyx_v_c_buf = __pyx_t_7;

    /* "adbc_driver_manager/_lib.pyx":732
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseGetOptionBytes(
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
*/
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":733
 *             c_buf = buf
 *             with nogil:
 *                 status = AdbcDatabaseGetOptionBytes(             # <<<<<<<<<<<<<<
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
*/
          __pyx_v_status = AdbcDatabaseGetOptionBytes((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_buf, (&__pyx_v_c_len), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":732
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseGetOptionBytes(
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "adbc_driver_manager/_lib.pyx":735
 *                 status = AdbcDatabaseGetOptionBytes(
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *             if c_len <= len(buf):
 *                 # Entire value read
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 735, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":736
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 736, __pyx_L1_error)
    __pyx_t_8 = (__pyx_v_c_len <= __pyx_t_6);
    if (__pyx_t_8) {

      /* "adbc_driver_manager/_lib.pyx":738
 *             if c_len <= len(buf):
 *                 # Entire value read
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 # Buffer too small
*/
      goto __pyx_L4_break;

      /* "adbc_driver_manager/_lib.pyx":736
 *                     &self.database, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    }

    /* "adbc_driver_manager/_lib.pyx":741
 *             else:
 *                 # Buffer too small
 *                 new_len = len(buf) * 2             # <<<<<<<<<<<<<<
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
*/
    /*else*/ {
      __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 741, __pyx_L1_error)
      __pyx_v_new_len = (__pyx_t_6 * 2);

      /* "adbc_driver_manager/_lib.pyx":742
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      __pyx_t_2 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MAX_OPTION_SIZE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__pyx_t_8)) {

        /* "adbc_driver_manager/_lib.pyx":743
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
*/
        __pyx_t_4 = NULL;

        /* "adbc_driver_manager/_lib.pyx":745
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")             # <<<<<<<<<<<<<<
 *                 buf = bytearray(new_len)
 * 
*/
        __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 745, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 745, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_read_option;
        __pyx_t_10[1] = __pyx_v_key;
        __pyx_t_10[2] = __pyx_mstate_global->__pyx_kp_u_would_need_more_than;
        __pyx_t_10[3] = __pyx_t_2;
        __pyx_t_10[4] = __pyx_mstate_global->__pyx_kp_u_bytes;

        /* "adbc_driver_manager/_lib.pyx":744
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 5, 22 + __Pyx_PyUnicode_GET_LENGTH(__pyx_v_key) + 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 6, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_key));
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
          __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 743, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_Raise(__pyx_t_9, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 743, __pyx_L1_error)

        /* "adbc_driver_manager/_lib.pyx":742
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      }

      /* "adbc_driver_manager/_lib.pyx":746
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)             # <<<<<<<<<<<<<<
 * 
 *         return bytes(buf[:c_len])
*/
      __pyx_t_11 = NULL;
      __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 746, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_4};
        __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_DECREF_SET(__pyx_v_buf, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;
    }
  }
  __pyx_L4_break:;

  /* "adbc_driver_manager/_lib.pyx":748
 *                 buf = bytearray(new_len)
 * 
 *         return bytes(buf[:c_len])             # <<<<<<<<<<<<<<
 * 
 *     def get_option_float(self, key: str) -> float:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = NULL;
  __pyx_t_11 = PySequence_GetSlice(__pyx_v_buf, 0, __pyx_v_c_len); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
    __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
  }
  __pyx_r = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":719
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":750
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_9get_option_float(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_8get_option_float, "Get the value of a floating-point option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_9get_option_float = {"get_option_float", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_9get_option_float, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_8get_option_float};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_9get_option_float(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_float (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 750, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 750, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_float", 0) < (0)) __PYX_ERR(0, 750, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_float", 1, 1, 1, i); __PYX_ERR(0, 750, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 750, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_float", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 750, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 750, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_8get_option_float(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_8get_option_float(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  double __pyx_v_c_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_float", 0);

  /* "adbc_driver_manager/_lib.pyx":752
 *     def get_option_float(self, key: str) -> float:
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 752, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":754
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":755
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef double c_value = 0.0
 *         with nogil:
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 755, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":756
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcDatabaseGetOptionDouble(
*/
  __pyx_v_c_value = 0.0;

  /* "adbc_driver_manager/_lib.pyx":757
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseGetOptionDouble(
 *                 &self.database, c_key, &c_value, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":758
 *         cdef double c_value = 0.0
 *         with nogil:
 *             status = AdbcDatabaseGetOptionDouble(             # <<<<<<<<<<<<<<
 *                 &self.database, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcDatabaseGetOptionDouble((&__pyx_v_self->database), __pyx_v_c_key, (&__pyx_v_c_value), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":757
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseGetOptionDouble(
 *                 &self.database, c_key, &c_value, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":760
 *             status = AdbcDatabaseGetOptionDouble(
 *                 &self.database, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return c_value
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":761
 *                 &self.database, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
 *         return c_value             # <<<<<<<<<<<<<<
 * 
 *     def get_option_int(self, key: str) -> int:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":750
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":763
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_11get_option_int(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_10get_option_int, "Get the value of an integer option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_11get_option_int = {"get_option_int", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_11get_option_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_10get_option_int};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_11get_option_int(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_int (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 763, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 763, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_int", 0) < (0)) __PYX_ERR(0, 763, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_int", 1, 1, 1, i); __PYX_ERR(0, 763, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 763, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_int", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 763, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 763, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_10get_option_int(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_10get_option_int(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  int64_t __pyx_v_c_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_int", 0);

  /* "adbc_driver_manager/_lib.pyx":765
 *     def get_option_int(self, key: str) -> int:
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 765, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":767
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":768
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef int64_t c_value = 0
 *         with nogil:
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 768, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":769
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcDatabaseGetOptionInt(
*/
  __pyx_v_c_value = 0;

  /* "adbc_driver_manager/_lib.pyx":770
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseGetOptionInt(
 *                 &self.database, c_key, &c_value, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":771
 *         cdef int64_t c_value = 0
 *         with nogil:
 *             status = AdbcDatabaseGetOptionInt(             # <<<<<<<<<<<<<<
 *                 &self.database, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcDatabaseGetOptionInt((&__pyx_v_self->database), __pyx_v_c_key, (&__pyx_v_c_value), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":770
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcDatabaseGetOptionInt(
 *                 &self.database, c_key, &c_value, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":773
 *             status = AdbcDatabaseGetOptionInt(
 *                 &self.database, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return c_value
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 773, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":774
 *                 &self.database, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
 *         return c_value             # <<<<<<<<<<<<<<
 * 
 *     def set_options(self, **kwargs) -> None:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_c_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyInt_FromNumber(&__pyx_t_2, NULL, 0) < (0)) __PYX_ERR(0, 774, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":763
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.get_option_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":776
 *         return c_value
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Set arbitrary key-value options.
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_13set_options(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_12set_options, "\n        Set arbitrary key-value options.\n\n        Pass options as kwargs: ``set_options(**{\"some.option\": \"value\"})``.\n\n        Note, not all drivers support setting options after creation.\n\n        See Also\n        --------\n        adbc_driver_manager.DatabaseOptions : Standard option names.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_13set_options = {"set_options", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_13set_options, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_12set_options};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_13set_options(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_options (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("set_options", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("set_options", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_12set_options(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_12set_options(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  struct AdbcError __pyx_v_c_error;
  char *__pyx_v_c_key;
  char *__pyx_v_c_value;
  size_t __pyx_v_c_value_len;
  double __pyx_v_c_double;
  int64_t __pyx_v_c_int;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  char *__pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  double __pyx_t_12;
  int64_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17[6];
  PyObject *__pyx_t_18 = NULL;
  size_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_options", 0);

  /* "adbc_driver_manager/_lib.pyx":788
 *         adbc_driver_manager.DatabaseOptions : Standard option names.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef char* c_key = NULL
 *         cdef char* c_value = NULL
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":789
 *         """
 *         cdef CAdbcError c_error = empty_error()
 *         cdef char* c_key = NULL             # <<<<<<<<<<<<<<
 *         cdef char* c_value = NULL
 *         cdef size_t c_value_len = 0
*/
  __pyx_v_c_key = NULL;

  /* "adbc_driver_manager/_lib.pyx":790
 *         cdef CAdbcError c_error = empty_error()
 *         cdef char* c_key = NULL
 *         cdef char* c_value = NULL             # <<<<<<<<<<<<<<
 *         cdef size_t c_value_len = 0
 *         cdef double c_double
*/
  __pyx_v_c_value = NULL;

  /* "adbc_driver_manager/_lib.pyx":791
 *         cdef char* c_key = NULL
 *         cdef char* c_value = NULL
 *         cdef size_t c_value_len = 0             # <<<<<<<<<<<<<<
 *         cdef double c_double
 *         cdef int64_t c_int
*/
  __pyx_v_c_value_len = 0;

  /* "adbc_driver_manager/_lib.pyx":794
 *         cdef double c_double
 *         cdef int64_t c_int
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             key = _to_bytes(key, "option key")
 *             c_key = key
*/
  __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_2);
  __pyx_t_2 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
    if (unlikely(__pyx_t_8 == 0)) break;
    if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "adbc_driver_manager/_lib.pyx":795
 *         cdef int64_t c_int
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "option key")             # <<<<<<<<<<<<<<
 *             c_key = key
 * 
*/
    __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_kp_u_option_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "adbc_driver_manager/_lib.pyx":796
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "option key")
 *             c_key = key             # <<<<<<<<<<<<<<
 * 
 *             if value is None:
*/
    __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_key); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 796, __pyx_L1_error)
    __pyx_v_c_key = __pyx_t_9;

    /* "adbc_driver_manager/_lib.pyx":798
 *             c_key = key
 * 
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 c_value = NULL
 *                 status = AdbcDatabaseSetOption(
*/
    __pyx_t_10 = (__pyx_v_value == Py_None);
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":799
 * 
 *             if value is None:
 *                 c_value = NULL             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
*/
      __pyx_v_c_value = NULL;

      /* "adbc_driver_manager/_lib.pyx":800
 *             if value is None:
 *                 c_value = NULL
 *                 status = AdbcDatabaseSetOption(             # <<<<<<<<<<<<<<
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, str):
*/
      __pyx_v_status = AdbcDatabaseSetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

      /* "adbc_driver_manager/_lib.pyx":798
 *             c_key = key
 * 
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 c_value = NULL
 *                 status = AdbcDatabaseSetOption(
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":802
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
    __pyx_t_10 = PyUnicode_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":803
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, str):
 *                 value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 status = AdbcDatabaseSetOption(
*/
      __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "adbc_driver_manager/_lib.pyx":804
 *             elif isinstance(value, str):
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 804, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":805
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 status = AdbcDatabaseSetOption(             # <<<<<<<<<<<<<<
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
*/
      __pyx_v_status = AdbcDatabaseSetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

      /* "adbc_driver_manager/_lib.pyx":802
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":807
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
    __pyx_t_10 = PyBool_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":808
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 808, __pyx_L1_error)
      if (__pyx_t_10) {

        /* "adbc_driver_manager/_lib.pyx":809
 *             elif isinstance(value, bool):
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED             # <<<<<<<<<<<<<<
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
*/
        __pyx_t_7 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_ENABLED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 809, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "adbc_driver_manager/_lib.pyx":808
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
        goto __pyx_L6;
      }

      /* "adbc_driver_manager/_lib.pyx":811
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
      /*else*/ {
        __pyx_t_7 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_DISABLED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 811, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
        __pyx_t_7 = 0;
      }
      __pyx_L6:;

      /* "adbc_driver_manager/_lib.pyx":812
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 status = AdbcDatabaseSetOption(
*/
      __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 812, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "adbc_driver_manager/_lib.pyx":813
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 813, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":814
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 status = AdbcDatabaseSetOption(             # <<<<<<<<<<<<<<
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):
*/
      __pyx_v_status = AdbcDatabaseSetOption((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

      /* "adbc_driver_manager/_lib.pyx":807
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":816
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
    __pyx_t_10 = PyBytes_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":817
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 c_value_len = len(value)
 *                 with nogil:
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 817, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":818
 *             elif isinstance(value, bytes):
 *                 c_value = value
 *                 c_value_len = len(value)             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcDatabaseSetOptionBytes(
*/
      __pyx_t_11 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 818, __pyx_L1_error)
      __pyx_v_c_value_len = __pyx_t_11;

      /* "adbc_driver_manager/_lib.pyx":819
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionBytes(
 *                         &self.database, c_key, <const uint8_t*> c_value,
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":820
 *                 c_value_len = len(value)
 *                 with nogil:
 *                     status = AdbcDatabaseSetOptionBytes(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
*/
            __pyx_v_status = AdbcDatabaseSetOptionBytes((&__pyx_v_self->database), __pyx_v_c_key, ((uint8_t const *)__pyx_v_c_value), __pyx_v_c_value_len, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":819
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionBytes(
 *                         &self.database, c_key, <const uint8_t*> c_value,
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L11;
            }
            __pyx_L11:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":816
 *                 status = AdbcDatabaseSetOption(
 *                     &self.database, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":823
 *                         &self.database, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_double = value
 *                 with nogil:
*/
    __pyx_t_10 = PyFloat_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":824
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):
 *                 c_double = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcDatabaseSetOptionDouble(
*/
      __pyx_t_12 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 824, __pyx_L1_error)
      __pyx_v_c_double = __pyx_t_12;

      /* "adbc_driver_manager/_lib.pyx":825
 *             elif isinstance(value, float):
 *                 c_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionDouble(
 *                         &self.database, c_key, c_double, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":826
 *                 c_double = value
 *                 with nogil:
 *                     status = AdbcDatabaseSetOptionDouble(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, c_double, &c_error)
 *             elif isinstance(value, int):
*/
            __pyx_v_status = AdbcDatabaseSetOptionDouble((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_double, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":825
 *             elif isinstance(value, float):
 *                 c_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionDouble(
 *                         &self.database, c_key, c_double, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L16;
            }
            __pyx_L16:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":823
 *                         &self.database, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_double = value
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":828
 *                     status = AdbcDatabaseSetOptionDouble(
 *                         &self.database, c_key, c_double, &c_error)
 *             elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_int = value
 *                 with nogil:
*/
    __pyx_t_10 = PyLong_Check(__pyx_v_value); 
    if (likely(__pyx_t_10)) {

      /* "adbc_driver_manager/_lib.pyx":829
 *                         &self.database, c_key, c_double, &c_error)
 *             elif isinstance(value, int):
 *                 c_int = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcDatabaseSetOptionInt(
*/
      __pyx_t_13 = __Pyx_PyLong_As_int64_t(__pyx_v_value); if (unlikely((__pyx_t_13 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 829, __pyx_L1_error)
      __pyx_v_c_int = __pyx_t_13;

      /* "adbc_driver_manager/_lib.pyx":830
 *             elif isinstance(value, int):
 *                 c_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionInt(
 *                         &self.database, c_key, c_int, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":831
 *                 c_int = value
 *                 with nogil:
 *                     status = AdbcDatabaseSetOptionInt(             # <<<<<<<<<<<<<<
 *                         &self.database, c_key, c_int, &c_error)
 *             else:
*/
            __pyx_v_status = AdbcDatabaseSetOptionInt((&__pyx_v_self->database), __pyx_v_c_key, __pyx_v_c_int, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":830
 *             elif isinstance(value, int):
 *                 c_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcDatabaseSetOptionInt(
 *                         &self.database, c_key, c_int, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L21;
            }
            __pyx_L21:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":828
 *                     status = AdbcDatabaseSetOptionDouble(
 *                         &self.database, c_key, c_double, &c_error)
 *             elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_int = value
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":834
 *                         &self.database, c_key, c_int, &c_error)
 *             else:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     f"Unsupported type {type(value)} for value {value!r} "
 *                     f"of option {key}")
*/
    /*else*/ {
      __pyx_t_6 = NULL;

      /* "adbc_driver_manager/_lib.pyx":835
 *             else:
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "             # <<<<<<<<<<<<<<
 *                     f"of option {key}")
 * 
*/
      __pyx_t_14 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_value)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_value), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);

      /* "adbc_driver_manager/_lib.pyx":836
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "
 *                     f"of option {key}")             # <<<<<<<<<<<<<<
 * 
 *             check_error(status, &c_error)
*/
      __pyx_t_16 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_17[0] = __pyx_mstate_global->__pyx_kp_u_Unsupported_type;
      __pyx_t_17[1] = __pyx_t_14;
      __pyx_t_17[2] = __pyx_mstate_global->__pyx_kp_u_for_value;
      __pyx_t_17[3] = __pyx_t_15;
      __pyx_t_17[4] = __pyx_mstate_global->__pyx_kp_u_of_option;
      __pyx_t_17[5] = __pyx_t_16;

      /* "adbc_driver_manager/_lib.pyx":835
 *             else:
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "             # <<<<<<<<<<<<<<
 *                     f"of option {key}")
 * 
*/
      __pyx_t_18 = __Pyx_PyUnicode_Join(__pyx_t_17, 6, 17 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14) + 11 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16));
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_19 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_18};
        __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_19, (2-__pyx_t_19) | (__pyx_t_19*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 834, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 834, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "adbc_driver_manager/_lib.pyx":838
 *                     f"of option {key}")
 * 
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":776
 *         return c_value
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Set arbitrary key-value options.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.set_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_14__reduce_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_database_cannot_be_converte, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_16__setstate_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_12AdbcDatabase_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_database_cannot_be_converte, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcDatabase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":859
 *         CAdbcConnection connection
 * 
 *     def __init__(self, AdbcDatabase database, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         super().__init__("AdbcStatement")
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static int __pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_database = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_database,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 859, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 859, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values, kwd_pos_args, __pyx_kwds_len, "__init__", 1) < (0)) __PYX_ERR(0, 859, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 859, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 859, __pyx_L3_error)
    }
    __pyx_v_database = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 859, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_database), __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, 1, "database", 0))) __PYX_ERR(0, 859, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection___init__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_database, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection___init__(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *__pyx_v_database, PyObject *__pyx_v_kwargs) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  char const *__pyx_v_c_key;
  char const *__pyx_v_c_value;
  int64_t __pyx_v_c_value_int;
  double __pyx_v_c_value_double;
  size_t __pyx_v_c_value_len;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  struct AdbcError __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  char const *__pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  int64_t __pyx_t_14;
  double __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "adbc_driver_manager/_lib.pyx":860
 * 
 *     def __init__(self, AdbcDatabase database, **kwargs) -> None:
 *         super().__init__("AdbcStatement")             # <<<<<<<<<<<<<<
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_AdbcStatement};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":861
 *     def __init__(self, AdbcDatabase database, **kwargs) -> None:
 *         super().__init__("AdbcStatement")
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef const char* c_key
*/
  __pyx_t_6 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 861, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_6;

  /* "adbc_driver_manager/_lib.pyx":867
 *         cdef int64_t c_value_int
 *         cdef double c_value_double
 *         cdef size_t c_value_len = 0             # <<<<<<<<<<<<<<
 * 
 *         self.database = database
*/
  __pyx_v_c_value_len = 0;

  /* "adbc_driver_manager/_lib.pyx":869
 *         cdef size_t c_value_len = 0
 * 
 *         self.database = database             # <<<<<<<<<<<<<<
 *         memset(&self.connection, 0, cython.sizeof(CAdbcConnection))
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_database);
  __Pyx_GIVEREF((PyObject *)__pyx_v_database);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->database);
  __Pyx_DECREF((PyObject *)__pyx_v_self->database);
  __pyx_v_self->database = __pyx_v_database;

  /* "adbc_driver_manager/_lib.pyx":870
 * 
 *         self.database = database
 *         memset(&self.connection, 0, cython.sizeof(CAdbcConnection))             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  (void)(memset((&__pyx_v_self->connection), 0, (sizeof(struct AdbcConnection))));

  /* "adbc_driver_manager/_lib.pyx":872
 *         memset(&self.connection, 0, cython.sizeof(CAdbcConnection))
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionNew(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":873
 * 
 *         with nogil:
 *             status = AdbcConnectionNew(&self.connection, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcConnectionNew((&__pyx_v_self->connection), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":872
 *         memset(&self.connection, 0, cython.sizeof(CAdbcConnection))
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionNew(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":874
 *         with nogil:
 *             status = AdbcConnectionNew(&self.connection, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         for key, value in kwargs.items():
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 874, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":876
 *         check_error(status, &c_error)
 * 
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             key = _to_bytes(key, "key")
 *             c_key = key
*/
  __pyx_t_7 = 0;
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_8), (&__pyx_t_9)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_8, &__pyx_t_7, &__pyx_t_3, &__pyx_t_2, NULL, __pyx_t_9);
    if (unlikely(__pyx_t_10 == 0)) break;
    if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":877
 * 
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *             c_key = key
 * 
*/
    __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":878
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "key")
 *             c_key = key             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(value, pathlib.Path):
*/
    __pyx_t_11 = __Pyx_PyObject_AsString(__pyx_v_key); if (unlikely((!__pyx_t_11) && PyErr_Occurred())) __PYX_ERR(0, 878, __pyx_L1_error)
    __pyx_v_c_key = __pyx_t_11;

    /* "adbc_driver_manager/_lib.pyx":880
 *             c_key = key
 * 
 *             if isinstance(value, pathlib.Path):             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             elif isinstance(value, enum.Enum):
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pathlib); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_12 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_12) {

      /* "adbc_driver_manager/_lib.pyx":881
 * 
 *             if isinstance(value, pathlib.Path):
 *                 value = str(value)             # <<<<<<<<<<<<<<
 *             elif isinstance(value, enum.Enum):
 *                 value = value.value
*/
      __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "adbc_driver_manager/_lib.pyx":880
 *             c_key = key
 * 
 *             if isinstance(value, pathlib.Path):             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             elif isinstance(value, enum.Enum):
*/
      goto __pyx_L8;
    }

    /* "adbc_driver_manager/_lib.pyx":882
 *             if isinstance(value, pathlib.Path):
 *                 value = str(value)
 *             elif isinstance(value, enum.Enum):             # <<<<<<<<<<<<<<
 *                 value = value.value
 * 
*/
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_enum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Enum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_12 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_12) {

      /* "adbc_driver_manager/_lib.pyx":883
 *                 value = str(value)
 *             elif isinstance(value, enum.Enum):
 *                 value = value.value             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(value, bool):
*/
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "adbc_driver_manager/_lib.pyx":882
 *             if isinstance(value, pathlib.Path):
 *                 value = str(value)
 *             elif isinstance(value, enum.Enum):             # <<<<<<<<<<<<<<
 *                 value = value.value
 * 
*/
    }
    __pyx_L8:;

    /* "adbc_driver_manager/_lib.pyx":885
 *                 value = value.value
 * 
 *             if isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
    __pyx_t_12 = PyBool_Check(__pyx_v_value); 
    if (__pyx_t_12) {

      /* "adbc_driver_manager/_lib.pyx":886
 * 
 *             if isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 886, __pyx_L1_error)
      if (__pyx_t_12) {

        /* "adbc_driver_manager/_lib.pyx":887
 *             if isinstance(value, bool):
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED             # <<<<<<<<<<<<<<
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
*/
        __pyx_t_2 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_ENABLED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 887, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "adbc_driver_manager/_lib.pyx":886
 * 
 *             if isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
        goto __pyx_L10;
      }

      /* "adbc_driver_manager/_lib.pyx":889
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
      /*else*/ {
        __pyx_t_2 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_DISABLED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;
      }
      __pyx_L10:;

      /* "adbc_driver_manager/_lib.pyx":890
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 with nogil:
*/
      __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "adbc_driver_manager/_lib.pyx":891
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(
*/
      __pyx_t_13 = __Pyx_PyObject_AsString(__pyx_v_value); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(0, 891, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_13;

      /* "adbc_driver_manager/_lib.pyx":892
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":893
 *                 c_value = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:
*/
            __pyx_v_status = AdbcConnectionSetOption((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

            /* "adbc_driver_manager/_lib.pyx":895
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             if isinstance(value, int):
*/
            __pyx_t_12 = (__pyx_v_status != ADBC_STATUS_OK);
            if (__pyx_t_12) {

              /* "adbc_driver_manager/_lib.pyx":896
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)             # <<<<<<<<<<<<<<
 *             if isinstance(value, int):
 *                 c_value_int = value
*/
              (void)(AdbcConnectionRelease((&__pyx_v_self->connection), NULL));

              /* "adbc_driver_manager/_lib.pyx":895
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             if isinstance(value, int):
*/
            }
          }

          /* "adbc_driver_manager/_lib.pyx":892
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L15;
            }
            __pyx_L15:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":885
 *                 value = value.value
 * 
 *             if isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
    }

    /* "adbc_driver_manager/_lib.pyx":897
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             if isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_value_int = value
 *                 with nogil:
*/
    __pyx_t_12 = PyLong_Check(__pyx_v_value); 
    if (__pyx_t_12) {

      /* "adbc_driver_manager/_lib.pyx":898
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             if isinstance(value, int):
 *                 c_value_int = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionInt(
*/
      __pyx_t_14 = __Pyx_PyLong_As_int64_t(__pyx_v_value); if (unlikely((__pyx_t_14 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L1_error)
      __pyx_v_c_value_int = __pyx_t_14;

      /* "adbc_driver_manager/_lib.pyx":899
 *             if isinstance(value, int):
 *                 c_value_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionInt(
 *                         &self.connection, c_key, c_value_int, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":900
 *                 c_value_int = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionInt(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_value_int, &c_error)
 *                     if status != ADBC_STATUS_OK:
*/
            __pyx_v_status = AdbcConnectionSetOptionInt((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_value_int, (&__pyx_v_c_error));

            /* "adbc_driver_manager/_lib.pyx":902
 *                     status = AdbcConnectionSetOptionInt(
 *                         &self.connection, c_key, c_value_int, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, float):
*/
            __pyx_t_12 = (__pyx_v_status != ADBC_STATUS_OK);
            if (__pyx_t_12) {

              /* "adbc_driver_manager/_lib.pyx":903
 *                         &self.connection, c_key, c_value_int, &c_error)
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)             # <<<<<<<<<<<<<<
 *             elif isinstance(value, float):
 *                 c_value_double = value
*/
              (void)(AdbcConnectionRelease((&__pyx_v_self->connection), NULL));

              /* "adbc_driver_manager/_lib.pyx":902
 *                     status = AdbcConnectionSetOptionInt(
 *                         &self.connection, c_key, c_value_int, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, float):
*/
            }
          }

          /* "adbc_driver_manager/_lib.pyx":899
 *             if isinstance(value, int):
 *                 c_value_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionInt(
 *                         &self.connection, c_key, c_value_int, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L22;
            }
            __pyx_L22:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":897
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             if isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_value_int = value
 *                 with nogil:
*/
      goto __pyx_L17;
    }

    /* "adbc_driver_manager/_lib.pyx":904
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_value_double = value
 *                 with nogil:
*/
    __pyx_t_12 = PyFloat_Check(__pyx_v_value); 
    if (__pyx_t_12) {

      /* "adbc_driver_manager/_lib.pyx":905
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, float):
 *                 c_value_double = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionDouble(
*/
      __pyx_t_15 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L1_error)
      __pyx_v_c_value_double = __pyx_t_15;

      /* "adbc_driver_manager/_lib.pyx":906
 *             elif isinstance(value, float):
 *                 c_value_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_value_double, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":907
 *                 c_value_double = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionDouble(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_value_double, &c_error)
 *                     if status != ADBC_STATUS_OK:
*/
            __pyx_v_status = AdbcConnectionSetOptionDouble((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_value_double, (&__pyx_v_c_error));

            /* "adbc_driver_manager/_lib.pyx":909
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_value_double, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, bytes):
*/
            __pyx_t_12 = (__pyx_v_status != ADBC_STATUS_OK);
            if (__pyx_t_12) {

              /* "adbc_driver_manager/_lib.pyx":910
 *                         &self.connection, c_key, c_value_double, &c_error)
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)             # <<<<<<<<<<<<<<
 *             elif isinstance(value, bytes):
 *                 c_value = value
*/
              (void)(AdbcConnectionRelease((&__pyx_v_self->connection), NULL));

              /* "adbc_driver_manager/_lib.pyx":909
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_value_double, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, bytes):
*/
            }
          }

          /* "adbc_driver_manager/_lib.pyx":906
 *             elif isinstance(value, float):
 *                 c_value_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_value_double, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L28;
            }
            __pyx_L28:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":904
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_value_double = value
 *                 with nogil:
*/
      goto __pyx_L17;
    }

    /* "adbc_driver_manager/_lib.pyx":911
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
    __pyx_t_12 = PyBytes_Check(__pyx_v_value); 
    if (__pyx_t_12) {

      /* "adbc_driver_manager/_lib.pyx":912
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, bytes):
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 c_value_len = len(value)
 *                 with nogil:
*/
      __pyx_t_13 = __Pyx_PyObject_AsString(__pyx_v_value); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_13;

      /* "adbc_driver_manager/_lib.pyx":913
 *             elif isinstance(value, bytes):
 *                 c_value = value
 *                 c_value_len = len(value)             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionBytes(
*/
      __pyx_t_16 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 913, __pyx_L1_error)
      __pyx_v_c_value_len = __pyx_t_16;

      /* "adbc_driver_manager/_lib.pyx":914
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionBytes(
 *                         &self.connection, c_key, <const uint8_t*> c_value,
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":915
 *                 c_value_len = len(value)
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionBytes(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
*/
            __pyx_v_status = AdbcConnectionSetOptionBytes((&__pyx_v_self->connection), __pyx_v_c_key, ((uint8_t const *)__pyx_v_c_value), __pyx_v_c_value_len, (&__pyx_v_c_error));

            /* "adbc_driver_manager/_lib.pyx":918
 *                         &self.connection, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             else:
*/
            __pyx_t_12 = (__pyx_v_status != ADBC_STATUS_OK);
            if (__pyx_t_12) {

              /* "adbc_driver_manager/_lib.pyx":919
 *                         c_value_len, &c_error)
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)             # <<<<<<<<<<<<<<
 *             else:
 *                 value = _to_bytes(value, "value")
*/
              (void)(AdbcConnectionRelease((&__pyx_v_self->connection), NULL));

              /* "adbc_driver_manager/_lib.pyx":918
 *                         &self.connection, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             else:
*/
            }
          }

          /* "adbc_driver_manager/_lib.pyx":914
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionBytes(
 *                         &self.connection, c_key, <const uint8_t*> c_value,
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L34;
            }
            __pyx_L34:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":911
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
      goto __pyx_L17;
    }

    /* "adbc_driver_manager/_lib.pyx":921
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             else:
 *                 value = _to_bytes(value, "value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 with nogil:
*/
    /*else*/ {
      __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 921, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "adbc_driver_manager/_lib.pyx":922
 *             else:
 *                 value = _to_bytes(value, "value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(
*/
      __pyx_t_13 = __Pyx_PyObject_AsString(__pyx_v_value); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(0, 922, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_13;

      /* "adbc_driver_manager/_lib.pyx":923
 *                 value = _to_bytes(value, "value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":924
 *                 c_value = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:
*/
            __pyx_v_status = AdbcConnectionSetOption((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));

            /* "adbc_driver_manager/_lib.pyx":926
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             check_error(status, &c_error)
*/
            __pyx_t_12 = (__pyx_v_status != ADBC_STATUS_OK);
            if (__pyx_t_12) {

              /* "adbc_driver_manager/_lib.pyx":927
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)             # <<<<<<<<<<<<<<
 *             check_error(status, &c_error)
 * 
*/
              (void)(AdbcConnectionRelease((&__pyx_v_self->connection), NULL));

              /* "adbc_driver_manager/_lib.pyx":926
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *                     if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             check_error(status, &c_error)
*/
            }
          }

          /* "adbc_driver_manager/_lib.pyx":923
 *                 value = _to_bytes(value, "value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L40;
            }
            __pyx_L40:;
          }
      }
    }
    __pyx_L17:;

    /* "adbc_driver_manager/_lib.pyx":928
 *                     if status != ADBC_STATUS_OK:
 *                         AdbcConnectionRelease(&self.connection, NULL)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 928, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":930
 *             check_error(status, &c_error)
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionInit(&self.connection, &database.database, &c_error)
 *             if status != ADBC_STATUS_OK:
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":931
 * 
 *         with nogil:
 *             status = AdbcConnectionInit(&self.connection, &database.database, &c_error)             # <<<<<<<<<<<<<<
 *             if status != ADBC_STATUS_OK:
 *                 AdbcConnectionRelease(&self.connection, NULL)
*/
        __pyx_v_status = AdbcConnectionInit((&__pyx_v_self->connection), (&__pyx_v_database->database), (&__pyx_v_c_error));

        /* "adbc_driver_manager/_lib.pyx":932
 *         with nogil:
 *             status = AdbcConnectionInit(&self.connection, &database.database, &c_error)
 *             if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                 AdbcConnectionRelease(&self.connection, NULL)
 *         check_error(status, &c_error)
*/
        __pyx_t_12 = (__pyx_v_status != ADBC_STATUS_OK);
        if (__pyx_t_12) {

          /* "adbc_driver_manager/_lib.pyx":933
 *             status = AdbcConnectionInit(&self.connection, &database.database, &c_error)
 *             if status != ADBC_STATUS_OK:
 *                 AdbcConnectionRelease(&self.connection, NULL)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
          (void)(AdbcConnectionRelease((&__pyx_v_self->connection), NULL));

          /* "adbc_driver_manager/_lib.pyx":932
 *         with nogil:
 *             status = AdbcConnectionInit(&self.connection, &database.database, &c_error)
 *             if status != ADBC_STATUS_OK:             # <<<<<<<<<<<<<<
 *                 AdbcConnectionRelease(&self.connection, NULL)
 *         check_error(status, &c_error)
*/
        }
      }

      /* "adbc_driver_manager/_lib.pyx":930
 *             check_error(status, &c_error)
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionInit(&self.connection, &database.database, &c_error)
 *             if status != ADBC_STATUS_OK:
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L44;
        }
        __pyx_L44:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":934
 *             if status != ADBC_STATUS_OK:
 *                 AdbcConnectionRelease(&self.connection, NULL)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         database._open_child()
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 934, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":936
 *         check_error(status, &c_error)
 * 
 *         database._open_child()             # <<<<<<<<<<<<<<
 * 
 *     def cancel(self) -> None:
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_database->__pyx_base.__pyx_vtab)->__pyx_base._open_child(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_database)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":859
 *         CAdbcConnection connection
 * 
 *     def __init__(self, AdbcDatabase database, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         super().__init__("AdbcStatement")
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":938
 *         database._open_child()
 * 
 *     def cancel(self) -> None:             # <<<<<<<<<<<<<<
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_3cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_2cancel, "Attempt to cancel any ongoing operations on the connection.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_3cancel = {"cancel", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_3cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_2cancel};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_3cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("cancel", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("cancel", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_2cancel(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_2cancel(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 0);

  /* "adbc_driver_manager/_lib.pyx":940
 *     def cancel(self) -> None:
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         with nogil:
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 940, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":942
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionCancel(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":943
 *         cdef CAdbcStatusCode status
 *         with nogil:
 *             status = AdbcConnectionCancel(&self.connection, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcConnectionCancel((&__pyx_v_self->connection), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":942
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionCancel(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":944
 *         with nogil:
 *             status = AdbcConnectionCancel(&self.connection, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def commit(self) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 944, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":938
 *         database._open_child()
 * 
 *     def cancel(self) -> None:             # <<<<<<<<<<<<<<
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":946
 *         check_error(status, &c_error)
 * 
 *     def commit(self) -> None:             # <<<<<<<<<<<<<<
 *         """Commit the current transaction."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_5commit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_4commit, "Commit the current transaction.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_5commit = {"commit", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_5commit, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_4commit};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_5commit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("commit (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("commit", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("commit", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_4commit(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_4commit(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("commit", 0);

  /* "adbc_driver_manager/_lib.pyx":948
 *     def commit(self) -> None:
 *         """Commit the current transaction."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         with nogil:
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":950
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionCommit(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":951
 *         cdef CAdbcStatusCode status
 *         with nogil:
 *             status = AdbcConnectionCommit(&self.connection, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcConnectionCommit((&__pyx_v_self->connection), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":950
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionCommit(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":952
 *         with nogil:
 *             status = AdbcConnectionCommit(&self.connection, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def get_info(self, info_codes=None) -> ArrowArrayStreamHandle:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 952, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":946
 *         check_error(status, &c_error)
 * 
 *     def commit(self) -> None:             # <<<<<<<<<<<<<<
 *         """Commit the current transaction."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.commit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":954
 *         check_error(status, &c_error)
 * 
 *     def get_info(self, info_codes=None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get metadata about the database/driver.
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_7get_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_6get_info, "\n        Get metadata about the database/driver.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_7get_info = {"get_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_7get_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_6get_info};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_7get_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_info_codes = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_info_codes,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 954, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 954, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_info", 0) < (0)) __PYX_ERR(0, 954, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 954, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_info_codes = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_info", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 954, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_6get_info(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_info_codes);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_6get_info(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_info_codes) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_stream = 0;
  std::vector<uint32_t>  __pyx_v_c_info_codes;
  PyObject *__pyx_v_info_code = NULL;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  uint32_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_info", 0);

  /* "adbc_driver_manager/_lib.pyx":958
 *         Get metadata about the database/driver.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 958, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":960
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()             # <<<<<<<<<<<<<<
 *         cdef c_vector[uint32_t] c_info_codes
 * 
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 960, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_stream = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":963
 *         cdef c_vector[uint32_t] c_info_codes
 * 
 *         if info_codes:             # <<<<<<<<<<<<<<
 *             for info_code in info_codes:
 *                 if isinstance(info_code, int):
*/
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_info_codes); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 963, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "adbc_driver_manager/_lib.pyx":964
 * 
 *         if info_codes:
 *             for info_code in info_codes:             # <<<<<<<<<<<<<<
 *                 if isinstance(info_code, int):
 *                     c_info_codes.push_back(info_code)
*/
    if (likely(PyList_CheckExact(__pyx_v_info_codes)) || PyTuple_CheckExact(__pyx_v_info_codes)) {
      __pyx_t_2 = __pyx_v_info_codes; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_info_codes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 964, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 964, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 964, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          __pyx_t_3 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_6;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 964, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6));
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6);
          #endif
          ++__pyx_t_6;
        }
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 964, __pyx_L1_error)
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 964, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_info_code, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "adbc_driver_manager/_lib.pyx":965
 *         if info_codes:
 *             for info_code in info_codes:
 *                 if isinstance(info_code, int):             # <<<<<<<<<<<<<<
 *                     c_info_codes.push_back(info_code)
 *                 else:
*/
      __pyx_t_5 = PyLong_Check(__pyx_v_info_code); 
      if (__pyx_t_5) {

        /* "adbc_driver_manager/_lib.pyx":966
 *             for info_code in info_codes:
 *                 if isinstance(info_code, int):
 *                     c_info_codes.push_back(info_code)             # <<<<<<<<<<<<<<
 *                 else:
 *                     c_info_codes.push_back(info_code.value)
*/
        __pyx_t_8 = __Pyx_PyLong_As_uint32_t(__pyx_v_info_code); if (unlikely((__pyx_t_8 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L1_error)
        try {
          __pyx_v_c_info_codes.push_back(__pyx_t_8);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 966, __pyx_L1_error)
        }

        /* "adbc_driver_manager/_lib.pyx":965
 *         if info_codes:
 *             for info_code in info_codes:
 *                 if isinstance(info_code, int):             # <<<<<<<<<<<<<<
 *                     c_info_codes.push_back(info_code)
 *                 else:
*/
        goto __pyx_L6;
      }

      /* "adbc_driver_manager/_lib.pyx":968
 *                     c_info_codes.push_back(info_code)
 *                 else:
 *                     c_info_codes.push_back(info_code.value)             # <<<<<<<<<<<<<<
 * 
 *             with nogil:
*/
      /*else*/ {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_info_code, __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 968, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = __Pyx_PyLong_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_8 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 968, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        try {
          __pyx_v_c_info_codes.push_back(__pyx_t_8);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 968, __pyx_L1_error)
        }
      }
      __pyx_L6:;

      /* "adbc_driver_manager/_lib.pyx":964
 * 
 *         if info_codes:
 *             for info_code in info_codes:             # <<<<<<<<<<<<<<
 *                 if isinstance(info_code, int):
 *                     c_info_codes.push_back(info_code)
*/
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":970
 *                     c_info_codes.push_back(info_code.value)
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetInfo(
 *                     &self.connection,
*/
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":971
 * 
 *             with nogil:
 *                 status = AdbcConnectionGetInfo(             # <<<<<<<<<<<<<<
 *                     &self.connection,
 *                     c_info_codes.data(),
*/
          __pyx_v_status = AdbcConnectionGetInfo((&__pyx_v_self->connection), __pyx_v_c_info_codes.data(), __pyx_v_c_info_codes.size(), (&__pyx_v_stream->stream), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":970
 *                     c_info_codes.push_back(info_code.value)
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetInfo(
 *                     &self.connection,
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L10;
          }
          __pyx_L10:;
        }
    }

    /* "adbc_driver_manager/_lib.pyx":963
 *         cdef c_vector[uint32_t] c_info_codes
 * 
 *         if info_codes:             # <<<<<<<<<<<<<<
 *             for info_code in info_codes:
 *                 if isinstance(info_code, int):
*/
    goto __pyx_L3;
  }

  /* "adbc_driver_manager/_lib.pyx":978
 *                     &c_error)
 *         else:
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetInfo(
 *                     &self.connection,
*/
  /*else*/ {
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":979
 *         else:
 *             with nogil:
 *                 status = AdbcConnectionGetInfo(             # <<<<<<<<<<<<<<
 *                     &self.connection,
 *                     NULL,
*/
          __pyx_v_status = AdbcConnectionGetInfo((&__pyx_v_self->connection), NULL, 0, (&__pyx_v_stream->stream), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":978
 *                     &c_error)
 *         else:
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetInfo(
 *                     &self.connection,
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L13;
          }
          __pyx_L13:;
        }
    }
  }
  __pyx_L3:;

  /* "adbc_driver_manager/_lib.pyx":986
 *                     &c_error)
 * 
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return stream
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 986, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":987
 * 
 *         check_error(status, &c_error)
 *         return stream             # <<<<<<<<<<<<<<
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __pyx_r = __pyx_v_stream;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":954
 *         check_error(status, &c_error)
 * 
 *     def get_info(self, info_codes=None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get metadata about the database/driver.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XDECREF(__pyx_v_info_code);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":989
 *         return stream
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,             # <<<<<<<<<<<<<<
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:
 *         """
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_9get_objects(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_8get_objects, "\n        Get a hierarchical view of database objects.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_9get_objects = {"get_objects", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_9get_objects, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_8get_objects};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_9get_objects(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_depth = 0;
  PyObject *__pyx_v_catalog = 0;
  PyObject *__pyx_v_db_schema = 0;
  PyObject *__pyx_v_table_name = 0;
  CYTHON_UNUSED PyObject *__pyx_v_table_types = 0;
  PyObject *__pyx_v_column_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_objects (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_depth,&__pyx_mstate_global->__pyx_n_u_catalog,&__pyx_mstate_global->__pyx_n_u_db_schema,&__pyx_mstate_global->__pyx_n_u_table_name,&__pyx_mstate_global->__pyx_n_u_table_types,&__pyx_mstate_global->__pyx_n_u_column_name,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 989, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_objects", 0) < (0)) __PYX_ERR(0, 989, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":990
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get a hierarchical view of database objects.
*/
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_objects", 0, 1, 6, i); __PYX_ERR(0, 989, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 989, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 989, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }

      /* "adbc_driver_manager/_lib.pyx":989
 *         return stream
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,             # <<<<<<<<<<<<<<
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:
 *         """
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":990
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get a hierarchical view of database objects.
*/
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_depth = values[0];
    __pyx_v_catalog = values[1];
    __pyx_v_db_schema = values[2];
    __pyx_v_table_name = values[3];
    __pyx_v_table_types = values[4];
    __pyx_v_column_name = values[5];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_objects", 0, 1, 6, __pyx_nargs); __PYX_ERR(0, 989, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_objects", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_8get_objects(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_depth, __pyx_v_catalog, __pyx_v_db_schema, __pyx_v_table_name, __pyx_v_table_types, __pyx_v_column_name);

  /* "adbc_driver_manager/_lib.pyx":989
 *         return stream
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,             # <<<<<<<<<<<<<<
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:
 *         """
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_8get_objects(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_depth, PyObject *__pyx_v_catalog, PyObject *__pyx_v_db_schema, PyObject *__pyx_v_table_name, CYTHON_UNUSED PyObject *__pyx_v_table_types, PyObject *__pyx_v_column_name) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_stream = 0;
  int __pyx_v_c_depth;
  char *__pyx_v_c_catalog;
  char *__pyx_v_c_db_schema;
  char *__pyx_v_c_table_name;
  char *__pyx_v_c_column_name;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  char *__pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_objects", 0);
  __Pyx_INCREF(__pyx_v_catalog);
  __Pyx_INCREF(__pyx_v_db_schema);
  __Pyx_INCREF(__pyx_v_table_name);
  __Pyx_INCREF(__pyx_v_column_name);

  /* "adbc_driver_manager/_lib.pyx":994
 *         Get a hierarchical view of database objects.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 994, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":996
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()             # <<<<<<<<<<<<<<
 *         cdef int c_depth = GetObjectsDepth(depth).value
 * 
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_stream = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":997
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 *         cdef int c_depth = GetObjectsDepth(depth).value             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_catalog = NULL
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_GetObjectsDepth); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_depth};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_t_5); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_c_depth = __pyx_t_6;

  /* "adbc_driver_manager/_lib.pyx":999
 *         cdef int c_depth = GetObjectsDepth(depth).value
 * 
 *         cdef char* c_catalog = NULL             # <<<<<<<<<<<<<<
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")
*/
  __pyx_v_c_catalog = NULL;

  /* "adbc_driver_manager/_lib.pyx":1000
 * 
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:             # <<<<<<<<<<<<<<
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog
*/
  __pyx_t_7 = (__pyx_v_catalog != Py_None);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":1001
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")             # <<<<<<<<<<<<<<
 *             c_catalog = catalog
 * 
*/
    __pyx_t_5 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_catalog, __pyx_mstate_global->__pyx_n_u_catalog); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF_SET(__pyx_v_catalog, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "adbc_driver_manager/_lib.pyx":1002
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_db_schema = NULL
*/
    __pyx_t_8 = __Pyx_PyObject_AsWritableString(__pyx_v_catalog); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1002, __pyx_L1_error)
    __pyx_v_c_catalog = __pyx_t_8;

    /* "adbc_driver_manager/_lib.pyx":1000
 * 
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:             # <<<<<<<<<<<<<<
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1004
 *             c_catalog = catalog
 * 
 *         cdef char* c_db_schema = NULL             # <<<<<<<<<<<<<<
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")
*/
  __pyx_v_c_db_schema = NULL;

  /* "adbc_driver_manager/_lib.pyx":1005
 * 
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:             # <<<<<<<<<<<<<<
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema
*/
  __pyx_t_7 = (__pyx_v_db_schema != Py_None);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":1006
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")             # <<<<<<<<<<<<<<
 *             c_db_schema = db_schema
 * 
*/
    __pyx_t_5 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_db_schema, __pyx_mstate_global->__pyx_n_u_db_schema); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF_SET(__pyx_v_db_schema, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "adbc_driver_manager/_lib.pyx":1007
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_table_name = NULL
*/
    __pyx_t_8 = __Pyx_PyObject_AsWritableString(__pyx_v_db_schema); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1007, __pyx_L1_error)
    __pyx_v_c_db_schema = __pyx_t_8;

    /* "adbc_driver_manager/_lib.pyx":1005
 * 
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:             # <<<<<<<<<<<<<<
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1009
 *             c_db_schema = db_schema
 * 
 *         cdef char* c_table_name = NULL             # <<<<<<<<<<<<<<
 *         if table_name is not None:
 *             table_name = _to_bytes(table_name, "table_name")
*/
  __pyx_v_c_table_name = NULL;

  /* "adbc_driver_manager/_lib.pyx":1010
 * 
 *         cdef char* c_table_name = NULL
 *         if table_name is not None:             # <<<<<<<<<<<<<<
 *             table_name = _to_bytes(table_name, "table_name")
 *             c_table_name = table_name
*/
  __pyx_t_7 = (__pyx_v_table_name != Py_None);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":1011
 *         cdef char* c_table_name = NULL
 *         if table_name is not None:
 *             table_name = _to_bytes(table_name, "table_name")             # <<<<<<<<<<<<<<
 *             c_table_name = table_name
 * 
*/
    __pyx_t_5 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_table_name, __pyx_mstate_global->__pyx_n_u_table_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF_SET(__pyx_v_table_name, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "adbc_driver_manager/_lib.pyx":1012
 *         if table_name is not None:
 *             table_name = _to_bytes(table_name, "table_name")
 *             c_table_name = table_name             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_column_name = NULL
*/
    __pyx_t_8 = __Pyx_PyObject_AsWritableString(__pyx_v_table_name); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1012, __pyx_L1_error)
    __pyx_v_c_table_name = __pyx_t_8;

    /* "adbc_driver_manager/_lib.pyx":1010
 * 
 *         cdef char* c_table_name = NULL
 *         if table_name is not None:             # <<<<<<<<<<<<<<
 *             table_name = _to_bytes(table_name, "table_name")
 *             c_table_name = table_name
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1014
 *             c_table_name = table_name
 * 
 *         cdef char* c_column_name = NULL             # <<<<<<<<<<<<<<
 *         if column_name is not None:
 *             column_name = _to_bytes(column_name, "column_name")
*/
  __pyx_v_c_column_name = NULL;

  /* "adbc_driver_manager/_lib.pyx":1015
 * 
 *         cdef char* c_column_name = NULL
 *         if column_name is not None:             # <<<<<<<<<<<<<<
 *             column_name = _to_bytes(column_name, "column_name")
 *             c_column_name = column_name
*/
  __pyx_t_7 = (__pyx_v_column_name != Py_None);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":1016
 *         cdef char* c_column_name = NULL
 *         if column_name is not None:
 *             column_name = _to_bytes(column_name, "column_name")             # <<<<<<<<<<<<<<
 *             c_column_name = column_name
 * 
*/
    __pyx_t_5 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_column_name, __pyx_mstate_global->__pyx_n_u_column_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF_SET(__pyx_v_column_name, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "adbc_driver_manager/_lib.pyx":1017
 *         if column_name is not None:
 *             column_name = _to_bytes(column_name, "column_name")
 *             c_column_name = column_name             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
    __pyx_t_8 = __Pyx_PyObject_AsWritableString(__pyx_v_column_name); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L1_error)
    __pyx_v_c_column_name = __pyx_t_8;

    /* "adbc_driver_manager/_lib.pyx":1015
 * 
 *         cdef char* c_column_name = NULL
 *         if column_name is not None:             # <<<<<<<<<<<<<<
 *             column_name = _to_bytes(column_name, "column_name")
 *             c_column_name = column_name
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1019
 *             c_column_name = column_name
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetObjects(
 *                 &self.connection,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1020
 * 
 *         with nogil:
 *             status = AdbcConnectionGetObjects(             # <<<<<<<<<<<<<<
 *                 &self.connection,
 *                 c_depth,
*/
        __pyx_v_status = AdbcConnectionGetObjects((&__pyx_v_self->connection), __pyx_v_c_depth, __pyx_v_c_catalog, __pyx_v_c_db_schema, __pyx_v_c_table_name, NULL, __pyx_v_c_column_name, (&__pyx_v_stream->stream), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1019
 *             c_column_name = column_name
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetObjects(
 *                 &self.connection,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1030
 *                 &stream.stream,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         return stream
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1030, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1032
 *         check_error(status, &c_error)
 * 
 *         return stream             # <<<<<<<<<<<<<<
 * 
 *     def get_option(
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __pyx_r = __pyx_v_stream;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":989
 *         return stream
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,             # <<<<<<<<<<<<<<
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:
 *         """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_objects", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XDECREF(__pyx_v_catalog);
  __Pyx_XDECREF(__pyx_v_db_schema);
  __Pyx_XDECREF(__pyx_v_table_name);
  __Pyx_XDECREF(__pyx_v_column_name);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1034
 *         return stream
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_11get_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_10get_option, "\n        Get the value of a string option.\n\n        Parameters\n        ----------\n        key : str or bytes\n            The option to get.\n        encoding : str\n            The encoding of the option value.  This should almost\n            always be UTF-8.\n        errors : str\n            What to do about errors when decoding the option value\n            (see bytes.decode).\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_11get_option = {"get_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_11get_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_10get_option};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_11get_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_encoding = 0;
  PyObject *__pyx_v_errors = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,&__pyx_mstate_global->__pyx_n_u_encoding,&__pyx_mstate_global->__pyx_n_u_errors,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1034, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1034, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option", 0) < (0)) __PYX_ERR(0, 1034, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_kp_u_utf_8));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_strict));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option", 1, 1, 1, i); __PYX_ERR(0, 1034, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1034, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_kp_u_utf_8));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_strict));
    }
    __pyx_v_key = values[0];
    __pyx_v_encoding = values[1];
    __pyx_v_errors = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1034, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_10get_option(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_key, __pyx_v_encoding, __pyx_v_errors);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_10get_option(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  size_t __pyx_v_c_len;
  char *__pyx_v_c_buf;
  PyObject *__pyx_v_buf = NULL;
  Py_ssize_t __pyx_v_new_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10[5];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option", 0);

  /* "adbc_driver_manager/_lib.pyx":1055
 *             (see bytes.decode).
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1055, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1057
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1058
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef size_t c_len = 0
 *         cdef char* c_buf = NULL
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1058, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1059
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0             # <<<<<<<<<<<<<<
 *         cdef char* c_buf = NULL
 * 
*/
  __pyx_v_c_len = 0;

  /* "adbc_driver_manager/_lib.pyx":1060
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
 *         cdef char* c_buf = NULL             # <<<<<<<<<<<<<<
 * 
 *         buf = bytearray(1024)
*/
  __pyx_v_c_buf = NULL;

  /* "adbc_driver_manager/_lib.pyx":1062
 *         cdef char* c_buf = NULL
 * 
 *         buf = bytearray(1024)             # <<<<<<<<<<<<<<
 *         while True:
 *             c_len = len(buf)
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_int_1024};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_buf = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1063
 * 
 *         buf = bytearray(1024)
 *         while True:             # <<<<<<<<<<<<<<
 *             c_len = len(buf)
 *             c_buf = buf
*/
  while (1) {

    /* "adbc_driver_manager/_lib.pyx":1064
 *         buf = bytearray(1024)
 *         while True:
 *             c_len = len(buf)             # <<<<<<<<<<<<<<
 *             c_buf = buf
 *             with nogil:
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1064, __pyx_L1_error)
    __pyx_v_c_len = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":1065
 *         while True:
 *             c_len = len(buf)
 *             c_buf = buf             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 status = AdbcConnectionGetOption(
*/
    __pyx_t_3 = __Pyx_PyObject_AsWritableString(__pyx_v_buf); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1065, __pyx_L1_error)
    __pyx_v_c_buf = __pyx_t_3;

    /* "adbc_driver_manager/_lib.pyx":1066
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetOption(
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
*/
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1067
 *             c_buf = buf
 *             with nogil:
 *                 status = AdbcConnectionGetOption(             # <<<<<<<<<<<<<<
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
*/
          __pyx_v_status = AdbcConnectionGetOption((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_buf, (&__pyx_v_c_len), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":1066
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetOption(
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "adbc_driver_manager/_lib.pyx":1069
 *                 status = AdbcConnectionGetOption(
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *             if c_len <= len(buf):
 *                 # Entire value read
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":1070
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1070, __pyx_L1_error)
    __pyx_t_7 = (__pyx_v_c_len <= __pyx_t_6);
    if (__pyx_t_7) {

      /* "adbc_driver_manager/_lib.pyx":1072
 *             if c_len <= len(buf):
 *                 # Entire value read
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 # Buffer too small
*/
      goto __pyx_L4_break;

      /* "adbc_driver_manager/_lib.pyx":1070
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    }

    /* "adbc_driver_manager/_lib.pyx":1075
 *             else:
 *                 # Buffer too small
 *                 new_len = len(buf) * 2             # <<<<<<<<<<<<<<
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
*/
    /*else*/ {
      __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1075, __pyx_L1_error)
      __pyx_v_new_len = (__pyx_t_6 * 2);

      /* "adbc_driver_manager/_lib.pyx":1076
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      __pyx_t_2 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MAX_OPTION_SIZE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1076, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1076, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__pyx_t_7)) {

        /* "adbc_driver_manager/_lib.pyx":1077
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
*/
        __pyx_t_4 = NULL;

        /* "adbc_driver_manager/_lib.pyx":1078
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "adbc_driver_manager/_lib.pyx":1079
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")             # <<<<<<<<<<<<<<
 *                 buf = bytearray(new_len)
 * 
*/
        __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1079, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_read_option;
        __pyx_t_10[1] = __pyx_t_2;
        __pyx_t_10[2] = __pyx_mstate_global->__pyx_kp_u_would_need_more_than;
        __pyx_t_10[3] = __pyx_t_9;
        __pyx_t_10[4] = __pyx_mstate_global->__pyx_kp_u_bytes;

        /* "adbc_driver_manager/_lib.pyx":1078
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 5, 22 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 6, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1078, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
          __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1077, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_Raise(__pyx_t_8, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 1077, __pyx_L1_error)

        /* "adbc_driver_manager/_lib.pyx":1076
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      }

      /* "adbc_driver_manager/_lib.pyx":1080
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)             # <<<<<<<<<<<<<<
 * 
 *         # Remove trailing null terminator
*/
      __pyx_t_11 = NULL;
      __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1080, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_4};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1080, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_DECREF_SET(__pyx_v_buf, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;
    }
  }
  __pyx_L4_break:;

  /* "adbc_driver_manager/_lib.pyx":1083
 * 
 *         # Remove trailing null terminator
 *         if c_len > 0:             # <<<<<<<<<<<<<<
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)
*/
  __pyx_t_7 = (__pyx_v_c_len > 0);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":1084
 *         # Remove trailing null terminator
 *         if c_len > 0:
 *             c_len -= 1             # <<<<<<<<<<<<<<
 *         return buf[:c_len].decode(encoding, errors)
 * 
*/
    __pyx_v_c_len = (__pyx_v_c_len - 1);

    /* "adbc_driver_manager/_lib.pyx":1083
 * 
 *         # Remove trailing null terminator
 *         if c_len > 0:             # <<<<<<<<<<<<<<
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1085
 *         if c_len > 0:
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)             # <<<<<<<<<<<<<<
 * 
 *     def get_option_bytes(self, key: str) -> bytes:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_11 = PySequence_GetSlice(__pyx_v_buf, 0, __pyx_v_c_len); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = __pyx_t_11;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_encoding, __pyx_v_errors};
    __pyx_t_8 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_decode, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
  }
  __pyx_r = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1034
 *         return stream
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1087
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_13get_option_bytes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_12get_option_bytes, "Get the value of a binary option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_13get_option_bytes = {"get_option_bytes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_13get_option_bytes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_12get_option_bytes};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_13get_option_bytes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_bytes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1087, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1087, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_bytes", 0) < (0)) __PYX_ERR(0, 1087, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_bytes", 1, 1, 1, i); __PYX_ERR(0, 1087, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1087, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_bytes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1087, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 1087, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_12get_option_bytes(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_12get_option_bytes(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  size_t __pyx_v_c_len;
  uint8_t *__pyx_v_c_buf;
  PyObject *__pyx_v_buf = NULL;
  Py_ssize_t __pyx_v_new_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  uint8_t *__pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10[5];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_bytes", 0);

  /* "adbc_driver_manager/_lib.pyx":1089
 *     def get_option_bytes(self, key: str) -> bytes:
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1089, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1091
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1092
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef size_t c_len = 0
 *         cdef uint8_t* c_buf = NULL
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1092, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1093
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0             # <<<<<<<<<<<<<<
 *         cdef uint8_t* c_buf = NULL
 * 
*/
  __pyx_v_c_len = 0;

  /* "adbc_driver_manager/_lib.pyx":1094
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
 *         cdef uint8_t* c_buf = NULL             # <<<<<<<<<<<<<<
 * 
 *         buf = bytearray(1024)
*/
  __pyx_v_c_buf = NULL;

  /* "adbc_driver_manager/_lib.pyx":1096
 *         cdef uint8_t* c_buf = NULL
 * 
 *         buf = bytearray(1024)             # <<<<<<<<<<<<<<
 *         while True:
 *             c_len = len(buf)
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_int_1024};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_buf = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1097
 * 
 *         buf = bytearray(1024)
 *         while True:             # <<<<<<<<<<<<<<
 *             c_len = len(buf)
 *             c_buf = buf
*/
  while (1) {

    /* "adbc_driver_manager/_lib.pyx":1098
 *         buf = bytearray(1024)
 *         while True:
 *             c_len = len(buf)             # <<<<<<<<<<<<<<
 *             c_buf = buf
 *             with nogil:
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1098, __pyx_L1_error)
    __pyx_v_c_len = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":1099
 *         while True:
 *             c_len = len(buf)
 *             c_buf = buf             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 status = AdbcConnectionGetOptionBytes(
*/
    __pyx_t_7 = __Pyx_PyObject_AsWritableUString(__pyx_v_buf); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 1099, __pyx_L1_error)
    __pyx_v_c_buf = __pyx_t_7;

    /* "adbc_driver_manager/_lib.pyx":1100
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetOptionBytes(
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
*/
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1101
 *             c_buf = buf
 *             with nogil:
 *                 status = AdbcConnectionGetOptionBytes(             # <<<<<<<<<<<<<<
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
*/
          __pyx_v_status = AdbcConnectionGetOptionBytes((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_buf, (&__pyx_v_c_len), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":1100
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionGetOptionBytes(
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "adbc_driver_manager/_lib.pyx":1103
 *                 status = AdbcConnectionGetOptionBytes(
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *             if c_len <= len(buf):
 *                 # Entire value read
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1103, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":1104
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1104, __pyx_L1_error)
    __pyx_t_8 = (__pyx_v_c_len <= __pyx_t_6);
    if (__pyx_t_8) {

      /* "adbc_driver_manager/_lib.pyx":1106
 *             if c_len <= len(buf):
 *                 # Entire value read
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 # Buffer too small
*/
      goto __pyx_L4_break;

      /* "adbc_driver_manager/_lib.pyx":1104
 *                     &self.connection, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    }

    /* "adbc_driver_manager/_lib.pyx":1109
 *             else:
 *                 # Buffer too small
 *                 new_len = len(buf) * 2             # <<<<<<<<<<<<<<
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
*/
    /*else*/ {
      __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1109, __pyx_L1_error)
      __pyx_v_new_len = (__pyx_t_6 * 2);

      /* "adbc_driver_manager/_lib.pyx":1110
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      __pyx_t_2 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MAX_OPTION_SIZE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1110, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1110, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__pyx_t_8)) {

        /* "adbc_driver_manager/_lib.pyx":1111
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
*/
        __pyx_t_4 = NULL;

        /* "adbc_driver_manager/_lib.pyx":1113
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")             # <<<<<<<<<<<<<<
 *                 buf = bytearray(new_len)
 * 
*/
        __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1113, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1113, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_read_option;
        __pyx_t_10[1] = __pyx_v_key;
        __pyx_t_10[2] = __pyx_mstate_global->__pyx_kp_u_would_need_more_than;
        __pyx_t_10[3] = __pyx_t_2;
        __pyx_t_10[4] = __pyx_mstate_global->__pyx_kp_u_bytes;

        /* "adbc_driver_manager/_lib.pyx":1112
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 5, 22 + __Pyx_PyUnicode_GET_LENGTH(__pyx_v_key) + 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 6, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_key));
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1112, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
          __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_Raise(__pyx_t_9, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 1111, __pyx_L1_error)

        /* "adbc_driver_manager/_lib.pyx":1110
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      }

      /* "adbc_driver_manager/_lib.pyx":1114
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)             # <<<<<<<<<<<<<<
 * 
 *         return bytes(buf[:c_len])
*/
      __pyx_t_11 = NULL;
      __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1114, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_4};
        __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1114, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_DECREF_SET(__pyx_v_buf, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;
    }
  }
  __pyx_L4_break:;

  /* "adbc_driver_manager/_lib.pyx":1116
 *                 buf = bytearray(new_len)
 * 
 *         return bytes(buf[:c_len])             # <<<<<<<<<<<<<<
 * 
 *     def get_option_float(self, key: str) -> float:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = NULL;
  __pyx_t_11 = PySequence_GetSlice(__pyx_v_buf, 0, __pyx_v_c_len); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
    __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
  }
  __pyx_r = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1087
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1118
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_15get_option_float(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_14get_option_float, "Get the value of a floating-point option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_15get_option_float = {"get_option_float", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_15get_option_float, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_14get_option_float};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_15get_option_float(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_float (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1118, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1118, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_float", 0) < (0)) __PYX_ERR(0, 1118, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_float", 1, 1, 1, i); __PYX_ERR(0, 1118, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1118, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_float", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1118, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 1118, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_14get_option_float(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_14get_option_float(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  double __pyx_v_c_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_float", 0);

  /* "adbc_driver_manager/_lib.pyx":1120
 *     def get_option_float(self, key: str) -> float:
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1120, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1122
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1123
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef double c_value = 0.0
 *         with nogil:
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1123, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1123, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1124
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcConnectionGetOptionDouble(
*/
  __pyx_v_c_value = 0.0;

  /* "adbc_driver_manager/_lib.pyx":1125
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetOptionDouble(
 *                 &self.connection, c_key, &c_value, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1126
 *         cdef double c_value = 0.0
 *         with nogil:
 *             status = AdbcConnectionGetOptionDouble(             # <<<<<<<<<<<<<<
 *                 &self.connection, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcConnectionGetOptionDouble((&__pyx_v_self->connection), __pyx_v_c_key, (&__pyx_v_c_value), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1125
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetOptionDouble(
 *                 &self.connection, c_key, &c_value, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1128
 *             status = AdbcConnectionGetOptionDouble(
 *                 &self.connection, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return c_value
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1128, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1129
 *                 &self.connection, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
 *         return c_value             # <<<<<<<<<<<<<<
 * 
 *     def get_option_int(self, key: str) -> int:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1118
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1131
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_17get_option_int(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_16get_option_int, "Get the value of an integer option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_17get_option_int = {"get_option_int", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_17get_option_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_16get_option_int};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_17get_option_int(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_int (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1131, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1131, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_int", 0) < (0)) __PYX_ERR(0, 1131, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_int", 1, 1, 1, i); __PYX_ERR(0, 1131, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1131, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_int", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1131, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 1131, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_16get_option_int(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_16get_option_int(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  int64_t __pyx_v_c_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_int", 0);

  /* "adbc_driver_manager/_lib.pyx":1133
 *     def get_option_int(self, key: str) -> int:
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1133, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1135
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1136
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef int64_t c_value = 0
 *         with nogil:
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1136, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1136, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1137
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcConnectionGetOptionInt(
*/
  __pyx_v_c_value = 0;

  /* "adbc_driver_manager/_lib.pyx":1138
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetOptionInt(
 *                 &self.connection, c_key, &c_value, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1139
 *         cdef int64_t c_value = 0
 *         with nogil:
 *             status = AdbcConnectionGetOptionInt(             # <<<<<<<<<<<<<<
 *                 &self.connection, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcConnectionGetOptionInt((&__pyx_v_self->connection), __pyx_v_c_key, (&__pyx_v_c_value), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1138
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetOptionInt(
 *                 &self.connection, c_key, &c_value, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1141
 *             status = AdbcConnectionGetOptionInt(
 *                 &self.connection, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return c_value
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1141, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1142
 *                 &self.connection, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
 *         return c_value             # <<<<<<<<<<<<<<
 * 
 *     def get_statistics(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_c_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyInt_FromNumber(&__pyx_t_2, NULL, 0) < (0)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1131
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_option_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1144
 *         return c_value
 * 
 *     def get_statistics(             # <<<<<<<<<<<<<<
 *         self,
 *         catalog=None,
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_19get_statistics(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_18get_statistics, "\n        Get statistics about data distribution in table(s).\n\n        Parameters\n        ----------\n        catalog : str, optional\n            The catalog to filter by. May be a search pattern or None.\n        db_schema : str, optional\n            The database schema to filter by. May be a search pattern or None.\n        table_name : str, optional\n            The table name to filter by. May be a search pattern or None.\n        approximate : bool, default True\n            If True, allow approximate or cached statistics. If False, request\n            exact statistics (may be expensive or unsupported).\n\n        Returns\n        -------\n        ArrowArrayStreamHandle\n            A stream of statistics data with nested structure.\n\n        Notes\n        -----\n        Available since ADBC API revision 1.1.0. Drivers may return\n        approximate values even when exact values are requested, as\n        indicated in the result's statistic_is_approximate column.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_19get_statistics = {"get_statistics", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_19get_statistics, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_18get_statistics};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_19get_statistics(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_catalog = 0;
  PyObject *__pyx_v_db_schema = 0;
  PyObject *__pyx_v_table_name = 0;
  PyObject *__pyx_v_approximate = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_statistics (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_catalog,&__pyx_mstate_global->__pyx_n_u_db_schema,&__pyx_mstate_global->__pyx_n_u_table_name,&__pyx_mstate_global->__pyx_n_u_approximate,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1144, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_statistics", 0) < (0)) __PYX_ERR(0, 1144, __pyx_L3_error)

      /* "adbc_driver_manager/_lib.pyx":1146
 *     def get_statistics(
 *         self,
 *         catalog=None,             # <<<<<<<<<<<<<<
 *         db_schema=None,
 *         table_name=None,
*/
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":1147
 *         self,
 *         catalog=None,
 *         db_schema=None,             # <<<<<<<<<<<<<<
 *         table_name=None,
 *         approximate=True
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":1148
 *         catalog=None,
 *         db_schema=None,
 *         table_name=None,             # <<<<<<<<<<<<<<
 *         approximate=True
 *     ) -> ArrowArrayStreamHandle:
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":1149
 *         db_schema=None,
 *         table_name=None,
 *         approximate=True             # <<<<<<<<<<<<<<
 *     ) -> ArrowArrayStreamHandle:
 *         """
*/
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_True));
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }

      /* "adbc_driver_manager/_lib.pyx":1146
 *     def get_statistics(
 *         self,
 *         catalog=None,             # <<<<<<<<<<<<<<
 *         db_schema=None,
 *         table_name=None,
*/
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":1147
 *         self,
 *         catalog=None,
 *         db_schema=None,             # <<<<<<<<<<<<<<
 *         table_name=None,
 *         approximate=True
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":1148
 *         catalog=None,
 *         db_schema=None,
 *         table_name=None,             # <<<<<<<<<<<<<<
 *         approximate=True
 *     ) -> ArrowArrayStreamHandle:
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "adbc_driver_manager/_lib.pyx":1149
 *         db_schema=None,
 *         table_name=None,
 *         approximate=True             # <<<<<<<<<<<<<<
 *     ) -> ArrowArrayStreamHandle:
 *         """
*/
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_True));
    }
    __pyx_v_catalog = values[0];
    __pyx_v_db_schema = values[1];
    __pyx_v_table_name = values[2];
    __pyx_v_approximate = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_statistics", 0, 0, 4, __pyx_nargs); __PYX_ERR(0, 1144, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_statistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_18get_statistics(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_catalog, __pyx_v_db_schema, __pyx_v_table_name, __pyx_v_approximate);

  /* "adbc_driver_manager/_lib.pyx":1144
 *         return c_value
 * 
 *     def get_statistics(             # <<<<<<<<<<<<<<
 *         self,
 *         catalog=None,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_18get_statistics(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_catalog, PyObject *__pyx_v_db_schema, PyObject *__pyx_v_table_name, PyObject *__pyx_v_approximate) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_stream = 0;
  char *__pyx_v_c_catalog;
  char *__pyx_v_c_db_schema;
  char *__pyx_v_c_table_name;
  char __pyx_v_c_approximate;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  char *__pyx_t_6;
  char __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_statistics", 0);
  __Pyx_INCREF(__pyx_v_catalog);
  __Pyx_INCREF(__pyx_v_db_schema);
  __Pyx_INCREF(__pyx_v_table_name);

  /* "adbc_driver_manager/_lib.pyx":1177
 *         indicated in the result's statistic_is_approximate column.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1177, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1179
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_catalog = NULL
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_stream = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1181
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 * 
 *         cdef char* c_catalog = NULL             # <<<<<<<<<<<<<<
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")
*/
  __pyx_v_c_catalog = NULL;

  /* "adbc_driver_manager/_lib.pyx":1182
 * 
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:             # <<<<<<<<<<<<<<
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog
*/
  __pyx_t_5 = (__pyx_v_catalog != Py_None);
  if (__pyx_t_5) {

    /* "adbc_driver_manager/_lib.pyx":1183
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")             # <<<<<<<<<<<<<<
 *             c_catalog = catalog
 * 
*/
    __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_catalog, __pyx_mstate_global->__pyx_n_u_catalog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_catalog, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":1184
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_db_schema = NULL
*/
    __pyx_t_6 = __Pyx_PyObject_AsWritableString(__pyx_v_catalog); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1184, __pyx_L1_error)
    __pyx_v_c_catalog = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":1182
 * 
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:             # <<<<<<<<<<<<<<
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1186
 *             c_catalog = catalog
 * 
 *         cdef char* c_db_schema = NULL             # <<<<<<<<<<<<<<
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")
*/
  __pyx_v_c_db_schema = NULL;

  /* "adbc_driver_manager/_lib.pyx":1187
 * 
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:             # <<<<<<<<<<<<<<
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema
*/
  __pyx_t_5 = (__pyx_v_db_schema != Py_None);
  if (__pyx_t_5) {

    /* "adbc_driver_manager/_lib.pyx":1188
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")             # <<<<<<<<<<<<<<
 *             c_db_schema = db_schema
 * 
*/
    __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_db_schema, __pyx_mstate_global->__pyx_n_u_db_schema); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_db_schema, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":1189
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_table_name = NULL
*/
    __pyx_t_6 = __Pyx_PyObject_AsWritableString(__pyx_v_db_schema); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1189, __pyx_L1_error)
    __pyx_v_c_db_schema = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":1187
 * 
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:             # <<<<<<<<<<<<<<
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1191
 *             c_db_schema = db_schema
 * 
 *         cdef char* c_table_name = NULL             # <<<<<<<<<<<<<<
 *         if table_name is not None:
 *             table_name = _to_bytes(table_name, "table_name")
*/
  __pyx_v_c_table_name = NULL;

  /* "adbc_driver_manager/_lib.pyx":1192
 * 
 *         cdef char* c_table_name = NULL
 *         if table_name is not None:             # <<<<<<<<<<<<<<
 *             table_name = _to_bytes(table_name, "table_name")
 *             c_table_name = table_name
*/
  __pyx_t_5 = (__pyx_v_table_name != Py_None);
  if (__pyx_t_5) {

    /* "adbc_driver_manager/_lib.pyx":1193
 *         cdef char* c_table_name = NULL
 *         if table_name is not None:
 *             table_name = _to_bytes(table_name, "table_name")             # <<<<<<<<<<<<<<
 *             c_table_name = table_name
 * 
*/
    __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_table_name, __pyx_mstate_global->__pyx_n_u_table_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_table_name, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":1194
 *         if table_name is not None:
 *             table_name = _to_bytes(table_name, "table_name")
 *             c_table_name = table_name             # <<<<<<<<<<<<<<
 * 
 *         cdef char c_approximate = 1 if approximate else 0
*/
    __pyx_t_6 = __Pyx_PyObject_AsWritableString(__pyx_v_table_name); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L1_error)
    __pyx_v_c_table_name = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":1192
 * 
 *         cdef char* c_table_name = NULL
 *         if table_name is not None:             # <<<<<<<<<<<<<<
 *             table_name = _to_bytes(table_name, "table_name")
 *             c_table_name = table_name
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1196
 *             c_table_name = table_name
 * 
 *         cdef char c_approximate = 1 if approximate else 0             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_approximate); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1196, __pyx_L1_error)
  if (__pyx_t_5) {
    __pyx_t_7 = 1;
  } else {
    __pyx_t_7 = 0;
  }
  __pyx_v_c_approximate = __pyx_t_7;

  /* "adbc_driver_manager/_lib.pyx":1198
 *         cdef char c_approximate = 1 if approximate else 0
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetStatistics(
 *                 &self.connection,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1199
 * 
 *         with nogil:
 *             status = AdbcConnectionGetStatistics(             # <<<<<<<<<<<<<<
 *                 &self.connection,
 *                 c_catalog,
*/
        __pyx_v_status = AdbcConnectionGetStatistics((&__pyx_v_self->connection), __pyx_v_c_catalog, __pyx_v_c_db_schema, __pyx_v_c_table_name, __pyx_v_c_approximate, (&__pyx_v_stream->stream), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1198
 *         cdef char c_approximate = 1 if approximate else 0
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetStatistics(
 *                 &self.connection,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1207
 *                 &stream.stream,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return stream
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1207, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1208
 *                 &c_error)
 *         check_error(status, &c_error)
 *         return stream             # <<<<<<<<<<<<<<
 * 
 *     def get_statistic_names(self) -> ArrowArrayStreamHandle:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __pyx_r = __pyx_v_stream;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1144
 *         return c_value
 * 
 *     def get_statistics(             # <<<<<<<<<<<<<<
 *         self,
 *         catalog=None,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_statistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XDECREF(__pyx_v_catalog);
  __Pyx_XDECREF(__pyx_v_db_schema);
  __Pyx_XDECREF(__pyx_v_table_name);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1210
 *         return stream
 * 
 *     def get_statistic_names(self) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get custom statistic names defined by this driver.
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_21get_statistic_names(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_20get_statistic_names, "\n        Get custom statistic names defined by this driver.\n\n        Returns\n        -------\n        ArrowArrayStreamHandle\n            A stream with columns: statistic_name (utf8), statistic_key (int16).\n\n        Notes\n        -----\n        Available since ADBC API revision 1.1.0. Returns driver-specific\n        statistic names and their keys. Standard ADBC statistics (keys 0-1023)\n        are not included - only driver-specific statistics.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_21get_statistic_names = {"get_statistic_names", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_21get_statistic_names, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_20get_statistic_names};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_21get_statistic_names(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_statistic_names (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_statistic_names", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_statistic_names", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_20get_statistic_names(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_20get_statistic_names(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_stream = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_statistic_names", 0);

  /* "adbc_driver_manager/_lib.pyx":1225
 *         are not included - only driver-specific statistics.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1225, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1227
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_stream = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1229
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetStatisticNames(
 *                 &self.connection,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1230
 * 
 *         with nogil:
 *             status = AdbcConnectionGetStatisticNames(             # <<<<<<<<<<<<<<
 *                 &self.connection,
 *                 &stream.stream,
*/
        __pyx_v_status = AdbcConnectionGetStatisticNames((&__pyx_v_self->connection), (&__pyx_v_stream->stream), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1229
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetStatisticNames(
 *                 &self.connection,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1234
 *                 &stream.stream,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return stream
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1234, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1235
 *                 &c_error)
 *         check_error(status, &c_error)
 *         return stream             # <<<<<<<<<<<<<<
 * 
 *     def get_table_schema(self, catalog, db_schema, table_name) -> ArrowSchemaHandle:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __pyx_r = __pyx_v_stream;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1210
 *         return stream
 * 
 *     def get_statistic_names(self) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get custom statistic names defined by this driver.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_statistic_names", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1237
 *         return stream
 * 
 *     def get_table_schema(self, catalog, db_schema, table_name) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the Arrow schema of a table.
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_23get_table_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_22get_table_schema, "\n        Get the Arrow schema of a table.\n\n        Returns\n        -------\n        ArrowSchemaHandle\n            A C Data Interface ArrowSchema struct containing the schema.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_23get_table_schema = {"get_table_schema", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_23get_table_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_22get_table_schema};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_23get_table_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_catalog = 0;
  PyObject *__pyx_v_db_schema = 0;
  PyObject *__pyx_v_table_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_table_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_catalog,&__pyx_mstate_global->__pyx_n_u_db_schema,&__pyx_mstate_global->__pyx_n_u_table_name,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1237, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1237, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1237, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1237, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_table_schema", 0) < (0)) __PYX_ERR(0, 1237, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_table_schema", 1, 3, 3, i); __PYX_ERR(0, 1237, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1237, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1237, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1237, __pyx_L3_error)
    }
    __pyx_v_catalog = values[0];
    __pyx_v_db_schema = values[1];
    __pyx_v_table_name = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_table_schema", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1237, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_table_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_22get_table_schema(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_catalog, __pyx_v_db_schema, __pyx_v_table_name);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_22get_table_schema(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_catalog, PyObject *__pyx_v_db_schema, PyObject *__pyx_v_table_name) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_handle = 0;
  char *__pyx_v_c_table_name;
  char *__pyx_v_c_catalog;
  char *__pyx_v_c_db_schema;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_table_schema", 0);
  __Pyx_INCREF(__pyx_v_catalog);
  __Pyx_INCREF(__pyx_v_db_schema);
  __Pyx_INCREF(__pyx_v_table_name);

  /* "adbc_driver_manager/_lib.pyx":1246
 *             A C Data Interface ArrowSchema struct containing the schema.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1246, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1248
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()             # <<<<<<<<<<<<<<
 *         table_name = _to_bytes(table_name, "table_name")
 *         cdef char* c_table_name = table_name
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1248, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_handle = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1249
 *         cdef CAdbcStatusCode status
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()
 *         table_name = _to_bytes(table_name, "table_name")             # <<<<<<<<<<<<<<
 *         cdef char* c_table_name = table_name
 * 
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_table_name, __pyx_mstate_global->__pyx_n_u_table_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_table_name, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1250
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()
 *         table_name = _to_bytes(table_name, "table_name")
 *         cdef char* c_table_name = table_name             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_catalog = NULL
*/
  __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_table_name); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 1250, __pyx_L1_error)
  __pyx_v_c_table_name = __pyx_t_5;

  /* "adbc_driver_manager/_lib.pyx":1252
 *         cdef char* c_table_name = table_name
 * 
 *         cdef char* c_catalog = NULL             # <<<<<<<<<<<<<<
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")
*/
  __pyx_v_c_catalog = NULL;

  /* "adbc_driver_manager/_lib.pyx":1253
 * 
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:             # <<<<<<<<<<<<<<
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog
*/
  __pyx_t_6 = (__pyx_v_catalog != Py_None);
  if (__pyx_t_6) {

    /* "adbc_driver_manager/_lib.pyx":1254
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")             # <<<<<<<<<<<<<<
 *             c_catalog = catalog
 * 
*/
    __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_catalog, __pyx_mstate_global->__pyx_n_u_catalog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_catalog, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":1255
 *         if catalog is not None:
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_db_schema = NULL
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_catalog); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L1_error)
    __pyx_v_c_catalog = __pyx_t_5;

    /* "adbc_driver_manager/_lib.pyx":1253
 * 
 *         cdef char* c_catalog = NULL
 *         if catalog is not None:             # <<<<<<<<<<<<<<
 *             catalog = _to_bytes(catalog, "catalog")
 *             c_catalog = catalog
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1257
 *             c_catalog = catalog
 * 
 *         cdef char* c_db_schema = NULL             # <<<<<<<<<<<<<<
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")
*/
  __pyx_v_c_db_schema = NULL;

  /* "adbc_driver_manager/_lib.pyx":1258
 * 
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:             # <<<<<<<<<<<<<<
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema
*/
  __pyx_t_6 = (__pyx_v_db_schema != Py_None);
  if (__pyx_t_6) {

    /* "adbc_driver_manager/_lib.pyx":1259
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")             # <<<<<<<<<<<<<<
 *             c_db_schema = db_schema
 * 
*/
    __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_db_schema, __pyx_mstate_global->__pyx_n_u_db_schema); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_db_schema, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":1260
 *         if db_schema is not None:
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_db_schema); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 1260, __pyx_L1_error)
    __pyx_v_c_db_schema = __pyx_t_5;

    /* "adbc_driver_manager/_lib.pyx":1258
 * 
 *         cdef char* c_db_schema = NULL
 *         if db_schema is not None:             # <<<<<<<<<<<<<<
 *             db_schema = _to_bytes(db_schema, "db_schema")
 *             c_db_schema = db_schema
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1262
 *             c_db_schema = db_schema
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetTableSchema(
 *                 &self.connection,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1263
 * 
 *         with nogil:
 *             status = AdbcConnectionGetTableSchema(             # <<<<<<<<<<<<<<
 *                 &self.connection,
 *                 c_catalog,
*/
        __pyx_v_status = AdbcConnectionGetTableSchema((&__pyx_v_self->connection), __pyx_v_c_catalog, __pyx_v_c_db_schema, __pyx_v_c_table_name, (&__pyx_v_handle->schema), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1262
 *             c_db_schema = db_schema
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetTableSchema(
 *                 &self.connection,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1270
 *                 &handle.schema,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return handle
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1270, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1271
 *                 &c_error)
 *         check_error(status, &c_error)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     def get_table_types(self) -> ArrowArrayStreamHandle:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1237
 *         return stream
 * 
 *     def get_table_schema(self, catalog, db_schema, table_name) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the Arrow schema of a table.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_table_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XDECREF(__pyx_v_catalog);
  __Pyx_XDECREF(__pyx_v_db_schema);
  __Pyx_XDECREF(__pyx_v_table_name);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1273
 *         return handle
 * 
 *     def get_table_types(self) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the list of supported table types.
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_25get_table_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_24get_table_types, "\n        Get the list of supported table types.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_25get_table_types = {"get_table_types", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_25get_table_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_24get_table_types};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_25get_table_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_table_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_table_types", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_table_types", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_24get_table_types(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_24get_table_types(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_stream = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_table_types", 0);

  /* "adbc_driver_manager/_lib.pyx":1277
 *         Get the list of supported table types.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1277, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1279
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_stream = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1281
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetTableTypes(
 *                 &self.connection, &stream.stream, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1282
 * 
 *         with nogil:
 *             status = AdbcConnectionGetTableTypes(             # <<<<<<<<<<<<<<
 *                 &self.connection, &stream.stream, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcConnectionGetTableTypes((&__pyx_v_self->connection), (&__pyx_v_stream->stream), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1281
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionGetTableTypes(
 *                 &self.connection, &stream.stream, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1284
 *             status = AdbcConnectionGetTableTypes(
 *                 &self.connection, &stream.stream, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return stream
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1284, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1285
 *                 &self.connection, &stream.stream, &c_error)
 *         check_error(status, &c_error)
 *         return stream             # <<<<<<<<<<<<<<
 * 
 *     def read_partition(self, bytes partition not None) -> ArrowArrayStreamHandle:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __pyx_r = __pyx_v_stream;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1273
 *         return handle
 * 
 *     def get_table_types(self) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the list of supported table types.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.get_table_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1287
 *         return stream
 * 
 *     def read_partition(self, bytes partition not None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """Fetch a single partition from execute_partitions."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_27read_partition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_26read_partition, "Fetch a single partition from execute_partitions.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_27read_partition = {"read_partition", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_27read_partition, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_26read_partition};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_27read_partition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_partition = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_partition (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_partition,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1287, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1287, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "read_partition", 0) < (0)) __PYX_ERR(0, 1287, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("read_partition", 1, 1, 1, i); __PYX_ERR(0, 1287, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1287, __pyx_L3_error)
    }
    __pyx_v_partition = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("read_partition", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1287, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.read_partition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_partition), (&PyBytes_Type), 0, "partition", 1))) __PYX_ERR(0, 1287, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_26read_partition(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_partition);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_26read_partition(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_partition) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_stream = 0;
  uint8_t const *__pyx_v_data;
  size_t __pyx_v_length;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  uint8_t const *__pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_partition", 0);

  /* "adbc_driver_manager/_lib.pyx":1289
 *     def read_partition(self, bytes partition not None) -> ArrowArrayStreamHandle:
 *         """Fetch a single partition from execute_partitions."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1291
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()             # <<<<<<<<<<<<<<
 *         cdef const uint8_t* data = <const uint8_t*> partition
 *         cdef size_t length = len(partition)
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1291, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_stream = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1292
 *         cdef CAdbcStatusCode status
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 *         cdef const uint8_t* data = <const uint8_t*> partition             # <<<<<<<<<<<<<<
 *         cdef size_t length = len(partition)
 * 
*/
  __pyx_t_5 = __Pyx_PyBytes_AsUString(__pyx_v_partition); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 1292, __pyx_L1_error)
  __pyx_v_data = ((uint8_t const *)__pyx_t_5);

  /* "adbc_driver_manager/_lib.pyx":1293
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 *         cdef const uint8_t* data = <const uint8_t*> partition
 *         cdef size_t length = len(partition)             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  __pyx_t_6 = __Pyx_PyBytes_GET_SIZE(__pyx_v_partition); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1293, __pyx_L1_error)
  __pyx_v_length = __pyx_t_6;

  /* "adbc_driver_manager/_lib.pyx":1295
 *         cdef size_t length = len(partition)
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionReadPartition(
 *                 &self.connection,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1296
 * 
 *         with nogil:
 *             status = AdbcConnectionReadPartition(             # <<<<<<<<<<<<<<
 *                 &self.connection,
 *                 data,
*/
        __pyx_v_status = AdbcConnectionReadPartition((&__pyx_v_self->connection), __pyx_v_data, __pyx_v_length, (&__pyx_v_stream->stream), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1295
 *         cdef size_t length = len(partition)
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionReadPartition(
 *                 &self.connection,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1302
 *                 &stream.stream,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return stream
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1302, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1303
 *                 &c_error)
 *         check_error(status, &c_error)
 *         return stream             # <<<<<<<<<<<<<<
 * 
 *     def rollback(self) -> None:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __pyx_r = __pyx_v_stream;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1287
 *         return stream
 * 
 *     def read_partition(self, bytes partition not None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """Fetch a single partition from execute_partitions."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.read_partition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1305
 *         return stream
 * 
 *     def rollback(self) -> None:             # <<<<<<<<<<<<<<
 *         """Rollback the current transaction."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_29rollback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_28rollback, "Rollback the current transaction.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_29rollback = {"rollback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_29rollback, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_28rollback};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_29rollback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rollback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("rollback", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("rollback", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_28rollback(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_28rollback(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rollback", 0);

  /* "adbc_driver_manager/_lib.pyx":1307
 *     def rollback(self) -> None:
 *         """Rollback the current transaction."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         with nogil:
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1307, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1309
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionRollback(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1310
 *         cdef CAdbcStatusCode status
 *         with nogil:
 *             status = AdbcConnectionRollback(&self.connection, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcConnectionRollback((&__pyx_v_self->connection), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1309
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionRollback(&self.connection, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1311
 *         with nogil:
 *             status = AdbcConnectionRollback(&self.connection, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def set_autocommit(self, bint enabled) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1311, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1305
 *         return stream
 * 
 *     def rollback(self) -> None:             # <<<<<<<<<<<<<<
 *         """Rollback the current transaction."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.rollback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1313
 *         check_error(status, &c_error)
 * 
 *     def set_autocommit(self, bint enabled) -> None:             # <<<<<<<<<<<<<<
 *         """Toggle whether autocommit is enabled."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_31set_autocommit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_30set_autocommit, "Toggle whether autocommit is enabled.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_31set_autocommit = {"set_autocommit", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_31set_autocommit, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_30set_autocommit};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_31set_autocommit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_enabled;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_autocommit (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_enabled,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1313, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1313, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_autocommit", 0) < (0)) __PYX_ERR(0, 1313, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_autocommit", 1, 1, 1, i); __PYX_ERR(0, 1313, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1313, __pyx_L3_error)
    }
    __pyx_v_enabled = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_enabled == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1313, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_autocommit", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1313, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.set_autocommit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_30set_autocommit(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_enabled);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_30set_autocommit(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, int __pyx_v_enabled) {
  struct AdbcError __pyx_v_c_error;
  char const *__pyx_v_value;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_autocommit", 0);

  /* "adbc_driver_manager/_lib.pyx":1315
 *     def set_autocommit(self, bint enabled) -> None:
 *         """Toggle whether autocommit is enabled."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         if enabled:
 *             value = ADBC_OPTION_VALUE_ENABLED
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1315, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1316
 *         """Toggle whether autocommit is enabled."""
 *         cdef CAdbcError c_error = empty_error()
 *         if enabled:             # <<<<<<<<<<<<<<
 *             value = ADBC_OPTION_VALUE_ENABLED
 *         else:
*/
  if (__pyx_v_enabled) {

    /* "adbc_driver_manager/_lib.pyx":1317
 *         cdef CAdbcError c_error = empty_error()
 *         if enabled:
 *             value = ADBC_OPTION_VALUE_ENABLED             # <<<<<<<<<<<<<<
 *         else:
 *             value = ADBC_OPTION_VALUE_DISABLED
*/
    __pyx_v_value = ADBC_OPTION_VALUE_ENABLED;

    /* "adbc_driver_manager/_lib.pyx":1316
 *         """Toggle whether autocommit is enabled."""
 *         cdef CAdbcError c_error = empty_error()
 *         if enabled:             # <<<<<<<<<<<<<<
 *             value = ADBC_OPTION_VALUE_ENABLED
 *         else:
*/
    goto __pyx_L3;
  }

  /* "adbc_driver_manager/_lib.pyx":1319
 *             value = ADBC_OPTION_VALUE_ENABLED
 *         else:
 *             value = ADBC_OPTION_VALUE_DISABLED             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  /*else*/ {
    __pyx_v_value = ADBC_OPTION_VALUE_DISABLED;
  }
  __pyx_L3:;

  /* "adbc_driver_manager/_lib.pyx":1321
 *             value = ADBC_OPTION_VALUE_DISABLED
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionSetOption(
 *                 &self.connection,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1322
 * 
 *         with nogil:
 *             status = AdbcConnectionSetOption(             # <<<<<<<<<<<<<<
 *                 &self.connection,
 *                 ADBC_CONNECTION_OPTION_AUTOCOMMIT,
*/
        __pyx_v_status = AdbcConnectionSetOption((&__pyx_v_self->connection), ADBC_CONNECTION_OPTION_AUTOCOMMIT, __pyx_v_value, (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1321
 *             value = ADBC_OPTION_VALUE_DISABLED
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcConnectionSetOption(
 *                 &self.connection,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1327
 *                 value,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def set_options(self, **kwargs) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1327, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1313
 *         check_error(status, &c_error)
 * 
 *     def set_autocommit(self, bint enabled) -> None:             # <<<<<<<<<<<<<<
 *         """Toggle whether autocommit is enabled."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.set_autocommit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1329
 *         check_error(status, &c_error)
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Set arbitrary key-value options.
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_33set_options(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_32set_options, "\n        Set arbitrary key-value options.\n\n        Pass options as kwargs: ``set_options(**{\"some.option\": \"value\"})``.\n\n        Note, not all drivers support setting options after creation.\n\n        See Also\n        --------\n        adbc_driver_manager.ConnectionOptions : Standard option names.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_33set_options = {"set_options", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_33set_options, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_32set_options};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_33set_options(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_options (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("set_options", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("set_options", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_32set_options(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_32set_options(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  struct AdbcError __pyx_v_c_error;
  char *__pyx_v_c_key;
  char *__pyx_v_c_value;
  size_t __pyx_v_c_value_len;
  double __pyx_v_c_double;
  int64_t __pyx_v_c_int;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  char *__pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  double __pyx_t_12;
  int64_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17[6];
  PyObject *__pyx_t_18 = NULL;
  size_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_options", 0);

  /* "adbc_driver_manager/_lib.pyx":1341
 *         adbc_driver_manager.ConnectionOptions : Standard option names.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef char* c_key = NULL
 *         cdef char* c_value = NULL
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1341, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1342
 *         """
 *         cdef CAdbcError c_error = empty_error()
 *         cdef char* c_key = NULL             # <<<<<<<<<<<<<<
 *         cdef char* c_value = NULL
 *         cdef size_t c_value_len
*/
  __pyx_v_c_key = NULL;

  /* "adbc_driver_manager/_lib.pyx":1343
 *         cdef CAdbcError c_error = empty_error()
 *         cdef char* c_key = NULL
 *         cdef char* c_value = NULL             # <<<<<<<<<<<<<<
 *         cdef size_t c_value_len
 *         cdef double c_double
*/
  __pyx_v_c_value = NULL;

  /* "adbc_driver_manager/_lib.pyx":1347
 *         cdef double c_double
 *         cdef int64_t c_int
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             key = _to_bytes(key, "option key")
 *             c_key = key
*/
  __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_2);
  __pyx_t_2 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
    if (unlikely(__pyx_t_8 == 0)) break;
    if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "adbc_driver_manager/_lib.pyx":1348
 *         cdef int64_t c_int
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "option key")             # <<<<<<<<<<<<<<
 *             c_key = key
 * 
*/
    __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_kp_u_option_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "adbc_driver_manager/_lib.pyx":1349
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "option key")
 *             c_key = key             # <<<<<<<<<<<<<<
 * 
 *             if value is None:
*/
    __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_key); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L1_error)
    __pyx_v_c_key = __pyx_t_9;

    /* "adbc_driver_manager/_lib.pyx":1351
 *             c_key = key
 * 
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 c_value = NULL
 *                 with nogil:
*/
    __pyx_t_10 = (__pyx_v_value == Py_None);
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1352
 * 
 *             if value is None:
 *                 c_value = NULL             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(
*/
      __pyx_v_c_value = NULL;

      /* "adbc_driver_manager/_lib.pyx":1353
 *             if value is None:
 *                 c_value = NULL
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1354
 *                 c_value = NULL
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, str):
*/
            __pyx_v_status = AdbcConnectionSetOption((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1353
 *             if value is None:
 *                 c_value = NULL
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L10;
            }
            __pyx_L10:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1351
 *             c_key = key
 * 
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 c_value = NULL
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1356
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
    __pyx_t_10 = PyUnicode_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1357
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, str):
 *                 value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 with nogil:
*/
      __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "adbc_driver_manager/_lib.pyx":1358
 *             elif isinstance(value, str):
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1358, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":1359
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1360
 *                 c_value = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
*/
            __pyx_v_status = AdbcConnectionSetOption((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1359
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L15;
            }
            __pyx_L15:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1356
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1362
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
    __pyx_t_10 = PyBool_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1363
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1363, __pyx_L1_error)
      if (__pyx_t_10) {

        /* "adbc_driver_manager/_lib.pyx":1364
 *             elif isinstance(value, bool):
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED             # <<<<<<<<<<<<<<
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
*/
        __pyx_t_7 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_ENABLED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "adbc_driver_manager/_lib.pyx":1363
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
        goto __pyx_L16;
      }

      /* "adbc_driver_manager/_lib.pyx":1366
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
      /*else*/ {
        __pyx_t_7 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_DISABLED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
        __pyx_t_7 = 0;
      }
      __pyx_L16:;

      /* "adbc_driver_manager/_lib.pyx":1367
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 with nogil:
*/
      __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "adbc_driver_manager/_lib.pyx":1368
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1368, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":1369
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1370
 *                 c_value = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOption(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):
*/
            __pyx_v_status = AdbcConnectionSetOption((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1369
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L21;
            }
            __pyx_L21:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1362
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1372
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
    __pyx_t_10 = PyBytes_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1373
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 c_value_len = len(value)
 *                 with nogil:
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1373, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":1374
 *             elif isinstance(value, bytes):
 *                 c_value = value
 *                 c_value_len = len(value)             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionBytes(
*/
      __pyx_t_11 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1374, __pyx_L1_error)
      __pyx_v_c_value_len = __pyx_t_11;

      /* "adbc_driver_manager/_lib.pyx":1375
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionBytes(
 *                         &self.connection, c_key, <const uint8_t*> c_value,
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1376
 *                 c_value_len = len(value)
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionBytes(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
*/
            __pyx_v_status = AdbcConnectionSetOptionBytes((&__pyx_v_self->connection), __pyx_v_c_key, ((uint8_t const *)__pyx_v_c_value), __pyx_v_c_value_len, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1375
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionBytes(
 *                         &self.connection, c_key, <const uint8_t*> c_value,
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L26;
            }
            __pyx_L26:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1372
 *                     status = AdbcConnectionSetOption(
 *                         &self.connection, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1379
 *                         &self.connection, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_double = value
 *                 with nogil:
*/
    __pyx_t_10 = PyFloat_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1380
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):
 *                 c_double = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionDouble(
*/
      __pyx_t_12 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L1_error)
      __pyx_v_c_double = __pyx_t_12;

      /* "adbc_driver_manager/_lib.pyx":1381
 *             elif isinstance(value, float):
 *                 c_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_double, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1382
 *                 c_double = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionDouble(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_double, &c_error)
 *             elif isinstance(value, int):
*/
            __pyx_v_status = AdbcConnectionSetOptionDouble((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_double, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1381
 *             elif isinstance(value, float):
 *                 c_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_double, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L31;
            }
            __pyx_L31:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1379
 *                         &self.connection, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_double = value
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1384
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_double, &c_error)
 *             elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_int = value
 *                 with nogil:
*/
    __pyx_t_10 = PyLong_Check(__pyx_v_value); 
    if (likely(__pyx_t_10)) {

      /* "adbc_driver_manager/_lib.pyx":1385
 *                         &self.connection, c_key, c_double, &c_error)
 *             elif isinstance(value, int):
 *                 c_int = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionInt(
*/
      __pyx_t_13 = __Pyx_PyLong_As_int64_t(__pyx_v_value); if (unlikely((__pyx_t_13 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1385, __pyx_L1_error)
      __pyx_v_c_int = __pyx_t_13;

      /* "adbc_driver_manager/_lib.pyx":1386
 *             elif isinstance(value, int):
 *                 c_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionInt(
 *                         &self.connection, c_key, c_int, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1387
 *                 c_int = value
 *                 with nogil:
 *                     status = AdbcConnectionSetOptionInt(             # <<<<<<<<<<<<<<
 *                         &self.connection, c_key, c_int, &c_error)
 *             else:
*/
            __pyx_v_status = AdbcConnectionSetOptionInt((&__pyx_v_self->connection), __pyx_v_c_key, __pyx_v_c_int, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1386
 *             elif isinstance(value, int):
 *                 c_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcConnectionSetOptionInt(
 *                         &self.connection, c_key, c_int, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L36;
            }
            __pyx_L36:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1384
 *                     status = AdbcConnectionSetOptionDouble(
 *                         &self.connection, c_key, c_double, &c_error)
 *             elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_int = value
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1390
 *                         &self.connection, c_key, c_int, &c_error)
 *             else:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     f"Unsupported type {type(value)} for value {value!r} "
 *                     f"of option {key}")
*/
    /*else*/ {
      __pyx_t_6 = NULL;

      /* "adbc_driver_manager/_lib.pyx":1391
 *             else:
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "             # <<<<<<<<<<<<<<
 *                     f"of option {key}")
 * 
*/
      __pyx_t_14 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_value)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_value), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);

      /* "adbc_driver_manager/_lib.pyx":1392
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "
 *                     f"of option {key}")             # <<<<<<<<<<<<<<
 * 
 *             check_error(status, &c_error)
*/
      __pyx_t_16 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_17[0] = __pyx_mstate_global->__pyx_kp_u_Unsupported_type;
      __pyx_t_17[1] = __pyx_t_14;
      __pyx_t_17[2] = __pyx_mstate_global->__pyx_kp_u_for_value;
      __pyx_t_17[3] = __pyx_t_15;
      __pyx_t_17[4] = __pyx_mstate_global->__pyx_kp_u_of_option;
      __pyx_t_17[5] = __pyx_t_16;

      /* "adbc_driver_manager/_lib.pyx":1391
 *             else:
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "             # <<<<<<<<<<<<<<
 *                     f"of option {key}")
 * 
*/
      __pyx_t_18 = __Pyx_PyUnicode_Join(__pyx_t_17, 6, 17 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14) + 11 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16));
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_19 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_18};
        __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_19, (2-__pyx_t_19) | (__pyx_t_19*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 1390, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "adbc_driver_manager/_lib.pyx":1394
 *                     f"of option {key}")
 * 
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def close(self) -> None:
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1394, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1329
 *         check_error(status, &c_error)
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Set arbitrary key-value options.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.set_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1396
 *             check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_35close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_34close, "Release the handle to the connection.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_35close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_35close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_34close};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_35close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_34close(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_34close(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "adbc_driver_manager/_lib.pyx":1398
 *     def close(self) -> None:
 *         """Release the handle to the connection."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1400
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             if self.connection.private_data == NULL:
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._check_open_children(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1401
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self.connection.private_data == NULL:
 *                 return
*/
  /*with:*/ {
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->__pyx_base._lock, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->__pyx_base._lock, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1401, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1401, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1402
 *         self._check_open_children()
 *         with self._lock:
 *             if self.connection.private_data == NULL:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          __pyx_t_10 = (__pyx_v_self->connection.private_data == NULL);
          if (__pyx_t_10) {

            /* "adbc_driver_manager/_lib.pyx":1403
 *         with self._lock:
 *             if self.connection.private_data == NULL:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             with nogil:
*/
            __Pyx_XDECREF(__pyx_r);
            __pyx_r = Py_None; __Pyx_INCREF(Py_None);
            goto __pyx_L11_try_return;

            /* "adbc_driver_manager/_lib.pyx":1402
 *         self._check_open_children()
 *         with self._lock:
 *             if self.connection.private_data == NULL:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          }

          /* "adbc_driver_manager/_lib.pyx":1405
 *                 return
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionRelease(&self.connection, &c_error)
 *             check_error(status, &c_error)
*/
          {
              PyThreadState * _save;
              _save = PyEval_SaveThread();
              __Pyx_FastGIL_Remember();
              /*try:*/ {

                /* "adbc_driver_manager/_lib.pyx":1406
 * 
 *             with nogil:
 *                 status = AdbcConnectionRelease(&self.connection, &c_error)             # <<<<<<<<<<<<<<
 *             check_error(status, &c_error)
 *             self.database._close_child()
*/
                __pyx_v_status = AdbcConnectionRelease((&__pyx_v_self->connection), (&__pyx_v_c_error));
              }

              /* "adbc_driver_manager/_lib.pyx":1405
 *                 return
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcConnectionRelease(&self.connection, &c_error)
 *             check_error(status, &c_error)
*/
              /*finally:*/ {
                /*normal exit:*/{
                  __Pyx_FastGIL_Forget();
                  PyEval_RestoreThread(_save);
                  goto __pyx_L16;
                }
                __pyx_L16:;
              }
          }

          /* "adbc_driver_manager/_lib.pyx":1407
 *             with nogil:
 *                 status = AdbcConnectionRelease(&self.connection, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *             self.database._close_child()
 * 
*/
          __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L7_error)

          /* "adbc_driver_manager/_lib.pyx":1408
 *                 status = AdbcConnectionRelease(&self.connection, &c_error)
 *             check_error(status, &c_error)
 *             self.database._close_child()             # <<<<<<<<<<<<<<
 * 
 * 
*/
          __pyx_t_2 = ((struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcDatabase *)__pyx_v_self->database->__pyx_base.__pyx_vtab)->__pyx_base._close_child(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self->database)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1408, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "adbc_driver_manager/_lib.pyx":1401
 *         cdef CAdbcStatusCode status
 *         self._check_open_children()
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self.connection.private_data == NULL:
 *                 return
*/
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 1401, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_4);
          __pyx_t_11 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1401, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1401, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_10 < (0)) __PYX_ERR(0, 1401, __pyx_L9_except_error)
          __pyx_t_13 = (!__pyx_t_10);
          if (unlikely(__pyx_t_13)) {
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_XGIVEREF(__pyx_t_4);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_5, __pyx_t_4);
            __pyx_t_2 = 0;  __pyx_t_5 = 0;  __pyx_t_4 = 0; 
            __PYX_ERR(0, 1401, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L11_try_return:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L4_return;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1401, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L4_return: {
        __pyx_t_9 = __pyx_r;
        __pyx_r = 0;
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1401, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_r = __pyx_t_9;
        __pyx_t_9 = 0;
        goto __pyx_L0;
      }
      __pyx_L6:;
    }
    goto __pyx_L20;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L20:;
  }

  /* "adbc_driver_manager/_lib.pyx":1396
 *             check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_37__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_37__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_37__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_37__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_36__reduce_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_connection_cannot_be_conver, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_39__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_39__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_39__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_39__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_38__setstate_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_14AdbcConnection_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_connection_cannot_be_conver, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcConnection.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1427
 *         CAdbcStatement statement
 * 
 *     def __init__(self, AdbcConnection connection) -> None:             # <<<<<<<<<<<<<<
 *         super().__init__("(no child type)")
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static int __pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_connection = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_connection,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1427, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1427, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 1427, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 1427, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1427, __pyx_L3_error)
    }
    __pyx_v_connection = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1427, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_connection), __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, 1, "connection", 0))) __PYX_ERR(0, 1427, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement___init__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_connection);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement___init__(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *__pyx_v_connection) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  struct AdbcError __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "adbc_driver_manager/_lib.pyx":1428
 * 
 *     def __init__(self, AdbcConnection connection) -> None:
 *         super().__init__("(no child type)")             # <<<<<<<<<<<<<<
 *         cdef CAdbcError c_error = empty_error()
 *         self.connection = connection
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_no_child_type};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":1429
 *     def __init__(self, AdbcConnection connection) -> None:
 *         super().__init__("(no child type)")
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         self.connection = connection
 *         memset(&self.statement, 0, cython.sizeof(CAdbcStatement))
*/
  __pyx_t_6 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1429, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_6;

  /* "adbc_driver_manager/_lib.pyx":1430
 *         super().__init__("(no child type)")
 *         cdef CAdbcError c_error = empty_error()
 *         self.connection = connection             # <<<<<<<<<<<<<<
 *         memset(&self.statement, 0, cython.sizeof(CAdbcStatement))
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_connection);
  __Pyx_GIVEREF((PyObject *)__pyx_v_connection);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->connection);
  __Pyx_DECREF((PyObject *)__pyx_v_self->connection);
  __pyx_v_self->connection = __pyx_v_connection;

  /* "adbc_driver_manager/_lib.pyx":1431
 *         cdef CAdbcError c_error = empty_error()
 *         self.connection = connection
 *         memset(&self.statement, 0, cython.sizeof(CAdbcStatement))             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  (void)(memset((&__pyx_v_self->statement), 0, (sizeof(struct AdbcStatement))));

  /* "adbc_driver_manager/_lib.pyx":1433
 *         memset(&self.statement, 0, cython.sizeof(CAdbcStatement))
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementNew(
 *                 &connection.connection,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1434
 * 
 *         with nogil:
 *             status = AdbcStatementNew(             # <<<<<<<<<<<<<<
 *                 &connection.connection,
 *                 &self.statement,
*/
        __pyx_v_status = AdbcStatementNew((&__pyx_v_connection->connection), (&__pyx_v_self->statement), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1433
 *         memset(&self.statement, 0, cython.sizeof(CAdbcStatement))
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementNew(
 *                 &connection.connection,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1438
 *                 &self.statement,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         connection._open_child()
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1438, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1440
 *         check_error(status, &c_error)
 * 
 *         connection._open_child()             # <<<<<<<<<<<<<<
 * 
 *     def bind(self, data, schema=None) -> None:
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_connection->__pyx_base.__pyx_vtab)->__pyx_base._open_child(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_connection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "adbc_driver_manager/_lib.pyx":1427
 *         CAdbcStatement statement
 * 
 *     def __init__(self, AdbcConnection connection) -> None:             # <<<<<<<<<<<<<<
 *         super().__init__("(no child type)")
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1442
 *         connection._open_child()
 * 
 *     def bind(self, data, schema=None) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Bind an ArrowArray to this statement.
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_3bind(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_2bind, "\n        Bind an ArrowArray to this statement.\n\n        Parameters\n        ----------\n        data : PyCapsule or int or ArrowArrayHandle\n        schema : PyCapsule or int or ArrowSchemaHandle\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_3bind = {"bind", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_3bind, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_2bind};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_3bind(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_schema = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bind (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_data,&__pyx_mstate_global->__pyx_n_u_schema,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1442, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1442, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1442, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bind", 0) < (0)) __PYX_ERR(0, 1442, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bind", 0, 1, 2, i); __PYX_ERR(0, 1442, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1442, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1442, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_data = values[0];
    __pyx_v_schema = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bind", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1442, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_2bind(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_data, __pyx_v_schema);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_2bind(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_schema) {
  struct AdbcError __pyx_v_c_error;
  struct ArrowArray *__pyx_v_c_array;
  struct ArrowSchema *__pyx_v_c_schema;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  void *__pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind", 0);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_INCREF(__pyx_v_schema);

  /* "adbc_driver_manager/_lib.pyx":1451
 *         schema : PyCapsule or int or ArrowSchemaHandle
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CArrowArray* c_array
 *         cdef CArrowSchema* c_schema
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1451, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1455
 *         cdef CArrowSchema* c_schema
 * 
 *         if (hasattr(data, "__arrow_c_array__") and             # <<<<<<<<<<<<<<
 *                 not isinstance(data, ArrowArrayHandle)):
 *             if schema is not None:
*/
  __pyx_t_3 = __Pyx_HasAttr(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_arrow_c_array); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1455, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":1456
 * 
 *         if (hasattr(data, "__arrow_c_array__") and
 *                 not isinstance(data, ArrowArrayHandle)):             # <<<<<<<<<<<<<<
 *             if schema is not None:
 *                 raise ValueError(
*/
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_data, __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle); 
  __pyx_t_4 = (!__pyx_t_3);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "adbc_driver_manager/_lib.pyx":1455
 *         cdef CArrowSchema* c_schema
 * 
 *         if (hasattr(data, "__arrow_c_array__") and             # <<<<<<<<<<<<<<
 *                 not isinstance(data, ArrowArrayHandle)):
 *             if schema is not None:
*/
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1457
 *         if (hasattr(data, "__arrow_c_array__") and
 *                 not isinstance(data, ArrowArrayHandle)):
 *             if schema is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     "Can not provide a schema when passing Arrow-compatible "
*/
    __pyx_t_2 = (__pyx_v_schema != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "adbc_driver_manager/_lib.pyx":1458
 *                 not isinstance(data, ArrowArrayHandle)):
 *             if schema is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     "Can not provide a schema when passing Arrow-compatible "
 *                     "data that implements the Arrow PyCapsule Protocol"
*/
      __pyx_t_6 = NULL;
      __pyx_t_7 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_Can_not_provide_a_schema_when_pa};
        __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1458, __pyx_L1_error)

      /* "adbc_driver_manager/_lib.pyx":1457
 *         if (hasattr(data, "__arrow_c_array__") and
 *                 not isinstance(data, ArrowArrayHandle)):
 *             if schema is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     "Can not provide a schema when passing Arrow-compatible "
*/
    }

    /* "adbc_driver_manager/_lib.pyx":1462
 *                     "data that implements the Arrow PyCapsule Protocol"
 *                 )
 *             schema, data = data.__arrow_c_array__()             # <<<<<<<<<<<<<<
 * 
 *         if is_pycapsule(data, b"arrow_array"):
*/
    __pyx_t_6 = __pyx_v_data;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_arrow_c_array, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1462, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1);
        __Pyx_INCREF(__pyx_t_8);
      } else {
        __pyx_t_6 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1462, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __pyx_t_8 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1462, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_8);
      }
      #else
      __pyx_t_6 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
      index = 0; __pyx_t_6 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < (0)) __PYX_ERR(0, 1462, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1462, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }
    __Pyx_DECREF_SET(__pyx_v_schema, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "adbc_driver_manager/_lib.pyx":1455
 *         cdef CArrowSchema* c_schema
 * 
 *         if (hasattr(data, "__arrow_c_array__") and             # <<<<<<<<<<<<<<
 *                 not isinstance(data, ArrowArrayHandle)):
 *             if schema is not None:
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1464
 *             schema, data = data.__arrow_c_array__()
 * 
 *         if is_pycapsule(data, b"arrow_array"):             # <<<<<<<<<<<<<<
 *             c_array = <CArrowArray*> PyCapsule_GetPointer(data, "arrow_array")
 *         elif isinstance(data, ArrowArrayHandle):
*/
  __pyx_t_8 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_is_pycapsule); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    assert(__pyx_t_8);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_v_data, __pyx_mstate_global->__pyx_n_b_arrow_array};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1465
 * 
 *         if is_pycapsule(data, b"arrow_array"):
 *             c_array = <CArrowArray*> PyCapsule_GetPointer(data, "arrow_array")             # <<<<<<<<<<<<<<
 *         elif isinstance(data, ArrowArrayHandle):
 *             c_array = &(<ArrowArrayHandle> data).array
*/
    __pyx_t_11 = PyCapsule_GetPointer(__pyx_v_data, __pyx_k_arrow_array); if (unlikely(__pyx_t_11 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 1465, __pyx_L1_error)
    __pyx_v_c_array = ((struct ArrowArray *)__pyx_t_11);

    /* "adbc_driver_manager/_lib.pyx":1464
 *             schema, data = data.__arrow_c_array__()
 * 
 *         if is_pycapsule(data, b"arrow_array"):             # <<<<<<<<<<<<<<
 *             c_array = <CArrowArray*> PyCapsule_GetPointer(data, "arrow_array")
 *         elif isinstance(data, ArrowArrayHandle):
*/
    goto __pyx_L9;
  }

  /* "adbc_driver_manager/_lib.pyx":1466
 *         if is_pycapsule(data, b"arrow_array"):
 *             c_array = <CArrowArray*> PyCapsule_GetPointer(data, "arrow_array")
 *         elif isinstance(data, ArrowArrayHandle):             # <<<<<<<<<<<<<<
 *             c_array = &(<ArrowArrayHandle> data).array
 *         elif isinstance(data, int):
*/
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_data, __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle); 
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1467
 *             c_array = <CArrowArray*> PyCapsule_GetPointer(data, "arrow_array")
 *         elif isinstance(data, ArrowArrayHandle):
 *             c_array = &(<ArrowArrayHandle> data).array             # <<<<<<<<<<<<<<
 *         elif isinstance(data, int):
 *             c_array = <CArrowArray*> data
*/
    __pyx_v_c_array = (&((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)__pyx_v_data)->array);

    /* "adbc_driver_manager/_lib.pyx":1466
 *         if is_pycapsule(data, b"arrow_array"):
 *             c_array = <CArrowArray*> PyCapsule_GetPointer(data, "arrow_array")
 *         elif isinstance(data, ArrowArrayHandle):             # <<<<<<<<<<<<<<
 *             c_array = &(<ArrowArrayHandle> data).array
 *         elif isinstance(data, int):
*/
    goto __pyx_L9;
  }

  /* "adbc_driver_manager/_lib.pyx":1468
 *         elif isinstance(data, ArrowArrayHandle):
 *             c_array = &(<ArrowArrayHandle> data).array
 *         elif isinstance(data, int):             # <<<<<<<<<<<<<<
 *             c_array = <CArrowArray*> data
 *         else:
*/
  __pyx_t_2 = PyLong_Check(__pyx_v_data); 
  if (likely(__pyx_t_2)) {

    /* "adbc_driver_manager/_lib.pyx":1469
 *             c_array = &(<ArrowArrayHandle> data).array
 *         elif isinstance(data, int):
 *             c_array = <CArrowArray*> data             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(
*/
    __pyx_v_c_array = ((struct ArrowArray *)__pyx_v_data);

    /* "adbc_driver_manager/_lib.pyx":1468
 *         elif isinstance(data, ArrowArrayHandle):
 *             c_array = &(<ArrowArrayHandle> data).array
 *         elif isinstance(data, int):             # <<<<<<<<<<<<<<
 *             c_array = <CArrowArray*> data
 *         else:
*/
    goto __pyx_L9;
  }

  /* "adbc_driver_manager/_lib.pyx":1471
 *             c_array = <CArrowArray*> data
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 "data must be Arrow-compatible data (implementing the Arrow PyCapsule "
 *                 f"Protocol), a PyCapsule, int or ArrowArrayHandle, not {type(data)}"
*/
  /*else*/ {
    __pyx_t_6 = NULL;

    /* "adbc_driver_manager/_lib.pyx":1473
 *             raise TypeError(
 *                 "data must be Arrow-compatible data (implementing the Arrow PyCapsule "
 *                 f"Protocol), a PyCapsule, int or ArrowArrayHandle, not {type(data)}"             # <<<<<<<<<<<<<<
 *             )
 * 
*/
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_data)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "adbc_driver_manager/_lib.pyx":1472
 *         else:
 *             raise TypeError(
 *                 "data must be Arrow-compatible data (implementing the Arrow PyCapsule "             # <<<<<<<<<<<<<<
 *                 f"Protocol), a PyCapsule, int or ArrowArrayHandle, not {type(data)}"
 *             )
*/
    __pyx_t_9 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_data_must_be_Arrow_compatible_da, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_9};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1471, __pyx_L1_error)
  }
  __pyx_L9:;

  /* "adbc_driver_manager/_lib.pyx":1476
 *             )
 * 
 *         if is_pycapsule(schema, b"arrow_schema"):             # <<<<<<<<<<<<<<
 *             c_schema = <CArrowSchema*> PyCapsule_GetPointer(schema, "arrow_schema")
 *         elif isinstance(schema, ArrowSchemaHandle):
*/
  __pyx_t_9 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_is_pycapsule); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
    assert(__pyx_t_9);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_v_schema, __pyx_mstate_global->__pyx_n_b_arrow_schema};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1477
 * 
 *         if is_pycapsule(schema, b"arrow_schema"):
 *             c_schema = <CArrowSchema*> PyCapsule_GetPointer(schema, "arrow_schema")             # <<<<<<<<<<<<<<
 *         elif isinstance(schema, ArrowSchemaHandle):
 *             c_schema = &(<ArrowSchemaHandle> schema).schema
*/
    __pyx_t_11 = PyCapsule_GetPointer(__pyx_v_schema, __pyx_k_arrow_schema); if (unlikely(__pyx_t_11 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 1477, __pyx_L1_error)
    __pyx_v_c_schema = ((struct ArrowSchema *)__pyx_t_11);

    /* "adbc_driver_manager/_lib.pyx":1476
 *             )
 * 
 *         if is_pycapsule(schema, b"arrow_schema"):             # <<<<<<<<<<<<<<
 *             c_schema = <CArrowSchema*> PyCapsule_GetPointer(schema, "arrow_schema")
 *         elif isinstance(schema, ArrowSchemaHandle):
*/
    goto __pyx_L10;
  }

  /* "adbc_driver_manager/_lib.pyx":1478
 *         if is_pycapsule(schema, b"arrow_schema"):
 *             c_schema = <CArrowSchema*> PyCapsule_GetPointer(schema, "arrow_schema")
 *         elif isinstance(schema, ArrowSchemaHandle):             # <<<<<<<<<<<<<<
 *             c_schema = &(<ArrowSchemaHandle> schema).schema
 *         elif isinstance(schema, int):
*/
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_schema, __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle); 
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1479
 *             c_schema = <CArrowSchema*> PyCapsule_GetPointer(schema, "arrow_schema")
 *         elif isinstance(schema, ArrowSchemaHandle):
 *             c_schema = &(<ArrowSchemaHandle> schema).schema             # <<<<<<<<<<<<<<
 *         elif isinstance(schema, int):
 *             c_schema = <CArrowSchema*> schema
*/
    __pyx_v_c_schema = (&((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_v_schema)->schema);

    /* "adbc_driver_manager/_lib.pyx":1478
 *         if is_pycapsule(schema, b"arrow_schema"):
 *             c_schema = <CArrowSchema*> PyCapsule_GetPointer(schema, "arrow_schema")
 *         elif isinstance(schema, ArrowSchemaHandle):             # <<<<<<<<<<<<<<
 *             c_schema = &(<ArrowSchemaHandle> schema).schema
 *         elif isinstance(schema, int):
*/
    goto __pyx_L10;
  }

  /* "adbc_driver_manager/_lib.pyx":1480
 *         elif isinstance(schema, ArrowSchemaHandle):
 *             c_schema = &(<ArrowSchemaHandle> schema).schema
 *         elif isinstance(schema, int):             # <<<<<<<<<<<<<<
 *             c_schema = <CArrowSchema*> schema
 *         else:
*/
  __pyx_t_2 = PyLong_Check(__pyx_v_schema); 
  if (likely(__pyx_t_2)) {

    /* "adbc_driver_manager/_lib.pyx":1481
 *             c_schema = &(<ArrowSchemaHandle> schema).schema
 *         elif isinstance(schema, int):
 *             c_schema = <CArrowSchema*> schema             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("schema must be a PyCapsule, int or ArrowSchemaHandle, "
*/
    __pyx_v_c_schema = ((struct ArrowSchema *)__pyx_v_schema);

    /* "adbc_driver_manager/_lib.pyx":1480
 *         elif isinstance(schema, ArrowSchemaHandle):
 *             c_schema = &(<ArrowSchemaHandle> schema).schema
 *         elif isinstance(schema, int):             # <<<<<<<<<<<<<<
 *             c_schema = <CArrowSchema*> schema
 *         else:
*/
    goto __pyx_L10;
  }

  /* "adbc_driver_manager/_lib.pyx":1483
 *             c_schema = <CArrowSchema*> schema
 *         else:
 *             raise TypeError("schema must be a PyCapsule, int or ArrowSchemaHandle, "             # <<<<<<<<<<<<<<
 *                             f"not {type(schema)}")
 * 
*/
  /*else*/ {
    __pyx_t_6 = NULL;

    /* "adbc_driver_manager/_lib.pyx":1484
 *         else:
 *             raise TypeError("schema must be a PyCapsule, int or ArrowSchemaHandle, "
 *                             f"not {type(schema)}")             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
    __pyx_t_9 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_schema)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "adbc_driver_manager/_lib.pyx":1483
 *             c_schema = <CArrowSchema*> schema
 *         else:
 *             raise TypeError("schema must be a PyCapsule, int or ArrowSchemaHandle, "             # <<<<<<<<<<<<<<
 *                             f"not {type(schema)}")
 * 
*/
    __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_schema_must_be_a_PyCapsule_int_o, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_8};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1483, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "adbc_driver_manager/_lib.pyx":1486
 *                             f"not {type(schema)}")
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementBind(
 *                 &self.statement,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1487
 * 
 *         with nogil:
 *             status = AdbcStatementBind(             # <<<<<<<<<<<<<<
 *                 &self.statement,
 *                 c_array,
*/
        __pyx_v_status = AdbcStatementBind((&__pyx_v_self->statement), __pyx_v_c_array, __pyx_v_c_schema, (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1486
 *                             f"not {type(schema)}")
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementBind(
 *                 &self.statement,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L13;
        }
        __pyx_L13:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1492
 *                 c_schema,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def bind_stream(self, stream) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1492, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1442
 *         connection._open_child()
 * 
 *     def bind(self, data, schema=None) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Bind an ArrowArray to this statement.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_schema);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1494
 *         check_error(status, &c_error)
 * 
 *     def bind_stream(self, stream) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Bind an ArrowArrayStream to this statement.
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_5bind_stream(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_4bind_stream, "\n        Bind an ArrowArrayStream to this statement.\n\n        Parameters\n        ----------\n        stream : PyCapsule or int or ArrowArrayStreamHandle\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_5bind_stream = {"bind_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_5bind_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_4bind_stream};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_5bind_stream(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_stream = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bind_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1494, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bind_stream", 0) < (0)) __PYX_ERR(0, 1494, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bind_stream", 1, 1, 1, i); __PYX_ERR(0, 1494, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1494, __pyx_L3_error)
    }
    __pyx_v_stream = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bind_stream", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1494, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.bind_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_4bind_stream(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_4bind_stream(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_stream) {
  struct AdbcError __pyx_v_c_error;
  struct ArrowArrayStream *__pyx_v_c_stream;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  void *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind_stream", 0);
  __Pyx_INCREF(__pyx_v_stream);

  /* "adbc_driver_manager/_lib.pyx":1502
 *         stream : PyCapsule or int or ArrowArrayStreamHandle
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CArrowArrayStream* c_stream
 * 
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1502, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1506
 * 
 *         if (
 *             hasattr(stream, "__arrow_c_stream__")             # <<<<<<<<<<<<<<
 *             and not isinstance(stream, ArrowArrayStreamHandle)
 *         ):
*/
  __pyx_t_3 = __Pyx_HasAttr(__pyx_v_stream, __pyx_mstate_global->__pyx_n_u_arrow_c_stream); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1506, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "adbc_driver_manager/_lib.pyx":1507
 *         if (
 *             hasattr(stream, "__arrow_c_stream__")
 *             and not isinstance(stream, ArrowArrayStreamHandle)             # <<<<<<<<<<<<<<
 *         ):
 *             stream = stream.__arrow_c_stream__()
*/
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_stream, __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle); 
  __pyx_t_4 = (!__pyx_t_3);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "adbc_driver_manager/_lib.pyx":1505
 *         cdef CArrowArrayStream* c_stream
 * 
 *         if (             # <<<<<<<<<<<<<<
 *             hasattr(stream, "__arrow_c_stream__")
 *             and not isinstance(stream, ArrowArrayStreamHandle)
*/
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1509
 *             and not isinstance(stream, ArrowArrayStreamHandle)
 *         ):
 *             stream = stream.__arrow_c_stream__()             # <<<<<<<<<<<<<<
 * 
 *         if is_pycapsule(stream, b"arrow_array_stream"):
*/
    __pyx_t_6 = __pyx_v_stream;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_arrow_c_stream, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_DECREF_SET(__pyx_v_stream, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "adbc_driver_manager/_lib.pyx":1505
 *         cdef CArrowArrayStream* c_stream
 * 
 *         if (             # <<<<<<<<<<<<<<
 *             hasattr(stream, "__arrow_c_stream__")
 *             and not isinstance(stream, ArrowArrayStreamHandle)
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1511
 *             stream = stream.__arrow_c_stream__()
 * 
 *         if is_pycapsule(stream, b"arrow_array_stream"):             # <<<<<<<<<<<<<<
 *             c_stream = <CArrowArrayStream*> PyCapsule_GetPointer(
 *                 stream, "arrow_array_stream"
*/
  __pyx_t_6 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_is_pycapsule); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
    assert(__pyx_t_6);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_stream, __pyx_mstate_global->__pyx_n_b_arrow_array_stream};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1512
 * 
 *         if is_pycapsule(stream, b"arrow_array_stream"):
 *             c_stream = <CArrowArrayStream*> PyCapsule_GetPointer(             # <<<<<<<<<<<<<<
 *                 stream, "arrow_array_stream"
 *             )
*/
    __pyx_t_9 = PyCapsule_GetPointer(__pyx_v_stream, __pyx_k_arrow_array_stream); if (unlikely(__pyx_t_9 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 1512, __pyx_L1_error)
    __pyx_v_c_stream = ((struct ArrowArrayStream *)__pyx_t_9);

    /* "adbc_driver_manager/_lib.pyx":1511
 *             stream = stream.__arrow_c_stream__()
 * 
 *         if is_pycapsule(stream, b"arrow_array_stream"):             # <<<<<<<<<<<<<<
 *             c_stream = <CArrowArrayStream*> PyCapsule_GetPointer(
 *                 stream, "arrow_array_stream"
*/
    goto __pyx_L6;
  }

  /* "adbc_driver_manager/_lib.pyx":1515
 *                 stream, "arrow_array_stream"
 *             )
 *         elif isinstance(stream, ArrowArrayStreamHandle):             # <<<<<<<<<<<<<<
 *             c_stream = &(<ArrowArrayStreamHandle> stream).stream
 *         elif isinstance(stream, int):
*/
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_stream, __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle); 
  if (__pyx_t_2) {

    /* "adbc_driver_manager/_lib.pyx":1516
 *             )
 *         elif isinstance(stream, ArrowArrayStreamHandle):
 *             c_stream = &(<ArrowArrayStreamHandle> stream).stream             # <<<<<<<<<<<<<<
 *         elif isinstance(stream, int):
 *             c_stream = <CArrowArrayStream*> stream
*/
    __pyx_v_c_stream = (&((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_v_stream)->stream);

    /* "adbc_driver_manager/_lib.pyx":1515
 *                 stream, "arrow_array_stream"
 *             )
 *         elif isinstance(stream, ArrowArrayStreamHandle):             # <<<<<<<<<<<<<<
 *             c_stream = &(<ArrowArrayStreamHandle> stream).stream
 *         elif isinstance(stream, int):
*/
    goto __pyx_L6;
  }

  /* "adbc_driver_manager/_lib.pyx":1517
 *         elif isinstance(stream, ArrowArrayStreamHandle):
 *             c_stream = &(<ArrowArrayStreamHandle> stream).stream
 *         elif isinstance(stream, int):             # <<<<<<<<<<<<<<
 *             c_stream = <CArrowArrayStream*> stream
 *         else:
*/
  __pyx_t_2 = PyLong_Check(__pyx_v_stream); 
  if (likely(__pyx_t_2)) {

    /* "adbc_driver_manager/_lib.pyx":1518
 *             c_stream = &(<ArrowArrayStreamHandle> stream).stream
 *         elif isinstance(stream, int):
 *             c_stream = <CArrowArrayStream*> stream             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(f"data must be a PyCapsule, int or ArrowArrayStreamHandle, "
*/
    __pyx_v_c_stream = ((struct ArrowArrayStream *)__pyx_v_stream);

    /* "adbc_driver_manager/_lib.pyx":1517
 *         elif isinstance(stream, ArrowArrayStreamHandle):
 *             c_stream = &(<ArrowArrayStreamHandle> stream).stream
 *         elif isinstance(stream, int):             # <<<<<<<<<<<<<<
 *             c_stream = <CArrowArrayStream*> stream
 *         else:
*/
    goto __pyx_L6;
  }

  /* "adbc_driver_manager/_lib.pyx":1520
 *             c_stream = <CArrowArrayStream*> stream
 *         else:
 *             raise TypeError(f"data must be a PyCapsule, int or ArrowArrayStreamHandle, "             # <<<<<<<<<<<<<<
 *                             f"not {type(stream)}")
 * 
*/
  /*else*/ {
    __pyx_t_8 = NULL;

    /* "adbc_driver_manager/_lib.pyx":1521
 *         else:
 *             raise TypeError(f"data must be a PyCapsule, int or ArrowArrayStreamHandle, "
 *                             f"not {type(stream)}")             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_stream)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "adbc_driver_manager/_lib.pyx":1520
 *             c_stream = <CArrowArrayStream*> stream
 *         else:
 *             raise TypeError(f"data must be a PyCapsule, int or ArrowArrayStreamHandle, "             # <<<<<<<<<<<<<<
 *                             f"not {type(stream)}")
 * 
*/
    __pyx_t_10 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_data_must_be_a_PyCapsule_int_or, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_10};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1520, __pyx_L1_error)
  }
  __pyx_L6:;

  /* "adbc_driver_manager/_lib.pyx":1523
 *                             f"not {type(stream)}")
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementBindStream(
 *                 &self.statement,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1524
 * 
 *         with nogil:
 *             status = AdbcStatementBindStream(             # <<<<<<<<<<<<<<
 *                 &self.statement,
 *                 c_stream,
*/
        __pyx_v_status = AdbcStatementBindStream((&__pyx_v_self->statement), __pyx_v_c_stream, (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1523
 *                             f"not {type(stream)}")
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementBindStream(
 *                 &self.statement,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1528
 *                 c_stream,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def cancel(self) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1528, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1494
 *         check_error(status, &c_error)
 * 
 *     def bind_stream(self, stream) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Bind an ArrowArrayStream to this statement.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.bind_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_stream);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1530
 *         check_error(status, &c_error)
 * 
 *     def cancel(self) -> None:             # <<<<<<<<<<<<<<
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_7cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_6cancel, "Attempt to cancel any ongoing operations on the connection.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_7cancel = {"cancel", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_7cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_6cancel};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_7cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("cancel", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("cancel", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_6cancel(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_6cancel(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 0);

  /* "adbc_driver_manager/_lib.pyx":1532
 *     def cancel(self) -> None:
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         with nogil:
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1532, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1534
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementCancel(&self.statement, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1535
 *         cdef CAdbcStatusCode status
 *         with nogil:
 *             status = AdbcStatementCancel(&self.statement, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcStatementCancel((&__pyx_v_self->statement), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1534
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementCancel(&self.statement, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1536
 *         with nogil:
 *             status = AdbcStatementCancel(&self.statement, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def close(self) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1530
 *         check_error(status, &c_error)
 * 
 *     def cancel(self) -> None:             # <<<<<<<<<<<<<<
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1538
 *         check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the statement."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_9close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_8close, "Release the handle to the statement.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_9close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_9close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_8close};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_9close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_8close(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_8close(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "adbc_driver_manager/_lib.pyx":1540
 *     def close(self) -> None:
 *         """Release the handle to the statement."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         self.connection._close_child()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1540, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1542
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         self.connection._close_child()             # <<<<<<<<<<<<<<
 *         with self._lock:
 *             if self.statement.private_data == NULL:
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcConnection *)__pyx_v_self->connection->__pyx_base.__pyx_vtab)->__pyx_base._close_child(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v_self->connection)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1543
 *         cdef CAdbcStatusCode status
 *         self.connection._close_child()
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self.statement.private_data == NULL:
 *                 return
*/
  /*with:*/ {
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->__pyx_base._lock, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_v_self->__pyx_base._lock, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1543, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1543, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1544
 *         self.connection._close_child()
 *         with self._lock:
 *             if self.statement.private_data == NULL:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          __pyx_t_10 = (__pyx_v_self->statement.private_data == NULL);
          if (__pyx_t_10) {

            /* "adbc_driver_manager/_lib.pyx":1545
 *         with self._lock:
 *             if self.statement.private_data == NULL:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             with nogil:
*/
            __Pyx_XDECREF(__pyx_r);
            __pyx_r = Py_None; __Pyx_INCREF(Py_None);
            goto __pyx_L11_try_return;

            /* "adbc_driver_manager/_lib.pyx":1544
 *         self.connection._close_child()
 *         with self._lock:
 *             if self.statement.private_data == NULL:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          }

          /* "adbc_driver_manager/_lib.pyx":1547
 *                 return
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcStatementRelease(&self.statement, &c_error)
 *             check_error(status, &c_error)
*/
          {
              PyThreadState * _save;
              _save = PyEval_SaveThread();
              __Pyx_FastGIL_Remember();
              /*try:*/ {

                /* "adbc_driver_manager/_lib.pyx":1548
 * 
 *             with nogil:
 *                 status = AdbcStatementRelease(&self.statement, &c_error)             # <<<<<<<<<<<<<<
 *             check_error(status, &c_error)
 * 
*/
                __pyx_v_status = AdbcStatementRelease((&__pyx_v_self->statement), (&__pyx_v_c_error));
              }

              /* "adbc_driver_manager/_lib.pyx":1547
 *                 return
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcStatementRelease(&self.statement, &c_error)
 *             check_error(status, &c_error)
*/
              /*finally:*/ {
                /*normal exit:*/{
                  __Pyx_FastGIL_Forget();
                  PyEval_RestoreThread(_save);
                  goto __pyx_L16;
                }
                __pyx_L16:;
              }
          }

          /* "adbc_driver_manager/_lib.pyx":1549
 *             with nogil:
 *                 status = AdbcStatementRelease(&self.statement, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def execute_query(self) -> Tuple[ArrowArrayStreamHandle, int]:
*/
          __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1549, __pyx_L7_error)

          /* "adbc_driver_manager/_lib.pyx":1543
 *         cdef CAdbcStatusCode status
 *         self.connection._close_child()
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             if self.statement.private_data == NULL:
 *                 return
*/
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 1543, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_4);
          __pyx_t_11 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1543, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1543, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_10 < (0)) __PYX_ERR(0, 1543, __pyx_L9_except_error)
          __pyx_t_13 = (!__pyx_t_10);
          if (unlikely(__pyx_t_13)) {
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_XGIVEREF(__pyx_t_4);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_5, __pyx_t_4);
            __pyx_t_2 = 0;  __pyx_t_5 = 0;  __pyx_t_4 = 0; 
            __PYX_ERR(0, 1543, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L11_try_return:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L4_return;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1543, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L4_return: {
        __pyx_t_9 = __pyx_r;
        __pyx_r = 0;
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1543, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_r = __pyx_t_9;
        __pyx_t_9 = 0;
        goto __pyx_L0;
      }
      __pyx_L6:;
    }
    goto __pyx_L20;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L20:;
  }

  /* "adbc_driver_manager/_lib.pyx":1538
 *         check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the statement."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1551
 *             check_error(status, &c_error)
 * 
 *     def execute_query(self) -> Tuple[ArrowArrayStreamHandle, int]:             # <<<<<<<<<<<<<<
 *         """
 *         Execute the query and get the result set.
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_11execute_query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_10execute_query, "\n        Execute the query and get the result set.\n\n        Returns\n        -------\n        ArrowArrayStreamHandle\n            The result set.\n        int\n            The number of rows if known, else -1.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_11execute_query = {"execute_query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_11execute_query, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_10execute_query};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_11execute_query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_query (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("execute_query", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("execute_query", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_10execute_query(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_10execute_query(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *__pyx_v_stream = 0;
  int64_t __pyx_v_rows_affected;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_query", 0);

  /* "adbc_driver_manager/_lib.pyx":1562
 *             The number of rows if known, else -1.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 *         cdef int64_t rows_affected = 0
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1562, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1563
 *         """
 *         cdef CAdbcError c_error = empty_error()
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()             # <<<<<<<<<<<<<<
 *         cdef int64_t rows_affected = 0
 *         with nogil:
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1563, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_stream = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1564
 *         cdef CAdbcError c_error = empty_error()
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 *         cdef int64_t rows_affected = 0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementExecuteQuery(
*/
  __pyx_v_rows_affected = 0;

  /* "adbc_driver_manager/_lib.pyx":1565
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 *         cdef int64_t rows_affected = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecuteQuery(
 *                 &self.statement,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1566
 *         cdef int64_t rows_affected = 0
 *         with nogil:
 *             status = AdbcStatementExecuteQuery(             # <<<<<<<<<<<<<<
 *                 &self.statement,
 *                 &stream.stream,
*/
        __pyx_v_status = AdbcStatementExecuteQuery((&__pyx_v_self->statement), (&__pyx_v_stream->stream), (&__pyx_v_rows_affected), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1565
 *         cdef ArrowArrayStreamHandle stream = ArrowArrayStreamHandle()
 *         cdef int64_t rows_affected = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecuteQuery(
 *                 &self.statement,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1571
 *                 &rows_affected,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return (stream, rows_affected)
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1571, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1572
 *                 &c_error)
 *         check_error(status, &c_error)
 *         return (stream, rows_affected)             # <<<<<<<<<<<<<<
 * 
 *     def execute_partitions(self) -> \
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_rows_affected); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __Pyx_GIVEREF((PyObject *)__pyx_v_stream);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_stream)) != (0)) __PYX_ERR(0, 1572, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 1572, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1551
 *             check_error(status, &c_error)
 * 
 *     def execute_query(self) -> Tuple[ArrowArrayStreamHandle, int]:             # <<<<<<<<<<<<<<
 *         """
 *         Execute the query and get the result set.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.execute_query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1574
 *         return (stream, rows_affected)
 * 
 *     def execute_partitions(self) -> \             # <<<<<<<<<<<<<<
 *             Tuple[List[bytes], Optional[ArrowSchemaHandle], int]:
 *         """
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_13execute_partitions(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_12execute_partitions, "\n        Execute the query and get the partitions of the result set.\n\n        Not all drivers will support this.\n\n        Returns\n        -------\n        list of byte\n            The partitions of the distributed result set.\n        ArrowSchemaHandle or None\n            The schema of the result set.  May be None if incremental\n            execution is enabled and the server does not return a schema.\n        int\n            The number of rows if known, else -1.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_13execute_partitions = {"execute_partitions", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_13execute_partitions, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_12execute_partitions};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_13execute_partitions(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_partitions (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("execute_partitions", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("execute_partitions", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_12execute_partitions(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_12execute_partitions(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_schema = 0;
  struct AdbcPartitions __pyx_v_c_partitions;
  int64_t __pyx_v_rows_affected;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_partitions = NULL;
  size_t __pyx_v_i;
  size_t __pyx_v_length;
  char const *__pyx_v_data;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  struct AdbcPartitions __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_partitions", 0);

  /* "adbc_driver_manager/_lib.pyx":1591
 *             The number of rows if known, else -1.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()
 *         cdef CAdbcPartitions c_partitions = CAdbcPartitions(
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1591, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1592
 *         """
 *         cdef CAdbcError c_error = empty_error()
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()             # <<<<<<<<<<<<<<
 *         cdef CAdbcPartitions c_partitions = CAdbcPartitions(
 *             0, NULL, NULL, NULL, NULL)
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1592, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_schema = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1593
 *         cdef CAdbcError c_error = empty_error()
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()
 *         cdef CAdbcPartitions c_partitions = CAdbcPartitions(             # <<<<<<<<<<<<<<
 *             0, NULL, NULL, NULL, NULL)
 *         cdef int64_t rows_affected = 0
*/
  __pyx_t_5.num_partitions = 0;

  /* "adbc_driver_manager/_lib.pyx":1594
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()
 *         cdef CAdbcPartitions c_partitions = CAdbcPartitions(
 *             0, NULL, NULL, NULL, NULL)             # <<<<<<<<<<<<<<
 *         cdef int64_t rows_affected = 0
 * 
*/
  __pyx_t_5.partitions = NULL;
  __pyx_t_5.partition_lengths = NULL;
  __pyx_t_5.private_data = NULL;
  __pyx_t_5.release = NULL;
  __pyx_v_c_partitions = __pyx_t_5;

  /* "adbc_driver_manager/_lib.pyx":1595
 *         cdef CAdbcPartitions c_partitions = CAdbcPartitions(
 *             0, NULL, NULL, NULL, NULL)
 *         cdef int64_t rows_affected = 0             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  __pyx_v_rows_affected = 0;

  /* "adbc_driver_manager/_lib.pyx":1597
 *         cdef int64_t rows_affected = 0
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecutePartitions(
 *                 &self.statement,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1598
 * 
 *         with nogil:
 *             status = AdbcStatementExecutePartitions(             # <<<<<<<<<<<<<<
 *                 &self.statement,
 *                 &schema.schema,
*/
        __pyx_v_status = AdbcStatementExecutePartitions((&__pyx_v_self->statement), (&__pyx_v_schema->schema), (&__pyx_v_c_partitions), (&__pyx_v_rows_affected), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1597
 *         cdef int64_t rows_affected = 0
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecutePartitions(
 *                 &self.statement,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1604
 *                 &rows_affected,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *         partitions = []
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1604, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1606
 *         check_error(status, &c_error)
 * 
 *         partitions = []             # <<<<<<<<<<<<<<
 *         for i in range(c_partitions.num_partitions):
 *             length = c_partitions.partition_lengths[i]
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_partitions = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1607
 * 
 *         partitions = []
 *         for i in range(c_partitions.num_partitions):             # <<<<<<<<<<<<<<
 *             length = c_partitions.partition_lengths[i]
 *             data = <const char*> c_partitions.partitions[i]
*/
  __pyx_t_4 = __pyx_v_c_partitions.num_partitions;
  __pyx_t_6 = __pyx_t_4;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "adbc_driver_manager/_lib.pyx":1608
 *         partitions = []
 *         for i in range(c_partitions.num_partitions):
 *             length = c_partitions.partition_lengths[i]             # <<<<<<<<<<<<<<
 *             data = <const char*> c_partitions.partitions[i]
 *             partitions.append(PyBytes_FromStringAndSize(data, length))
*/
    __pyx_v_length = (__pyx_v_c_partitions.partition_lengths[__pyx_v_i]);

    /* "adbc_driver_manager/_lib.pyx":1609
 *         for i in range(c_partitions.num_partitions):
 *             length = c_partitions.partition_lengths[i]
 *             data = <const char*> c_partitions.partitions[i]             # <<<<<<<<<<<<<<
 *             partitions.append(PyBytes_FromStringAndSize(data, length))
 *         c_partitions.release(&c_partitions)
*/
    __pyx_v_data = ((char const *)(__pyx_v_c_partitions.partitions[__pyx_v_i]));

    /* "adbc_driver_manager/_lib.pyx":1610
 *             length = c_partitions.partition_lengths[i]
 *             data = <const char*> c_partitions.partitions[i]
 *             partitions.append(PyBytes_FromStringAndSize(data, length))             # <<<<<<<<<<<<<<
 *         c_partitions.release(&c_partitions)
 * 
*/
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_data, __pyx_v_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_partitions, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1610, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "adbc_driver_manager/_lib.pyx":1611
 *             data = <const char*> c_partitions.partitions[i]
 *             partitions.append(PyBytes_FromStringAndSize(data, length))
 *         c_partitions.release(&c_partitions)             # <<<<<<<<<<<<<<
 * 
 *         if schema.schema.release == NULL:
*/
  __pyx_v_c_partitions.release((&__pyx_v_c_partitions));

  /* "adbc_driver_manager/_lib.pyx":1613
 *         c_partitions.release(&c_partitions)
 * 
 *         if schema.schema.release == NULL:             # <<<<<<<<<<<<<<
 *             return partitions, None, rows_affected
 *         return partitions, schema, rows_affected
*/
  __pyx_t_9 = (__pyx_v_schema->schema.release == NULL);
  if (__pyx_t_9) {

    /* "adbc_driver_manager/_lib.pyx":1614
 * 
 *         if schema.schema.release == NULL:
 *             return partitions, None, rows_affected             # <<<<<<<<<<<<<<
 *         return partitions, schema, rows_affected
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_rows_affected); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_partitions);
    __Pyx_GIVEREF(__pyx_v_partitions);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_partitions) != (0)) __PYX_ERR(0, 1614, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, Py_None) != (0)) __PYX_ERR(0, 1614, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2) != (0)) __PYX_ERR(0, 1614, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "adbc_driver_manager/_lib.pyx":1613
 *         c_partitions.release(&c_partitions)
 * 
 *         if schema.schema.release == NULL:             # <<<<<<<<<<<<<<
 *             return partitions, None, rows_affected
 *         return partitions, schema, rows_affected
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1615
 *         if schema.schema.release == NULL:
 *             return partitions, None, rows_affected
 *         return partitions, schema, rows_affected             # <<<<<<<<<<<<<<
 * 
 *     def execute_schema(self) -> ArrowSchemaHandle:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyLong_From_int64_t(__pyx_v_rows_affected); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_partitions);
  __Pyx_GIVEREF(__pyx_v_partitions);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_partitions) != (0)) __PYX_ERR(0, 1615, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_schema);
  __Pyx_GIVEREF((PyObject *)__pyx_v_schema);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_schema)) != (0)) __PYX_ERR(0, 1615, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3) != (0)) __PYX_ERR(0, 1615, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1574
 *         return (stream, rows_affected)
 * 
 *     def execute_partitions(self) -> \             # <<<<<<<<<<<<<<
 *             Tuple[List[bytes], Optional[ArrowSchemaHandle], int]:
 *         """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.execute_partitions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_schema);
  __Pyx_XDECREF(__pyx_v_partitions);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1617
 *         return partitions, schema, rows_affected
 * 
 *     def execute_schema(self) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the schema of the result set without executing the query.
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_15execute_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_14execute_schema, "\n        Get the schema of the result set without executing the query.\n\n        Returns\n        -------\n        ArrowSchemaHandle\n            The schema of the result set.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_15execute_schema = {"execute_schema", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_15execute_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_14execute_schema};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_15execute_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("execute_schema", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("execute_schema", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_14execute_schema(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_14execute_schema(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_schema = 0;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_schema", 0);

  /* "adbc_driver_manager/_lib.pyx":1626
 *             The schema of the result set.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()
 *         with nogil:
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1626, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1627
 *         """
 *         cdef CAdbcError c_error = empty_error()
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementExecuteSchema(
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1627, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_schema = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1628
 *         cdef CAdbcError c_error = empty_error()
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecuteSchema(
 *                 &self.statement,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1629
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()
 *         with nogil:
 *             status = AdbcStatementExecuteSchema(             # <<<<<<<<<<<<<<
 *                 &self.statement,
 *                 &schema.schema,
*/
        __pyx_v_status = AdbcStatementExecuteSchema((&__pyx_v_self->statement), (&__pyx_v_schema->schema), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1628
 *         cdef CAdbcError c_error = empty_error()
 *         cdef ArrowSchemaHandle schema = ArrowSchemaHandle()
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecuteSchema(
 *                 &self.statement,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1633
 *                 &schema.schema,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return schema
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1633, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1634
 *                 &c_error)
 *         check_error(status, &c_error)
 *         return schema             # <<<<<<<<<<<<<<
 * 
 *     def execute_update(self) -> int:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_schema);
  __pyx_r = __pyx_v_schema;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1617
 *         return partitions, schema, rows_affected
 * 
 *     def execute_schema(self) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the schema of the result set without executing the query.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.execute_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_schema);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1636
 *         return schema
 * 
 *     def execute_update(self) -> int:             # <<<<<<<<<<<<<<
 *         """
 *         Execute the query without a result set.
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_17execute_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_16execute_update, "\n        Execute the query without a result set.\n\n        Returns\n        -------\n        int\n            The number of affected rows if known, else -1.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_17execute_update = {"execute_update", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_17execute_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_16execute_update};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_17execute_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("execute_update", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("execute_update", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_16execute_update(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_16execute_update(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  int64_t __pyx_v_rows_affected;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_update", 0);

  /* "adbc_driver_manager/_lib.pyx":1645
 *             The number of affected rows if known, else -1.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef int64_t rows_affected = 0
 *         with nogil:
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1645, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1646
 *         """
 *         cdef CAdbcError c_error = empty_error()
 *         cdef int64_t rows_affected = 0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementExecuteQuery(
*/
  __pyx_v_rows_affected = 0;

  /* "adbc_driver_manager/_lib.pyx":1647
 *         cdef CAdbcError c_error = empty_error()
 *         cdef int64_t rows_affected = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecuteQuery(
 *                 &self.statement,
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1648
 *         cdef int64_t rows_affected = 0
 *         with nogil:
 *             status = AdbcStatementExecuteQuery(             # <<<<<<<<<<<<<<
 *                 &self.statement,
 *                 NULL,
*/
        __pyx_v_status = AdbcStatementExecuteQuery((&__pyx_v_self->statement), NULL, (&__pyx_v_rows_affected), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1647
 *         cdef CAdbcError c_error = empty_error()
 *         cdef int64_t rows_affected = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementExecuteQuery(
 *                 &self.statement,
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1653
 *                 &rows_affected,
 *                 &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return rows_affected
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1653, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1654
 *                 &c_error)
 *         check_error(status, &c_error)
 *         return rows_affected             # <<<<<<<<<<<<<<
 * 
 *     def get_option(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_rows_affected); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyInt_FromNumber(&__pyx_t_2, NULL, 0) < (0)) __PYX_ERR(0, 1654, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1636
 *         return schema
 * 
 *     def execute_update(self) -> int:             # <<<<<<<<<<<<<<
 *         """
 *         Execute the query without a result set.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.execute_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1656
 *         return rows_affected
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_19get_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_18get_option, "\n        Get the value of a string option.\n\n        Parameters\n        ----------\n        key : str or bytes\n            The option to get.\n        encoding : str\n            The encoding of the option value.  This should almost\n            always be UTF-8.\n        errors : str\n            What to do about errors when decoding the option value\n            (see bytes.decode).\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_19get_option = {"get_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_19get_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_18get_option};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_19get_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_encoding = 0;
  PyObject *__pyx_v_errors = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,&__pyx_mstate_global->__pyx_n_u_encoding,&__pyx_mstate_global->__pyx_n_u_errors,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1656, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1656, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option", 0) < (0)) __PYX_ERR(0, 1656, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_kp_u_utf_8));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_strict));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option", 1, 1, 1, i); __PYX_ERR(0, 1656, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1656, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_kp_u_utf_8));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_strict));
    }
    __pyx_v_key = values[0];
    __pyx_v_encoding = values[1];
    __pyx_v_errors = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1656, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_18get_option(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_key, __pyx_v_encoding, __pyx_v_errors);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_18get_option(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  size_t __pyx_v_c_len;
  char *__pyx_v_c_buf;
  PyObject *__pyx_v_buf = NULL;
  Py_ssize_t __pyx_v_new_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10[5];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option", 0);

  /* "adbc_driver_manager/_lib.pyx":1677
 *             (see bytes.decode).
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1677, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1679
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1680
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef size_t c_len = 0
 *         cdef char* c_buf = NULL
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1680, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1680, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1681
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0             # <<<<<<<<<<<<<<
 *         cdef char* c_buf = NULL
 * 
*/
  __pyx_v_c_len = 0;

  /* "adbc_driver_manager/_lib.pyx":1682
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
 *         cdef char* c_buf = NULL             # <<<<<<<<<<<<<<
 * 
 *         buf = bytearray(1024)
*/
  __pyx_v_c_buf = NULL;

  /* "adbc_driver_manager/_lib.pyx":1684
 *         cdef char* c_buf = NULL
 * 
 *         buf = bytearray(1024)             # <<<<<<<<<<<<<<
 *         while True:
 *             c_len = len(buf)
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_int_1024};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_buf = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1685
 * 
 *         buf = bytearray(1024)
 *         while True:             # <<<<<<<<<<<<<<
 *             c_len = len(buf)
 *             c_buf = buf
*/
  while (1) {

    /* "adbc_driver_manager/_lib.pyx":1686
 *         buf = bytearray(1024)
 *         while True:
 *             c_len = len(buf)             # <<<<<<<<<<<<<<
 *             c_buf = buf
 *             with nogil:
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1686, __pyx_L1_error)
    __pyx_v_c_len = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":1687
 *         while True:
 *             c_len = len(buf)
 *             c_buf = buf             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 status = AdbcStatementGetOption(
*/
    __pyx_t_3 = __Pyx_PyObject_AsWritableString(__pyx_v_buf); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1687, __pyx_L1_error)
    __pyx_v_c_buf = __pyx_t_3;

    /* "adbc_driver_manager/_lib.pyx":1688
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcStatementGetOption(
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
*/
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1689
 *             c_buf = buf
 *             with nogil:
 *                 status = AdbcStatementGetOption(             # <<<<<<<<<<<<<<
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
*/
          __pyx_v_status = AdbcStatementGetOption((&__pyx_v_self->statement), __pyx_v_c_key, __pyx_v_c_buf, (&__pyx_v_c_len), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":1688
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcStatementGetOption(
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "adbc_driver_manager/_lib.pyx":1691
 *                 status = AdbcStatementGetOption(
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *             if c_len <= len(buf):
 *                 # Entire value read
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1691, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":1692
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1692, __pyx_L1_error)
    __pyx_t_7 = (__pyx_v_c_len <= __pyx_t_6);
    if (__pyx_t_7) {

      /* "adbc_driver_manager/_lib.pyx":1694
 *             if c_len <= len(buf):
 *                 # Entire value read
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 # Buffer too small
*/
      goto __pyx_L4_break;

      /* "adbc_driver_manager/_lib.pyx":1692
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    }

    /* "adbc_driver_manager/_lib.pyx":1697
 *             else:
 *                 # Buffer too small
 *                 new_len = len(buf) * 2             # <<<<<<<<<<<<<<
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
*/
    /*else*/ {
      __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1697, __pyx_L1_error)
      __pyx_v_new_len = (__pyx_t_6 * 2);

      /* "adbc_driver_manager/_lib.pyx":1698
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      __pyx_t_2 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MAX_OPTION_SIZE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1698, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1698, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__pyx_t_7)) {

        /* "adbc_driver_manager/_lib.pyx":1699
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
*/
        __pyx_t_4 = NULL;

        /* "adbc_driver_manager/_lib.pyx":1700
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1700, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "adbc_driver_manager/_lib.pyx":1701
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")             # <<<<<<<<<<<<<<
 *                 buf = bytearray(new_len)
 * 
*/
        __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1701, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1701, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_read_option;
        __pyx_t_10[1] = __pyx_t_2;
        __pyx_t_10[2] = __pyx_mstate_global->__pyx_kp_u_would_need_more_than;
        __pyx_t_10[3] = __pyx_t_9;
        __pyx_t_10[4] = __pyx_mstate_global->__pyx_kp_u_bytes;

        /* "adbc_driver_manager/_lib.pyx":1700
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 5, 22 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 6, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1700, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
          __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1699, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_Raise(__pyx_t_8, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 1699, __pyx_L1_error)

        /* "adbc_driver_manager/_lib.pyx":1698
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      }

      /* "adbc_driver_manager/_lib.pyx":1702
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)             # <<<<<<<<<<<<<<
 * 
 *         # Remove trailing null terminator
*/
      __pyx_t_11 = NULL;
      __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_4};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_DECREF_SET(__pyx_v_buf, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;
    }
  }
  __pyx_L4_break:;

  /* "adbc_driver_manager/_lib.pyx":1705
 * 
 *         # Remove trailing null terminator
 *         if c_len > 0:             # <<<<<<<<<<<<<<
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)
*/
  __pyx_t_7 = (__pyx_v_c_len > 0);
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":1706
 *         # Remove trailing null terminator
 *         if c_len > 0:
 *             c_len -= 1             # <<<<<<<<<<<<<<
 *         return buf[:c_len].decode(encoding, errors)
 * 
*/
    __pyx_v_c_len = (__pyx_v_c_len - 1);

    /* "adbc_driver_manager/_lib.pyx":1705
 * 
 *         # Remove trailing null terminator
 *         if c_len > 0:             # <<<<<<<<<<<<<<
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1707
 *         if c_len > 0:
 *             c_len -= 1
 *         return buf[:c_len].decode(encoding, errors)             # <<<<<<<<<<<<<<
 * 
 *     def get_option_bytes(self, key: str) -> bytes:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_11 = PySequence_GetSlice(__pyx_v_buf, 0, __pyx_v_c_len); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = __pyx_t_11;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_encoding, __pyx_v_errors};
    __pyx_t_8 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_decode, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
  }
  __pyx_r = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1656
 *         return rows_affected
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1709
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_21get_option_bytes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_20get_option_bytes, "Get the value of a binary option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_21get_option_bytes = {"get_option_bytes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_21get_option_bytes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_20get_option_bytes};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_21get_option_bytes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_bytes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1709, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1709, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_bytes", 0) < (0)) __PYX_ERR(0, 1709, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_bytes", 1, 1, 1, i); __PYX_ERR(0, 1709, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1709, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_bytes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1709, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 1709, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_20get_option_bytes(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_20get_option_bytes(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  size_t __pyx_v_c_len;
  uint8_t *__pyx_v_c_buf;
  PyObject *__pyx_v_buf = NULL;
  Py_ssize_t __pyx_v_new_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  uint8_t *__pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10[5];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_bytes", 0);

  /* "adbc_driver_manager/_lib.pyx":1711
 *     def get_option_bytes(self, key: str) -> bytes:
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1711, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1713
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1714
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef size_t c_len = 0
 *         cdef uint8_t* c_buf = NULL
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1714, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1714, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1715
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0             # <<<<<<<<<<<<<<
 *         cdef uint8_t* c_buf = NULL
 * 
*/
  __pyx_v_c_len = 0;

  /* "adbc_driver_manager/_lib.pyx":1716
 *         cdef char* c_key = key_bytes
 *         cdef size_t c_len = 0
 *         cdef uint8_t* c_buf = NULL             # <<<<<<<<<<<<<<
 * 
 *         buf = bytearray(1024)
*/
  __pyx_v_c_buf = NULL;

  /* "adbc_driver_manager/_lib.pyx":1718
 *         cdef uint8_t* c_buf = NULL
 * 
 *         buf = bytearray(1024)             # <<<<<<<<<<<<<<
 *         while True:
 *             c_len = len(buf)
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_int_1024};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_buf = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1719
 * 
 *         buf = bytearray(1024)
 *         while True:             # <<<<<<<<<<<<<<
 *             c_len = len(buf)
 *             c_buf = buf
*/
  while (1) {

    /* "adbc_driver_manager/_lib.pyx":1720
 *         buf = bytearray(1024)
 *         while True:
 *             c_len = len(buf)             # <<<<<<<<<<<<<<
 *             c_buf = buf
 *             with nogil:
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1720, __pyx_L1_error)
    __pyx_v_c_len = __pyx_t_6;

    /* "adbc_driver_manager/_lib.pyx":1721
 *         while True:
 *             c_len = len(buf)
 *             c_buf = buf             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 status = AdbcStatementGetOptionBytes(
*/
    __pyx_t_7 = __Pyx_PyObject_AsWritableUString(__pyx_v_buf); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 1721, __pyx_L1_error)
    __pyx_v_c_buf = __pyx_t_7;

    /* "adbc_driver_manager/_lib.pyx":1722
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcStatementGetOptionBytes(
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
*/
    {
        PyThreadState * _save;
        _save = PyEval_SaveThread();
        __Pyx_FastGIL_Remember();
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1723
 *             c_buf = buf
 *             with nogil:
 *                 status = AdbcStatementGetOptionBytes(             # <<<<<<<<<<<<<<
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
*/
          __pyx_v_status = AdbcStatementGetOptionBytes((&__pyx_v_self->statement), __pyx_v_c_key, __pyx_v_c_buf, (&__pyx_v_c_len), (&__pyx_v_c_error));
        }

        /* "adbc_driver_manager/_lib.pyx":1722
 *             c_len = len(buf)
 *             c_buf = buf
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 status = AdbcStatementGetOptionBytes(
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
*/
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_FastGIL_Forget();
            PyEval_RestoreThread(_save);
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "adbc_driver_manager/_lib.pyx":1725
 *                 status = AdbcStatementGetOptionBytes(
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *             if c_len <= len(buf):
 *                 # Entire value read
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1725, __pyx_L1_error)

    /* "adbc_driver_manager/_lib.pyx":1726
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1726, __pyx_L1_error)
    __pyx_t_8 = (__pyx_v_c_len <= __pyx_t_6);
    if (__pyx_t_8) {

      /* "adbc_driver_manager/_lib.pyx":1728
 *             if c_len <= len(buf):
 *                 # Entire value read
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 # Buffer too small
*/
      goto __pyx_L4_break;

      /* "adbc_driver_manager/_lib.pyx":1726
 *                     &self.statement, c_key, c_buf, &c_len, &c_error)
 *             check_error(status, &c_error)
 *             if c_len <= len(buf):             # <<<<<<<<<<<<<<
 *                 # Entire value read
 *                 break
*/
    }

    /* "adbc_driver_manager/_lib.pyx":1731
 *             else:
 *                 # Buffer too small
 *                 new_len = len(buf) * 2             # <<<<<<<<<<<<<<
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
*/
    /*else*/ {
      __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1731, __pyx_L1_error)
      __pyx_v_new_len = (__pyx_t_6 * 2);

      /* "adbc_driver_manager/_lib.pyx":1732
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      __pyx_t_2 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MAX_OPTION_SIZE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1732, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1732, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__pyx_t_8)) {

        /* "adbc_driver_manager/_lib.pyx":1733
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
*/
        __pyx_t_4 = NULL;

        /* "adbc_driver_manager/_lib.pyx":1735
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")             # <<<<<<<<<<<<<<
 *                 buf = bytearray(new_len)
 * 
*/
        __pyx_t_6 = __Pyx_PyByteArray_GET_SIZE(__pyx_v_buf); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1735, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1735, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_read_option;
        __pyx_t_10[1] = __pyx_v_key;
        __pyx_t_10[2] = __pyx_mstate_global->__pyx_kp_u_would_need_more_than;
        __pyx_t_10[3] = __pyx_t_2;
        __pyx_t_10[4] = __pyx_mstate_global->__pyx_kp_u_bytes;

        /* "adbc_driver_manager/_lib.pyx":1734
 *                 if new_len > _MAX_OPTION_SIZE:
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "             # <<<<<<<<<<<<<<
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)
*/
        __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 5, 22 + __Pyx_PyUnicode_GET_LENGTH(__pyx_v_key) + 23 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 6, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_key));
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1734, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
          __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1733, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_Raise(__pyx_t_9, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 1733, __pyx_L1_error)

        /* "adbc_driver_manager/_lib.pyx":1732
 *                 # Buffer too small
 *                 new_len = len(buf) * 2
 *                 if new_len > _MAX_OPTION_SIZE:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         f"Could not read option {key}: "
*/
      }

      /* "adbc_driver_manager/_lib.pyx":1736
 *                         f"Could not read option {key}: "
 *                         f"would need more than {len(buf)} bytes")
 *                 buf = bytearray(new_len)             # <<<<<<<<<<<<<<
 * 
 *         return bytes(buf[:c_len])
*/
      __pyx_t_11 = NULL;
      __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_new_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1736, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_4};
        __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1736, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_DECREF_SET(__pyx_v_buf, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;
    }
  }
  __pyx_L4_break:;

  /* "adbc_driver_manager/_lib.pyx":1738
 *                 buf = bytearray(new_len)
 * 
 *         return bytes(buf[:c_len])             # <<<<<<<<<<<<<<
 * 
 *     def get_option_float(self, key: str) -> float:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = NULL;
  __pyx_t_11 = PySequence_GetSlice(__pyx_v_buf, 0, __pyx_v_c_len); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
    __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
  }
  __pyx_r = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1709
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1740
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_23get_option_float(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_22get_option_float, "Get the value of a floating-point option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_23get_option_float = {"get_option_float", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_23get_option_float, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_22get_option_float};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_23get_option_float(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_float (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1740, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1740, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_float", 0) < (0)) __PYX_ERR(0, 1740, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_float", 1, 1, 1, i); __PYX_ERR(0, 1740, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1740, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_float", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1740, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 1740, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_22get_option_float(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_22get_option_float(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  double __pyx_v_c_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_float", 0);

  /* "adbc_driver_manager/_lib.pyx":1742
 *     def get_option_float(self, key: str) -> float:
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1742, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1744
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1745
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef double c_value = 0.0
 *         with nogil:
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1745, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1745, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1746
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementGetOptionDouble(
*/
  __pyx_v_c_value = 0.0;

  /* "adbc_driver_manager/_lib.pyx":1747
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementGetOptionDouble(
 *                 &self.statement, c_key, &c_value, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1748
 *         cdef double c_value = 0.0
 *         with nogil:
 *             status = AdbcStatementGetOptionDouble(             # <<<<<<<<<<<<<<
 *                 &self.statement, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcStatementGetOptionDouble((&__pyx_v_self->statement), __pyx_v_c_key, (&__pyx_v_c_value), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1747
 *         cdef char* c_key = key_bytes
 *         cdef double c_value = 0.0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementGetOptionDouble(
 *                 &self.statement, c_key, &c_value, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1750
 *             status = AdbcStatementGetOptionDouble(
 *                 &self.statement, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return c_value
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1750, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1751
 *                 &self.statement, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
 *         return c_value             # <<<<<<<<<<<<<<
 * 
 *     def get_option_int(self, key: str) -> int:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1740
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1753
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_25get_option_int(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_24get_option_int, "Get the value of an integer option.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_25get_option_int = {"get_option_int", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_25get_option_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_24get_option_int};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_25get_option_int(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_option_int (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1753, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1753, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_option_int", 0) < (0)) __PYX_ERR(0, 1753, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_option_int", 1, 1, 1, i); __PYX_ERR(0, 1753, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1753, __pyx_L3_error)
    }
    __pyx_v_key = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_option_int", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1753, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 0, "key", 2))) __PYX_ERR(0, 1753, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_24get_option_int(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_24get_option_int(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_key) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_v_key_bytes = NULL;
  char *__pyx_v_c_key;
  int64_t __pyx_v_c_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_option_int", 0);

  /* "adbc_driver_manager/_lib.pyx":1755
 *     def get_option_int(self, key: str) -> int:
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1755, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1757
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")             # <<<<<<<<<<<<<<
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
*/
  __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_key_bytes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1758
 *         cdef CAdbcStatusCode status
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes             # <<<<<<<<<<<<<<
 *         cdef int64_t c_value = 0
 *         with nogil:
*/
  if (unlikely(__pyx_v_key_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1758, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1758, __pyx_L1_error)
  __pyx_v_c_key = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1759
 *         key_bytes = _to_bytes(key, "key")
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementGetOptionInt(
*/
  __pyx_v_c_value = 0;

  /* "adbc_driver_manager/_lib.pyx":1760
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementGetOptionInt(
 *                 &self.statement, c_key, &c_value, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1761
 *         cdef int64_t c_value = 0
 *         with nogil:
 *             status = AdbcStatementGetOptionInt(             # <<<<<<<<<<<<<<
 *                 &self.statement, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcStatementGetOptionInt((&__pyx_v_self->statement), __pyx_v_c_key, (&__pyx_v_c_value), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1760
 *         cdef char* c_key = key_bytes
 *         cdef int64_t c_value = 0
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementGetOptionInt(
 *                 &self.statement, c_key, &c_value, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1763
 *             status = AdbcStatementGetOptionInt(
 *                 &self.statement, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return c_value
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1763, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1764
 *                 &self.statement, c_key, &c_value, &c_error)
 *         check_error(status, &c_error)
 *         return c_value             # <<<<<<<<<<<<<<
 * 
 *     def get_parameter_schema(self) -> ArrowSchemaHandle:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_c_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyInt_FromNumber(&__pyx_t_2, NULL, 0) < (0)) __PYX_ERR(0, 1764, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1753
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_option_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1766
 *         return c_value
 * 
 *     def get_parameter_schema(self) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """Get the Arrow schema for bound parameters.
 * 
*/

/* Python wrapper */
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_27get_parameter_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_26get_parameter_schema, "Get the Arrow schema for bound parameters.\n\n        This retrieves an Arrow schema describing the number, names,\n        and types of the parameters in a parameterized statement.  The\n        fields of the schema should be in order of the ordinal\n        position of the parameters; named parameters should appear\n        only once.\n\n        If the parameter does not have a name, or the name cannot be\n        determined, the name of the corresponding field in the schema\n        will be an empty string.  If the type cannot be determined,\n        the type of the corresponding field will be NA (NullType).\n\n        This should be called after :meth:`prepare`.\n\n        Raises\n        ------\n        NotSupportedError\n            If the schema could not be determined.\n\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_27get_parameter_schema = {"get_parameter_schema", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_27get_parameter_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_26get_parameter_schema};
static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_27get_parameter_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_parameter_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_parameter_schema", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_parameter_schema", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_26get_parameter_schema(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_26get_parameter_schema(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_v_handle = 0;
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_parameter_schema", 0);

  /* "adbc_driver_manager/_lib.pyx":1788
 * 
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef CAdbcStatusCode status
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1788, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1790
 *         cdef CAdbcError c_error = empty_error()
 *         cdef CAdbcStatusCode status
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1790, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_handle = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1792
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementGetParameterSchema(
 *                 &self.statement, &handle.schema, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1793
 * 
 *         with nogil:
 *             status = AdbcStatementGetParameterSchema(             # <<<<<<<<<<<<<<
 *                 &self.statement, &handle.schema, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcStatementGetParameterSchema((&__pyx_v_self->statement), (&__pyx_v_handle->schema), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1792
 *         cdef ArrowSchemaHandle handle = ArrowSchemaHandle()
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementGetParameterSchema(
 *                 &self.statement, &handle.schema, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1795
 *             status = AdbcStatementGetParameterSchema(
 *                 &self.statement, &handle.schema, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 *         return handle
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1795, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1796
 *                 &self.statement, &handle.schema, &c_error)
 *         check_error(status, &c_error)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     def prepare(self) -> None:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1766
 *         return c_value
 * 
 *     def get_parameter_schema(self) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """Get the Arrow schema for bound parameters.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.get_parameter_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1798
 *         return handle
 * 
 *     def prepare(self) -> None:             # <<<<<<<<<<<<<<
 *         """Turn this statement into a prepared statement."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_29prepare(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_28prepare, "Turn this statement into a prepared statement.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_29prepare = {"prepare", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_29prepare, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_28prepare};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_29prepare(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prepare (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("prepare", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("prepare", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_28prepare(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_28prepare(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  struct AdbcError __pyx_v_c_error;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 0);

  /* "adbc_driver_manager/_lib.pyx":1800
 *     def prepare(self) -> None:
 *         """Turn this statement into a prepared statement."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementPrepare(&self.statement, &c_error)
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1800, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1801
 *         """Turn this statement into a prepared statement."""
 *         cdef CAdbcError c_error = empty_error()
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementPrepare(&self.statement, &c_error)
 *         check_error(status, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1802
 *         cdef CAdbcError c_error = empty_error()
 *         with nogil:
 *             status = AdbcStatementPrepare(&self.statement, &c_error)             # <<<<<<<<<<<<<<
 *         check_error(status, &c_error)
 * 
*/
        __pyx_v_status = AdbcStatementPrepare((&__pyx_v_self->statement), (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1801
 *         """Turn this statement into a prepared statement."""
 *         cdef CAdbcError c_error = empty_error()
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementPrepare(&self.statement, &c_error)
 *         check_error(status, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1803
 *         with nogil:
 *             status = AdbcStatementPrepare(&self.statement, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def set_options(self, **kwargs) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1803, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1798
 *         return handle
 * 
 *     def prepare(self) -> None:             # <<<<<<<<<<<<<<
 *         """Turn this statement into a prepared statement."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1805
 *         check_error(status, &c_error)
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """Set arbitrary key-value options for this statement only.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_31set_options(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_30set_options, "Set arbitrary key-value options for this statement only.\n\n        Pass options as kwargs: ``set_options(**{\"some.option\": \"value\"})``.\n\n        Note, not all drivers support setting options after creation.\n\n        See Also\n        --------\n        adbc_driver_manager.StatementOptions : Standard option names.\n        ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_31set_options = {"set_options", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_31set_options, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_30set_options};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_31set_options(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_options (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("set_options", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("set_options", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_30set_options(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_30set_options(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  struct AdbcError __pyx_v_c_error;
  char *__pyx_v_c_key;
  char *__pyx_v_c_value;
  size_t __pyx_v_c_value_len;
  double __pyx_v_c_double;
  int64_t __pyx_v_c_int;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  char *__pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  double __pyx_t_12;
  int64_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17[6];
  PyObject *__pyx_t_18 = NULL;
  size_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_options", 0);

  /* "adbc_driver_manager/_lib.pyx":1816
 *         adbc_driver_manager.StatementOptions : Standard option names.
 *         """
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef char* c_key = NULL
 *         cdef char* c_value = NULL
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1816, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1817
 *         """
 *         cdef CAdbcError c_error = empty_error()
 *         cdef char* c_key = NULL             # <<<<<<<<<<<<<<
 *         cdef char* c_value = NULL
 *         cdef size_t c_value_len
*/
  __pyx_v_c_key = NULL;

  /* "adbc_driver_manager/_lib.pyx":1818
 *         cdef CAdbcError c_error = empty_error()
 *         cdef char* c_key = NULL
 *         cdef char* c_value = NULL             # <<<<<<<<<<<<<<
 *         cdef size_t c_value_len
 *         cdef double c_double
*/
  __pyx_v_c_value = NULL;

  /* "adbc_driver_manager/_lib.pyx":1822
 *         cdef double c_double
 *         cdef int64_t c_int
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             key = _to_bytes(key, "option key")
 *             c_key = key
*/
  __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_2);
  __pyx_t_2 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
    if (unlikely(__pyx_t_8 == 0)) break;
    if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 1822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "adbc_driver_manager/_lib.pyx":1823
 *         cdef int64_t c_int
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "option key")             # <<<<<<<<<<<<<<
 *             c_key = key
 * 
*/
    __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_key, __pyx_mstate_global->__pyx_kp_u_option_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "adbc_driver_manager/_lib.pyx":1824
 *         for key, value in kwargs.items():
 *             key = _to_bytes(key, "option key")
 *             c_key = key             # <<<<<<<<<<<<<<
 * 
 *             if value is None:
*/
    __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_key); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1824, __pyx_L1_error)
    __pyx_v_c_key = __pyx_t_9;

    /* "adbc_driver_manager/_lib.pyx":1826
 *             c_key = key
 * 
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 c_value = NULL
 *                 with nogil:
*/
    __pyx_t_10 = (__pyx_v_value == Py_None);
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1827
 * 
 *             if value is None:
 *                 c_value = NULL             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcStatementSetOption(
*/
      __pyx_v_c_value = NULL;

      /* "adbc_driver_manager/_lib.pyx":1828
 *             if value is None:
 *                 c_value = NULL
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1829
 *                 c_value = NULL
 *                 with nogil:
 *                     status = AdbcStatementSetOption(             # <<<<<<<<<<<<<<
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, str):
*/
            __pyx_v_status = AdbcStatementSetOption((&__pyx_v_self->statement), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1828
 *             if value is None:
 *                 c_value = NULL
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L10;
            }
            __pyx_L10:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1826
 *             c_key = key
 * 
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 c_value = NULL
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1831
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
    __pyx_t_10 = PyUnicode_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1832
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, str):
 *                 value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 with nogil:
*/
      __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1832, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "adbc_driver_manager/_lib.pyx":1833
 *             elif isinstance(value, str):
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcStatementSetOption(
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1833, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":1834
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1835
 *                 c_value = value
 *                 with nogil:
 *                     status = AdbcStatementSetOption(             # <<<<<<<<<<<<<<
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
*/
            __pyx_v_status = AdbcStatementSetOption((&__pyx_v_self->statement), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1834
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L15;
            }
            __pyx_L15:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1831
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1837
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
    __pyx_t_10 = PyBool_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1838
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1838, __pyx_L1_error)
      if (__pyx_t_10) {

        /* "adbc_driver_manager/_lib.pyx":1839
 *             elif isinstance(value, bool):
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED             # <<<<<<<<<<<<<<
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
*/
        __pyx_t_7 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_ENABLED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1839, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "adbc_driver_manager/_lib.pyx":1838
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):
 *                 if value:             # <<<<<<<<<<<<<<
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
*/
        goto __pyx_L16;
      }

      /* "adbc_driver_manager/_lib.pyx":1841
 *                     value = ADBC_OPTION_VALUE_ENABLED
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED             # <<<<<<<<<<<<<<
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
*/
      /*else*/ {
        __pyx_t_7 = __Pyx_PyBytes_FromString(ADBC_OPTION_VALUE_DISABLED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1841, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
        __pyx_t_7 = 0;
      }
      __pyx_L16:;

      /* "adbc_driver_manager/_lib.pyx":1842
 *                 else:
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 with nogil:
*/
      __pyx_t_7 = __pyx_f_19adbc_driver_manager_4_lib__to_bytes(__pyx_v_value, __pyx_mstate_global->__pyx_kp_u_option_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1842, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "adbc_driver_manager/_lib.pyx":1843
 *                     value = ADBC_OPTION_VALUE_DISABLED
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcStatementSetOption(
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1843, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":1844
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1845
 *                 c_value = value
 *                 with nogil:
 *                     status = AdbcStatementSetOption(             # <<<<<<<<<<<<<<
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):
*/
            __pyx_v_status = AdbcStatementSetOption((&__pyx_v_self->statement), __pyx_v_c_key, __pyx_v_c_value, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1844
 *                 value = _to_bytes(value, "option value")
 *                 c_value = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L21;
            }
            __pyx_L21:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1837
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bool):             # <<<<<<<<<<<<<<
 *                 if value:
 *                     value = ADBC_OPTION_VALUE_ENABLED
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1847
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
    __pyx_t_10 = PyBytes_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1848
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):
 *                 c_value = value             # <<<<<<<<<<<<<<
 *                 c_value_len = len(value)
 *                 with nogil:
*/
      __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1848, __pyx_L1_error)
      __pyx_v_c_value = __pyx_t_9;

      /* "adbc_driver_manager/_lib.pyx":1849
 *             elif isinstance(value, bytes):
 *                 c_value = value
 *                 c_value_len = len(value)             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcStatementSetOptionBytes(
*/
      __pyx_t_11 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1849, __pyx_L1_error)
      __pyx_v_c_value_len = __pyx_t_11;

      /* "adbc_driver_manager/_lib.pyx":1850
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOptionBytes(
 *                         &self.statement, c_key, <const uint8_t*> c_value,
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1851
 *                 c_value_len = len(value)
 *                 with nogil:
 *                     status = AdbcStatementSetOptionBytes(             # <<<<<<<<<<<<<<
 *                         &self.statement, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
*/
            __pyx_v_status = AdbcStatementSetOptionBytes((&__pyx_v_self->statement), __pyx_v_c_key, ((uint8_t const *)__pyx_v_c_value), __pyx_v_c_value_len, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1850
 *                 c_value = value
 *                 c_value_len = len(value)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOptionBytes(
 *                         &self.statement, c_key, <const uint8_t*> c_value,
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L26;
            }
            __pyx_L26:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1847
 *                     status = AdbcStatementSetOption(
 *                         &self.statement, c_key, c_value, &c_error)
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 c_value = value
 *                 c_value_len = len(value)
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1854
 *                         &self.statement, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_double = value
 *                 with nogil:
*/
    __pyx_t_10 = PyFloat_Check(__pyx_v_value); 
    if (__pyx_t_10) {

      /* "adbc_driver_manager/_lib.pyx":1855
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):
 *                 c_double = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcStatementSetOptionDouble(
*/
      __pyx_t_12 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1855, __pyx_L1_error)
      __pyx_v_c_double = __pyx_t_12;

      /* "adbc_driver_manager/_lib.pyx":1856
 *             elif isinstance(value, float):
 *                 c_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOptionDouble(
 *                         &self.statement, c_key, c_double, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1857
 *                 c_double = value
 *                 with nogil:
 *                     status = AdbcStatementSetOptionDouble(             # <<<<<<<<<<<<<<
 *                         &self.statement, c_key, c_double, &c_error)
 *             elif isinstance(value, int):
*/
            __pyx_v_status = AdbcStatementSetOptionDouble((&__pyx_v_self->statement), __pyx_v_c_key, __pyx_v_c_double, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1856
 *             elif isinstance(value, float):
 *                 c_double = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOptionDouble(
 *                         &self.statement, c_key, c_double, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L31;
            }
            __pyx_L31:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1854
 *                         &self.statement, c_key, <const uint8_t*> c_value,
 *                         c_value_len, &c_error)
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 c_double = value
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1859
 *                     status = AdbcStatementSetOptionDouble(
 *                         &self.statement, c_key, c_double, &c_error)
 *             elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_int = value
 *                 with nogil:
*/
    __pyx_t_10 = PyLong_Check(__pyx_v_value); 
    if (likely(__pyx_t_10)) {

      /* "adbc_driver_manager/_lib.pyx":1860
 *                         &self.statement, c_key, c_double, &c_error)
 *             elif isinstance(value, int):
 *                 c_int = value             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     status = AdbcStatementSetOptionInt(
*/
      __pyx_t_13 = __Pyx_PyLong_As_int64_t(__pyx_v_value); if (unlikely((__pyx_t_13 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1860, __pyx_L1_error)
      __pyx_v_c_int = __pyx_t_13;

      /* "adbc_driver_manager/_lib.pyx":1861
 *             elif isinstance(value, int):
 *                 c_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOptionInt(
 *                         &self.statement, c_key, c_int, &c_error)
*/
      {
          PyThreadState * _save;
          _save = PyEval_SaveThread();
          __Pyx_FastGIL_Remember();
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1862
 *                 c_int = value
 *                 with nogil:
 *                     status = AdbcStatementSetOptionInt(             # <<<<<<<<<<<<<<
 *                         &self.statement, c_key, c_int, &c_error)
 *             else:
*/
            __pyx_v_status = AdbcStatementSetOptionInt((&__pyx_v_self->statement), __pyx_v_c_key, __pyx_v_c_int, (&__pyx_v_c_error));
          }

          /* "adbc_driver_manager/_lib.pyx":1861
 *             elif isinstance(value, int):
 *                 c_int = value
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     status = AdbcStatementSetOptionInt(
 *                         &self.statement, c_key, c_int, &c_error)
*/
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_FastGIL_Forget();
              PyEval_RestoreThread(_save);
              goto __pyx_L36;
            }
            __pyx_L36:;
          }
      }

      /* "adbc_driver_manager/_lib.pyx":1859
 *                     status = AdbcStatementSetOptionDouble(
 *                         &self.statement, c_key, c_double, &c_error)
 *             elif isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 c_int = value
 *                 with nogil:
*/
      goto __pyx_L5;
    }

    /* "adbc_driver_manager/_lib.pyx":1865
 *                         &self.statement, c_key, c_int, &c_error)
 *             else:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     f"Unsupported type {type(value)} for value {value!r} "
 *                     f"of option {key}")
*/
    /*else*/ {
      __pyx_t_6 = NULL;

      /* "adbc_driver_manager/_lib.pyx":1866
 *             else:
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "             # <<<<<<<<<<<<<<
 *                     f"of option {key}")
 * 
*/
      __pyx_t_14 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_value)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_value), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);

      /* "adbc_driver_manager/_lib.pyx":1867
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "
 *                     f"of option {key}")             # <<<<<<<<<<<<<<
 * 
 *             check_error(status, &c_error)
*/
      __pyx_t_16 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_17[0] = __pyx_mstate_global->__pyx_kp_u_Unsupported_type;
      __pyx_t_17[1] = __pyx_t_14;
      __pyx_t_17[2] = __pyx_mstate_global->__pyx_kp_u_for_value;
      __pyx_t_17[3] = __pyx_t_15;
      __pyx_t_17[4] = __pyx_mstate_global->__pyx_kp_u_of_option;
      __pyx_t_17[5] = __pyx_t_16;

      /* "adbc_driver_manager/_lib.pyx":1866
 *             else:
 *                 raise ValueError(
 *                     f"Unsupported type {type(value)} for value {value!r} "             # <<<<<<<<<<<<<<
 *                     f"of option {key}")
 * 
*/
      __pyx_t_18 = __Pyx_PyUnicode_Join(__pyx_t_17, 6, 17 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14) + 11 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16));
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_19 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_18};
        __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_19, (2-__pyx_t_19) | (__pyx_t_19*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 1865, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "adbc_driver_manager/_lib.pyx":1869
 *                     f"of option {key}")
 * 
 *             check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def set_sql_query(self, str query not None) -> None:
*/
    __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1869, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1805
 *         check_error(status, &c_error)
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """Set arbitrary key-value options for this statement only.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.set_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1871
 *             check_error(status, &c_error)
 * 
 *     def set_sql_query(self, str query not None) -> None:             # <<<<<<<<<<<<<<
 *         """Set a SQL query to be executed."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_33set_sql_query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_32set_sql_query, "Set a SQL query to be executed.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_33set_sql_query = {"set_sql_query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_33set_sql_query, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_32set_sql_query};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_33set_sql_query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_query = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_sql_query (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_query,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1871, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1871, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_sql_query", 0) < (0)) __PYX_ERR(0, 1871, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_sql_query", 1, 1, 1, i); __PYX_ERR(0, 1871, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1871, __pyx_L3_error)
    }
    __pyx_v_query = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_sql_query", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1871, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.set_sql_query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_query), (&PyUnicode_Type), 0, "query", 1))) __PYX_ERR(0, 1871, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_32set_sql_query(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_query);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_32set_sql_query(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_query) {
  struct AdbcError __pyx_v_c_error;
  PyObject *__pyx_v_query_data = 0;
  char const *__pyx_v_c_query;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char const *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_sql_query", 0);

  /* "adbc_driver_manager/_lib.pyx":1873
 *     def set_sql_query(self, str query not None) -> None:
 *         """Set a SQL query to be executed."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef bytes query_data = query.encode("utf-8")
 *         cdef const char* c_query = query_data
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1873, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1874
 *         """Set a SQL query to be executed."""
 *         cdef CAdbcError c_error = empty_error()
 *         cdef bytes query_data = query.encode("utf-8")             # <<<<<<<<<<<<<<
 *         cdef const char* c_query = query_data
 *         with nogil:
*/
  __pyx_t_2 = PyUnicode_AsUTF8String(__pyx_v_query); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_query_data = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1875
 *         cdef CAdbcError c_error = empty_error()
 *         cdef bytes query_data = query.encode("utf-8")
 *         cdef const char* c_query = query_data             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementSetSqlQuery(
*/
  __pyx_t_3 = __Pyx_PyBytes_AsString(__pyx_v_query_data); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1875, __pyx_L1_error)
  __pyx_v_c_query = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1876
 *         cdef bytes query_data = query.encode("utf-8")
 *         cdef const char* c_query = query_data
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementSetSqlQuery(
 *                 &self.statement, c_query, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1877
 *         cdef const char* c_query = query_data
 *         with nogil:
 *             status = AdbcStatementSetSqlQuery(             # <<<<<<<<<<<<<<
 *                 &self.statement, c_query, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcStatementSetSqlQuery((&__pyx_v_self->statement), __pyx_v_c_query, (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1876
 *         cdef bytes query_data = query.encode("utf-8")
 *         cdef const char* c_query = query_data
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementSetSqlQuery(
 *                 &self.statement, c_query, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1879
 *             status = AdbcStatementSetSqlQuery(
 *                 &self.statement, c_query, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 *     def set_substrait_plan(self, bytes plan not None) -> None:
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1879, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1871
 *             check_error(status, &c_error)
 * 
 *     def set_sql_query(self, str query not None) -> None:             # <<<<<<<<<<<<<<
 *         """Set a SQL query to be executed."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.set_sql_query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_query_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1881
 *         check_error(status, &c_error)
 * 
 *     def set_substrait_plan(self, bytes plan not None) -> None:             # <<<<<<<<<<<<<<
 *         """Set a Substrait plan to be executed."""
 *         cdef CAdbcError c_error = empty_error()
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_35set_substrait_plan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_34set_substrait_plan, "Set a Substrait plan to be executed.");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_35set_substrait_plan = {"set_substrait_plan", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_35set_substrait_plan, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_34set_substrait_plan};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_35set_substrait_plan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_plan = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_substrait_plan (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_plan,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1881, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1881, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_substrait_plan", 0) < (0)) __PYX_ERR(0, 1881, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_substrait_plan", 1, 1, 1, i); __PYX_ERR(0, 1881, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1881, __pyx_L3_error)
    }
    __pyx_v_plan = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_substrait_plan", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1881, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.set_substrait_plan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_plan), (&PyBytes_Type), 0, "plan", 1))) __PYX_ERR(0, 1881, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_34set_substrait_plan(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v_plan);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_34set_substrait_plan(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, PyObject *__pyx_v_plan) {
  struct AdbcError __pyx_v_c_error;
  uint8_t const *__pyx_v_c_plan;
  size_t __pyx_v_length;
  AdbcStatusCode __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AdbcError __pyx_t_1;
  uint8_t const *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_substrait_plan", 0);

  /* "adbc_driver_manager/_lib.pyx":1883
 *     def set_substrait_plan(self, bytes plan not None) -> None:
 *         """Set a Substrait plan to be executed."""
 *         cdef CAdbcError c_error = empty_error()             # <<<<<<<<<<<<<<
 *         cdef const uint8_t* c_plan = <const uint8_t*> plan
 *         cdef size_t length = len(plan)
*/
  __pyx_t_1 = __pyx_f_19adbc_driver_manager_4_lib_empty_error(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1883, __pyx_L1_error)
  __pyx_v_c_error = __pyx_t_1;

  /* "adbc_driver_manager/_lib.pyx":1884
 *         """Set a Substrait plan to be executed."""
 *         cdef CAdbcError c_error = empty_error()
 *         cdef const uint8_t* c_plan = <const uint8_t*> plan             # <<<<<<<<<<<<<<
 *         cdef size_t length = len(plan)
 *         with nogil:
*/
  __pyx_t_2 = __Pyx_PyBytes_AsUString(__pyx_v_plan); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L1_error)
  __pyx_v_c_plan = ((uint8_t const *)__pyx_t_2);

  /* "adbc_driver_manager/_lib.pyx":1885
 *         cdef CAdbcError c_error = empty_error()
 *         cdef const uint8_t* c_plan = <const uint8_t*> plan
 *         cdef size_t length = len(plan)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             status = AdbcStatementSetSubstraitPlan(
*/
  __pyx_t_3 = __Pyx_PyBytes_GET_SIZE(__pyx_v_plan); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1885, __pyx_L1_error)
  __pyx_v_length = __pyx_t_3;

  /* "adbc_driver_manager/_lib.pyx":1886
 *         cdef const uint8_t* c_plan = <const uint8_t*> plan
 *         cdef size_t length = len(plan)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementSetSubstraitPlan(
 *                 &self.statement, c_plan, length, &c_error)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1887
 *         cdef size_t length = len(plan)
 *         with nogil:
 *             status = AdbcStatementSetSubstraitPlan(             # <<<<<<<<<<<<<<
 *                 &self.statement, c_plan, length, &c_error)
 *         check_error(status, &c_error)
*/
        __pyx_v_status = AdbcStatementSetSubstraitPlan((&__pyx_v_self->statement), __pyx_v_c_plan, __pyx_v_length, (&__pyx_v_c_error));
      }

      /* "adbc_driver_manager/_lib.pyx":1886
 *         cdef const uint8_t* c_plan = <const uint8_t*> plan
 *         cdef size_t length = len(plan)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             status = AdbcStatementSetSubstraitPlan(
 *                 &self.statement, c_plan, length, &c_error)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1889
 *             status = AdbcStatementSetSubstraitPlan(
 *                 &self.statement, c_plan, length, &c_error)
 *         check_error(status, &c_error)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_19adbc_driver_manager_4_lib_check_error(__pyx_v_status, (&__pyx_v_c_error)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1889, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1881
 *         check_error(status, &c_error)
 * 
 *     def set_substrait_plan(self, bytes plan not None) -> None:             # <<<<<<<<<<<<<<
 *         """Set a Substrait plan to be executed."""
 *         cdef CAdbcError c_error = empty_error()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.set_substrait_plan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_37__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_37__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_37__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_37__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_36__reduce_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_statement_cannot_be_convert, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_39__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_39__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_39__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_39__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_38__setstate_cython__(((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_13AdbcStatement_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_statement_cannot_be_convert, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("adbc_driver_manager._lib.AdbcStatement.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1892
 * 
 * 
 * cdef const CAdbcError* PyAdbcErrorFromArrayStream(             # <<<<<<<<<<<<<<
 *         CArrowArrayStream* stream, CAdbcStatusCode* status):
 *     return AdbcErrorFromArrayStream(stream, status)
*/

static struct AdbcError const *__pyx_f_19adbc_driver_manager_4_lib_PyAdbcErrorFromArrayStream(struct ArrowArrayStream *__pyx_v_stream, AdbcStatusCode *__pyx_v_status) {
  struct AdbcError const *__pyx_r;

  /* "adbc_driver_manager/_lib.pyx":1894
 * cdef const CAdbcError* PyAdbcErrorFromArrayStream(
 *         CArrowArrayStream* stream, CAdbcStatusCode* status):
 *     return AdbcErrorFromArrayStream(stream, status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = AdbcErrorFromArrayStream(__pyx_v_stream, __pyx_v_status);
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1892
 * 
 * 
 * cdef const CAdbcError* PyAdbcErrorFromArrayStream(             # <<<<<<<<<<<<<<
 *         CArrowArrayStream* stream, CAdbcStatusCode* status):
 *     return AdbcErrorFromArrayStream(stream, status)
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1905
 * 
 * 
 * @functools.lru_cache             # <<<<<<<<<<<<<<
 * def _init_blocking_call():
 *     error = bytes(CInitBlockingCallback()).decode("utf-8")
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_5_init_blocking_call(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_5_init_blocking_call = {"_init_blocking_call", (PyCFunction)__pyx_pw_19adbc_driver_manager_4_lib_5_init_blocking_call, METH_NOARGS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_5_init_blocking_call(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_init_blocking_call (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_4_init_blocking_call(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_4_init_blocking_call(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init_blocking_call", 0);

  /* "adbc_driver_manager/_lib.pyx":1907
 * @functools.lru_cache
 * def _init_blocking_call():
 *     error = bytes(CInitBlockingCallback()).decode("utf-8")             # <<<<<<<<<<<<<<
 *     if error:
 *         warnings.warn(
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(pyadbc_driver_manager::InitBlockingCallback()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_3 = __Pyx_decode_bytes(__pyx_t_1, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_error = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "adbc_driver_manager/_lib.pyx":1908
 * def _init_blocking_call():
 *     error = bytes(CInitBlockingCallback()).decode("utf-8")
 *     if error:             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             f"Failed to initialize KeyboardInterrupt support: {error}",
*/
  {
    Py_ssize_t __pyx_temp = __Pyx_PyUnicode_IS_TRUE(__pyx_v_error);
    if (unlikely(((!CYTHON_ASSUME_SAFE_SIZE) && __pyx_temp < 0))) __PYX_ERR(0, 1908, __pyx_L1_error)
    __pyx_t_5 = (__pyx_temp != 0);
  }

  if (__pyx_t_5) {

    /* "adbc_driver_manager/_lib.pyx":1909
 *     error = bytes(CInitBlockingCallback()).decode("utf-8")
 *     if error:
 *         warnings.warn(             # <<<<<<<<<<<<<<
 *             f"Failed to initialize KeyboardInterrupt support: {error}",
 *             RuntimeWarning,
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1909, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_warn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1909, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "adbc_driver_manager/_lib.pyx":1910
 *     if error:
 *         warnings.warn(
 *             f"Failed to initialize KeyboardInterrupt support: {error}",             # <<<<<<<<<<<<<<
 *             RuntimeWarning,
 *         )
*/
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Failed_to_initialize_KeyboardInt, __pyx_v_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "adbc_driver_manager/_lib.pyx":1911
 *         warnings.warn(
 *             f"Failed to initialize KeyboardInterrupt support: {error}",
 *             RuntimeWarning,             # <<<<<<<<<<<<<<
 *         )
 * 
*/
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_t_2, ((PyObject *)(((PyTypeObject*)PyExc_RuntimeWarning)))};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "adbc_driver_manager/_lib.pyx":1908
 * def _init_blocking_call():
 *     error = bytes(CInitBlockingCallback()).decode("utf-8")
 *     if error:             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             f"Failed to initialize KeyboardInterrupt support: {error}",
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1905
 * 
 * 
 * @functools.lru_cache             # <<<<<<<<<<<<<<
 * def _init_blocking_call():
 *     error = bytes(CInitBlockingCallback()).decode("utf-8")
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("adbc_driver_manager._lib._init_blocking_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1919
 * 
 * 
 * def _blocking_call_impl(func, args, kwargs, cancel):             # <<<<<<<<<<<<<<
 *     """
 *     Run functions that are expected to block with a native SIGINT handler.
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_7_blocking_call_impl(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_19adbc_driver_manager_4_lib_6_blocking_call_impl, "\n    Run functions that are expected to block with a native SIGINT handler.\n\n    Parameters\n    ----------\n    ");
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_7_blocking_call_impl = {"_blocking_call_impl", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_7_blocking_call_impl, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_6_blocking_call_impl};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_7_blocking_call_impl(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_v_cancel = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_blocking_call_impl (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_func,&__pyx_mstate_global->__pyx_n_u_args,&__pyx_mstate_global->__pyx_n_u_kwargs,&__pyx_mstate_global->__pyx_n_u_cancel,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1919, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1919, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1919, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1919, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1919, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_blocking_call_impl", 0) < (0)) __PYX_ERR(0, 1919, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_blocking_call_impl", 1, 4, 4, i); __PYX_ERR(0, 1919, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1919, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1919, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1919, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1919, __pyx_L3_error)
    }
    __pyx_v_func = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwargs = values[2];
    __pyx_v_cancel = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_blocking_call_impl", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 1919, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_6_blocking_call_impl(__pyx_self, __pyx_v_func, __pyx_v_args, __pyx_v_kwargs, __pyx_v_cancel);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_6_blocking_call_impl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, PyObject *__pyx_v_cancel) {
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  char const *__pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  char const *__pyx_t_23;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_blocking_call_impl", 0);

  /* "adbc_driver_manager/_lib.pyx":1928
 *     global _blocking_exc
 * 
 *     if threading.current_thread() is not threading.main_thread():             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_current_thread); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_main_thread); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_7 = (__pyx_t_1 != __pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_7) {

    /* "adbc_driver_manager/_lib.pyx":1929
 * 
 *     if threading.current_thread() is not threading.main_thread():
 *         return func(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     _init_blocking_call()
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
      __PYX_ERR(0, 1929, __pyx_L1_error)
    }
    if (likely(PyDict_CheckExact(__pyx_v_kwargs))) {
      __pyx_t_1 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg((PyObject*)&PyDict_Type, __pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_v_func, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "adbc_driver_manager/_lib.pyx":1928
 *     global _blocking_exc
 * 
 *     if threading.current_thread() is not threading.main_thread():             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 * 
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1931
 *         return func(*args, **kwargs)
 * 
 *     _init_blocking_call()             # <<<<<<<<<<<<<<
 * 
 *     with _blocking_lock:
*/
  __pyx_t_1 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_init_blocking_call); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_1);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "adbc_driver_manager/_lib.pyx":1933
 *     _init_blocking_call()
 * 
 *     with _blocking_lock:             # <<<<<<<<<<<<<<
 *         if _blocking_exc:
 *             _blocking_exc = None
*/
  /*with:*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_blocking_lock); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = NULL;
    __pyx_t_2 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1933, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1933, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1934
 * 
 *     with _blocking_lock:
 *         if _blocking_exc:             # <<<<<<<<<<<<<<
 *             _blocking_exc = None
 * 
*/
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_blocking_exc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1934, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1934, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_7) {

            /* "adbc_driver_manager/_lib.pyx":1935
 *     with _blocking_lock:
 *         if _blocking_exc:
 *             _blocking_exc = None             # <<<<<<<<<<<<<<
 * 
 *     # Set the callback for the background thread and save the signal handler
*/
            if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_exc, Py_None) < (0)) __PYX_ERR(0, 1935, __pyx_L8_error)

            /* "adbc_driver_manager/_lib.pyx":1934
 * 
 *     with _blocking_lock:
 *         if _blocking_exc:             # <<<<<<<<<<<<<<
 *             _blocking_exc = None
 * 
*/
          }

          /* "adbc_driver_manager/_lib.pyx":1933
 *     _init_blocking_call()
 * 
 *     with _blocking_lock:             # <<<<<<<<<<<<<<
 *         if _blocking_exc:
 *             _blocking_exc = None
*/
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L13_try_end;
        __pyx_L8_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(0, 1933, __pyx_L10_except_error)
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_2);
          __pyx_t_1 = PyTuple_Pack(3, __pyx_t_6, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1933, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_1, NULL);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1933, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_7 < (0)) __PYX_ERR(0, 1933, __pyx_L10_except_error)
          __pyx_t_13 = (!__pyx_t_7);
          if (unlikely(__pyx_t_13)) {
            __Pyx_GIVEREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_XGIVEREF(__pyx_t_2);
            __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_4, __pyx_t_2);
            __pyx_t_6 = 0;  __pyx_t_4 = 0;  __pyx_t_2 = 0; 
            __PYX_ERR(0, 1933, __pyx_L10_except_error)
          }
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L9_exception_handled;
        }
        __pyx_L10_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L9_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        __pyx_L13_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_8) {
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_mstate_global->__pyx_tuple[0], NULL);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1933, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        goto __pyx_L7;
      }
      __pyx_L7:;
    }
    goto __pyx_L18;
    __pyx_L4_error:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L1_error;
    __pyx_L18:;
  }

  /* "adbc_driver_manager/_lib.pyx":1939
 *     # Set the callback for the background thread and save the signal handler
 *     # TODO: ideally this would be no-op if already set
 *     error = bytes(             # <<<<<<<<<<<<<<
 *         CSetBlockingCallback(&_handle_blocking_call, <void*>cancel)
 *     ).decode("utf-8")
*/
  __pyx_t_4 = NULL;

  /* "adbc_driver_manager/_lib.pyx":1940
 *     # TODO: ideally this would be no-op if already set
 *     error = bytes(
 *         CSetBlockingCallback(&_handle_blocking_call, <void*>cancel)             # <<<<<<<<<<<<<<
 *     ).decode("utf-8")
 *     if error:
*/
  __pyx_t_6 = __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(pyadbc_driver_manager::SetBlockingCallback((&__pyx_f_19adbc_driver_manager_4_lib__handle_blocking_call), ((void *)__pyx_v_cancel))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }

  /* "adbc_driver_manager/_lib.pyx":1941
 *     error = bytes(
 *         CSetBlockingCallback(&_handle_blocking_call, <void*>cancel)
 *     ).decode("utf-8")             # <<<<<<<<<<<<<<
 *     if error:
 *         warnings.warn(
*/
  __pyx_t_6 = __Pyx_decode_bytes(__pyx_t_2, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_error = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "adbc_driver_manager/_lib.pyx":1942
 *         CSetBlockingCallback(&_handle_blocking_call, <void*>cancel)
 *     ).decode("utf-8")
 *     if error:             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             f"Failed to set SIGINT handler: {error}",
*/
  {
    Py_ssize_t __pyx_temp = __Pyx_PyUnicode_IS_TRUE(__pyx_v_error);
    if (unlikely(((!CYTHON_ASSUME_SAFE_SIZE) && __pyx_temp < 0))) __PYX_ERR(0, 1942, __pyx_L1_error)
    __pyx_t_13 = (__pyx_temp != 0);
  }

  if (__pyx_t_13) {

    /* "adbc_driver_manager/_lib.pyx":1943
 *     ).decode("utf-8")
 *     if error:
 *         warnings.warn(             # <<<<<<<<<<<<<<
 *             f"Failed to set SIGINT handler: {error}",
 *             RuntimeWarning,
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_warnings); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":1944
 *     if error:
 *         warnings.warn(
 *             f"Failed to set SIGINT handler: {error}",             # <<<<<<<<<<<<<<
 *             RuntimeWarning,
 *         )
*/
    __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Failed_to_set_SIGINT_handler, __pyx_v_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "adbc_driver_manager/_lib.pyx":1945
 *         warnings.warn(
 *             f"Failed to set SIGINT handler: {error}",
 *             RuntimeWarning,             # <<<<<<<<<<<<<<
 *         )
 * 
*/
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_4, ((PyObject *)(((PyTypeObject*)PyExc_RuntimeWarning)))};
      __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "adbc_driver_manager/_lib.pyx":1942
 *         CSetBlockingCallback(&_handle_blocking_call, <void*>cancel)
 *     ).decode("utf-8")
 *     if error:             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             f"Failed to set SIGINT handler: {error}",
*/
  }

  /* "adbc_driver_manager/_lib.pyx":1948
 *         )
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 *     except BaseException as e:
*/
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_11, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1949
 * 
 *     try:
 *         return func(*args, **kwargs)             # <<<<<<<<<<<<<<
 *     except BaseException as e:
 *         with _blocking_lock:
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1949, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__pyx_v_kwargs == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
          __PYX_ERR(0, 1949, __pyx_L23_error)
        }
        if (likely(PyDict_CheckExact(__pyx_v_kwargs))) {
          __pyx_t_1 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1949, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg((PyObject*)&PyDict_Type, __pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1949, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_func, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1949, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L27_try_return;

        /* "adbc_driver_manager/_lib.pyx":1948
 *         )
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 *     except BaseException as e:
*/
      }
      __pyx_L23_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "adbc_driver_manager/_lib.pyx":1950
 *     try:
 *         return func(*args, **kwargs)
 *     except BaseException as e:             # <<<<<<<<<<<<<<
 *         with _blocking_lock:
 *             if _blocking_exc:
*/
      __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_BaseException))));
      if (__pyx_t_14) {
        __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 1950, __pyx_L25_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_v_e = __pyx_t_1;
        /*try:*/ {

          /* "adbc_driver_manager/_lib.pyx":1951
 *         return func(*args, **kwargs)
 *     except BaseException as e:
 *         with _blocking_lock:             # <<<<<<<<<<<<<<
 *             if _blocking_exc:
 *                 exc = _blocking_exc
*/
          /*with:*/ {
            __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_blocking_lock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1951, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1951, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_15 = NULL;
            __pyx_t_16 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1951, __pyx_L36_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_5 = 1;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_16))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
              assert(__pyx_t_15);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_16, __pyx__function);
              __pyx_t_5 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
              __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_16, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1951, __pyx_L36_error)
              __Pyx_GOTREF(__pyx_t_3);
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            /*try:*/ {
              {
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_17, &__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_12);
                __Pyx_XGOTREF(__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_18);
                /*try:*/ {

                  /* "adbc_driver_manager/_lib.pyx":1952
 *     except BaseException as e:
 *         with _blocking_lock:
 *             if _blocking_exc:             # <<<<<<<<<<<<<<
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
*/
                  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_blocking_exc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1952, __pyx_L42_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1952, __pyx_L42_error)
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  if (unlikely(__pyx_t_13)) {

                    /* "adbc_driver_manager/_lib.pyx":1953
 *         with _blocking_lock:
 *             if _blocking_exc:
 *                 exc = _blocking_exc             # <<<<<<<<<<<<<<
 *                 _blocking_exc = None
 *                 raise e from exc[1].with_traceback(exc[2])
*/
                    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_blocking_exc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1953, __pyx_L42_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __pyx_v_exc = __pyx_t_2;
                    __pyx_t_2 = 0;

                    /* "adbc_driver_manager/_lib.pyx":1954
 *             if _blocking_exc:
 *                 exc = _blocking_exc
 *                 _blocking_exc = None             # <<<<<<<<<<<<<<
 *                 raise e from exc[1].with_traceback(exc[2])
 *         raise e
*/
                    if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_exc, Py_None) < (0)) __PYX_ERR(0, 1954, __pyx_L42_error)

                    /* "adbc_driver_manager/_lib.pyx":1955
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
 *                 raise e from exc[1].with_traceback(exc[2])             # <<<<<<<<<<<<<<
 *         raise e
 *     finally:
*/
                    __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_exc, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1955, __pyx_L42_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    __pyx_t_3 = __pyx_t_16;
                    __Pyx_INCREF(__pyx_t_3);
                    __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_exc, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1955, __pyx_L42_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    __pyx_t_5 = 0;
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_15};
                      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_with_traceback, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1955, __pyx_L42_error)
                      __Pyx_GOTREF(__pyx_t_2);
                    }
                    __Pyx_Raise(__pyx_v_e, 0, 0, __pyx_t_2);
                    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                    __PYX_ERR(0, 1955, __pyx_L42_error)

                    /* "adbc_driver_manager/_lib.pyx":1952
 *     except BaseException as e:
 *         with _blocking_lock:
 *             if _blocking_exc:             # <<<<<<<<<<<<<<
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
*/
                  }

                  /* "adbc_driver_manager/_lib.pyx":1951
 *         return func(*args, **kwargs)
 *     except BaseException as e:
 *         with _blocking_lock:             # <<<<<<<<<<<<<<
 *             if _blocking_exc:
 *                 exc = _blocking_exc
*/
                }
                __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                goto __pyx_L49_try_end;
                __pyx_L42_error:;
                __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                /*except:*/ {
                  __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
                  if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_16, &__pyx_t_15) < 0) __PYX_ERR(0, 1951, __pyx_L44_except_error)
                  __Pyx_XGOTREF(__pyx_t_2);
                  __Pyx_XGOTREF(__pyx_t_16);
                  __Pyx_XGOTREF(__pyx_t_15);
                  __pyx_t_3 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_16, __pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1951, __pyx_L44_except_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_3, NULL);
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1951, __pyx_L44_except_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_19);
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  if (__pyx_t_13 < (0)) __PYX_ERR(0, 1951, __pyx_L44_except_error)
                  __pyx_t_7 = (!__pyx_t_13);
                  if (unlikely(__pyx_t_7)) {
                    __Pyx_GIVEREF(__pyx_t_2);
                    __Pyx_GIVEREF(__pyx_t_16);
                    __Pyx_XGIVEREF(__pyx_t_15);
                    __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_16, __pyx_t_15);
                    __pyx_t_2 = 0;  __pyx_t_16 = 0;  __pyx_t_15 = 0; 
                    __PYX_ERR(0, 1951, __pyx_L44_except_error)
                  }
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                  goto __pyx_L43_exception_handled;
                }
                __pyx_L44_except_error:;
                __Pyx_XGIVEREF(__pyx_t_12);
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_17, __pyx_t_18);
                goto __pyx_L34_error;
                __pyx_L43_exception_handled:;
                __Pyx_XGIVEREF(__pyx_t_12);
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_17, __pyx_t_18);
                __pyx_L49_try_end:;
              }
            }
            /*finally:*/ {
              /*normal exit:*/{
                if (__pyx_t_9) {
                  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_mstate_global->__pyx_tuple[0], NULL);
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                  if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1951, __pyx_L34_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                }
                goto __pyx_L41;
              }
              __pyx_L41:;
            }
            goto __pyx_L54;
            __pyx_L36_error:;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            goto __pyx_L34_error;
            __pyx_L54:;
          }

          /* "adbc_driver_manager/_lib.pyx":1956
 *                 _blocking_exc = None
 *                 raise e from exc[1].with_traceback(exc[2])
 *         raise e             # <<<<<<<<<<<<<<
 *     finally:
 *         # Restore the signal handler
*/
          __Pyx_Raise(__pyx_v_e, 0, 0, 0);
          __PYX_ERR(0, 1956, __pyx_L34_error)
        }

        /* "adbc_driver_manager/_lib.pyx":1950
 *     try:
 *         return func(*args, **kwargs)
 *     except BaseException as e:             # <<<<<<<<<<<<<<
 *         with _blocking_lock:
 *             if _blocking_exc:
*/
        /*finally:*/ {
          __pyx_L34_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_9 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_12 = 0; __pyx_t_19 = 0; __pyx_t_22 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
             __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_19, &__pyx_t_22);
            if ( unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_18, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_18, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_22);
            __pyx_t_14 = __pyx_lineno; __pyx_t_20 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_19, __pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_9, __pyx_t_18, __pyx_t_17);
            __pyx_t_9 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_12 = 0; __pyx_t_19 = 0; __pyx_t_22 = 0;
            __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_20; __pyx_filename = __pyx_t_21;
            goto __pyx_L25_except_error;
          }
        }
      }
      goto __pyx_L25_except_error;

      /* "adbc_driver_manager/_lib.pyx":1948
 *         )
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 *     except BaseException as e:
*/
      __pyx_L25_except_error:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_11, __pyx_t_10);
      goto __pyx_L21_error;
      __pyx_L27_try_return:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_11, __pyx_t_10);
      goto __pyx_L20_return;
    }
  }

  /* "adbc_driver_manager/_lib.pyx":1959
 *     finally:
 *         # Restore the signal handler
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")             # <<<<<<<<<<<<<<
 *         if error:
 *             warnings.warn(
*/
  /*finally:*/ {
    __pyx_L21_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_8 = 0; __pyx_t_22 = 0; __pyx_t_19 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_19, &__pyx_t_12);
      if ( unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_22);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_20 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_23 = __pyx_filename;
      {
        __pyx_t_1 = NULL;
        __pyx_t_4 = __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(pyadbc_driver_manager::ClearBlockingCallback()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1959, __pyx_L60_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_4};
          __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1959, __pyx_L60_error)
          __Pyx_GOTREF(__pyx_t_6);
        }
        __pyx_t_4 = __Pyx_decode_bytes(__pyx_t_6, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1959, __pyx_L60_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_error, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "adbc_driver_manager/_lib.pyx":1960
 *         # Restore the signal handler
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")
 *         if error:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",
*/
        {
          Py_ssize_t __pyx_temp = __Pyx_PyUnicode_IS_TRUE(__pyx_v_error);
          if (unlikely(((!CYTHON_ASSUME_SAFE_SIZE) && __pyx_temp < 0))) __PYX_ERR(0, 1960, __pyx_L60_error)
          __pyx_t_7 = (__pyx_temp != 0);
        }

        if (__pyx_t_7) {

          /* "adbc_driver_manager/_lib.pyx":1961
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")
 *         if error:
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"Failed to restore SIGINT handler: {error}",
 *                 RuntimeWarning,
*/
          __pyx_t_6 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1961, __pyx_L60_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_warn); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1961, __pyx_L60_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "adbc_driver_manager/_lib.pyx":1962
 *         if error:
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",             # <<<<<<<<<<<<<<
 *                 RuntimeWarning,
 *             )
*/
          __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Failed_to_restore_SIGINT_handler, __pyx_v_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L60_error)
          __Pyx_GOTREF(__pyx_t_1);

          /* "adbc_driver_manager/_lib.pyx":1963
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",
 *                 RuntimeWarning,             # <<<<<<<<<<<<<<
 *             )
 *         with _blocking_lock:
*/
          __pyx_t_5 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_15))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_15);
            assert(__pyx_t_6);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_15);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_15, __pyx__function);
            __pyx_t_5 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_1, ((PyObject *)(((PyTypeObject*)PyExc_RuntimeWarning)))};
            __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1961, __pyx_L60_error)
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "adbc_driver_manager/_lib.pyx":1960
 *         # Restore the signal handler
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")
 *         if error:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",
*/
        }

        /* "adbc_driver_manager/_lib.pyx":1965
 *                 RuntimeWarning,
 *             )
 *         with _blocking_lock:             # <<<<<<<<<<<<<<
 *             if _blocking_exc:
 *                 exc = _blocking_exc
*/
        /*with:*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_blocking_lock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1965, __pyx_L60_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_17 = __Pyx_PyObject_LookupSpecial(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1965, __pyx_L60_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_1 = NULL;
          __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1965, __pyx_L62_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = 1;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
            assert(__pyx_t_1);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
            __pyx_t_5 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
            __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1965, __pyx_L62_error)
            __Pyx_GOTREF(__pyx_t_15);
          }
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          /*try:*/ {
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_18, &__pyx_t_9, &__pyx_t_24);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_9);
              __Pyx_XGOTREF(__pyx_t_24);
              /*try:*/ {

                /* "adbc_driver_manager/_lib.pyx":1966
 *             )
 *         with _blocking_lock:
 *             if _blocking_exc:             # <<<<<<<<<<<<<<
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
*/
                __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_blocking_exc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1966, __pyx_L66_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1966, __pyx_L66_error)
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(__pyx_t_7)) {

                  /* "adbc_driver_manager/_lib.pyx":1967
 *         with _blocking_lock:
 *             if _blocking_exc:
 *                 exc = _blocking_exc             # <<<<<<<<<<<<<<
 *                 _blocking_exc = None
 *                 raise exc[1].with_traceback(exc[2]) from KeyboardInterrupt
*/
                  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_blocking_exc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1967, __pyx_L66_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_XDECREF_SET(__pyx_v_exc, __pyx_t_4);
                  __pyx_t_4 = 0;

                  /* "adbc_driver_manager/_lib.pyx":1968
 *             if _blocking_exc:
 *                 exc = _blocking_exc
 *                 _blocking_exc = None             # <<<<<<<<<<<<<<
 *                 raise exc[1].with_traceback(exc[2]) from KeyboardInterrupt
 * 
*/
                  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_exc, Py_None) < (0)) __PYX_ERR(0, 1968, __pyx_L66_error)

                  /* "adbc_driver_manager/_lib.pyx":1969
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
 *                 raise exc[1].with_traceback(exc[2]) from KeyboardInterrupt             # <<<<<<<<<<<<<<
 * 
 * 
*/
                  __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_exc, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1969, __pyx_L66_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __pyx_t_15 = __pyx_t_6;
                  __Pyx_INCREF(__pyx_t_15);
                  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_exc, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1969, __pyx_L66_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __pyx_t_5 = 0;
                  {
                    PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_1};
                    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_with_traceback, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1969, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_4);
                  }
                  __Pyx_Raise(__pyx_t_4, 0, 0, ((PyObject *)(((PyTypeObject*)PyExc_KeyboardInterrupt))));
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __PYX_ERR(0, 1969, __pyx_L66_error)

                  /* "adbc_driver_manager/_lib.pyx":1966
 *             )
 *         with _blocking_lock:
 *             if _blocking_exc:             # <<<<<<<<<<<<<<
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
*/
                }

                /* "adbc_driver_manager/_lib.pyx":1965
 *                 RuntimeWarning,
 *             )
 *         with _blocking_lock:             # <<<<<<<<<<<<<<
 *             if _blocking_exc:
 *                 exc = _blocking_exc
*/
              }
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
              goto __pyx_L71_try_end;
              __pyx_L66_error:;
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              /*except:*/ {
                __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_1) < 0) __PYX_ERR(0, 1965, __pyx_L68_except_error)
                __Pyx_XGOTREF(__pyx_t_4);
                __Pyx_XGOTREF(__pyx_t_6);
                __Pyx_XGOTREF(__pyx_t_1);
                __pyx_t_15 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1965, __pyx_L68_except_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_15, NULL);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1965, __pyx_L68_except_error)
                __Pyx_GOTREF(__pyx_t_25);
                __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_25);
                __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                if (__pyx_t_7 < (0)) __PYX_ERR(0, 1965, __pyx_L68_except_error)
                __pyx_t_13 = (!__pyx_t_7);
                if (unlikely(__pyx_t_13)) {
                  __Pyx_GIVEREF(__pyx_t_4);
                  __Pyx_GIVEREF(__pyx_t_6);
                  __Pyx_XGIVEREF(__pyx_t_1);
                  __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_6, __pyx_t_1);
                  __pyx_t_4 = 0;  __pyx_t_6 = 0;  __pyx_t_1 = 0; 
                  __PYX_ERR(0, 1965, __pyx_L68_except_error)
                }
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                goto __pyx_L67_exception_handled;
              }
              __pyx_L68_except_error:;
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_9);
              __Pyx_XGIVEREF(__pyx_t_24);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_9, __pyx_t_24);
              goto __pyx_L60_error;
              __pyx_L67_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_9);
              __Pyx_XGIVEREF(__pyx_t_24);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_9, __pyx_t_24);
              __pyx_L71_try_end:;
            }
          }
          /*finally:*/ {
            /*normal exit:*/{
              if (__pyx_t_17) {
                __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1965, __pyx_L60_error)
                __Pyx_GOTREF(__pyx_t_24);
                __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
              }
              goto __pyx_L65;
            }
            __pyx_L65:;
          }
          goto __pyx_L76;
          __pyx_L62_error:;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          goto __pyx_L60_error;
          __pyx_L76:;
        }
      }
      __Pyx_XGIVEREF(__pyx_t_22);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_19, __pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_8 = 0; __pyx_t_22 = 0; __pyx_t_19 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_20; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_23;
      goto __pyx_L1_error;
      __pyx_L60_error:;
      __Pyx_XGIVEREF(__pyx_t_22);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_19, __pyx_t_12);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_22 = 0; __pyx_t_19 = 0; __pyx_t_12 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L20_return: {
      __pyx_t_12 = __pyx_r;
      __pyx_r = 0;

      /* "adbc_driver_manager/_lib.pyx":1959
 *     finally:
 *         # Restore the signal handler
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")             # <<<<<<<<<<<<<<
 *         if error:
 *             warnings.warn(
*/
      __pyx_t_6 = NULL;
      __pyx_t_4 = __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(pyadbc_driver_manager::ClearBlockingCallback()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1959, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1959, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_4 = __Pyx_decode_bytes(__pyx_t_1, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1959, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_error, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "adbc_driver_manager/_lib.pyx":1960
 *         # Restore the signal handler
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")
 *         if error:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",
*/
      {
        Py_ssize_t __pyx_temp = __Pyx_PyUnicode_IS_TRUE(__pyx_v_error);
        if (unlikely(((!CYTHON_ASSUME_SAFE_SIZE) && __pyx_temp < 0))) __PYX_ERR(0, 1960, __pyx_L1_error)
        __pyx_t_13 = (__pyx_temp != 0);
      }

      if (__pyx_t_13) {

        /* "adbc_driver_manager/_lib.pyx":1961
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")
 *         if error:
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"Failed to restore SIGINT handler: {error}",
 *                 RuntimeWarning,
*/
        __pyx_t_1 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_warnings); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1961, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_warn); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1961, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "adbc_driver_manager/_lib.pyx":1962
 *         if error:
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",             # <<<<<<<<<<<<<<
 *                 RuntimeWarning,
 *             )
*/
        __pyx_t_6 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Failed_to_restore_SIGINT_handler, __pyx_v_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1962, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "adbc_driver_manager/_lib.pyx":1963
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",
 *                 RuntimeWarning,             # <<<<<<<<<<<<<<
 *             )
 *         with _blocking_lock:
*/
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_15))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_15);
          assert(__pyx_t_1);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_15);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_15, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_t_6, ((PyObject *)(((PyTypeObject*)PyExc_RuntimeWarning)))};
          __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1961, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "adbc_driver_manager/_lib.pyx":1960
 *         # Restore the signal handler
 *         error = bytes(CClearBlockingCallback()).decode("utf-8")
 *         if error:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"Failed to restore SIGINT handler: {error}",
*/
      }

      /* "adbc_driver_manager/_lib.pyx":1965
 *                 RuntimeWarning,
 *             )
 *         with _blocking_lock:             # <<<<<<<<<<<<<<
 *             if _blocking_exc:
 *                 exc = _blocking_exc
*/
      /*with:*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_blocking_lock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1965, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_19 = __Pyx_PyObject_LookupSpecial(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1965, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        __pyx_t_6 = NULL;
        __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1965, __pyx_L78_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
          assert(__pyx_t_6);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
          __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1965, __pyx_L78_error)
          __Pyx_GOTREF(__pyx_t_15);
        }
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_22, &__pyx_t_8, &__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_8);
            __Pyx_XGOTREF(__pyx_t_11);
            /*try:*/ {

              /* "adbc_driver_manager/_lib.pyx":1966
 *             )
 *         with _blocking_lock:
 *             if _blocking_exc:             # <<<<<<<<<<<<<<
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
*/
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_blocking_exc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1966, __pyx_L82_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1966, __pyx_L82_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (unlikely(__pyx_t_13)) {

                /* "adbc_driver_manager/_lib.pyx":1967
 *         with _blocking_lock:
 *             if _blocking_exc:
 *                 exc = _blocking_exc             # <<<<<<<<<<<<<<
 *                 _blocking_exc = None
 *                 raise exc[1].with_traceback(exc[2]) from KeyboardInterrupt
*/
                __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_blocking_exc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1967, __pyx_L82_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_v_exc = __pyx_t_4;
                __pyx_t_4 = 0;

                /* "adbc_driver_manager/_lib.pyx":1968
 *             if _blocking_exc:
 *                 exc = _blocking_exc
 *                 _blocking_exc = None             # <<<<<<<<<<<<<<
 *                 raise exc[1].with_traceback(exc[2]) from KeyboardInterrupt
 * 
*/
                if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_exc, Py_None) < (0)) __PYX_ERR(0, 1968, __pyx_L82_error)

                /* "adbc_driver_manager/_lib.pyx":1969
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
 *                 raise exc[1].with_traceback(exc[2]) from KeyboardInterrupt             # <<<<<<<<<<<<<<
 * 
 * 
*/
                __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_exc, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1969, __pyx_L82_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_15 = __pyx_t_1;
                __Pyx_INCREF(__pyx_t_15);
                __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_exc, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1969, __pyx_L82_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_5 = 0;
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_6};
                  __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_with_traceback, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1969, __pyx_L82_error)
                  __Pyx_GOTREF(__pyx_t_4);
                }
                __Pyx_Raise(__pyx_t_4, 0, 0, ((PyObject *)(((PyTypeObject*)PyExc_KeyboardInterrupt))));
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __PYX_ERR(0, 1969, __pyx_L82_error)

                /* "adbc_driver_manager/_lib.pyx":1966
 *             )
 *         with _blocking_lock:
 *             if _blocking_exc:             # <<<<<<<<<<<<<<
 *                 exc = _blocking_exc
 *                 _blocking_exc = None
*/
              }

              /* "adbc_driver_manager/_lib.pyx":1965
 *                 RuntimeWarning,
 *             )
 *         with _blocking_lock:             # <<<<<<<<<<<<<<
 *             if _blocking_exc:
 *                 exc = _blocking_exc
*/
            }
            __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            goto __pyx_L87_try_end;
            __pyx_L82_error:;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 1965, __pyx_L84_except_error)
              __Pyx_XGOTREF(__pyx_t_4);
              __Pyx_XGOTREF(__pyx_t_1);
              __Pyx_XGOTREF(__pyx_t_6);
              __pyx_t_15 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1965, __pyx_L84_except_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_15, NULL);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1965, __pyx_L84_except_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              if (__pyx_t_13 < (0)) __PYX_ERR(0, 1965, __pyx_L84_except_error)
              __pyx_t_7 = (!__pyx_t_13);
              if (unlikely(__pyx_t_7)) {
                __Pyx_GIVEREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_1);
                __Pyx_XGIVEREF(__pyx_t_6);
                __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_1, __pyx_t_6);
                __pyx_t_4 = 0;  __pyx_t_1 = 0;  __pyx_t_6 = 0; 
                __PYX_ERR(0, 1965, __pyx_L84_except_error)
              }
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              goto __pyx_L83_exception_handled;
            }
            __pyx_L84_except_error:;
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_8);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_8, __pyx_t_11);
            goto __pyx_L1_error;
            __pyx_L83_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_8);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_8, __pyx_t_11);
            __pyx_L87_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_19) {
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_mstate_global->__pyx_tuple[0], NULL);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1965, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            }
            goto __pyx_L81;
          }
          __pyx_L81:;
        }
        goto __pyx_L92;
        __pyx_L78_error:;
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
        goto __pyx_L1_error;
        __pyx_L92:;
      }
      __pyx_r = __pyx_t_12;
      __pyx_t_12 = 0;
      goto __pyx_L0;
    }
  }

  /* "adbc_driver_manager/_lib.pyx":1919
 * 
 * 
 * def _blocking_call_impl(func, args, kwargs, cancel):             # <<<<<<<<<<<<<<
 *     """
 *     Run functions that are expected to block with a native SIGINT handler.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1976
 *     _blocking_call = _blocking_call_impl
 * else:
 *     def _blocking_call(func, args, kwargs, cancel):             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_9_blocking_call(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_9_blocking_call = {"_blocking_call", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_9_blocking_call, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_9_blocking_call(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_cancel = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_blocking_call (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_func,&__pyx_mstate_global->__pyx_n_u_args,&__pyx_mstate_global->__pyx_n_u_kwargs,&__pyx_mstate_global->__pyx_n_u_cancel,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1976, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1976, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1976, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1976, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1976, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_blocking_call", 0) < (0)) __PYX_ERR(0, 1976, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_blocking_call", 1, 4, 4, i); __PYX_ERR(0, 1976, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1976, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1976, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1976, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1976, __pyx_L3_error)
    }
    __pyx_v_func = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwargs = values[2];
    __pyx_v_cancel = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_blocking_call", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 1976, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_8_blocking_call(__pyx_self, __pyx_v_func, __pyx_v_args, __pyx_v_kwargs, __pyx_v_cancel);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_8_blocking_call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_cancel) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_blocking_call", 0);

  /* "adbc_driver_manager/_lib.pyx":1977
 * else:
 *     def _blocking_call(func, args, kwargs, cancel):
 *         return func(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
    __PYX_ERR(0, 1977, __pyx_L1_error)
  }
  if (likely(PyDict_CheckExact(__pyx_v_kwargs))) {
    __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg((PyObject*)&PyDict_Type, __pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_func, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "adbc_driver_manager/_lib.pyx":1976
 *     _blocking_call = _blocking_call_impl
 * else:
 *     def _blocking_call(func, args, kwargs, cancel):             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib._blocking_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "adbc_driver_manager/_lib.pyx":1980
 * 
 * 
 * cdef void _handle_blocking_call(void* c_cancel) noexcept nogil:             # <<<<<<<<<<<<<<
 *     with gil:
 *         try:
*/

static void __pyx_f_19adbc_driver_manager_4_lib__handle_blocking_call(void *__pyx_v_c_cancel) {
  PyObject *__pyx_v_cancel = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_handle_blocking_call", 1);

  /* "adbc_driver_manager/_lib.pyx":1981
 * 
 * cdef void _handle_blocking_call(void* c_cancel) noexcept nogil:
 *     with gil:             # <<<<<<<<<<<<<<
 *         try:
 *             cancel = <object> c_cancel
*/
  {
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      /*try:*/ {

        /* "adbc_driver_manager/_lib.pyx":1982
 * cdef void _handle_blocking_call(void* c_cancel) noexcept nogil:
 *     with gil:
 *         try:             # <<<<<<<<<<<<<<
 *             cancel = <object> c_cancel
 *             cancel()
*/
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_3);
          /*try:*/ {

            /* "adbc_driver_manager/_lib.pyx":1983
 *     with gil:
 *         try:
 *             cancel = <object> c_cancel             # <<<<<<<<<<<<<<
 *             cancel()
 *         except:  # no-cython-lint
*/
            __pyx_t_4 = ((PyObject *)__pyx_v_c_cancel);
            __Pyx_INCREF(__pyx_t_4);
            __pyx_v_cancel = __pyx_t_4;
            __pyx_t_4 = 0;

            /* "adbc_driver_manager/_lib.pyx":1984
 *         try:
 *             cancel = <object> c_cancel
 *             cancel()             # <<<<<<<<<<<<<<
 *         except:  # no-cython-lint
 *             with _blocking_lock:
*/
            __pyx_t_5 = NULL;
            __Pyx_INCREF(__pyx_v_cancel);
            __pyx_t_6 = __pyx_v_cancel; 
            __pyx_t_7 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
              assert(__pyx_t_5);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
              __pyx_t_7 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
              __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1984, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "adbc_driver_manager/_lib.pyx":1982
 * cdef void _handle_blocking_call(void* c_cancel) noexcept nogil:
 *     with gil:
 *         try:             # <<<<<<<<<<<<<<
 *             cancel = <object> c_cancel
 *             cancel()
*/
          }
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L11_try_end;
          __pyx_L6_error:;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "adbc_driver_manager/_lib.pyx":1985
 *             cancel = <object> c_cancel
 *             cancel()
 *         except:  # no-cython-lint             # <<<<<<<<<<<<<<
 *             with _blocking_lock:
 *                 global _blocking_exc
*/
          /*except:*/ {
            __Pyx_AddTraceback("adbc_driver_manager._lib._handle_blocking_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 1985, __pyx_L8_except_error)
            __Pyx_XGOTREF(__pyx_t_4);
            __Pyx_XGOTREF(__pyx_t_6);
            __Pyx_XGOTREF(__pyx_t_5);

            /* "adbc_driver_manager/_lib.pyx":1986
 *             cancel()
 *         except:  # no-cython-lint
 *             with _blocking_lock:             # <<<<<<<<<<<<<<
 *                 global _blocking_exc
 *                 _blocking_exc = sys.exc_info()
*/
            /*with:*/ {
              __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_blocking_lock); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1986, __pyx_L8_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1986, __pyx_L8_except_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_11 = NULL;
              __pyx_t_12 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1986, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_7 = 1;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_12))) {
                __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
                assert(__pyx_t_11);
                PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_11);
                __Pyx_INCREF(__pyx__function);
                __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
                __pyx_t_7 = 0;
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_11, NULL};
                __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1986, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_10);
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              /*try:*/ {
                {
                  __Pyx_PyThreadState_declare
                  __Pyx_PyThreadState_assign
                  __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
                  __Pyx_XGOTREF(__pyx_t_13);
                  __Pyx_XGOTREF(__pyx_t_14);
                  __Pyx_XGOTREF(__pyx_t_15);
                  /*try:*/ {

                    /* "adbc_driver_manager/_lib.pyx":1988
 *             with _blocking_lock:
 *                 global _blocking_exc
 *                 _blocking_exc = sys.exc_info()             # <<<<<<<<<<<<<<
*/
                    __pyx_t_10 = NULL;
                    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1988, __pyx_L20_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_exc_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1988, __pyx_L20_error)
                    __Pyx_GOTREF(__pyx_t_11);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __pyx_t_7 = 1;
                    #if CYTHON_UNPACK_METHODS
                    if (unlikely(PyMethod_Check(__pyx_t_11))) {
                      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
                      assert(__pyx_t_10);
                      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
                      __Pyx_INCREF(__pyx_t_10);
                      __Pyx_INCREF(__pyx__function);
                      __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
                      __pyx_t_7 = 0;
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
                      __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1988, __pyx_L20_error)
                      __Pyx_GOTREF(__pyx_t_8);
                    }
                    if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_exc, __pyx_t_8) < (0)) __PYX_ERR(0, 1988, __pyx_L20_error)
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

                    /* "adbc_driver_manager/_lib.pyx":1986
 *             cancel()
 *         except:  # no-cython-lint
 *             with _blocking_lock:             # <<<<<<<<<<<<<<
 *                 global _blocking_exc
 *                 _blocking_exc = sys.exc_info()
*/
                  }
                  __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                  goto __pyx_L27_try_end;
                  __pyx_L20_error:;
                  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                  /*except:*/ {
                    __Pyx_AddTraceback("adbc_driver_manager._lib._handle_blocking_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
                    if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_11, &__pyx_t_10) < 0) __PYX_ERR(0, 1986, __pyx_L22_except_error)
                    __Pyx_XGOTREF(__pyx_t_8);
                    __Pyx_XGOTREF(__pyx_t_11);
                    __Pyx_XGOTREF(__pyx_t_10);
                    __pyx_t_12 = PyTuple_Pack(3, __pyx_t_8, __pyx_t_11, __pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1986, __pyx_L22_except_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL);
                    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1986, __pyx_L22_except_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                    if (__pyx_t_17 < (0)) __PYX_ERR(0, 1986, __pyx_L22_except_error)
                    __pyx_t_18 = (!__pyx_t_17);
                    if (unlikely(__pyx_t_18)) {
                      __Pyx_GIVEREF(__pyx_t_8);
                      __Pyx_GIVEREF(__pyx_t_11);
                      __Pyx_XGIVEREF(__pyx_t_10);
                      __Pyx_ErrRestoreWithState(__pyx_t_8, __pyx_t_11, __pyx_t_10);
                      __pyx_t_8 = 0;  __pyx_t_11 = 0;  __pyx_t_10 = 0; 
                      __PYX_ERR(0, 1986, __pyx_L22_except_error)
                    }
                    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                    goto __pyx_L21_exception_handled;
                  }
                  __pyx_L22_except_error:;
                  __Pyx_XGIVEREF(__pyx_t_13);
                  __Pyx_XGIVEREF(__pyx_t_14);
                  __Pyx_XGIVEREF(__pyx_t_15);
                  __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
                  goto __pyx_L8_except_error;
                  __pyx_L21_exception_handled:;
                  __Pyx_XGIVEREF(__pyx_t_13);
                  __Pyx_XGIVEREF(__pyx_t_14);
                  __Pyx_XGIVEREF(__pyx_t_15);
                  __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
                  __pyx_L27_try_end:;
                }
              }
              /*finally:*/ {
                /*normal exit:*/{
                  if (__pyx_t_9) {
                    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_mstate_global->__pyx_tuple[0], NULL);
                    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                    if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1986, __pyx_L8_except_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                  }
                  goto __pyx_L19;
                }
                __pyx_L19:;
              }
              goto __pyx_L31;
              __pyx_L14_error:;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              goto __pyx_L8_except_error;
              __pyx_L31:;
            }
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            goto __pyx_L7_exception_handled;
          }

          /* "adbc_driver_manager/_lib.pyx":1982
 * cdef void _handle_blocking_call(void* c_cancel) noexcept nogil:
 *     with gil:
 *         try:             # <<<<<<<<<<<<<<
 *             cancel = <object> c_cancel
 *             cancel()
*/
          __pyx_L8_except_error:;
          __Pyx_XGIVEREF(__pyx_t_1);
          __Pyx_XGIVEREF(__pyx_t_2);
          __Pyx_XGIVEREF(__pyx_t_3);
          __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
          goto __pyx_L4_error;
          __pyx_L7_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_1);
          __Pyx_XGIVEREF(__pyx_t_2);
          __Pyx_XGIVEREF(__pyx_t_3);
          __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
          __pyx_L11_try_end:;
        }
      }

      /* "adbc_driver_manager/_lib.pyx":1981
 * 
 * cdef void _handle_blocking_call(void* c_cancel) noexcept nogil:
 *     with gil:             # <<<<<<<<<<<<<<
 *         try:
 *             cancel = <object> c_cancel
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_PyGILState_Release(__pyx_gilstate_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_PyGILState_Release(__pyx_gilstate_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "adbc_driver_manager/_lib.pyx":1980
 * 
 * 
 * cdef void _handle_blocking_call(void* c_cancel) noexcept nogil:             # <<<<<<<<<<<<<<
 *     with gil:
 *         try:
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_WriteUnraisable("adbc_driver_manager._lib._handle_blocking_call", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_v_cancel);
  __Pyx_RefNannyFinishContext();
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__AdbcHandle(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, b'_child_type, _lock, _open_children')
*/

/* Python wrapper */
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11__pyx_unpickle__AdbcHandle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_19adbc_driver_manager_4_lib_11__pyx_unpickle__AdbcHandle = {"__pyx_unpickle__AdbcHandle", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11__pyx_unpickle__AdbcHandle, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_19adbc_driver_manager_4_lib_11__pyx_unpickle__AdbcHandle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__AdbcHandle (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle__AdbcHandle", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__AdbcHandle", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__AdbcHandle", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("adbc_driver_manager._lib.__pyx_unpickle__AdbcHandle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_19adbc_driver_manager_4_lib_10__pyx_unpickle__AdbcHandle(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_19adbc_driver_manager_4_lib_10__pyx_unpickle__AdbcHandle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__AdbcHandle", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle__AdbcHandle(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, b'_child_type, _lock, _open_children')             # <<<<<<<<<<<<<<
 *     __pyx_result = _AdbcHandle.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, __pyx_k_child_type__lock__open_children); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, b'_child_type, _lock, _open_children')
 *     __pyx_result = _AdbcHandle.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__AdbcHandle__set_state(<_AdbcHandle> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, b'_child_type, _lock, _open_children')
 *     __pyx_result = _AdbcHandle.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__AdbcHandle__set_state(<_AdbcHandle> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = _AdbcHandle.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__AdbcHandle__set_state(<_AdbcHandle> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__AdbcHandle__set_state(_AdbcHandle __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_19adbc_driver_manager_4_lib___pyx_unpickle__AdbcHandle__set_state(((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, b'_child_type, _lock, _open_children')
 *     __pyx_result = _AdbcHandle.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__AdbcHandle__set_state(<_AdbcHandle> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__AdbcHandle__set_state(<_AdbcHandle> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__AdbcHandle__set_state(_AdbcHandle __pyx_result, __pyx_state: tuple):
 *     __pyx_result._child_type = __pyx_state[0]; __pyx_result._lock = __pyx_state[1]; __pyx_result._open_children = __pyx_state[2]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__AdbcHandle(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, b'_child_type, _lock, _open_children')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("adbc_driver_manager._lib.__pyx_unpickle__AdbcHandle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__AdbcHandle__set_state(<_AdbcHandle> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__AdbcHandle__set_state(_AdbcHandle __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._child_type = __pyx_state[0]; __pyx_result._lock = __pyx_state[1]; __pyx_result._open_children = __pyx_state[2]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)
*/

static PyObject *__pyx_f_19adbc_driver_manager_4_lib___pyx_unpickle__AdbcHandle__set_state(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__AdbcHandle__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__AdbcHandle__set_state(_AdbcHandle __pyx_result, __pyx_state: tuple):
 *     __pyx_result._child_type = __pyx_state[0]; __pyx_result._lock = __pyx_state[1]; __pyx_result._open_children = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_child_type);
  __Pyx_DECREF(__pyx_v___pyx_result->_child_type);
  __pyx_v___pyx_result->_child_type = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_lock);
  __Pyx_DECREF(__pyx_v___pyx_result->_lock);
  __pyx_v___pyx_result->_lock = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_open_children = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__AdbcHandle__set_state(_AdbcHandle __pyx_result, __pyx_state: tuple):
 *     __pyx_result._child_type = __pyx_state[0]; __pyx_result._lock = __pyx_state[1]; __pyx_result._open_children = __pyx_state[2]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 3); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle__AdbcHandle__set_state(<_AdbcHandle> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__AdbcHandle__set_state(_AdbcHandle __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._child_type = __pyx_state[0]; __pyx_result._lock = __pyx_state[1]; __pyx_result._open_children = __pyx_state[2]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("adbc_driver_manager._lib.__pyx_unpickle__AdbcHandle__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle __pyx_vtable_19adbc_driver_manager_4_lib__AdbcHandle;

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib__AdbcHandle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)o);
  p->__pyx_vtab = __pyx_vtabptr_19adbc_driver_manager_4_lib__AdbcHandle;
  p->_lock = Py_None; Py_INCREF(Py_None);
  p->_child_type = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib__AdbcHandle(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *p = (struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib__AdbcHandle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_lock);
  Py_CLEAR(p->_child_type);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_19adbc_driver_manager_4_lib__AdbcHandle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *p = (struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_lock) {
    e = (*v)(p->_lock, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_19adbc_driver_manager_4_lib__AdbcHandle(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *p = (struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *)o;
  tmp = ((PyObject*)p->_lock);
  p->_lock = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_19adbc_driver_manager_4_lib__AdbcHandle[] = {
  {"__enter__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_3__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__exit__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_5__exit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib__AdbcHandle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib__AdbcHandle},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Base class for ADBC handles, which are context managers.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_19adbc_driver_manager_4_lib__AdbcHandle},
  {Py_tp_clear, (void *)__pyx_tp_clear_19adbc_driver_manager_4_lib__AdbcHandle},
  {Py_tp_methods, (void *)__pyx_methods_19adbc_driver_manager_4_lib__AdbcHandle},
  {Py_tp_init, (void *)__pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib__AdbcHandle},
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib__AdbcHandle_spec = {
  "adbc_driver_manager._lib._AdbcHandle",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_19adbc_driver_manager_4_lib__AdbcHandle_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib__AdbcHandle = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""_AdbcHandle", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib__AdbcHandle, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Base class for ADBC handles, which are context managers.\n    "), /*tp_doc*/
  __pyx_tp_traverse_19adbc_driver_manager_4_lib__AdbcHandle, /*tp_traverse*/
  __pyx_tp_clear_19adbc_driver_manager_4_lib__AdbcHandle, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_19adbc_driver_manager_4_lib__AdbcHandle, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_19adbc_driver_manager_4_lib_11_AdbcHandle_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib__AdbcHandle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowSchemaHandle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)o);
  __Pyx_default_placement_construct(&(p->schema));
  return o;
}

#if CYTHON_USE_TP_FINALIZE

static void __pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowSchemaHandle(PyObject *o) {
  PyObject *etype, *eval, *etb;
  PyErr_Fetch(&etype, &eval, &etb);
  __pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_1__del__(o);
  PyErr_Restore(etype, eval, etb);
}
#endif

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowSchemaHandle(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowSchemaHandle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  __Pyx_call_destructor(p->schema);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_address(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7address_1__get__(o);
}

static PyObject *__pyx_getprop_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_is_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_8is_valid_1__get__(o);
}

static PyMethodDef __pyx_methods_19adbc_driver_manager_4_lib_ArrowSchemaHandle[] = {
  {"release", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_3release, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_2release},
  {"__arrow_c_schema__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_5__arrow_c_schema__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_4__arrow_c_schema__},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_19adbc_driver_manager_4_lib_ArrowSchemaHandle[] = {
  {"address", __pyx_getprop_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_address, 0, PyDoc_STR("The address of the ArrowSchema."), 0},
  {"is_valid", __pyx_getprop_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_is_valid, 0, PyDoc_STR("Check validility (object has non-NULL release pointer)."), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowSchemaHandle},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A wrapper for an allocated ArrowSchema.\n\n    This object implements the Arrow PyCapsule interface.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_19adbc_driver_manager_4_lib_ArrowSchemaHandle},
  {Py_tp_getset, (void *)__pyx_getsets_19adbc_driver_manager_4_lib_ArrowSchemaHandle},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowSchemaHandle},
  #if CYTHON_USE_TP_FINALIZE
  {Py_tp_finalize, (void *)__pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowSchemaHandle},
  #endif
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle_spec = {
  "adbc_driver_manager._lib.ArrowSchemaHandle",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""ArrowSchemaHandle", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowSchemaHandle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowSchemaHandle, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    A wrapper for an allocated ArrowSchema.\n\n    This object implements the Arrow PyCapsule interface.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_19adbc_driver_manager_4_lib_ArrowSchemaHandle, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_19adbc_driver_manager_4_lib_ArrowSchemaHandle, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib_ArrowSchemaHandle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  __pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowSchemaHandle, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayHandle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)o);
  __Pyx_default_placement_construct(&(p->array));
  return o;
}

#if CYTHON_USE_TP_FINALIZE

static void __pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowArrayHandle(PyObject *o) {
  PyObject *etype, *eval, *etb;
  PyErr_Fetch(&etype, &eval, &etb);
  __pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_1__del__(o);
  PyErr_Restore(etype, eval, etb);
}
#endif

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayHandle(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayHandle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  __Pyx_call_destructor(p->array);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_19adbc_driver_manager_4_lib_16ArrowArrayHandle_is_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_8is_valid_1__get__(o);
}

static PyObject *__pyx_getprop_19adbc_driver_manager_4_lib_16ArrowArrayHandle_address(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7address_1__get__(o);
}

static PyMethodDef __pyx_methods_19adbc_driver_manager_4_lib_ArrowArrayHandle[] = {
  {"release", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_3release, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_16ArrowArrayHandle_2release},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_19adbc_driver_manager_4_lib_ArrowArrayHandle[] = {
  {"is_valid", __pyx_getprop_19adbc_driver_manager_4_lib_16ArrowArrayHandle_is_valid, 0, PyDoc_STR("Check validility (object has non-NULL release pointer)."), 0},
  {"address", __pyx_getprop_19adbc_driver_manager_4_lib_16ArrowArrayHandle_address, 0, PyDoc_STR("\n        The address of the ArrowArray.\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayHandle},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A wrapper for an allocated ArrowArray.\n\n    This object implements the Arrow PyCapsule interface.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_19adbc_driver_manager_4_lib_ArrowArrayHandle},
  {Py_tp_getset, (void *)__pyx_getsets_19adbc_driver_manager_4_lib_ArrowArrayHandle},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayHandle},
  #if CYTHON_USE_TP_FINALIZE
  {Py_tp_finalize, (void *)__pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowArrayHandle},
  #endif
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle_spec = {
  "adbc_driver_manager._lib.ArrowArrayHandle",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""ArrowArrayHandle", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayHandle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayHandle, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    A wrapper for an allocated ArrowArray.\n\n    This object implements the Arrow PyCapsule interface.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_19adbc_driver_manager_4_lib_ArrowArrayHandle, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_19adbc_driver_manager_4_lib_ArrowArrayHandle, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayHandle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  __pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowArrayHandle, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)o);
  __Pyx_default_placement_construct(&(p->stream));
  return o;
}

#if CYTHON_USE_TP_FINALIZE

static void __pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle(PyObject *o) {
  PyObject *etype, *eval, *etb;
  PyErr_Fetch(&etype, &eval, &etb);
  __pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_1__del__(o);
  PyErr_Restore(etype, eval, etb);
}
#endif

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  __Pyx_call_destructor(p->stream);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_address(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7address_1__get__(o);
}

static PyObject *__pyx_getprop_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_is_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_8is_valid_1__get__(o);
}

static PyMethodDef __pyx_methods_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle[] = {
  {"release", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_3release, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_2release},
  {"__arrow_c_schema__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_5__arrow_c_schema__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_4__arrow_c_schema__},
  {"__arrow_c_stream__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7__arrow_c_stream__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_6__arrow_c_stream__},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle[] = {
  {"address", __pyx_getprop_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_address, 0, PyDoc_STR("The address of the ArrowArrayStream."), 0},
  {"is_valid", __pyx_getprop_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_is_valid, 0, PyDoc_STR("Check validility (object has non-NULL release pointer)."), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A wrapper for an allocated ArrowArrayStream.\n\n    This object implements the Arrow PyCapsule interface.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle},
  {Py_tp_getset, (void *)__pyx_getsets_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle},
  #if CYTHON_USE_TP_FINALIZE
  {Py_tp_finalize, (void *)__pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle},
  #endif
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle_spec = {
  "adbc_driver_manager._lib.ArrowArrayStreamHandle",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""ArrowArrayStreamHandle", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    A wrapper for an allocated ArrowArrayStream.\n\n    This object implements the Arrow PyCapsule interface.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  __pyx_tp_finalize_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcDatabase __pyx_vtable_19adbc_driver_manager_4_lib_AdbcDatabase;

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcDatabase(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *p;
  PyObject *o = __pyx_tp_new_19adbc_driver_manager_4_lib__AdbcHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle*)__pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcDatabase;
  __Pyx_default_placement_construct(&(p->database));
  return o;
}

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcDatabase(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcDatabase) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->database);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib__AdbcHandle(o);
}

static PyMethodDef __pyx_methods_19adbc_driver_manager_4_lib_AdbcDatabase[] = {
  {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_3close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_2close},
  {"get_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_5get_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_4get_option},
  {"get_option_bytes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_7get_option_bytes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_6get_option_bytes},
  {"get_option_float", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_9get_option_float, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_8get_option_float},
  {"get_option_int", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_11get_option_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_10get_option_int},
  {"set_options", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_13set_options, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_12AdbcDatabase_12set_options},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcDatabase},
  {Py_tp_doc, (void *)PyDoc_STR("\n    An instance of a database.\n\n    Parameters\n    ----------\n    kwargs : dict\n        String key-value options to pass to the underlying database.\n        Must include at least \"driver\" to identify the underlying\n        database driver to load.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_19adbc_driver_manager_4_lib__AdbcHandle},
  {Py_tp_clear, (void *)__pyx_tp_clear_19adbc_driver_manager_4_lib__AdbcHandle},
  {Py_tp_methods, (void *)__pyx_methods_19adbc_driver_manager_4_lib_AdbcDatabase},
  {Py_tp_init, (void *)__pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcDatabase},
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase_spec = {
  "adbc_driver_manager._lib.AdbcDatabase",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""AdbcDatabase", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcDatabase, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    An instance of a database.\n\n    Parameters\n    ----------\n    kwargs : dict\n        String key-value options to pass to the underlying database.\n        Must include at least \"driver\" to identify the underlying\n        database driver to load.\n    "), /*tp_doc*/
  __pyx_tp_traverse_19adbc_driver_manager_4_lib__AdbcHandle, /*tp_traverse*/
  __pyx_tp_clear_19adbc_driver_manager_4_lib__AdbcHandle, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_19adbc_driver_manager_4_lib_AdbcDatabase, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_19adbc_driver_manager_4_lib_12AdbcDatabase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib_AdbcDatabase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcConnection __pyx_vtable_19adbc_driver_manager_4_lib_AdbcConnection;

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcConnection(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *p;
  PyObject *o = __pyx_tp_new_19adbc_driver_manager_4_lib__AdbcHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle*)__pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcConnection;
  __Pyx_default_placement_construct(&(p->connection));
  p->database = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcConnection(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcConnection) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->connection);
  Py_CLEAR(p->database);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib__AdbcHandle(o);
}

static int __pyx_tp_traverse_19adbc_driver_manager_4_lib_AdbcConnection(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)o;
  e = __pyx_tp_traverse_19adbc_driver_manager_4_lib__AdbcHandle(o, v, a); if (e) return e;
  if (p->database) {
    e = (*v)(((PyObject *)p->database), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_19adbc_driver_manager_4_lib_AdbcConnection(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)o;
  __pyx_tp_clear_19adbc_driver_manager_4_lib__AdbcHandle(o);
  tmp = ((PyObject*)p->database);
  p->database = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcDatabase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_19adbc_driver_manager_4_lib_AdbcConnection[] = {
  {"cancel", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_3cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_2cancel},
  {"commit", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_5commit, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_4commit},
  {"get_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_7get_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_6get_info},
  {"get_objects", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_9get_objects, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_8get_objects},
  {"get_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_11get_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_10get_option},
  {"get_option_bytes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_13get_option_bytes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_12get_option_bytes},
  {"get_option_float", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_15get_option_float, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_14get_option_float},
  {"get_option_int", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_17get_option_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_16get_option_int},
  {"get_statistics", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_19get_statistics, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_18get_statistics},
  {"get_statistic_names", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_21get_statistic_names, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_20get_statistic_names},
  {"get_table_schema", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_23get_table_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_22get_table_schema},
  {"get_table_types", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_25get_table_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_24get_table_types},
  {"read_partition", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_27read_partition, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_26read_partition},
  {"rollback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_29rollback, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_28rollback},
  {"set_autocommit", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_31set_autocommit, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_30set_autocommit},
  {"set_options", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_33set_options, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_32set_options},
  {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_35close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_14AdbcConnection_34close},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_37__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_39__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib_AdbcConnection_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcConnection},
  {Py_tp_doc, (void *)PyDoc_STR("\n    An active database connection.\n\n    Connections are not thread-safe and clients should take care to\n    serialize accesses to a connection.\n\n    Parameters\n    ----------\n    database : AdbcDatabase\n        The database to connect to.\n    kwargs : dict\n        String key-value options to pass to the underlying database.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_19adbc_driver_manager_4_lib_AdbcConnection},
  {Py_tp_clear, (void *)__pyx_tp_clear_19adbc_driver_manager_4_lib_AdbcConnection},
  {Py_tp_methods, (void *)__pyx_methods_19adbc_driver_manager_4_lib_AdbcConnection},
  {Py_tp_init, (void *)__pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcConnection},
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib_AdbcConnection_spec = {
  "adbc_driver_manager._lib.AdbcConnection",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_19adbc_driver_manager_4_lib_AdbcConnection_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib_AdbcConnection = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""AdbcConnection", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcConnection, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    An active database connection.\n\n    Connections are not thread-safe and clients should take care to\n    serialize accesses to a connection.\n\n    Parameters\n    ----------\n    database : AdbcDatabase\n        The database to connect to.\n    kwargs : dict\n        String key-value options to pass to the underlying database.\n    "), /*tp_doc*/
  __pyx_tp_traverse_19adbc_driver_manager_4_lib_AdbcConnection, /*tp_traverse*/
  __pyx_tp_clear_19adbc_driver_manager_4_lib_AdbcConnection, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_19adbc_driver_manager_4_lib_AdbcConnection, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_19adbc_driver_manager_4_lib_14AdbcConnection_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib_AdbcConnection, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_19adbc_driver_manager_4_lib_AdbcStatement __pyx_vtable_19adbc_driver_manager_4_lib_AdbcStatement;

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcStatement(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *p;
  PyObject *o = __pyx_tp_new_19adbc_driver_manager_4_lib__AdbcHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_19adbc_driver_manager_4_lib__AdbcHandle*)__pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcStatement;
  __Pyx_default_placement_construct(&(p->statement));
  p->connection = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcStatement(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcStatement) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->statement);
  Py_CLEAR(p->connection);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib__AdbcHandle(o);
}

static int __pyx_tp_traverse_19adbc_driver_manager_4_lib_AdbcStatement(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)o;
  e = __pyx_tp_traverse_19adbc_driver_manager_4_lib__AdbcHandle(o, v, a); if (e) return e;
  if (p->connection) {
    e = (*v)(((PyObject *)p->connection), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_19adbc_driver_manager_4_lib_AdbcStatement(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *p = (struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement *)o;
  __pyx_tp_clear_19adbc_driver_manager_4_lib__AdbcHandle(o);
  tmp = ((PyObject*)p->connection);
  p->connection = ((struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcConnection *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_19adbc_driver_manager_4_lib_AdbcStatement[] = {
  {"bind", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_3bind, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_2bind},
  {"bind_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_5bind_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_4bind_stream},
  {"cancel", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_7cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_6cancel},
  {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_9close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_8close},
  {"execute_query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_11execute_query, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_10execute_query},
  {"execute_partitions", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_13execute_partitions, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_12execute_partitions},
  {"execute_schema", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_15execute_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_14execute_schema},
  {"execute_update", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_17execute_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_16execute_update},
  {"get_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_19get_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_18get_option},
  {"get_option_bytes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_21get_option_bytes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_20get_option_bytes},
  {"get_option_float", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_23get_option_float, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_22get_option_float},
  {"get_option_int", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_25get_option_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_24get_option_int},
  {"get_parameter_schema", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_27get_parameter_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_26get_parameter_schema},
  {"prepare", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_29prepare, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_28prepare},
  {"set_options", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_31set_options, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_30set_options},
  {"set_sql_query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_33set_sql_query, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_32set_sql_query},
  {"set_substrait_plan", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_35set_substrait_plan, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_19adbc_driver_manager_4_lib_13AdbcStatement_34set_substrait_plan},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_37__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_39__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib_AdbcStatement_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcStatement},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A database statement.\n\n    Statements are not thread-safe and clients should take care to\n    serialize accesses to a connection.\n\n    Parameters\n    ----------\n    connection : AdbcConnection\n        The connection to create the statement for.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_19adbc_driver_manager_4_lib_AdbcStatement},
  {Py_tp_clear, (void *)__pyx_tp_clear_19adbc_driver_manager_4_lib_AdbcStatement},
  {Py_tp_methods, (void *)__pyx_methods_19adbc_driver_manager_4_lib_AdbcStatement},
  {Py_tp_init, (void *)__pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib_AdbcStatement},
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib_AdbcStatement_spec = {
  "adbc_driver_manager._lib.AdbcStatement",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_19adbc_driver_manager_4_lib_AdbcStatement_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib_AdbcStatement = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""AdbcStatement", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib_AdbcStatement), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib_AdbcStatement, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A database statement.\n\n    Statements are not thread-safe and clients should take care to\n    serialize accesses to a connection.\n\n    Parameters\n    ----------\n    connection : AdbcConnection\n        The connection to create the statement for.\n    "), /*tp_doc*/
  __pyx_tp_traverse_19adbc_driver_manager_4_lib_AdbcStatement, /*tp_traverse*/
  __pyx_tp_clear_19adbc_driver_manager_4_lib_AdbcStatement, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_19adbc_driver_manager_4_lib_AdbcStatement, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_19adbc_driver_manager_4_lib_13AdbcStatement_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib_AdbcStatement, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error, sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error[--__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error, sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error))))
  {
    __pyx_mstate_global->__pyx_freelist_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error[__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error++] = ((struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error},
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error_spec = {
  "adbc_driver_manager._lib.__pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""__pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr, sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr[--__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr, sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr))))
  {
    __pyx_mstate_global->__pyx_freelist_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr[__pyx_mstate_global->__pyx_freecount_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr_spec = {
  "adbc_driver_manager._lib.__pyx_scope_struct_1_genexpr",
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "adbc_driver_manager._lib.""__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  {
    __pyx_t_1 = __Pyx_ApiExport_GetApiDict(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_export_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_PyObject_AdbcStatusCode_struct_A);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_export_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_export_name = __pyx_export_signature + 159;
    void (*const __pyx_export_pointers[])(void) = {(void (*)(void))&__pyx_f_19adbc_driver_manager_4_lib_convert_error, (void (*)(void))&__pyx_f_19adbc_driver_manager_4_lib_PyAdbcErrorFromArrayStream, (void (*)(void))&__pyx_f_19adbc_driver_manager_4_lib_check_error, (void (*)(void)) NULL};
    void (*const *__pyx_export_pointer)(void) = __pyx_export_pointers;
    const char *__pyx_export_current_signature = __pyx_export_signature;
    while (*__pyx_export_pointer) {
      if (__Pyx_ExportFunction(__pyx_t_1, __pyx_export_name, *__pyx_export_pointer, __pyx_export_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_export_pointer;
      __pyx_export_name = strchr(__pyx_export_name, '\0') + 1;
      __pyx_export_signature = strchr(__pyx_export_signature, '\0') + 1;
      if (*__pyx_export_signature != '\0') __pyx_export_current_signature = __pyx_export_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_19adbc_driver_manager_4_lib__AdbcHandle = &__pyx_vtable_19adbc_driver_manager_4_lib__AdbcHandle;
  __pyx_vtable_19adbc_driver_manager_4_lib__AdbcHandle._open_child = (PyObject *(*)(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *))__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__open_child;
  __pyx_vtable_19adbc_driver_manager_4_lib__AdbcHandle._close_child = (PyObject *(*)(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *))__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__close_child;
  __pyx_vtable_19adbc_driver_manager_4_lib__AdbcHandle._check_open_children = (PyObject *(*)(struct __pyx_obj_19adbc_driver_manager_4_lib__AdbcHandle *))__pyx_f_19adbc_driver_manager_4_lib_11_AdbcHandle__check_open_children;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib__AdbcHandle_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle)) __PYX_ERR(0, 303, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib__AdbcHandle_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle) < (0)) __PYX_ERR(0, 303, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle = &__pyx_type_19adbc_driver_manager_4_lib__AdbcHandle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle) < (0)) __PYX_ERR(0, 303, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle, __pyx_vtabptr_19adbc_driver_manager_4_lib__AdbcHandle) < (0)) __PYX_ERR(0, 303, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle) < (0)) __PYX_ERR(0, 303, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_AdbcHandle, (PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle) < (0)) __PYX_ERR(0, 303, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle) < (0)) __PYX_ERR(0, 303, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle)) __PYX_ERR(0, 366, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle) < (0)) __PYX_ERR(0, 366, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle = &__pyx_type_19adbc_driver_manager_4_lib_ArrowSchemaHandle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle) < (0)) __PYX_ERR(0, 366, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle, (PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle) < (0)) __PYX_ERR(0, 366, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle) < (0)) __PYX_ERR(0, 366, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle)) __PYX_ERR(0, 413, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle) < (0)) __PYX_ERR(0, 413, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle = &__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayHandle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle) < (0)) __PYX_ERR(0, 413, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ArrowArrayHandle, (PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle) < (0)) __PYX_ERR(0, 413, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle) < (0)) __PYX_ERR(0, 413, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle)) __PYX_ERR(0, 449, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 449, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle = &__pyx_type_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 449, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle, (PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 449, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 449, __pyx_L1_error)
  __pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcDatabase = &__pyx_vtable_19adbc_driver_manager_4_lib_AdbcDatabase;
  __pyx_vtable_19adbc_driver_manager_4_lib_AdbcDatabase.__pyx_base = *__pyx_vtabptr_19adbc_driver_manager_4_lib__AdbcHandle;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase)) __PYX_ERR(0, 537, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase) < (0)) __PYX_ERR(0, 537, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase = &__pyx_type_19adbc_driver_manager_4_lib_AdbcDatabase;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase->tp_base = __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase) < (0)) __PYX_ERR(0, 537, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, __pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcDatabase) < (0)) __PYX_ERR(0, 537, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase) < (0)) __PYX_ERR(0, 537, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_AdbcDatabase, (PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase) < (0)) __PYX_ERR(0, 537, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase) < (0)) __PYX_ERR(0, 537, __pyx_L1_error)
  __pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcConnection = &__pyx_vtable_19adbc_driver_manager_4_lib_AdbcConnection;
  __pyx_vtable_19adbc_driver_manager_4_lib_AdbcConnection.__pyx_base = *__pyx_vtabptr_19adbc_driver_manager_4_lib__AdbcHandle;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib_AdbcConnection_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection)) __PYX_ERR(0, 841, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib_AdbcConnection_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection) < (0)) __PYX_ERR(0, 841, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection = &__pyx_type_19adbc_driver_manager_4_lib_AdbcConnection;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection->tp_base = __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection) < (0)) __PYX_ERR(0, 841, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcConnection) < (0)) __PYX_ERR(0, 841, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection) < (0)) __PYX_ERR(0, 841, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_AdbcConnection, (PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection) < (0)) __PYX_ERR(0, 841, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection) < (0)) __PYX_ERR(0, 841, __pyx_L1_error)
  __pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcStatement = &__pyx_vtable_19adbc_driver_manager_4_lib_AdbcStatement;
  __pyx_vtable_19adbc_driver_manager_4_lib_AdbcStatement.__pyx_base = *__pyx_vtabptr_19adbc_driver_manager_4_lib__AdbcHandle;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib_AdbcStatement_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement)) __PYX_ERR(0, 1411, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib_AdbcStatement_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement) < (0)) __PYX_ERR(0, 1411, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement = &__pyx_type_19adbc_driver_manager_4_lib_AdbcStatement;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement->tp_base = __pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement) < (0)) __PYX_ERR(0, 1411, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_vtabptr_19adbc_driver_manager_4_lib_AdbcStatement) < (0)) __PYX_ERR(0, 1411, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement) < (0)) __PYX_ERR(0, 1411, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_AdbcStatement, (PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement) < (0)) __PYX_ERR(0, 1411, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement) < (0)) __PYX_ERR(0, 1411, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error)) __PYX_ERR(0, 180, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error) < (0)) __PYX_ERR(0, 180, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error = &__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error) < (0)) __PYX_ERR(0, 180, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct____pyx_f_19adbc_driver_manager_4_lib_convert_error->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr)) __PYX_ERR(0, 202, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr_spec, __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr) < (0)) __PYX_ERR(0, 202, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr = &__pyx_type_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr) < (0)) __PYX_ERR(0, 202, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr->tp_dictoffset && __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_19adbc_driver_manager_4_lib___pyx_scope_struct_1_genexpr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__lib(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__lib},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_lib",
      __pyx_k_Low_level_ADBC_API, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit__lib(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__lib(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__lib(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_lib' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_lib" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_NOT_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit__lib", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_adbc_driver_manager___lib) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "adbc_driver_manager._lib")) {
      if (unlikely((PyDict_SetItemString(modules, "adbc_driver_manager._lib", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_export_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "adbc_driver_manager/_lib.pyx":23
 * """Low-level ADBC API."""
 * 
 * import enum             # <<<<<<<<<<<<<<
 * import functools
 * import threading
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_enum, __pyx_t_2) < (0)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":24
 * 
 * import enum
 * import functools             # <<<<<<<<<<<<<<
 * import threading
 * import os
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_functools, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_functools, __pyx_t_2) < (0)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":25
 * import enum
 * import functools
 * import threading             # <<<<<<<<<<<<<<
 * import os
 * import pathlib
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_threading, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_threading, __pyx_t_2) < (0)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":26
 * import functools
 * import threading
 * import os             # <<<<<<<<<<<<<<
 * import pathlib
 * import typing
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_os, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_os, __pyx_t_2) < (0)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":27
 * import threading
 * import os
 * import pathlib             # <<<<<<<<<<<<<<
 * import typing
 * import sys
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pathlib, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pathlib, __pyx_t_2) < (0)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":28
 * import os
 * import pathlib
 * import typing             # <<<<<<<<<<<<<<
 * import sys
 * import warnings
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_typing, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_typing, __pyx_t_2) < (0)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":29
 * import pathlib
 * import typing
 * import sys             # <<<<<<<<<<<<<<
 * import warnings
 * from typing import List, Optional, Tuple
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_sys, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_sys, __pyx_t_2) < (0)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":30
 * import typing
 * import sys
 * import warnings             # <<<<<<<<<<<<<<
 * from typing import List, Optional, Tuple
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_warnings, __pyx_t_2) < (0)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":31
 * import sys
 * import warnings
 * from typing import List, Optional, Tuple             # <<<<<<<<<<<<<<
 * 
 * import cython
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_List,__pyx_mstate_global->__pyx_n_u_Optional,__pyx_mstate_global->__pyx_n_u_Tuple};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_typing, __pyx_imported_names, 3, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_List,__pyx_mstate_global->__pyx_n_u_Optional,__pyx_mstate_global->__pyx_n_u_Tuple};
    for (__pyx_t_3=0; __pyx_t_3 < 3; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 31, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 31, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":44
 * from libcpp.vector cimport vector as c_vector
 * 
 * if typing.TYPE_CHECKING:             # <<<<<<<<<<<<<<
 *     from typing import Self  # no-cython-lint
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_typing); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_TYPE_CHECKING); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "adbc_driver_manager/_lib.pyx":45
 * 
 * if typing.TYPE_CHECKING:
 *     from typing import Self  # no-cython-lint             # <<<<<<<<<<<<<<
 * 
 * 
*/
    {
      PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Self};
      __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_typing, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
    }
    __pyx_t_4 = __pyx_t_1;
    __Pyx_GOTREF(__pyx_t_4);
    {
      PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Self};
      __pyx_t_3 = 0; {
        __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_4, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_2) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":44
 * from libcpp.vector cimport vector as c_vector
 * 
 * if typing.TYPE_CHECKING:             # <<<<<<<<<<<<<<
 *     from typing import Self  # no-cython-lint
 * 
*/
  }

  /* "adbc_driver_manager/_lib.pyx":48
 * 
 * 
 * class AdbcStatusCode(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     """
 *     A status code indicating the type of error.
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_6, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_AdbcStatusCode, __pyx_mstate_global->__pyx_n_u_AdbcStatusCode, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_A_status_code_indicating_the_ty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_2 != __pyx_t_4) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_4) < 0))) __PYX_ERR(0, 48, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":53
 *     """
 * 
 *     OK = ADBC_STATUS_OK             # <<<<<<<<<<<<<<
 *     UNKNOWN = ADBC_STATUS_UNKNOWN
 *     NOT_IMPLEMENTED = ADBC_STATUS_NOT_IMPLEMENTED
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_OK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_OK, __pyx_t_4) < (0)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":54
 * 
 *     OK = ADBC_STATUS_OK
 *     UNKNOWN = ADBC_STATUS_UNKNOWN             # <<<<<<<<<<<<<<
 *     NOT_IMPLEMENTED = ADBC_STATUS_NOT_IMPLEMENTED
 *     NOT_FOUND = ADBC_STATUS_NOT_FOUND
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_UNKNOWN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_UNKNOWN, __pyx_t_4) < (0)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":55
 *     OK = ADBC_STATUS_OK
 *     UNKNOWN = ADBC_STATUS_UNKNOWN
 *     NOT_IMPLEMENTED = ADBC_STATUS_NOT_IMPLEMENTED             # <<<<<<<<<<<<<<
 *     NOT_FOUND = ADBC_STATUS_NOT_FOUND
 *     ALREADY_EXISTS = ADBC_STATUS_ALREADY_EXISTS
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_NOT_IMPLEMENTED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_NOT_IMPLEMENTED, __pyx_t_4) < (0)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":56
 *     UNKNOWN = ADBC_STATUS_UNKNOWN
 *     NOT_IMPLEMENTED = ADBC_STATUS_NOT_IMPLEMENTED
 *     NOT_FOUND = ADBC_STATUS_NOT_FOUND             # <<<<<<<<<<<<<<
 *     ALREADY_EXISTS = ADBC_STATUS_ALREADY_EXISTS
 *     INVALID_ARGUMENT = ADBC_STATUS_INVALID_ARGUMENT
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_NOT_FOUND); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_NOT_FOUND, __pyx_t_4) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":57
 *     NOT_IMPLEMENTED = ADBC_STATUS_NOT_IMPLEMENTED
 *     NOT_FOUND = ADBC_STATUS_NOT_FOUND
 *     ALREADY_EXISTS = ADBC_STATUS_ALREADY_EXISTS             # <<<<<<<<<<<<<<
 *     INVALID_ARGUMENT = ADBC_STATUS_INVALID_ARGUMENT
 *     INVALID_STATE = ADBC_STATUS_INVALID_STATE
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_ALREADY_EXISTS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ALREADY_EXISTS, __pyx_t_4) < (0)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":58
 *     NOT_FOUND = ADBC_STATUS_NOT_FOUND
 *     ALREADY_EXISTS = ADBC_STATUS_ALREADY_EXISTS
 *     INVALID_ARGUMENT = ADBC_STATUS_INVALID_ARGUMENT             # <<<<<<<<<<<<<<
 *     INVALID_STATE = ADBC_STATUS_INVALID_STATE
 *     INVALID_DATA = ADBC_STATUS_INVALID_DATA
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_INVALID_ARGUMENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INVALID_ARGUMENT, __pyx_t_4) < (0)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":59
 *     ALREADY_EXISTS = ADBC_STATUS_ALREADY_EXISTS
 *     INVALID_ARGUMENT = ADBC_STATUS_INVALID_ARGUMENT
 *     INVALID_STATE = ADBC_STATUS_INVALID_STATE             # <<<<<<<<<<<<<<
 *     INVALID_DATA = ADBC_STATUS_INVALID_DATA
 *     INTEGRITY = ADBC_STATUS_INTEGRITY
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_INVALID_STATE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INVALID_STATE, __pyx_t_4) < (0)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":60
 *     INVALID_ARGUMENT = ADBC_STATUS_INVALID_ARGUMENT
 *     INVALID_STATE = ADBC_STATUS_INVALID_STATE
 *     INVALID_DATA = ADBC_STATUS_INVALID_DATA             # <<<<<<<<<<<<<<
 *     INTEGRITY = ADBC_STATUS_INTEGRITY
 *     INTERNAL = ADBC_STATUS_INTERNAL
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_INVALID_DATA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INVALID_DATA, __pyx_t_4) < (0)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":61
 *     INVALID_STATE = ADBC_STATUS_INVALID_STATE
 *     INVALID_DATA = ADBC_STATUS_INVALID_DATA
 *     INTEGRITY = ADBC_STATUS_INTEGRITY             # <<<<<<<<<<<<<<
 *     INTERNAL = ADBC_STATUS_INTERNAL
 *     IO = ADBC_STATUS_IO
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_INTEGRITY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INTEGRITY, __pyx_t_4) < (0)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":62
 *     INVALID_DATA = ADBC_STATUS_INVALID_DATA
 *     INTEGRITY = ADBC_STATUS_INTEGRITY
 *     INTERNAL = ADBC_STATUS_INTERNAL             # <<<<<<<<<<<<<<
 *     IO = ADBC_STATUS_IO
 *     CANCELLED = ADBC_STATUS_CANCELLED
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_INTERNAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INTERNAL, __pyx_t_4) < (0)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":63
 *     INTEGRITY = ADBC_STATUS_INTEGRITY
 *     INTERNAL = ADBC_STATUS_INTERNAL
 *     IO = ADBC_STATUS_IO             # <<<<<<<<<<<<<<
 *     CANCELLED = ADBC_STATUS_CANCELLED
 *     TIMEOUT = ADBC_STATUS_TIMEOUT
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_IO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_IO, __pyx_t_4) < (0)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":64
 *     INTERNAL = ADBC_STATUS_INTERNAL
 *     IO = ADBC_STATUS_IO
 *     CANCELLED = ADBC_STATUS_CANCELLED             # <<<<<<<<<<<<<<
 *     TIMEOUT = ADBC_STATUS_TIMEOUT
 *     UNAUTHENTICATED = ADBC_STATUS_UNAUTHENTICATED
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_CANCELLED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CANCELLED, __pyx_t_4) < (0)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":65
 *     IO = ADBC_STATUS_IO
 *     CANCELLED = ADBC_STATUS_CANCELLED
 *     TIMEOUT = ADBC_STATUS_TIMEOUT             # <<<<<<<<<<<<<<
 *     UNAUTHENTICATED = ADBC_STATUS_UNAUTHENTICATED
 *     UNAUTHORIZED = ADBC_STATUS_UNAUTHORIZED
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_TIMEOUT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_TIMEOUT, __pyx_t_4) < (0)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":66
 *     CANCELLED = ADBC_STATUS_CANCELLED
 *     TIMEOUT = ADBC_STATUS_TIMEOUT
 *     UNAUTHENTICATED = ADBC_STATUS_UNAUTHENTICATED             # <<<<<<<<<<<<<<
 *     UNAUTHORIZED = ADBC_STATUS_UNAUTHORIZED
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_UNAUTHENTICATED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_UNAUTHENTICATED, __pyx_t_4) < (0)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":67
 *     TIMEOUT = ADBC_STATUS_TIMEOUT
 *     UNAUTHENTICATED = ADBC_STATUS_UNAUTHENTICATED
 *     UNAUTHORIZED = ADBC_STATUS_UNAUTHORIZED             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_uint8_t(ADBC_STATUS_UNAUTHORIZED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_UNAUTHORIZED, __pyx_t_4) < (0)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":48
 * 
 * 
 * class AdbcStatusCode(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     """
 *     A status code indicating the type of error.
*/
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_AdbcStatusCode, __pyx_t_2, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_AdbcStatusCode, __pyx_t_4) < (0)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":70
 * 
 * 
 * class AdbcInfoCode(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     VENDOR_NAME = ADBC_INFO_VENDOR_NAME
 *     VENDOR_VERSION = ADBC_INFO_VENDOR_VERSION
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_enum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_Pack(1, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_6, __pyx_mstate_global->__pyx_n_u_AdbcInfoCode, __pyx_mstate_global->__pyx_n_u_AdbcInfoCode, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_6 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_4, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 70, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":71
 * 
 * class AdbcInfoCode(enum.IntEnum):
 *     VENDOR_NAME = ADBC_INFO_VENDOR_NAME             # <<<<<<<<<<<<<<
 *     VENDOR_VERSION = ADBC_INFO_VENDOR_VERSION
 *     VENDOR_ARROW_VERSION = ADBC_INFO_VENDOR_ARROW_VERSION
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint32_t(ADBC_INFO_VENDOR_NAME); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_VENDOR_NAME, __pyx_t_2) < (0)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":72
 * class AdbcInfoCode(enum.IntEnum):
 *     VENDOR_NAME = ADBC_INFO_VENDOR_NAME
 *     VENDOR_VERSION = ADBC_INFO_VENDOR_VERSION             # <<<<<<<<<<<<<<
 *     VENDOR_ARROW_VERSION = ADBC_INFO_VENDOR_ARROW_VERSION
 *     DRIVER_NAME = ADBC_INFO_DRIVER_NAME
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint32_t(ADBC_INFO_VENDOR_VERSION); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_VENDOR_VERSION, __pyx_t_2) < (0)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":73
 *     VENDOR_NAME = ADBC_INFO_VENDOR_NAME
 *     VENDOR_VERSION = ADBC_INFO_VENDOR_VERSION
 *     VENDOR_ARROW_VERSION = ADBC_INFO_VENDOR_ARROW_VERSION             # <<<<<<<<<<<<<<
 *     DRIVER_NAME = ADBC_INFO_DRIVER_NAME
 *     DRIVER_VERSION = ADBC_INFO_DRIVER_VERSION
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint32_t(ADBC_INFO_VENDOR_ARROW_VERSION); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_VENDOR_ARROW_VERSION, __pyx_t_2) < (0)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":74
 *     VENDOR_VERSION = ADBC_INFO_VENDOR_VERSION
 *     VENDOR_ARROW_VERSION = ADBC_INFO_VENDOR_ARROW_VERSION
 *     DRIVER_NAME = ADBC_INFO_DRIVER_NAME             # <<<<<<<<<<<<<<
 *     DRIVER_VERSION = ADBC_INFO_DRIVER_VERSION
 *     DRIVER_ARROW_VERSION = ADBC_INFO_DRIVER_ARROW_VERSION
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint32_t(ADBC_INFO_DRIVER_NAME); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DRIVER_NAME, __pyx_t_2) < (0)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":75
 *     VENDOR_ARROW_VERSION = ADBC_INFO_VENDOR_ARROW_VERSION
 *     DRIVER_NAME = ADBC_INFO_DRIVER_NAME
 *     DRIVER_VERSION = ADBC_INFO_DRIVER_VERSION             # <<<<<<<<<<<<<<
 *     DRIVER_ARROW_VERSION = ADBC_INFO_DRIVER_ARROW_VERSION
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint32_t(ADBC_INFO_DRIVER_VERSION); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DRIVER_VERSION, __pyx_t_2) < (0)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":76
 *     DRIVER_NAME = ADBC_INFO_DRIVER_NAME
 *     DRIVER_VERSION = ADBC_INFO_DRIVER_VERSION
 *     DRIVER_ARROW_VERSION = ADBC_INFO_DRIVER_ARROW_VERSION             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint32_t(ADBC_INFO_DRIVER_ARROW_VERSION); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DRIVER_ARROW_VERSION, __pyx_t_2) < (0)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":70
 * 
 * 
 * class AdbcInfoCode(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     VENDOR_NAME = ADBC_INFO_VENDOR_NAME
 *     VENDOR_VERSION = ADBC_INFO_VENDOR_VERSION
*/
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_AdbcInfoCode, __pyx_t_6, __pyx_t_4, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_AdbcInfoCode, __pyx_t_2) < (0)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "adbc_driver_manager/_lib.pyx":79
 * 
 * 
 * class Warning(UserWarning):             # <<<<<<<<<<<<<<
 *     """
 *     PEP 249-compliant base warning class.
*/
  __pyx_t_6 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_UserWarning)))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PEP560_update_bases(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_7, __pyx_mstate_global->__pyx_n_u_Warning, __pyx_mstate_global->__pyx_n_u_Warning, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_PEP_249_compliant_base_warning); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_7 != __pyx_t_6) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_6) < 0))) __PYX_ERR(0, 79, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Warning, __pyx_t_7, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_6);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Warning, __pyx_t_6) < (0)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "adbc_driver_manager/_lib.pyx":85
 * 
 * 
 * class Error(Exception):             # <<<<<<<<<<<<<<
 *     """
 *     PEP 249-compliant base exception class.
*/
  __pyx_t_7 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_Error, __pyx_mstate_global->__pyx_n_u_Error, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_PEP_249_compliant_base_exceptio); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_4 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_6, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 85, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "adbc_driver_manager/_lib.pyx":101
 *     """
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         message,
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "adbc_driver_manager/_lib.pyx":106
 *         *,
 *         status_code,
 *         vendor_code=None,             # <<<<<<<<<<<<<<
 *         sqlstate=None,
 *         details=None,
*/
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_vendor_code, Py_None) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":107
 *         status_code,
 *         vendor_code=None,
 *         sqlstate=None,             # <<<<<<<<<<<<<<
 *         details=None,
 *     ):
*/
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_sqlstate, Py_None) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":108
 *         vendor_code=None,
 *         sqlstate=None,
 *         details=None,             # <<<<<<<<<<<<<<
 *     ):
 *         super().__init__(message)
*/
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_details, Py_None) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":101
 *     """
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         message,
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_5Error_1__init__, 0, __pyx_mstate_global->__pyx_n_u_Error___init, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  PyList_Append(__pyx_t_7, __pyx_t_9);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_9, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_9) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":85
 * 
 * 
 * class Error(Exception):             # <<<<<<<<<<<<<<
 *     """
 *     PEP 249-compliant base exception class.
*/
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Error, __pyx_t_4, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_7, __pyx_t_9) < (0)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Error, __pyx_t_9) < (0)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":117
 * 
 * 
 * class InterfaceError(Error):             # <<<<<<<<<<<<<<
 *     """Errors related to the database interface."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare(__pyx_t_6, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_InterfaceError, __pyx_mstate_global->__pyx_n_u_InterfaceError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_the_database_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__pyx_t_4 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_9, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 117, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_InterfaceError, __pyx_t_4, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_InterfaceError, __pyx_t_2) < (0)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":121
 * 
 * 
 * class DatabaseError(Error):             # <<<<<<<<<<<<<<
 *     """Errors related to the database."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_9, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError, __pyx_mstate_global->__pyx_n_u_DatabaseError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_the_database); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_4 != __pyx_t_6) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_6) < 0))) __PYX_ERR(0, 121, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_Py3ClassCreate(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_DatabaseError, __pyx_t_4, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_6);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_DatabaseError, __pyx_t_6) < (0)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":125
 * 
 * 
 * class DataError(DatabaseError):             # <<<<<<<<<<<<<<
 *     """Errors related to processed data."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_DataError, __pyx_mstate_global->__pyx_n_u_DataError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_processed_data); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_4 != __pyx_t_9) {
    if (unlikely((PyDict_SetItemString(__pyx_t_6, "__orig_bases__", __pyx_t_9) < 0))) __PYX_ERR(0, 125, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_DataError, __pyx_t_4, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_DataError, __pyx_t_9) < (0)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":129
 * 
 * 
 * class OperationalError(DatabaseError):             # <<<<<<<<<<<<<<
 *     """Errors related to database operation, not under user control."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare(__pyx_t_6, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_OperationalError, __pyx_mstate_global->__pyx_n_u_OperationalError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_database_opera); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__pyx_t_4 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_9, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 129, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_OperationalError, __pyx_t_4, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_OperationalError, __pyx_t_2) < (0)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":133
 * 
 * 
 * class IntegrityError(DatabaseError):             # <<<<<<<<<<<<<<
 *     """Errors related to relational integrity."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_9, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_IntegrityError, __pyx_mstate_global->__pyx_n_u_IntegrityError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_relational_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_4 != __pyx_t_6) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_6) < 0))) __PYX_ERR(0, 133, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_Py3ClassCreate(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_IntegrityError, __pyx_t_4, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_6);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IntegrityError, __pyx_t_6) < (0)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":137
 * 
 * 
 * class InternalError(DatabaseError):             # <<<<<<<<<<<<<<
 *     """Errors related to database-internal errors."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_InternalError, __pyx_mstate_global->__pyx_n_u_InternalError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_database_inter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_4 != __pyx_t_9) {
    if (unlikely((PyDict_SetItemString(__pyx_t_6, "__orig_bases__", __pyx_t_9) < 0))) __PYX_ERR(0, 137, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_InternalError, __pyx_t_4, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_InternalError, __pyx_t_9) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":141
 * 
 * 
 * class ProgrammingError(DatabaseError):             # <<<<<<<<<<<<<<
 *     """Errors related to user errors."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare(__pyx_t_6, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_ProgrammingError, __pyx_mstate_global->__pyx_n_u_ProgrammingError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_user_errors); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__pyx_t_4 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_9, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 141, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ProgrammingError, __pyx_t_4, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ProgrammingError, __pyx_t_2) < (0)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":145
 * 
 * 
 * class NotSupportedError(DatabaseError):             # <<<<<<<<<<<<<<
 *     """An operation or some functionality is not supported."""
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_9, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_NotSupportedError, __pyx_mstate_global->__pyx_n_u_NotSupportedError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_An_operation_or_some_functionali); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_4 != __pyx_t_6) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_6) < 0))) __PYX_ERR(0, 145, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "adbc_driver_manager/_lib.pyx":148
 *     """An operation or some functionality is not supported."""
 * 
 *     def __init__(self, message, *, vendor_code=None, sqlstate=None, details=None):             # <<<<<<<<<<<<<<
 *         super().__init__(
 *             message,
*/
  __pyx_t_7 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_vendor_code, Py_None) < (0)) __PYX_ERR(0, 148, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_sqlstate, Py_None) < (0)) __PYX_ERR(0, 148, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_details, Py_None) < (0)) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_17NotSupportedError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_NotSupportedError___init, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  PyList_Append(__pyx_t_6, __pyx_t_8);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_8, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_8) < (0)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":145
 * 
 * 
 * class NotSupportedError(DatabaseError):             # <<<<<<<<<<<<<<
 *     """An operation or some functionality is not supported."""
 * 
*/
  __pyx_t_8 = __Pyx_Py3ClassCreate(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_NotSupportedError, __pyx_t_4, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_6, __pyx_t_8) < (0)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_NotSupportedError, __pyx_t_8) < (0)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":161
 * # could also define the exceptions in __init__ but then we'd have a
 * # circular import situation
 * Error.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * InterfaceError.__module__ = "adbc_driver_manager"
 * DatabaseError.__module__ = "adbc_driver_manager"
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":162
 * # circular import situation
 * Error.__module__ = "adbc_driver_manager"
 * InterfaceError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * DatabaseError.__module__ = "adbc_driver_manager"
 * DataError.__module__ = "adbc_driver_manager"
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_InterfaceError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":163
 * Error.__module__ = "adbc_driver_manager"
 * InterfaceError.__module__ = "adbc_driver_manager"
 * DatabaseError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * DataError.__module__ = "adbc_driver_manager"
 * OperationalError.__module__ = "adbc_driver_manager"
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DatabaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":164
 * InterfaceError.__module__ = "adbc_driver_manager"
 * DatabaseError.__module__ = "adbc_driver_manager"
 * DataError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * OperationalError.__module__ = "adbc_driver_manager"
 * IntegrityError.__module__ = "adbc_driver_manager"
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DataError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":165
 * DatabaseError.__module__ = "adbc_driver_manager"
 * DataError.__module__ = "adbc_driver_manager"
 * OperationalError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * IntegrityError.__module__ = "adbc_driver_manager"
 * InternalError.__module__ = "adbc_driver_manager"
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_OperationalError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":166
 * DataError.__module__ = "adbc_driver_manager"
 * OperationalError.__module__ = "adbc_driver_manager"
 * IntegrityError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * InternalError.__module__ = "adbc_driver_manager"
 * ProgrammingError.__module__ = "adbc_driver_manager"
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_IntegrityError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":167
 * OperationalError.__module__ = "adbc_driver_manager"
 * IntegrityError.__module__ = "adbc_driver_manager"
 * InternalError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * ProgrammingError.__module__ = "adbc_driver_manager"
 * NotSupportedError.__module__ = "adbc_driver_manager"
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_InternalError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":168
 * IntegrityError.__module__ = "adbc_driver_manager"
 * InternalError.__module__ = "adbc_driver_manager"
 * ProgrammingError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * NotSupportedError.__module__ = "adbc_driver_manager"
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ProgrammingError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":169
 * InternalError.__module__ = "adbc_driver_manager"
 * ProgrammingError.__module__ = "adbc_driver_manager"
 * NotSupportedError.__module__ = "adbc_driver_manager"             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_NotSupportedError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_module, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager) < (0)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":172
 * 
 * 
 * INGEST_OPTION_MODE = ADBC_INGEST_OPTION_MODE.decode("utf-8")             # <<<<<<<<<<<<<<
 * INGEST_OPTION_MODE_APPEND = ADBC_INGEST_OPTION_MODE_APPEND.decode("utf-8")
 * INGEST_OPTION_MODE_CREATE = ADBC_INGEST_OPTION_MODE_CREATE.decode("utf-8")
*/
  __pyx_t_3 = __Pyx_ssize_strlen(ADBC_INGEST_OPTION_MODE); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 172, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(ADBC_INGEST_OPTION_MODE, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_INGEST_OPTION_MODE, __pyx_t_4) < (0)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":173
 * 
 * INGEST_OPTION_MODE = ADBC_INGEST_OPTION_MODE.decode("utf-8")
 * INGEST_OPTION_MODE_APPEND = ADBC_INGEST_OPTION_MODE_APPEND.decode("utf-8")             # <<<<<<<<<<<<<<
 * INGEST_OPTION_MODE_CREATE = ADBC_INGEST_OPTION_MODE_CREATE.decode("utf-8")
 * INGEST_OPTION_MODE_REPLACE = ADBC_INGEST_OPTION_MODE_REPLACE.decode("utf-8")
*/
  __pyx_t_3 = __Pyx_ssize_strlen(ADBC_INGEST_OPTION_MODE_APPEND); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 173, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(ADBC_INGEST_OPTION_MODE_APPEND, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_INGEST_OPTION_MODE_APPEND, __pyx_t_4) < (0)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":174
 * INGEST_OPTION_MODE = ADBC_INGEST_OPTION_MODE.decode("utf-8")
 * INGEST_OPTION_MODE_APPEND = ADBC_INGEST_OPTION_MODE_APPEND.decode("utf-8")
 * INGEST_OPTION_MODE_CREATE = ADBC_INGEST_OPTION_MODE_CREATE.decode("utf-8")             # <<<<<<<<<<<<<<
 * INGEST_OPTION_MODE_REPLACE = ADBC_INGEST_OPTION_MODE_REPLACE.decode("utf-8")
 * INGEST_OPTION_MODE_CREATE_APPEND = ADBC_INGEST_OPTION_MODE_CREATE_APPEND.decode("utf-8")
*/
  __pyx_t_3 = __Pyx_ssize_strlen(ADBC_INGEST_OPTION_MODE_CREATE); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(ADBC_INGEST_OPTION_MODE_CREATE, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_INGEST_OPTION_MODE_CREATE, __pyx_t_4) < (0)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":175
 * INGEST_OPTION_MODE_APPEND = ADBC_INGEST_OPTION_MODE_APPEND.decode("utf-8")
 * INGEST_OPTION_MODE_CREATE = ADBC_INGEST_OPTION_MODE_CREATE.decode("utf-8")
 * INGEST_OPTION_MODE_REPLACE = ADBC_INGEST_OPTION_MODE_REPLACE.decode("utf-8")             # <<<<<<<<<<<<<<
 * INGEST_OPTION_MODE_CREATE_APPEND = ADBC_INGEST_OPTION_MODE_CREATE_APPEND.decode("utf-8")
 * INGEST_OPTION_TARGET_TABLE = ADBC_INGEST_OPTION_TARGET_TABLE.decode("utf-8")
*/
  __pyx_t_3 = __Pyx_ssize_strlen(ADBC_INGEST_OPTION_MODE_REPLACE); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(ADBC_INGEST_OPTION_MODE_REPLACE, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_INGEST_OPTION_MODE_REPLACE, __pyx_t_4) < (0)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":176
 * INGEST_OPTION_MODE_CREATE = ADBC_INGEST_OPTION_MODE_CREATE.decode("utf-8")
 * INGEST_OPTION_MODE_REPLACE = ADBC_INGEST_OPTION_MODE_REPLACE.decode("utf-8")
 * INGEST_OPTION_MODE_CREATE_APPEND = ADBC_INGEST_OPTION_MODE_CREATE_APPEND.decode("utf-8")             # <<<<<<<<<<<<<<
 * INGEST_OPTION_TARGET_TABLE = ADBC_INGEST_OPTION_TARGET_TABLE.decode("utf-8")
 * 
*/
  __pyx_t_3 = __Pyx_ssize_strlen(ADBC_INGEST_OPTION_MODE_CREATE_APPEND); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 176, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(ADBC_INGEST_OPTION_MODE_CREATE_APPEND, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_INGEST_OPTION_MODE_CREATE_APPEND, __pyx_t_4) < (0)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":177
 * INGEST_OPTION_MODE_REPLACE = ADBC_INGEST_OPTION_MODE_REPLACE.decode("utf-8")
 * INGEST_OPTION_MODE_CREATE_APPEND = ADBC_INGEST_OPTION_MODE_CREATE_APPEND.decode("utf-8")
 * INGEST_OPTION_TARGET_TABLE = ADBC_INGEST_OPTION_TARGET_TABLE.decode("utf-8")             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = __Pyx_ssize_strlen(ADBC_INGEST_OPTION_TARGET_TABLE); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(ADBC_INGEST_OPTION_TARGET_TABLE, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_INGEST_OPTION_TARGET_TABLE, __pyx_t_4) < (0)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":281
 * 
 * 
 * def _test_error(status_code, message, vendor_code, sqlstate) -> Error:             # <<<<<<<<<<<<<<
 *     cdef CAdbcError error
 *     error.release = NULL
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_Error) < (0)) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_1_test_error, 0, __pyx_mstate_global->__pyx_n_u_test_error, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test_error, __pyx_t_9) < (0)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":317
 *         self._child_type = child_type
 * 
 *     def __enter__(self) -> "Self":             # <<<<<<<<<<<<<<
 *         return self
 * 
*/
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_kp_u_Self_2) < (0)) __PYX_ERR(0, 317, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_3__enter__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcHandle___enter, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle, __pyx_mstate_global->__pyx_n_u_enter, __pyx_t_4) < (0)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":320
 *         return self
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb) -> None:             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 320, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_5__exit__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcHandle___exit, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle, __pyx_mstate_global->__pyx_n_u_exit, __pyx_t_9) < (0)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcHandle___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__AdbcHandle, (type(self), 0x52640c0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__AdbcHandle__set_state(self, __pyx_state)
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_11_AdbcHandle_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcHandle___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib__AdbcHandle, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":342
 * 
 * 
 * def is_pycapsule(obj, bytes name) -> bool:             # <<<<<<<<<<<<<<
 *     """Check if an object is a PyCapsule of a specific type."""
 *     # Taken from nanoarrow
*/
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_3is_pycapsule, 0, __pyx_mstate_global->__pyx_n_u_is_pycapsule, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_is_pycapsule, __pyx_t_4) < (0)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":388
 *         return self.schema.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this schema without having to import it.
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_3release, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle_release, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle, __pyx_mstate_global->__pyx_n_u_release, __pyx_t_4) < (0)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":399
 *             self.schema.release = NULL
 * 
 *     def __arrow_c_schema__(self) -> object:             # <<<<<<<<<<<<<<
 *         """Consume this object to get a PyCapsule."""
 *         # Reference:
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_object) < (0)) __PYX_ERR(0, 399, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_5__arrow_c_schema__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle___arrow_c_sche, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowSchemaHandle, __pyx_mstate_global->__pyx_n_u_arrow_c_schema, __pyx_t_9) < (0)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle___reduce_cytho, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.schema cannot be converted to a Python object for pickling"
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_17ArrowSchemaHandle_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle___setstate_cyt, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":430
 *         return self.array.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this array without having to import it.
 * 
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_16ArrowArrayHandle_3release, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayHandle_release, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayHandle, __pyx_mstate_global->__pyx_n_u_release, __pyx_t_9) < (0)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_16ArrowArrayHandle_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayHandle___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.array cannot be converted to a Python object for pickling"
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_16ArrowArrayHandle_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayHandle___setstate_cyth, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":471
 *         return self.stream.release != NULL
 * 
 *     def release(self):             # <<<<<<<<<<<<<<
 *         """Release this stream without having to import it.
 * 
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_3release, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle_release, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_mstate_global->__pyx_n_u_release, __pyx_t_9) < (0)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":482
 *             self.stream.release = NULL
 * 
 *     def __arrow_c_schema__(self) -> object:             # <<<<<<<<<<<<<<
 *         """Get a PyCapsule without consuming this object."""
 *         cdef const char* err = NULL
*/
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_object) < (0)) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_5__arrow_c_schema__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle___arrow_c, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_mstate_global->__pyx_n_u_arrow_c_schema, __pyx_t_4) < (0)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":504
 *         return capsule
 * 
 *     def __arrow_c_stream__(self, requested_schema=None) -> object:             # <<<<<<<<<<<<<<
 *         """Consume this object to get a PyCapsule."""
 *         if requested_schema is not None:
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_object) < (0)) __PYX_ERR(0, 504, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_7__arrow_c_stream__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle___arrow_c_2, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_ArrowArrayStreamHandle, __pyx_mstate_global->__pyx_n_u_arrow_c_stream, __pyx_t_9) < (0)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle___reduce, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.stream cannot be converted to a Python object for pickling"
*/
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_22ArrowArrayStreamHandle_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle___setstat, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_9) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":520
 * 
 * 
 * class GetObjectsDepth(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     """
 *     How much data to fetch for adbc_get_objects.
*/
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_enum); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_IntEnum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyTuple_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_GetObjectsDepth, __pyx_mstate_global->__pyx_n_u_GetObjectsDepth, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_kp_u_How_much_data_to_fetch_for_adbc); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__pyx_t_4 != __pyx_t_9) {
    if (unlikely((PyDict_SetItemString(__pyx_t_8, "__orig_bases__", __pyx_t_9) < 0))) __PYX_ERR(0, 520, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":525
 *     """
 * 
 *     ALL = ADBC_OBJECT_DEPTH_ALL             # <<<<<<<<<<<<<<
 *     CATALOGS = ADBC_OBJECT_DEPTH_CATALOGS
 *     DB_SCHEMAS = ADBC_OBJECT_DEPTH_DB_SCHEMAS
*/
  __pyx_t_9 = __Pyx_PyLong_From_int(ADBC_OBJECT_DEPTH_ALL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_ALL, __pyx_t_9) < (0)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":526
 * 
 *     ALL = ADBC_OBJECT_DEPTH_ALL
 *     CATALOGS = ADBC_OBJECT_DEPTH_CATALOGS             # <<<<<<<<<<<<<<
 *     DB_SCHEMAS = ADBC_OBJECT_DEPTH_DB_SCHEMAS
 *     TABLES = ADBC_OBJECT_DEPTH_TABLES
*/
  __pyx_t_9 = __Pyx_PyLong_From_int(ADBC_OBJECT_DEPTH_CATALOGS); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_CATALOGS, __pyx_t_9) < (0)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":527
 *     ALL = ADBC_OBJECT_DEPTH_ALL
 *     CATALOGS = ADBC_OBJECT_DEPTH_CATALOGS
 *     DB_SCHEMAS = ADBC_OBJECT_DEPTH_DB_SCHEMAS             # <<<<<<<<<<<<<<
 *     TABLES = ADBC_OBJECT_DEPTH_TABLES
 *     COLUMNS = ADBC_OBJECT_DEPTH_COLUMNS
*/
  __pyx_t_9 = __Pyx_PyLong_From_int(ADBC_OBJECT_DEPTH_DB_SCHEMAS); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_DB_SCHEMAS, __pyx_t_9) < (0)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":528
 *     CATALOGS = ADBC_OBJECT_DEPTH_CATALOGS
 *     DB_SCHEMAS = ADBC_OBJECT_DEPTH_DB_SCHEMAS
 *     TABLES = ADBC_OBJECT_DEPTH_TABLES             # <<<<<<<<<<<<<<
 *     COLUMNS = ADBC_OBJECT_DEPTH_COLUMNS
 * 
*/
  __pyx_t_9 = __Pyx_PyLong_From_int(ADBC_OBJECT_DEPTH_TABLES); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_TABLES, __pyx_t_9) < (0)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":529
 *     DB_SCHEMAS = ADBC_OBJECT_DEPTH_DB_SCHEMAS
 *     TABLES = ADBC_OBJECT_DEPTH_TABLES
 *     COLUMNS = ADBC_OBJECT_DEPTH_COLUMNS             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_9 = __Pyx_PyLong_From_int(ADBC_OBJECT_DEPTH_COLUMNS); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_COLUMNS, __pyx_t_9) < (0)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "adbc_driver_manager/_lib.pyx":520
 * 
 * 
 * class GetObjectsDepth(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     """
 *     How much data to fetch for adbc_get_objects.
*/
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_GetObjectsDepth, __pyx_t_4, __pyx_t_8, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_9);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_GetObjectsDepth, __pyx_t_9) < (0)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":534
 * # Assume a driver won't return more than 128 MiB of option data at
 * # once.
 * _MAX_OPTION_SIZE = 2**27             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MAX_OPTION_SIZE, __pyx_mstate_global->__pyx_int_134217728) < (0)) __PYX_ERR(0, 534, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":653
 *         check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the database."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 653, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_3close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase_close, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":666
 *             check_error(status, &c_error)
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_encoding, __pyx_mstate_global->__pyx_kp_u_utf_8) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_errors, __pyx_mstate_global->__pyx_n_u_strict) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_kp_u_str_bytes) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_5get_option, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase_get_option, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_8, __pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, __pyx_mstate_global->__pyx_n_u_get_option, __pyx_t_8) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":719
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 719, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_bytes_2) < (0)) __PYX_ERR(0, 719, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_7get_option_bytes, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase_get_option_bytes, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, __pyx_mstate_global->__pyx_n_u_get_option_bytes, __pyx_t_4) < (0)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":750
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 750, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_float) < (0)) __PYX_ERR(0, 750, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_9get_option_float, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase_get_option_float, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, __pyx_mstate_global->__pyx_n_u_get_option_float, __pyx_t_8) < (0)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":763
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 763, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_int) < (0)) __PYX_ERR(0, 763, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_11get_option_int, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase_get_option_int, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, __pyx_mstate_global->__pyx_n_u_get_option_int, __pyx_t_4) < (0)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":776
 *         return c_value
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Set arbitrary key-value options.
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_13set_options, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase_set_options, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcDatabase, __pyx_mstate_global->__pyx_n_u_set_options, __pyx_t_8) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_8) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.database cannot be converted to a Python object for pickling"
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_12AdbcDatabase_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcDatabase___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_8) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":938
 *         database._open_child()
 * 
 *     def cancel(self) -> None:             # <<<<<<<<<<<<<<
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 938, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_3cancel, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_cancel, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_cancel, __pyx_t_4) < (0)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":946
 *         check_error(status, &c_error)
 * 
 *     def commit(self) -> None:             # <<<<<<<<<<<<<<
 *         """Commit the current transaction."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 946, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_5commit, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_commit, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_commit, __pyx_t_8) < (0)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":954
 *         check_error(status, &c_error)
 * 
 *     def get_info(self, info_codes=None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get metadata about the database/driver.
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 954, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_7get_info, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_info, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_info, __pyx_t_4) < (0)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":989
 *         return stream
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,             # <<<<<<<<<<<<<<
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:
 *         """
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 989, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_9get_objects, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_objects, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_8, __pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_objects, __pyx_t_8) < (0)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1034
 *         return stream
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_encoding, __pyx_mstate_global->__pyx_kp_u_utf_8) < (0)) __PYX_ERR(0, 1034, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_errors, __pyx_mstate_global->__pyx_n_u_strict) < (0)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_kp_u_str_bytes) < (0)) __PYX_ERR(0, 1034, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_11get_option, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_option, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_2, __pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_option, __pyx_t_2) < (0)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1087
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1087, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_bytes_2) < (0)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_13get_option_bytes, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_option_bytes, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_option_bytes, __pyx_t_4) < (0)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1118
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1118, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_float) < (0)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_15get_option_float, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_option_float, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_option_float, __pyx_t_2) < (0)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1131
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1131, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_int) < (0)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_17get_option_int, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_option_int, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_option_int, __pyx_t_4) < (0)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1144
 *         return c_value
 * 
 *     def get_statistics(             # <<<<<<<<<<<<<<
 *         self,
 *         catalog=None,
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_19get_statistics, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_statistics, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_statistics, __pyx_t_2) < (0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1210
 *         return stream
 * 
 *     def get_statistic_names(self) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get custom statistic names defined by this driver.
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_21get_statistic_names, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_statistic_nam, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_statistic_names, __pyx_t_4) < (0)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1237
 *         return stream
 * 
 *     def get_table_schema(self, catalog, db_schema, table_name) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the Arrow schema of a table.
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle) < (0)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_23get_table_schema, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_table_schema, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_table_schema, __pyx_t_2) < (0)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1273
 *         return handle
 * 
 *     def get_table_types(self) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the list of supported table types.
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_25get_table_types, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_get_table_types, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_get_table_types, __pyx_t_4) < (0)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1287
 *         return stream
 * 
 *     def read_partition(self, bytes partition not None) -> ArrowArrayStreamHandle:             # <<<<<<<<<<<<<<
 *         """Fetch a single partition from execute_partitions."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowArrayStreamHandle) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_27read_partition, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_read_partition, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_read_partition, __pyx_t_2) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1305
 *         return stream
 * 
 *     def rollback(self) -> None:             # <<<<<<<<<<<<<<
 *         """Rollback the current transaction."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_29rollback, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_rollback, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_rollback, __pyx_t_4) < (0)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1313
 *         check_error(status, &c_error)
 * 
 *     def set_autocommit(self, bint enabled) -> None:             # <<<<<<<<<<<<<<
 *         """Toggle whether autocommit is enabled."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_31set_autocommit, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_set_autocommit, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_set_autocommit, __pyx_t_2) < (0)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1329
 *         check_error(status, &c_error)
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Set arbitrary key-value options.
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_33set_options, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_set_options, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_set_options, __pyx_t_4) < (0)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1396
 *             check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1396, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_35close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection_close, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcConnection, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 1396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_37__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.connection cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_14AdbcConnection_39__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcConnection___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1442
 *         connection._open_child()
 * 
 *     def bind(self, data, schema=None) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Bind an ArrowArray to this statement.
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1442, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_3bind, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_bind, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_bind, __pyx_t_4) < (0)) __PYX_ERR(0, 1442, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1494
 *         check_error(status, &c_error)
 * 
 *     def bind_stream(self, stream) -> None:             # <<<<<<<<<<<<<<
 *         """
 *         Bind an ArrowArrayStream to this statement.
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_5bind_stream, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_bind_stream, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_bind_stream, __pyx_t_2) < (0)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1530
 *         check_error(status, &c_error)
 * 
 *     def cancel(self) -> None:             # <<<<<<<<<<<<<<
 *         """Attempt to cancel any ongoing operations on the connection."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_7cancel, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_cancel, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_cancel, __pyx_t_4) < (0)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1538
 *         check_error(status, &c_error)
 * 
 *     def close(self) -> None:             # <<<<<<<<<<<<<<
 *         """Release the handle to the statement."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_9close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_close, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1551
 *             check_error(status, &c_error)
 * 
 *     def execute_query(self) -> Tuple[ArrowArrayStreamHandle, int]:             # <<<<<<<<<<<<<<
 *         """
 *         Execute the query and get the result set.
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_kp_u_Tuple_ArrowArrayStreamHandle_int) < (0)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_11execute_query, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_execute_query, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[52])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_execute_query, __pyx_t_4) < (0)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1574
 *         return (stream, rows_affected)
 * 
 *     def execute_partitions(self) -> \             # <<<<<<<<<<<<<<
 *             Tuple[List[bytes], Optional[ArrowSchemaHandle], int]:
 *         """
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_kp_u_Tuple_List_bytes_Optional_ArrowS) < (0)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_13execute_partitions, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_execute_partitions, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_execute_partitions, __pyx_t_2) < (0)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1617
 *         return partitions, schema, rows_affected
 * 
 *     def execute_schema(self) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """
 *         Get the schema of the result set without executing the query.
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle) < (0)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_15execute_schema, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_execute_schema, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[54])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_execute_schema, __pyx_t_4) < (0)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1636
 *         return schema
 * 
 *     def execute_update(self) -> int:             # <<<<<<<<<<<<<<
 *         """
 *         Execute the query without a result set.
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_int) < (0)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_17execute_update, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_execute_update, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[55])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_execute_update, __pyx_t_2) < (0)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "adbc_driver_manager/_lib.pyx":1656
 *         return rows_affected
 * 
 *     def get_option(             # <<<<<<<<<<<<<<
 *         self,
 *         key: str | bytes,
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_encoding, __pyx_mstate_global->__pyx_kp_u_utf_8) < (0)) __PYX_ERR(0, 1656, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_errors, __pyx_mstate_global->__pyx_n_u_strict) < (0)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_kp_u_str_bytes) < (0)) __PYX_ERR(0, 1656, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_19get_option, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_get_option, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[56])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_8, __pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_get_option, __pyx_t_8) < (0)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1709
 *         return buf[:c_len].decode(encoding, errors)
 * 
 *     def get_option_bytes(self, key: str) -> bytes:             # <<<<<<<<<<<<<<
 *         """Get the value of a binary option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1709, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_bytes_2) < (0)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_21get_option_bytes, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_get_option_bytes, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[57])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_get_option_bytes, __pyx_t_4) < (0)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1740
 *         return bytes(buf[:c_len])
 * 
 *     def get_option_float(self, key: str) -> float:             # <<<<<<<<<<<<<<
 *         """Get the value of a floating-point option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1740, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_float) < (0)) __PYX_ERR(0, 1740, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_23get_option_float, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_get_option_float, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[58])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_get_option_float, __pyx_t_8) < (0)) __PYX_ERR(0, 1740, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1753
 *         return c_value
 * 
 *     def get_option_int(self, key: str) -> int:             # <<<<<<<<<<<<<<
 *         """Get the value of an integer option."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_key, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1753, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_int) < (0)) __PYX_ERR(0, 1753, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_25get_option_int, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_get_option_int, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[59])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_get_option_int, __pyx_t_4) < (0)) __PYX_ERR(0, 1753, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1766
 *         return c_value
 * 
 *     def get_parameter_schema(self) -> ArrowSchemaHandle:             # <<<<<<<<<<<<<<
 *         """Get the Arrow schema for bound parameters.
 * 
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_ArrowSchemaHandle) < (0)) __PYX_ERR(0, 1766, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_27get_parameter_schema, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_get_parameter_sche, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[60])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_get_parameter_schema, __pyx_t_8) < (0)) __PYX_ERR(0, 1766, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1798
 *         return handle
 * 
 *     def prepare(self) -> None:             # <<<<<<<<<<<<<<
 *         """Turn this statement into a prepared statement."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1798, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_29prepare, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_prepare, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[61])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_prepare_2, __pyx_t_4) < (0)) __PYX_ERR(0, 1798, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1805
 *         check_error(status, &c_error)
 * 
 *     def set_options(self, **kwargs) -> None:             # <<<<<<<<<<<<<<
 *         """Set arbitrary key-value options for this statement only.
 * 
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1805, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_31set_options, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_set_options, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[62])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_set_options, __pyx_t_8) < (0)) __PYX_ERR(0, 1805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1871
 *             check_error(status, &c_error)
 * 
 *     def set_sql_query(self, str query not None) -> None:             # <<<<<<<<<<<<<<
 *         """Set a SQL query to be executed."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_33set_sql_query, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_set_sql_query, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[63])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_set_sql_query, __pyx_t_4) < (0)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1881
 *         check_error(status, &c_error)
 * 
 *     def set_substrait_plan(self, bytes plan not None) -> None:             # <<<<<<<<<<<<<<
 *         """Set a Substrait plan to be executed."""
 *         cdef CAdbcError c_error = empty_error()
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < (0)) __PYX_ERR(0, 1881, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_35set_substrait_plan, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement_set_substrait_plan, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[64])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_19adbc_driver_manager_4_lib_AdbcStatement, __pyx_mstate_global->__pyx_n_u_set_substrait_plan, __pyx_t_8) < (0)) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_37__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[65])); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_8) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.statement cannot be converted to a Python object for pickling"
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_13AdbcStatement_39__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AdbcStatement___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[66])); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_8) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1905
 * 
 * 
 * @functools.lru_cache             # <<<<<<<<<<<<<<
 * def _init_blocking_call():
 *     error = bytes(CInitBlockingCallback()).decode("utf-8")
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_functools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_lru_cache); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_5_init_blocking_call, 0, __pyx_mstate_global->__pyx_n_u_init_blocking_call, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[67])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __pyx_t_10 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_9, __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1905, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_init_blocking_call, __pyx_t_8) < (0)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1915
 * 
 * 
 * _blocking_lock = threading.Lock()             # <<<<<<<<<<<<<<
 * _blocking_exc = None
 * 
*/
  __pyx_t_9 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Lock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
    __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_10, (1-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_lock, __pyx_t_8) < (0)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1916
 * 
 * _blocking_lock = threading.Lock()
 * _blocking_exc = None             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_exc, Py_None) < (0)) __PYX_ERR(0, 1916, __pyx_L1_error)

  /* "adbc_driver_manager/_lib.pyx":1919
 * 
 * 
 * def _blocking_call_impl(func, args, kwargs, cancel):             # <<<<<<<<<<<<<<
 *     """
 *     Run functions that are expected to block with a native SIGINT handler.
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_7_blocking_call_impl, 0, __pyx_mstate_global->__pyx_n_u_blocking_call_impl, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[68])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_8);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_call_impl, __pyx_t_8) < (0)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "adbc_driver_manager/_lib.pyx":1972
 * 
 * 
 * if os.name != "nt":             # <<<<<<<<<<<<<<
 *     # https://github.com/apache/arrow-adbc/issues/1522
 *     _blocking_call = _blocking_call_impl
*/
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_nt, Py_NE)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "adbc_driver_manager/_lib.pyx":1974
 * if os.name != "nt":
 *     # https://github.com/apache/arrow-adbc/issues/1522
 *     _blocking_call = _blocking_call_impl             # <<<<<<<<<<<<<<
 * else:
 *     def _blocking_call(func, args, kwargs, cancel):
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_blocking_call_impl); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_call, __pyx_t_4) < (0)) __PYX_ERR(0, 1974, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "adbc_driver_manager/_lib.pyx":1972
 * 
 * 
 * if os.name != "nt":             # <<<<<<<<<<<<<<
 *     # https://github.com/apache/arrow-adbc/issues/1522
 *     _blocking_call = _blocking_call_impl
*/
    goto __pyx_L3;
  }

  /* "adbc_driver_manager/_lib.pyx":1976
 *     _blocking_call = _blocking_call_impl
 * else:
 *     def _blocking_call(func, args, kwargs, cancel):             # <<<<<<<<<<<<<<
 *         return func(*args, **kwargs)
 * 
*/
  /*else*/ {
    __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_9_blocking_call, 0, __pyx_mstate_global->__pyx_n_u_blocking_call, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[69])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
    PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
    #endif
    if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_blocking_call, __pyx_t_4) < (0)) __PYX_ERR(0, 1976, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__AdbcHandle(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x52640c0, 0x461d5d7, 0xad6f2dd, b'_child_type, _lock, _open_children')
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_19adbc_driver_manager_4_lib_11__pyx_unpickle__AdbcHandle, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__AdbcHandle, NULL, __pyx_mstate_global->__pyx_n_u_adbc_driver_manager__lib, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[70])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__AdbcHandle, __pyx_t_4) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "adbc_driver_manager/_lib.pyx":1
 * # Licensed to the Apache Software Foundation (ASF) under one             # <<<<<<<<<<<<<<
 * # or more contributor license agreements.  See the NOTICE file
 * # distributed with this work for additional information
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_4) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init adbc_driver_manager._lib", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init adbc_driver_manager._lib");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_super); if (!__pyx_builtin_super) __PYX_ERR(0, 110, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "adbc_driver_manager/_lib.pyx":324
 * 
 *     cdef _open_child(self):
 *         with self._lock:             # <<<<<<<<<<<<<<
 *             self._open_children += 1
 * 
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "adbc_driver_manager/_lib.pyx":504
 *         return capsule
 * 
 *     def __arrow_c_stream__(self, requested_schema=None) -> object:             # <<<<<<<<<<<<<<
 *         """Consume this object to get a PyCapsule."""
 *         if requested_schema is not None:
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "adbc_driver_manager/_lib.pyx":989
 *         return stream
 * 
 *     def get_objects(self, depth, catalog=None, db_schema=None, table_name=None,             # <<<<<<<<<<<<<<
 *                     table_types=None, column_name=None) -> ArrowArrayStreamHandle:
 *         """
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(5, Py_None, Py_None, Py_None, Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);

  /* "adbc_driver_manager/_lib.pyx":1144
 *         return c_value
 * 
 *     def get_statistics(             # <<<<<<<<<<<<<<
 *         self,
 *         catalog=None,
*/
  __pyx_mstate_global->__pyx_tuple[3] = PyTuple_Pack(4, Py_None, Py_None, Py_None, Py_True); if (unlikely(!__pyx_mstate_global->__pyx_tuple[3])) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<4; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 9; } index[] = {{2},{53},{52},{39},{104},{13},{22},{43},{61},{33},{39},{31},{41},{30},{23},{48},{34},{30},{54},{4},{179},{395},{47},{12},{6},{34},{52},{26},{17},{15},{1},{2},{1},{1},{28},{8},{6},{16},{122},{61},{7},{6},{16},{17},{11},{2},{9},{18},{21},{15},{11},{10},{12},{58},{62},{67},{65},{63},{66},{63},{11},{10},{15},{5},{15},{11},{23},{3},{14},{14},{32},{34},{21},{20},{21},{23},{26},{25},{31},{31},{29},{34},{29},{31},{30},{29},{23},{29},{26},{12},{30},{32},{18},{23},{29},{29},{27},{24},{11},{21},{20},{29},{31},{12},{13},{31},{33},{18},{25},{20},{19},{32},{27},{28},{28},{24},{30},{30},{28},{34},{21},{25},{27},{32},{14},{16},{34},{36},{24},{22},{41},{41},{40},{42},{30},{17},{36},{35},{37},{25},{9},{8},{7},{10},{20},{11},{14},{9},{13},{4},{5},{14},{15},{18},{25},{25},{32},{26},{26},{9},{8},{16},{12},{13},{2},{7},{14},{14},{13},{4},{4},{16},{9},{15},{4},{17},{26},{2},{16},{8},{4},{16},{20},{4},{6},{7},{13},{5},{15},{12},{7},{20},{11},{14},{7},{19},{24},{36},{35},{9},{11},{4},{17},{18},{18},{5},{18},{11},{4},{11},{14},{19},{13},{14},{4},{3},{5},{13},{7},{5},{9},{13},{11},{7},{8},{7},{12},{5},{5},{5},{12},{6},{7},{8},{8},{12},{7},{11},{6},{7},{7},{10},{9},{18},{5},{11},{6},{10},{30},{14},{4},{8},{9},{6},{5},{7},{8},{5},{7},{1},{7},{6},{8},{9},{4},{3},{5},{6},{3},{8},{6},{8},{7},{18},{13},{14},{14},{8},{5},{4},{8},{9},{7},{8},{11},{10},{16},{16},{14},{20},{19},{14},{16},{15},{12},{6},{1},{9},{10},{8},{19},{9},{3},{13},{12},{8},{5},{4},{3},{9},{6},{6},{10},{9},{8},{11},{7},{13},{10},{15},{8},{4},{7},{7},{4},{2},{14},{3},{6},{2},{9},{17},{10},{4},{7},{4},{3},{6},{11},{7},{12},{12},{14},{12},{11},{10},{26},{14},{12},{5},{10},{2},{14},{10},{17},{13},{7},{16},{6},{8},{13},{1},{6},{4},{4},{14},{12},{11},{13},{18},{10},{12},{19},{8},{5},{6},{11},{3},{6},{6},{5},{3},{10},{11},{8},{11},{9},{5},{6},{6},{12},{5},{6},{11},{4},{8},{14},{211},{5},{2},{165},{10},{94},{63},{12},{44},{230},{66},{41},{41},{192},{74},{195},{223},{223},{223},{60},{113},{166},{13},{153},{40},{192},{192},{192},{79},{79},{79},{9},{44},{44},{44},{68},{23},{66},{89},{281},{79},{79},{79},{171},{68},{66},{44},{115},{11},{67},{66},{55},{386},{386},{382},{80},{74},{73},{82},{250},{11},{18},{12}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (4900 bytes) */
const char* const cstring = "(\265/\375`\3454\325\230\000Z\224\204\037H\360\316\272\001\314\300\304\210\233\006\254Ll_\010!\205\244\002\226\ts\323BB\203K\024\355\214\332\264\036z\330()I\315\027\212\010,G(^B\321\355u\033\221A\360AA\361\271_\365\371\002\236H\022{\310C}\342%(\277\001\313\001\377\001\270\224\256\353*|\366\233\034o4\307\345\224^\032\271M\337\222<\321\355\033\261s]\273\373\0343N\337#\247u\t\022\264'w\356<_S\351\244\357\231\276$\022\213b\326\321\341R\260\004s\3272\231o\3304E\r)\267T\272/ H\232\234r\244\321\330\304\\,\033\223\214D\033w\214t_RH3\347\230\270(\332\330\364\355u3\022\247'\341;\037\273\337b(\003e1\326Ll\314u\334\245iN\343z\343:\nh2K\221\3348(\016)\220\327\270Y\312P1\360\276`\224\276\3759\363[\353\354\353\311\314G\227\025\225\273\373\270q\270\311\305K\302$r\223\367,u\227\266\272)}\344\347\254\347Q\317\356J\313\n\312+|]RsS7\356\031\037\267H\2266\257Cm_\2666Y\266\272\327\373\304\254\304X\317\304L\363\233f\323\347-\306\357n\332\256\254\265R\255bw\233\245\357\333\274/\270\323\\\211f.g\272\375\341n\024\334\363\223\373\342wm\344\271\203\3164\323\254]q\367\276\365X\232\276\237\327\334\"y\232R\272\271=\311Qv\317\263\322\337<\366P\034\317\022uw/G=\226\272\364\245\371:;.\243t\257\333=\271o{\304\335\370\3659\373,\316p\267R\243\364\236q\227\334\266\234\233o\333\246\373\0259Ni\342\310\214\022\367\355\354\244N\034\363\335\243H\346\230[\336\246\217\351r\016\230\"no\2268\334\331\257/\311\333\371\325y>fM\236\264\273oe\277fin\\y\311\024\2634\315}i\345\3358\235\277\333\346o\215\344>k.\222\374n\261\324\267\237>\373~\301\273q_\002\342\366}\233\033ks\3148\233<\3756\373\215\306\230\3430\306\225\270\216{\351I\237\26377KS,k\271\223\034\237f\026m\276\226sC\362\362D\347\"\327\341p\347\230\343\032\306\335\317|\337\033c]\212\342\026m\036{r\233\346\251M\234\316&3\364\021&\305\364\030c\254\265\326\232\334\343\276\301\332w=\316\316rF\336\316\261\254\225\032k\3329\033\313r\246u\266\306\343\014\367Mr\177\226f\3115\251\353\231s\3353\306\275\346\363\215\353\340r\326\3458\372\216\3036mf\3377\215\347\224\032\027""\210V\215\253\365\001\346\002Q12\032\260E\363\001\205\303\227\010\231\214\306\025\312\254h\033\346uv\013{\315S:\272\027\30652\2335\233\317/\327\007W,\025\206\254\227\314*\325\242qu\3002W+\034\261\212\205\000e\360\215.Yc&W+O\256\\\253\224\254\025\312\320\340\024\210\326\212\005\266`\"d+\230*\224\2618\310B\025\014\224\241@\326+\204\2050T\252\025\276^ \016S.\031,{\271\260,\026\252Z+\233\307\355\224\313o\025\311o'\031\2062`\253\345Z\305p\215\013\006\0031,\004\261\202\245b\240L\030\203\212a\016\251XK\206\206\345B\305T.\026(\003Q\251\220\305\nQ-\020\245q\336=\357q\233d\036y\236\235$\027q\247\316\334g\256o\177\352\255\263\261H\362\331\312s\035\214w\252\227\376\314\375F\347\331\310\275\231q\030\373\276o\231\304\232\373\214K\032\311=\277\215[c\254\361\326\247\270M\214SP1Yl\025\"\304\"R`\010c\201,\026\270\315\362\226C\332<\317\271\244\026\024\232?K\265l\236\350p\3326\003-\251\212\311'\371\376\351\007\370\261P\010\017\344\325\324\250\265TA\213\322\3709\224\247{U\322Q\240\305)m\371*R\234\017\223\032\010A-\364\362\377\310\340\345V\226\022\366\303#\241&H3\200\353\210-qJU\240\031\005AC\0138\273\002\274\013<\227\023\366\344\366\200-rj\346\315\360pV\026\022{\002\246\r>\353\204\017\257\205\304:\260\320\263\241\361a\2405\340\321\003[\214\322\024\037\271\257\203\n\3509_z\003\246\240\374mh\306\227\352/\002\305\331\255\020\350\001[|\253\210*2ZU>i\354+X-8`\035|<\227\336\227C\017\240=\360'Xe\322SZ\312?\007\205\351\2510\351(\035\371#\277\206^\030\215>\002\321\227j\200\277\006}\371\322Q\250\247\221\346>G\017\210\332\321gA[v\003B\013B\275o\265\240\341\223\334v\304\343\301\205u)\261%\020\013\242\037\2426\247\004\276\210\250\345\020\265\237\257\333\301v\034;x\255`\302\232\200\330\021\332af\255\000b\333\221\264\004\372\201~\035( ,\210\250%\300a\002\2525\014V\n\371_@U;\221\356\257\374\321O\261\313;\002+\270\256\224\260\027P\003\2426\367QP\217P\013~\022\364\013\265\0375\227\036HG\242V#\324\313\351L\352\300\203\365HU?\301\000v@\250w\342\225V\025\022\020\233\333\025\010\201?\006=\241\006B3\247\017b&""\305\3702\024\344\243\315j\312\010;\322\321z|=pu\351\235\324\221\272\221\336\217\366So\376\314\177A\rt\2643\237\003\355\020\265\247\235H\343\"\243\347\264\203BaQB\275\0315Bj\242q\021\3529\251\245\224\001\210=\251\027\267\322\354\240{]\\\221\2278\327\004\341^\224 \334\363\232N \\\335\253\367\335O\216\273\256\233\033w\366~\366L\333\242H\356\336\365\246\3212\356\244y\372m\006\371\266m\272\353\362Iu}\006\377\373\372\207\246\003\354\327R\320\220z\3242fv\340N\204\321V\350\231Hi\377\005P]\006\306\005\201\2541\332\371vdV\320\232SQ*\031\301d\242\266\276\213\210\205Y\275\232)\250&\376K\272\342\207\320\216\334\232\327\034\024\252\004V\200J\302%\004\301\211\035\261\021\031]\024\026TAj\357#`c\007\234\300\352_\322\361\321\361\265#\2155D-1\"\242\200U\360\321\036\000A\360\245aF\372\340\217\240\341\201\036\020\313\323\342\221jHE.\315\251\355\203Z\330\014{\263s\2017`\235p\331\317\324\323;\201\330\317\372\340\324\033\361\260\354\014\340\205\244TS\001\266\300\307C\332a\341\371\025\265\236\252\000\025\303\367\275\217\037\2425;\234\025\265%\365\277\204\000\306gT\033|\324\223\022P5}<\322\252\262\263\301\304\232HX\020: QF\307H=\311\351\203\225\324\204A.E\031aM\024\334\211\271\020\302\302\314\360\200< \236\030\325\351\227\312\265\202\032A\204\235\205p\264\237\235\007<\007l\231\177Fo\001\273\362U(\010Gz\"'\177\017\010\330\016\252\035\3075aMx/\026l\002\250\017R\032\021\306\205\307]\241B]QK\177$\363\024\254\262\223\032\001\215\260\333\000\265%Rj\363QC4\004\244$\332\321\007\000\242\347)\377\022P\215\311\3614`\240\035$\265\247$\017\366k\361\255'\025?\203x\364\214v8\204@UF;4\035=\030\2231\243\024P/\252\270\301\207\352\035\205uD\225\235\322\0012\022\004\324\303\223,@`\032\363O\021#\224f\207\261\013\201\332\023\013\316\265\020\240\001\352\201(\311N\003\324\373\354l\270`u\320\017\254\323H\375\324\014\337\365\375\336\206\306\305F\017\310;\361T\252\004#\036\313:ZYTUJ\002\252\020H\277\003\312\000\355\250\307\247#\300\232S%\210\210\230>\202\257\362\003\205@\250\243h\212&\265\214\251\221\221$Ia9\203p0p""\240P0\236P\352\363n\263t\032E\222dI\010\"\014\022c0\021\022\031\tD\002I\"\205\245I:5\340~@?i\177\246\307Oh\213g\322%\021\354:\376\335\207Z\345\213\224\313\036\306\251\031\3004\321$.\225L\201\230\244\322\177\262\261v\356\227&XUX<\005\234\275\274\2256J\316\225_\233\037\274XI\270\275I]u{H\033Q\034\334\231?9\024\342R\251\236\006\215\240\266~#\234ndL\007a\310\240wcL-\026\014\226\202\324\016\270\013\362N3fHm\177&\256^?Lk\rd<\36125\312:\373 \312V\244\020\331- \365\005\235\364\\\311E]\017 \310\275\345A\216L,G\231\355\374\250\021@\234\225\n\031\350\"\357G/V\177\0369\201d\305&E\2001m\026\365U6Y9\0174\271B+\241\232\233\315=\234\236\342\326\374\002\213W|Z32\024\014YH\007%a\177\271R\010'\013\351\004\321E\0145\357\235\302\227\222\344\251\237\253\316\302\320\202\324\230\267\00311\026ohvF\213P\370\001)\334x\216\306K\212\303M9\254ZF\036\307e*/\355\321\247\361t\331\245\\\271|O\005\021(\272\031\001.\265\216Dc~\324`\200\214\227\340\203Z\266\001\227,R\274\005\354\251bz\\\006\267\301Ht'\322\027(JP\327p\002sFZS\341\304\222H\022\234\377\033L\247\t\247D\231QA\300\323\366vd<\000\343\332E\365+\007N\2275\007\340\330\370\254&a\211EP0\002\022\310s\312\242\032\234\215\031\013s\326.\211s1b \272\333\255\0212\315Y\301\353\304\344\312nUAI\326\006K;\010\r\371r\201$\304 f\310*a\306a\363,=!\252\224/\230r\207\265\360y\247\263z\215\261\007.\352Yp\373\025\001h\253\002\334_h\325\235\304R\3674\351B\332\034\211\r\360\202\177\320\363\351\211\351F\270G\021\\\000|\020\023\315\210u\312\374\201\246d1c\242\232\343\330\003\337.\210\274j,\262\213o\301\361\372\266{\322E\214\231?\031(t)r'\255:\232\220\016^\250\324!\247\214\361\017\003\354\025\374\303\254\235.\336b\004\343\271x\\c\337\030t\207\254\344s\235\351S\212\013+\272 \233\257?\210\254\273\314\016Y\177\307T\n)\335&Z\345z\0355X<\317\215\322p\301y\303sq\006)\272\375\002O5\234\237\372\023jR\340\204\247=\000\305VH\330\007\276\035z\031\231\221\020!\030VB\312\241\255ix\233u\201$W\2009\322;%\234n\257\262\355>\301\020\330\200\260J\257.\217-5\001\337@\243\031|\276\324}""\250M\250\213\210\256fW>]LE\320d\25237\364[\202u\210}4\215&\014\264\374\245\234\020\365\356\355m\nP\264{\251\205\256\260\245\0351\373\r\256\301\361\341\321]5\034;\004A\006Oh\335\350j>\225o\010\272Qi\357\303\370\036\232\263\203W\255\032\233[\237\373$\313w\234\377\n\260\000r\243\242x-v\237\242\027\277\2316~G\251O6):%d\027W\307j\2618\314V\3141\301\337_.S\321\226\"\022\253P}\243\355\261\237\250e\232\027[\nVL\323\n\276F\235\277\275\243G\206l)\341zUb\251o\370\351\034\356\254\224\224\267\013\330\217V\213\005\312\022J\277d\326\024X\352\322\305\002\225HLmj\322:\244 \3726\245\310#f4#/i\020|\237w0\030?\221f\206\212\342\266\034\343_,\014\312\227\257\323;\352\211$\300r\323\236\211\307\241\210!\347\375\016\263\270\201x\225\225\236\226\300\217_<\203.\002\253h\374OOeU\210\371\227P\320%;q\254\334\351\004\306\223\364G\326.\231-\363f\230!>\250*\313\215\n\251\262\352\2372\246\336!H\021\336\031\232\2036\222\226\0138a\337|4\023\304\210&S)ch\006\343\234\231:Y\367(\210)\032\243\231\311\267\021\210\312t\224\235\2005\255\276[\237@\016`\232\340\315^\232\222Yn\313\271\335~^\232\225p>@J\004@\025\325\221T\201\3735\252\277!a\364\344\350\357j\260p\016\272\330\241\005\030\267\253\371\3207\244\003\216_@\223\360y\247K\330\263 \303\310\332Z\236'\370\207\241\3362\245\246\364W(#l\004\330\263\306{\036~\220\034\254\366qx\032\230a\317j_&\270\276\016c\n\010J\326\273\354\221&\2372\020\302\350_lS\2576!\010\275\214uvBg\255^\206\322bJ\3420\343\266=.\207c1\027D(\2522\275\312S\001l'\225\025\326!_\302\2531\274\375\314\333\220Ed\224\253^\322%m\321\020\232\016{sN\241{\304\337\276\225FF\030S\201\\\232E\327,\035\025T\002h\177\371\252R`\313\322;\317d\276$$\365\366j\316\232\331\035\370\244X\n\302\023\024aE\265L\270*%V\330\320\334\210V\313x\217\343\372\372Eg_\221\020\372\346\212\237/S?X7pu\226n\364\002\204pE\217\317\t)>b\247&\343\231\240\364\326{\3265IU\360\325\2619\203\357j\230\032\316\224\350\245Y\221\\\251\216h\213#5y\202\315r\346\017*\2545\322\343JB\343\263\254d\320\3345\2411\037\025\362\235[\334 \254\376 \3463\010Q~""\350\223\223\305\201f\320C[\362\316\033\2432\234\337%m^\2134\033B\340\247W\332A;>O\376M8$\341kh\326Y\360\344C\0360\213\356\0378\355)U\211Dv$\r\351\016\277\346\n\244\354~p\315\357\275\t\2256\255\034\352\247*\246\027\t`\317\241\215&32\"Pl\031\252\341-\014H\232\306\345\201F.\025\306\200\035\241\013\217\340Bn\336\360l\274\214\316\241\345Op{TX\000\226\370'\274\263\233\t\216Tr\244\006\207\343\276_\rK$\021\376\2406\333\\\325\n\352xL\032@\321\r\264\224\nYK'\032\233\365\216+\031RG%\305\322\013\253\227\321f\331X\271s\250\010\205\364\231i\3626\013\023PU\310)b\307]\266\035\020\231S\324\341c\370P\215\005\031\207<\361\220k\270\221\177G\231\307\334[\027)\031\246\007\010\002\027\354'\260\203\034\234\26379\236\233\021B\322\233\321\033\220I\201Y)f\031g\3517U\370\233\257\377\275F\232\261\346\266\375}x\340\262\034<o\360\367AJ\316\237h\026\360\262$\311\251%8D\310*)\210p\363[\315bO\177u\370>\000\226\302\234\0325\314\227\2768$\242\022T8\246\2203\327\222.\362\311\337TX?ZG\014\025X\256D\360\277=\2625\357\274\021JI\305\037\213\372\346\177~\353X\206c\266H\204x\312B&\205\033\337\207\000l\3373\371A\237\021\371V\026<\263\300Z{\261\026P\033J\223\246\245K\365\212C\361\327\355\250d\363S\307\266DH\214\243E\302\316C5\351g5\254TwU\304Q\253\356m\344w\373\314\006s[\376\263\035\376\375\t\363\205\320\312+\215\374\243&\201\212\022S\314m\314\310\313X\264\"-\220\227J\006\201iok\361A\237\237\341\031\300R\330`\025\246\017\004n\360\325)\007\r\021\2523\364\315\365\217|\236\353gw\361\250|\364\212\360h$#%\350<$:\003\233\001\224\035\r\355\005\243\3214m\234\304\005\037A\241>X\222\256\375?\363\014=\367\221\246\212\337\354+\2022\352us\005|\321F\2500(\271!\\\234k\332\361l\250\342G\353\341\371\220\320\313'\023\032\202\354Y\332 \217\346:\032D\327'\314\336\200\260U\270\201V\007\\\032~\256\226\247\177\311\226\2320\227\330\310\3677\323|?\001\207\255\223\036\320\313\245\277\207Y\214J\251\330<\236I\206\273\232\324\366Z\213\326\241x\260\37117UU\242:\354=F\367?\177\333b\177xy\353\215\301=@\277\300\362\356+\026\035k\214\366\237x""\253\037]\241\217\357`\215_m\022U\374:\240\275\340\321 \001\373(:\3445\356p\361-\254\345\000>\031\246\227@\204\204\200F\366\004P@z\005`{Y\007I\251\211;\374\3225\37344\0148\016\306;\331\242\207\242]\017\240\333\232\254Gq5\253\003\000\\C\372\005\\\202u\006\272\375\360\021\273]X\333\310Ou\343\225`\245:\206\006\330\335\\\323\023wQ\377)\t\364\264\3140\020\010\001U\354\245c+t\300\003Z\237\204\322\n\001\305\2610\246E\260l\205\260\262\02639XT>\3371@>>\201\227\213\311\000E\t@\357K\303\262\350\302\244#\020\273V\035P\304B\232s\0039E\253C\016\210D\273.\323\340\002-I\312\333\243\275\237\231\\$\375\224\306\324\007\367\223!\201\221\021\352/v\305I\321\013u\004\342-!W\245\213\333\350Hy4V\002\236\235\350-OF\376{\264\240:s\350\325J\323\265\r\007\364\356\230\324X\032\t\224\265\313X\345h\026P\333b\237\362\203\244\251AX\n\3265\254\302\312\271A\002\304!\226\210\271\227\253\220\026\333j\330\035\367\202]%\023\t\017 \235\371s\rP=\220\360OE\020f\0253\321}~U\3430\003\230\337ZfW3\266\350=|\366\235-j\325{BC\212\350\246\261XD\3063]1\037A\332)\305b/\205c\207I\336<\202Sc#R\320\321\"\365ue\243\342}\270K\326\333~\315\0308\216\242x\004\370\210>\333\204P\230\034\2610*6\3364\307\315\021\312\202p\325C\267H2(j \\'\020,&\365\270!\031\362\314\266`\212K\252\263\330\255\"\331&\010 \007\337X3y\242\260*\203\237I\221\226f\001hA\031(\244%\215\305\350\226\334\t\037\021,\354.\340U\216\357zDAUr\215\354.\355\257\035J\025\315\242\302nZ\371H\014\357~;\302?\201\031\224\364\006\026\306\316\020\023\221\003\225\252\026\036\350\351\276\210\326\237\025\226\372\307\254\303\321\017\351\2364\037\325%\274\362]\304\212\026\366\020|\253v\271j<.\372\223WrJ\010K\312\306\2711\"\211\366\326.\002\272\203\255-\250\232\r\177}\326q\002_\342\203 \2103\317\246D\001\332\246\022\274{\227mg\006\271@\255@\316\220\010\212+\240vO\216\341\233\257S\223S\256\370\000\332\000D\373\016@p\222#\013\000\255\033\300\010OmW\246\276\211\231V\210\236\340>_\014\342_\353\346=\266&'\003nRp\220Jl\224\322\350\000\2010\3017\341\006\354\311\365\351\275[8\032\322""\200@\270\320\316\"\347P\2377\257\332\035\0001";
    PyObject *data = __Pyx_DecompressString(cstring, 4900, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (4930 bytes) */
const char* const cstring = "BZh91AY&SYb\013~i\000\006\266\177\377\377\377\377\377\377\377\377\377\377\377\377\377\277\377\377\377@@@@@@@@@@@@@\000@\000`\025\237}\357\272R\253m\017|/Z\275\3579\346\351\215)\261\323M\266\206\232\246\336n\032\355\200K\266\016\273\262\335\324\334\325\350_\000\016\306A^\351\010\t\"A\004\310\311\246\224\375\t\350\304\320\233T\336\232\032)\342i\224y\023'\250i\241\243\322lD4b4\000\3654\323\324<\240hd\031\010\002h\na\n\032m h\001\241\240\000\r\006\2004\000\r\000\000\000\032\006\200\t \"*zQ\232\206\201\240z\217Q\246M\006\200\000\000\000\000\000\000\000\000\000\000h\022jBh\210Hy(a\241\250i\264\032\231\224\030\000M\000\304\310\014 \302\r\014\2324d\323#\000\230\002%\021\022@4h\000\032\032=OP\310\323\361)\372\231=F\220\r\006\203M4\036\243\312\007\222\036\243\311<j\233MM\221\250\365\000H\220@\010\021\246\243jf\246\n\236\2154OSL)\355S\3654\207\250\320\000\0000\214\200\0312\006\200\365\0035\007\365\265f\216\363\030$4I>\035\017\233\223\227Z\217j\317}\315\364R,\271/\232\006\322\317@\324\342\331P!I:\375\200\r\347\205\374\247\364[M\342\203i$\003~\350\313Q\307\205\020\016\254\224Y*Qe;B\221\210A6.7\03133E\025W\225)\202\201)\223\200S\344\3273?/\t\253>\\\314\317\232\001\364\244d\246\264\355;@\2413%\223\271\006)jhc\001\215\2647\030\252\242\221 ,\214V\004P\2120\210\260\200\306+\004I\004\210\212\212\016s\270\201\303\207\000\200\024)D\213\202\325pt \277\343>W\261\366D\030\263\306\377\315\177v\231\030\373\ttK\361\004Xt\230\345\227\277\031\016H\200#\026@Y\005\010\014d\223#5\223\227\246\202\311I\246\021`\310\2121#9L\206N\241Asy\303\231Rh.\313\332I+\226%\001\026\214L\032i\016\311\266$\000\336\301\"AB1\222C\027\020S\353\313\2011d&\222\317V\235\200\262L@^\263\240\007\004\266\"\324y&\215\\h3`\230mU(\005\225\231\232\264v\230\213r\240\030&\224\225FhT\304V\024\264\016\201\357|\033[f\315\260\031lX\216\220J\351!I\334\201\275\322E{j\201nM\010mLAWb\024\\$f\001\257k\231'\001\t\307\256\301\320\2224\354\220\226\016Q\246\362\352\310\304\211@/\364\244\250\027\024\004\302\355\002DXXkF\325@\214\014\204\027""\260S\3055y\t@[e@)\347IP\275j,\261\030+\204\256`\254.f\202\245;\r\310EtIZ\232\264x\254Q\036\226'\202\243N&e\202J\250\222\306\304\226\240\242(+\325\362\302i:r\020~\200\301\007`\203\ryb\261\343~6\376\013\244w[\241\203\341'\304z-\313q\247)\372q\256J\021\261\007\222C\233B51\3529\243\"\004\230B\344\213\t\002\224O* \237\311\235\017\363_%/\205\247\256\347\344\345\345\342l(\224&\t\314\335Z\004\214-\035Y#\241\031\254\001\222\227}\0036\017\324\257q\326\202\216\307\013\226\030OC\026\346\320E\226\364`U\373\370\271\240\271D\342l\350b\254\014\013<F\204\260\271\301c\037\323\332\367+\226w\213\205C\346\2276\307\330*\243\357\000A\t\n\227\306RhP\222Z\2640\245\266Z3\013n\345\236~o<\221\\3\224\307\220.\212\340\265\022\325\212t\344\345ul\346\034*\236]\\\016,\003\nS\316\010\223\n\234\212\341\231S\261\023\301\201\024\002\002\326H v\t\357$\254\004%/\255y\026\304\203\320\236\265\242a<885X\004\335J\035\207\301\251\302\342\005\300\354\327\214]p\025\345 \301\177\254\352vP\010\325\313\023n\317\314\306\026&\001R\206\021\206Bp\245\215lK\217\001\010\227[\013\226\225\252\270a\023,\354\331\356\032hO'\001\273\310\356\206-\336\236'k\301PG`=\201~\033\307\016\027K\221f'^\244\212J\303\\\\\314=\013\363{L\375\206\3542u5\271\nJe\270\354]oUn\213`\3763T\224DDDD\020\026G\373\312i\325\223\223$j\374\303b\201~\247\207\014\376\020\360\026\017\0337*B7\363\304\350x\002KtBs\206d\321y;\251\034TQ\304\323:>\257w\205`\252\222R\345c\026\343\247\202\037\030t\224\203\221I\323\250r\350+\221DQ\353p\211x1\006\242\331B\261\227*1\345\031[\227\305\365\305M\342\361q\333\347\330Z\3060\267h\367\312\\j\201:\323\321\226`=\302n\035\270\314j\327\310\334\261\2515\265\005%1\244\240\024x\224\245G\325\317\312\352\275\272\366\343p\246\250\023\347\301@0\264*\301oF\017P\307+b\347\306\362\264\225\226\217O\273\246az\307\335R\264?5\307\024#\201\275\332\"e\307\365\305\020\323]\244#\324\301\327s|\337[\312\333@\306\343G\361?v4\032\256\206\337.\302\322\226;\037\033kL\343dk:\0239R/<\305\233\372>4b\304U\026f8\353\371bS\007\016\023\225\236\322\357\355]""\251%\317\022\033\251t\010\367<LR\236/\245\242\214\216\231\2246\204I\001\025%\214\224\254\n\340 \206f\272\024&b0\014*`\250\\\367\2167\273\277\371\266\347\352\275\334K9\214\345\253YH\260\026\333t\021\\\252\226*\305\320J~\r\n>Da'e\006u\025%\014\004\032\224p\250\027\016E\371hB\353s5f\275\033\274\220\366M%,\310\330\211M\231\"R\371\022\222\254\371&:\301\322\244\236\276\226U\350\352B\304\215\332\275e\221tV\214\365N-9\026\014\262\242\270\253v\372\317/m)\215\241n@\036tyO\003\260\342p:\037\211\267\342%\022\221)J\"\"\034\201\243=\tO\236+\266'YY\303\322\341\253e\231\271&\017\3156e\310\247\004\022\346\254\210d\341\344U\3112\352\211vq\362e\233\222gS/;a\243\r\217\330\231GZ\344\351r&?.d\036$C*\337\t\224+a\rbZ\224\245\325\325\355P\233\030\327\313b\013\271\255\356\215\236\203\277\277\024p\306v;\244:W\260\026A\313I\361\014\374lJT\361\212sUs\020\2176*\230\307\240\225\354\371I\214\214M\247\023\347\035j\"E\204]\n\352\006k\026\230\035\212\215\034\"\202I\324\216I\030;%\004\311$\250\032FN\n\236\005$\003\270\231\331\341C\314\002\201\2016\235'\022\365P\211\034\271\030\355=\035k\233\177\220\241\010\223m\361\326xz/\277\225\333\314\002\220fF=\254^\264\276O\330\367\020.\364\215g\000]\013C\0134#\212\202m\261\277d\211>\344\242Zn\336LZ\375G>\317\336\340kH\351\343^\217?\034\236\321i\277]\234\376\2622\330\006\243\034\200\241\204\330\344r<3K\336u\214d\302\341\227D\233\350\2166,a\020'\361SuIMQ-\24231\353$\0275cB\242\247\233I\276\375\310\270\306\224G\272\376\264\257\364\333\205c\347\336\335\202?\034\032\322\021\357\373\360C3>}D3\307\006\271\014\267\356\3455\251\353\326\263\006\265\303\276\303c\342\250\304H\370\215\031q9'\332\331\254\202\267%Y\315\311)\335s\274\274rEK\270\303\231`i`\276\326\213*\273\3144\372j\"\221\242\002\371\025L\032\253\361fM\326\372\036\254\3247GVid\3467\000\362\310\312\034\330w\371\361\305f15QY\321*\025\302\347\220\n\006ri@\3674\331\235\363\212\"\236!\220m\361\362l\265\347\263D6\315_\"\031\307\313\204\340{bC`\350Fy:\331[T\345\004\262d\273\216^\027\373\0106@\342\010m\004@B\020\r\004""\321\207\030b+\220\021\266\324\013\260C\231\262\331Fr\271\212na\201c\r\376\240\265L\231D/\302?&\305^,\322WX\226B\013D\\\243\201\360e\222w+\375i\n\252\223\002\276\355\206\204\372vU\022?\001\211\257\204\226\023\352;Y\312\360\345\375\020\275m\333r\tx\300/MN'\254O^\3235\260^f&!\233\007\261\266\304p\300\275*tQ$U\332|X\212\265)dBK\313\250\222\021\356\367\236\342\350\0028\241\301]TH\341\020\241C\270H\354T1\n\005N\224\303&\254\271\265\231s\216^\317\225\215\202\357uI\001\336L\0256u\024\360\367\270 l\251p\0270\346G/=(gL\246\310_N8H\215\364\324rH\003/1\266\335\3335c\323;>%K';*Z\227\2139,\\\202\2749\032\207\001\206\321,\300\016\334\242\3067Tj\020\313\226\024s\317B\275\267\352\003\342I\244i\2739dJ\323*8h\315(\225\201g\240F\314\355V\010D(\304$\302\264\030/\200\003c\023\204\224\031?\033\t\344\006(M\264\"\3066(\234\001P\016<:\202\2647\233.\222A'\r]>,\224\312FM\206\031@V\311\366'Y\237\n\000\030\\\310f\320\330\310\035\215\266\327\031l\222\266\353\201\023c\346\2156-0$\253\016\014\035+\311\"O\340\021\242^\202\3146\357\177j4\033\217a\027F.\005\203\303m\267\td]\205\313\035\267\357\347_\032pI\261az\373B\220P`\306\310g[]n\263\033\213@\2552\232\365|&\353\34373u\025UQ\035xp\\\334\224\214;'2!@\360\026>\"\335\363\251\274\315\177is\273\004\215\220-\300/^@bD\0204\\\302\332N\367;\222(o\235\036_BUV\024\225\221=C\001\260\354\274M\326\315B\023B,\035\3622\253\241X\372\233\2257\025Z\211&\010\307Q\r\270\017=\346\222\251\234\356.$^\231.}\025\223,f8\234Fa\003\211\337k\003zR\306\362\n;\242\003\214`\367NV\340\245czB\350e1\243[\272G\275J\026@A\021\004CcI\311}\233\216\030\032\335s\300\273Q\217\035\271m\303\226\327\017\301\365^b{\317\347/\221\256\361\006k\202\365\345\265\213S\324\027P\333\0351Z\010\034A\226\004\304\021J\27428\023\020\323\313E,\323\212\301\t\226e\232\325S\253ekM\365\235\346\333A\275\023\234Ny\364N\010#7dB_\013\033\232\022\372\021\003\000,\335\344\226\303SP<\366\310\361\325(:\344P\333\327l\264[3\223\335\3452\346\006\327Gq\317\36505\324\237.AM*S\233(]:\350G?\227\237\251\025\274NT\264'\206\366""\022L\346\210=\021,D\200\267\236\2453'_\214\275Z\031\231\034\315\304\033\210\267\211:\251&+8(a\301K\333K\310#\240\r\0327\3105\013f\035\333s\235\004\254\263\253\363\266\245\320d8\r88>cd\303\001X\005K#\250\3423f\322\024$\213A\234e90\243U\0079\004\004\345Si\255\223+\235\250*P\216\231\345\200\262\333a@\314j\315\021\3776WR*^*\306\231\220\335\372\315d\346\224\3157\036\334\302u\204\r\215\272\000\261#g%a\223\027\353RB\254r9Nd\034\206\016\207%\215\321-\266\230\205U\324\302RN\212\\\275\020\227d\224\3115(\253\313\312\226y\213b\266\375\315,\013mmGL\213$N\340\217\222$$\250\212)q\022#2\r\n*8L\220[%,\315\215\244\351H\2212#w\0104I\264\275\346\233Z\024\223}\033n\212\222\010\273\331ITU(\3275\321\360\013T\200\267\227\261\276\251f\320\256\361\242\315\274Z6\013\323\265q\361P\263\302\005\200z\216\2641\214`t\345\212\264\260\224{\031Y'\272\007\004e\000j]\215\211\202M$\333\020\275\260r\310\020\340CH\331\246k\275\256\255\311(M\030\204-Ot\211%')\t1\331\266)y\002.C\332*\363\000h1\314\306\3776\251\242\337\\!\007q\250\351L\303\232\001\352 \346\226\303\222\321\022{\327\3019\007*P\026\022\240\332\221\001\200\206\262\030\212Lt\375[\344\250*]\202\267\035`\"L\340nm\010\314h\264\360h\303PJF\246\026z\021\264@v\244v\201\332\007Zu\301i*\252\250^\023\262\266hR2\300\016\030\032\311rQ\202[\247\024\206\312\363Y(j0\240H\"0\241\372\224q\234\373\207\024\"\263\\\356\006o\346\355\362.\274\022\016\356p:\t`\256Ar\2463Xe,\246\254\255j\247\005\241\\\311\366\205J\256\227\327E\352\273._\036\210\002\206E\222H\016\217%\215\375\250\002\034<S\237\030\204FD.B^\005m\001x\251\330\006D \337\310o).\000\323\311\352\232\031\251+\334hW\004D\233\22527\220h\021&C\2304\227b\305\217\026\"\240=\306:\327{y\020\202\374r@j\204!\261\242\372\022\244\353\212#\016\233\325\304\354,0\361\341\t\271\275{4\240\035.M\215\331\246\224\317\234v\202X?u\272\353\013\3232'\232(\202h\314\240j,tQ\236bh\006\351VS\r\026;(Z\016z\377\026<\306\027\013\236|\326\301\207\246\n\234R\247\223n\252\303\236{\265\345\334\273\027gM\030V\333\311\364\214bq\n\202\244\224""\212<\3349\000\265\337\334;\270\314f\370X \223L&hw\345\031\260>\205<\026\302\023\226\240\226\330\t\332q\341\014\027\265F\243\t\335\004\031\275\222\240\314`\313\253\234\211'T\002n#\"\032\327\006\245\022\032\332\306\243\014\365R\033\213\002L\235\232IF\340\220\300\226\353RD[\213\231d\314CV\006\023\303&\251h!YE\010\332\245QE\026*\225\020l\033\240_\245\323 F\205\3050l2\031\250c@\305\277\236\013\206\212\303\r\211\017\263\235\014\024\212J\212\300L\301f\260\026\261\263\263\252U\277L\203\026\204X\2141\302D\231\230\354V\026\"\306\265\363\220\333\251\336\215\233\027\222\277\002\361B\3514\227\215\210\341q\033\002`k\375}\206\263\241to\326}\\`y\0362;\366H7\201\214\354\364b\221\322t\225\004\260`7\341\215\363\203\027B\323CC>\347l\001\020\222\250\305C\0305\361\337\305\371\374C\024Q\020\301VpCq\336\343S\325`)\315\276\t\036\233\205\343\000i\240b\3240Ju S\243$!$n\315c\260\306\366\033\231\204]\327\217w\220\r\377\234\373>\336\214H\221I\010kH\302E\340Q\307\261z\255]\310\272\210:0\022I\357\034\"\277\177\006!\223\212\240\310\370\360?v\nV\001K\325j\323\333\316\326\311b\234\343e\027\304\363\354j\3206r\332\255d\270R\2219n\272\352\221\036\206\303\345\232\220\262\010\304~\275\316\035\251\013\001\241\303|\206\r\021\223\033\252\322\266\255#\243N\307\023\032P\346\323s\253D+\002\2200\216h\365\014\370\254\222\266\001P2\254\266\366\204\262\223*\376X\347\326\203\217\234\304\"\024D\242\023\263)\306\333\343qLt\210\024\\\310\177\207Y\336\r\260\013\371C\220\336l\220y\346\350\306\325t`\005:E\000\2749e\373+\203\006E\337\270N\274\020\335\245X5\313h\030\017\230\245\006\004dW\016\307#-\374A\332\0345^\234\320\245\0000\365l9\263\2227.6\205\224\270\255\332\364\267\314Y\247!\314\3533\341\240\333A\013d\323\3048c\272\037\233\022\312\373\374\016]\323\340& s\\\332\017\316j\205F\204\216\363\371\025j\351r\237\357\020\354\261\316I\201\371*7'1\226$\222\331K\241\275\314\316\332\316:\032g\242S\326jy5\240\222\367\n\217F*\312\224\025\030\014Pbw\377\267\302\267\363\005\324\351\301\322T\342\024\007\351\261\301\271\267\333\340""\363o\235T0\243\376\203\317\302m\300\335\027\021\271\365D\017\r\343\251\273\375\356\256\365\375\305\020\231\344\037\271\3214\327\021\262\376\243\007\037\001\302l\254\006_\366\322\204\014\363WY\347\032\351\205\322}\201qHmqK\321\360k\034\256\365\027\350\347GdNw\370{\224}-\231\335\377\330\252dNJL\367\351\311\317\301\230\324\315\031]z6\237\035\342v\211H\224\201\034\352\002\327\242Y\177i\343\307J\001C\2256\307\217\247\177\236XV\0268+\013M\367p\364\374\320F\242Z\354\300\372\032W\327\227\234\214\344\327\344\200\301~\372\277\356\027\240\301\311\352\241\241\240W\321\025\322\352\200\231u\356x\026E\302\240\354\3568\230!\341\201\241r\270\311\206Lw\260\032\261f\256X\303e\271\267f\372\313\034\262\177,SVZ\313\003\357\316+\3551d;\343le\036/\226\3130\262\313]\257{,\321\032\3223=\323^\332=\215\235S\256\323\257+mgS\316:k\363\3701\377\263*\255l\364\237ss\234\331\275o\232m\3724\2609\370$1S;;Ww\0368C\305\004\352\004Q\227n\332[\t\257\301!\244g\344t\304nzA\267\350\037\304BG[[\210\257\200\344\030\273\216\317\\\252w;\265\375\370;\206\237C\035v\022\364\352\313\263\277\215N\216\205;\021\330\216H\374qQ\3518\266\344\222)]e9\245\030O\023\230\262\323\352\351\016\356\023\324DRg2\364\257/s\225\213\326\020mUVm1\232\305\311\244I\244n\300\305\2521\227Q\032YS?\326\2673\245g\3233\224\032UCLD\266\251\315\232,\027\240\260\257hL\020\332\210ZV\251\312l^Y\236\250Uj\227\006\321:\225\235h\303T\271Xt)p\352Q]tf\005\212\212\226\306j\207L9\033^\373\261\212\315\3561\025PEE\024r\232^\322\322T\320\311-\203\022\355\325u&\261\206\270q\032E\251y\241m\t\233\341mkZ\252\2611\374,Z,\333:\353\304\371\365\265\356Mg\233$\253\227\233\310\213\212V\366\323YW3\252\304\313;\n\305\n$\2266\347-\016\367\206\266\315.:\260\321*\223\341h\362\310T\371\234Zrp\324Z\261PX\325\204\025I\231\331\354.SI\017WO4\251\222\233\231\336\247SA{\330\036\323\347M\254\2137k\022\322K\022\346+P\337*\327X\377\305\334\221N\024$\030\202\337\232@";
    PyObject *data = __Pyx_DecompressString(cstring, 4930, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (4964 bytes) */
const char* const cstring = "x\332\335Z\313W\324\330\272\227\333\250\240\240\024\342\203\266\365\244\264[\304\006\272\313\266}\337\356\233\206\0229\215<\013\355\276\335\336\254\220\n\220\266H\212$%\3249\236\273\0342\3140\303\0143\254a\r\031:\314\260\206\374\t\376\t\367\367\355\235\244RU\001Y\266\347\234\265.\013\223\235\275\277\375\355\357\375\330\362P8%\340G\024,[\266+\226\240\030EU\320\364\242\246\310\266\246\257\013\366\206*\330\325\262*\030k\202j\232\2069\301\341u\301(\253&`\014\214L\30126Ua\255\242+4!\2274\273*h\226\240\033\266`U\312e\303\264\325\342\204\210\355\333x\310\325e\333T\345\315\247\262^,\251\202\\\302G\261\212\223u\253\262\251\026'e\235m,\233\306k\r\304\310\202\245l\250\233\262\260\275\241\352BY\266,\242\213!\033W\214\3152hX\005\232\242l\313\240V\266\005m\263\\R7U\335\266\030\365\014RX\250N\312e\253\002\300\005\323\260\r\305(\341\034:F)\031\226*L\032\225R\221\035K\304\2007\306X\236\030\2660U\222\301\200`\033\354\224U\331R\3075\335VMp\312\205bM\034\014\332\024\324\030;\240\242\027US\250Xx\200e\3334J)\233\301\274\242Z\026>\010M\n\000\0332Y\013D\312\272\t\221\247\200\221\000\"B>\260\314\360\230k\262\222\006\310\310\rY}\"k%>\273\256\332\241r\036\n7\233\323\232\256\331\032\214\340o\252\360\263Z]5d\2638C\270\315J9\266\207\207B\023\336T-\3330Uayfzf\256 l0\2730\223\020\026\016j_ev\370\024\252\335\254(\033\241\372\raM\265\361\265\006\223\224\213\253\212\004\n%c\365\017U\261-n\270s\206\256\316\031\266\312Me\262jo@\3170\325\242Z\322VIOj\251\n_05\005\004\023\220.,\344\027\306\357\334\277#\340`\220\312p\201\213U\245\004ST-r\214\325\212V\202\2670G\261&\204\2315\241jT\004]\ru\t\270\344\006\233\354\2308\"\351\2170+d\252\224\260\035\246=\"\0245\023\207h\257U\332\375D.Au\214v\020\"\334\276\363\200Y}I\223u[`zSw\024\225\333+;a\342\024\367P\033L\254Vl\325b\237\343\361\017\373\344\336.1o\177(\210\020\3252\233\231\304\004\003\240\237\002\3503Lm]#+\023\247~\232L\006\tN\322kU/\032f\204\007\0264\026\372\216\\\212\321\210!\324\270UV\025mMSZc\315\032\214]\265\340\256\034\243\265U\242eB\0075\244\240#\252""\226\027g\227\013b!\237\216\242\250\332\260\034\013\030J\232e\377fW\020\017~c\310V\253\220\307\313\227iD\026\213Z\350O\314\314#$c\035\330\017S\305\266l\352\024\235BE\020\344DL\354CadY-\255\215\024\030=\351\321p\214d\370\222C\314\022\361\234\3421a>$\230\357[fN\307\267\274\344{V(\254\254\225\340\017\232\316\202\032\217\3360\355\025=\216\301<\224O\010\317\231B\230\364\036\n\243\243\302\304\217\314[\212&\214\316\2246e]^W\315o$\270\304D\271\272#\027\213\222N>\303\210\031\021\024\0369WU\346M\314\3636+\026\233H\017\3137\343\220\034\245\224\203\202\362\350\030\342}<\3158\243\364\322\024V$&\"\240\345\344\303\267\265\312\230m\326,\031\004\252:{\332\3127$\201o\270\004\254\370\033qx\rA\310b\021\345\265\\\252\250\302\272\242Y|W\361\225Zm\023FL\016\314\215(`\022\273\251\033\202\262\241\225\270\370G)b\204\t&|\001O8bG\204\371\356\203\234%\255\2003e\301\276&db8A\027\262\014x\n\0039!c1\217GE\306WYS^\225\240\031\266\033\320\272\252\360p\362\221(\342\214\362\261\010B\001|\364v\212\037dm\177\002\003\231\313Gm\207\336\337p\265[Uk\002\221cM\333\271\tt(\217Ly\235\250\032\255\330k\343\367\2711\321V\262\242\021a[\2637\250T\320\021\266\266y5B\371c\223\022#KC\342\354\2548\273\224\027\247~\225\362\277\314,\027\226)jO\306\332j\375\232\220$S-V\024UR\030\251\222\324\261\016#g\202:\010\002\314+j\251}\222\252\245\3669csS\263\333&)\361j\372\232\2212\035\346\343\264\225r\n'\315\005\211\211\365\340eD?\331>x\031\256\223\262H2@\240\325\024I\2277S\261\307\020i\2136\005\002\211\033\354\201\313\254.h[\245:S*\313\246\255\245\360\214\242\260\264*+\257\332\246\2412I\256P\240L\221\270\025s\312\216\232\n\23509N\267\212\304j\252M\304\353\261\362\343\231V\245\245L7U\226\266\030+,m1TW\274\224\340\217\221\305\343^b\010\372U*3%\251mrG\263\333\347\332\304\320\272\230*\205\031\0303UGQ\255\304\002L\313G\272t\223\313\251\210\233\000\253\350\273:g$\036\214Z\027\232\316\231\230\213\324\323\234RwT\005%`\323\316\254\364\365\255\212jV\323\227\232\246\335\271V)#\324\267\035\331j\022i\363M\233H]\215\215\"uU\323S\326\300\036|""\227T\237F-b0\000\332\310l\363\226\326\005\324\240i\"aK\225U\350C\206E\225K\262\336Z8\267W)\355\337)\006\322\t\321i#\3550h\313Tr\354\324\342&}\026xeZ\220\224PBI\264\007\001\262\351C\000\017\344\245\r\356\020\216Z \223|%K\233\216\211\203\270i\203I\243\257\r$\235\264\026\240\220\252Iqn2?;\233\237\232\024\013\342\354\374\364\362\344\374\354\312\263\271\345\251\237\244\345\311\247\371g\342\362\324\322\314\363\374\222$.-\315\277\2200Z\236\231\237\013\347\346\304g\371p\030- \252\261.;\no\354#\257W6\371\200\335\264H\022u\322\2224\255\332\363<gN\241\321\333\230\231\233\316/\027\244\371\205\002\020I\317\346\247\362\2353\222\270\260\220\237\233JY\230D\tQ\310\037\270p\360\306\245\374\302\2548\331\266\263 .M\347\013x\3754\213\225B~zi\246\360+\r\226\346\304\331\231\271\347xLA$\323+\317\362s\205\350{\n\"\214\306\2543\232\231\237\321mb~&\272\312`\002\230\211.$\232_\350\200\330\007\365F\263\206\362Jz&\376\022\321\262<\363\337\371\271\371\202\364d~en\212\0063\317\026f\363tp~*l\373\227\243F\210!\351\230\210%>\377\363|ti\023\036\030\365_\013\262\275\201fe\035\001g\023\345\036[\223\244\205\352\016\376Mi\212-\315\251;\366\222\272F\235\036\223\312ra\346Y~~\245P\370u\0012~\232\237\374\031\022d\r\336\312\234\270RxJr\201M\345\247\370\347\374\022\270\300\370\347\271\371\027s\317\241\211\3716\223\n\347\310\244\302a\270\360\202\367\236)\215\\\312\324\004\365vr\334\363\322\264\266\246Z\210o\252l*\033\210\250\366\206D\335s\002&\354\205:AJ%C\241\213\"\271\014\220\035m\023c\331\\\267\232^\312:\022d\331\016\267\355\2146\262\245h\232lUuE3PY\232F\005=\243j\221\213H\274\230\246\234\230\310\213\322*N\177\005\306%\005\204\264}Q\337\331\234Rw\224\346\007\275W\r\243\264ZYc\371\010T6\311\017I\006xeM\001*\360h\254c`\224*\233:+\024!\320\325\220\r\014\311-\3612*\250\366\024\211] (\254\374e7#\026\215mEB\241\257H%UW\022\371XayD\341\351F\2111r\336\224\260~\344\007\262\216!|14\254\006Pxg\030\321Hm&\2537%\211]B\320\013\002\004\001\022\322\226\242RU\311\n\205$/\254\234l6~a\263\303\371\230xL\332-Y? \323\352\352N""\331T*\246\211l(\331\033T\302R\257\027\365{\261D\212*\261\315\244\022^\243HR\221\234#z\025\r\005uY\330E\203\023@\263')&*\342T\304\003P\300\210\3407\217t\321\265\303\345Jo{\225=\301+\275!\226\316j\247\245\276i\255hZk\230\250Jd\305\007\335cc\030?m\230\211\025r\037\3655\211F\246Y\235\264\3278\355UMk\035\223V\271\244t$\255-H{\317\321\326dHt\343\311S\232\304oJ\265\330\014\233\366\030\2059\376jq\0316CL\203@I\243\033\302\320\0051.WCc\303\030\342\326\020\031\324M\353\017C\323a\331\370\345L\277\332&\347\207}\256\333\033`\272\010\326e|\233\025\340\007\311\022b\r\214\221?\271\r\201I\013\214`\016\306\022Z->\214\"\216b#\323 \24305\225-\220T\342\247\272\315\0378\017\372\261u\033V\2230\000\250\210k\311\260\342\311x q\"\233+\026\3054\372\207\340HnY6\312<\346HRXD\306\2032\302)I\231l\037\223\210\377\220\215F\344\261\361\206\252\274\262@\t\3732U\010\315\346\343P74\344^J\243\212\316\256\017\324d\037\302W^s\335\342g\253\"\2278\317\314\230\331\203\235n*\255\255d\\\370H\235mN4\241\356\320\220\0256\246\nT\026\242whP\246jWL=j@\021\234-I^[\203\004\325\242\305!\350\206\304R\365bk7\312\212\251P+\211\342\272\245\234\356,\2401ST\327d&\236\270\030\223R\n\263\350Z8~T\254\304\03560\362\210\311o\356\255\n\262\267U\265\232\321\263\305E\010\201\024\276Xd\341f\250\2211\200a~\r\317\343B\305RcZX\340\345\027t\211ko\272\362\r\257}-\272\303i\206\331\205*\257\332\204[7[\373\2041\272\035\254`\201\246Y\005!\334\032=\3261G\377!F\233\243\205\266\252Y\2705\326vqOX^\033ZQ8\322y-Q\376\330B5^|b\032\233\211s\2161c\016\301\350\307\317\005\335\327\274Q?\367\276\347X\357u\357k\177q\277g<\030\237\334\313\275?q\254\267o7\327\350\033p\372\335\305\306\3009\347\241\273\350\276\3623\215!\301;\345g\375\334~8xP\023\3673\203\215\201\341`x\302\227\033C\347\235\202\313\240\216{\213\361\353\274\363\334\315\321k1\r\366<\377m\302\364\234sr\316SW\304\311=gw\267\202\356\313X\341\243\333\265\321zn\277\347tp\372\013w\332\003\201\203\301\340W\336\242'\357\367\3646\372\262\336\005o\313\357j\364d\234\313\256\330\350\031t\306\335\255F\3379""\347Qp\365NM\254\001\365\231]y\367\357n\326\275\343V\t\260ow\332y\344\312!\366\357j\300}\215\004\321\350!\322_\370\"\r\257\343\240\3763\215\276\363\301\371q|\017d\234!\347wo\230\304\320h\322\032t\177_[$\214O\234l\320}=\270\376}M\332\003\021\375\273K\273U\267\253\321\327\337\350\273\344\236\361dz\341\023\2346\372\206\0341\350\036\rF\037\004\017\376\372.\333\020\030\177\035\352h\364\\v\177\304qwk\231\375\236\277\200\245\236\323\273\367\235{n\266\321w\301\321\334-\257\027\232!\254\373=Y/C\253\017\235\027\020\000\246z\275\014\343\247\357s7K\274ei\365\221\263\016\331\366\r\273\303^\326{\354CD\220\360>\343\023\253\217\235mH\004\033F\275\234\367\254\006\314_\270\342~\273\000\350\325\371\353\\p\266\350\225I\010f\237\013\367A=\023\013\367\253\340\253\2735H\375\212\273\354u\201\342\020\371\355Z6\222\356%\377R-\203\317&\232\267\342\3733d\225wv_;\277\272\nv\301:O8\033\256\014\001t{\"\377|\345e\232\200\333\020\216\005\226\261r\322\331\001X\2277\004{d\237\177\367\262o\305}\020=x\256qn\210L\260\367\324\373\273m\202\377gJ\373\232\367U0r\277\336\025I\366\333ZW\272d\323d\032\032\354\227\236H\306r3\030\271W{]_\254\313\251\246\017g\363\267\0165\375\267\240\276\377\275\020r\337\350\271\031\334\204\272\032=\217\352p\261\263\020\370s\367.\261\nW\213\310\205\2574\311m\222\330|$\r\200l\035\007\346\235\363\216\354\274!\277=\357\374\036d\201\243&\022\221?x_\370\277\325\273p$D\014i\005W\357\325\266\352'\210\237!\216\303\205\207\237\336\275\353\\\204\214\317yl\323c\367\265\307\004\361\306\275\357\345\336\222\211Rx9;@\232\274\032kr\310\231\301v\325c^\204\307\027$\036\246\2013\273\232\263\005W\244\3201\344L\2621yd#\014R\244\021\026\244\256x\353\376\013\304\217\027u\261\016M^\004d\017\263[&\233e\36737\347\222-!lf.\273\327\331q\267=\0267\357\273\267\241\255\241\317\33518\376%`\204?]\372\306\337\252}\306x\037\340\026\325\305\003\356]g\320\311\361\310\000\266,\347\232\363\304\035\201I\335\362\305\177\002{q\014\276\354M\373\367j\331\332\275z\226\364\375\377\204=P\301\331\373\002q|\032\344L\327su\361""\337\305\336\373S\307zO\021\255=\275\373QJ@\362\203S\314\273?\370\031?\313\2079z\315\272\314=\177F\310B\364\n\216\375\327^\327~w\357\356eG\334\357\346\\#\3164\242\231\223o\267(\311H\340!|\361)\242\376\036\"N\216\345\323Fw\317\356\361\335\025'\213!\300~\205\037V\334\247\236\010\207\356\246\020%;\377\2134\252S\222da\205E>r\271;\2736q5\350\334\342\256<r\244X\323\277[\200\253\236F\002\021\275\202?\\c\201\360\273]\005q\001\250\276\204\217K\224[\375\"V\240\006\333\371\316\221I\234%\360{\3751\271\331\376\000)\2252Qrze/\273\227\213BV\017m\274\355\374B\271\367\247\275\354\373\301\330j\242\370u\036\231Bv\337P\236F0\376S\321<\014{\271Z&%J\037\030\242\337Bs\024\3672\301\340\210g\371#,\247q\375\0148\307\251\002\370\261\016\321\365\223\314\237\372b\230 \0170\373\253\024\004?d\366\024\227?\255W\347\322\215\236\322z\305\021\201\351?\334\354\247\243\037u\332'v\333\177-\375T\004~\332\244\361)\350\007\351\215\366:\3437\357$\252\345\336\2662+\256,\216\210.v\210\337\275\213~\027\271z-\367'\320\305%\340Ko\004aq\224\261\334\341T\215\356S\273_;p\373\260\316\"\254!\202\257\375\034\"K\241v\245.\327\267Z\312\363a7.@C\330Q\n\274\376\035\377\r3\261$,\213\007-x#\330\1770u&a\005$\200\301\366\322)\334EI(\245Jj\227\3171\024V\260\203\343}\324\327\210\356\212w\303\377\214\2723D\244\216f\240\243^|\211 \177\367 \301?\254\307\002\212\233\t\364&\344eQ\365}H\001zH\351)P\340\214\371&*\373QP\346\234'\301\347\214\351!\347\257\010\275T\241\301\336\267Y\t\014\313\375\303\243\222\n\254\366\237w\246\335;\301_r\324\322P\317\261\345\236\240\256\343\202c\004_A\200\365\033{]\215\376\213\315\351\325\340\322\327\376\367\265\326I\003\351\013\371d\324\245Z\2641\375\364]\367;\361\335\"O\307\217\271\303\301\363\202\213HZ(e\353O\321\217\365_\214\347\0353\030\036'\037MN\002\330\333\n\221\302\351\257]\367\356PO\031\t\351\006<\262]H\231\026s\270\346\335\364\263\007\232:e\240\366V\352PG<2:*\347>\354\210GF\027[\356!\216\030t\223\335\016D6p\202J\234\276\020\204\272\254nf\003;$O\264\367\301e\220\030);\022\366X0\206\362\2461@\016""\327\246\232\333\265\021r\244\244j\306\020D\016T\r\026[T\323\246\027\"\365\352\001\276D\356p\0047msu`\354\020~\363\354t\014\235\201\345&\224\225\361\277\364K,O'\372\335.8\325\3613(\244\256\272\266w\037\241g\253\321\335\307}\214%\006\224\021\257w_8yw\210\260\017:0\036\310j\037\017\333\375O\377\244_\251M\326\354\372\275\275\221w\231\260\3448\033\234\035\363\247P\305^\254\237\336\273\367\016v\3202\223\243\300Jq\356F\255+\350f\312\n\373\365\323t\027\204$\267\317\036\007\310\273S\344p\226\367cmEY\207\310\177G\225u#\335\372\251z\202I\006\302\267\265\323\365G{\257\202\305\245F7\312Lw\303\223\275-b\352\r\253\344z\340J\276^\237{\207\332\n\244\006\335p\341\367\027b=1\323\276\342\206A*\357\014\2737\274\023t\017\303-)O9\2311\307\0233\254\021i\245q\376Bc\350Jp\205\352\251K\303\220\361\357\376\305Z/\002S\266\201t\235c\267O\003\274F\204\177\034}\013\325\342Ct\241B\217\217\302@p\003W\351\006\304\2137\334\257w\305\033\202o\305=T\014\354Z,\261\215\244\020BS\202\212\321\237bv\227\344\251+\202\243;\232\030\356$:\362\256\375(\206C\302~\267/\006\343\250_\341\277\0022\325~\263\306\341\032\270\364i4@\226\302\250x\211\372y\270\203\3324\371\035\272\345H\0328\024C\232\006b\231bC\220C;r\250\006\220\226\233\350?g\345G\252\006\356\325\266\232p\027\351\206\344Oj\2005B\024\251\217\256\206\216\273\202\007\324%\036\254\203\017\302\037\256\200\017nO\223~,\247\337\374\223\3017\224Q\016\223~\354-\200\366\377\326\341\323\261\364\251\226\217\341\266krm\353\310\322G\330JD\240\250\362:(c\264\335\237%\303\340\200\323\343\"\032\263\310\036fm\272\032\370\207\207\252\205`Y\367\374\320\375\037\377;\2060\272`\247\210(\373v\3551#/IY\013\246\240\217\232\336&\242\007\356\234\177\235\325%\021\242\261\260O\377\232\265\374GF\364\310\325\375IVB$\256\374\321\303t\327\236\355e\340\034\311~h\300\351E\267$#\341\2359v\374\004]_T\235\263\256\005\223\030\256\215\325\345D\237\277\317\016\245\213\024\226\221`>\224#\311\2243\254G\243\233\303\357\251\343j\034\357\333\025w\027[[m*\270\372\372\251\316>\221\300\211\224\301\252n\246""\030$\035*S!I\244z\032\255\263\032y\320\373\021\345\324wd\001\r\272G%\007\312\020\322\340\302\030\330\334\346\013a\226\314\273P#\231\316\000Ov\0358\0257\343^s\347\321\250@B\024\336\371\337N\260\277TH\014\303\277\035\3403\374\177\354\376\017\325Si\254";
    PyObject *data = __Pyx_DecompressString(cstring, 4964, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (13797 bytes) */
const char* const bytes = ": \n    A status code indicating the type of error.\n    An operation or some functionality is not supported.ArrowArrayStreamHandle already consumedCan not provide a schema when passing Arrow-compatible data that implements the Arrow PyCapsule ProtocolCannot close Could not read option Errors related to database-internal errors.Errors related to database operation, not under user control.Errors related to processed data.Errors related to relational integrity.Errors related to the database.Errors related to the database interface.Errors related to user errors.Failed to get schema: (Failed to initialize KeyboardInterrupt support: Failed to restore SIGINT handler: Failed to set SIGINT handler: \n    How much data to fetch for adbc_get_objects.\n    NoneNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.\n    PEP 249-compliant base exception class.\n\n    Attributes\n    ----------\n    status_code : AdbcStatusCode\n        The original ADBC status code.\n    vendor_code : int, optional\n        A vendor-specific status code if present.\n    sqlstate : str, optional\n        The SQLSTATE code if present.\n    details : list[tuple[str, bytes]], optional\n        Additional error details, if present.\n    \n    PEP 249-compliant base warning class.\n    . SQLSTATE: 'Self'Tuple[ArrowArrayStreamHandle, int]Tuple[List[bytes], Optional[ArrowSchemaHandle], int]Underflow in closing this Unsupported type . Vendor code: )) .?adbc_driver_manager/_lib.pyxadd_note bytes' cannot be Nonedata must be Arrow-compatible data (implementing the Arrow PyCapsule Protocol), a PyCapsule, int or ArrowArrayHandle, not data must be a PyCapsule, int or ArrowArrayStreamHandle, not disableenableetc/adbc/driversetc/adbc/profiles for value gcisenabledkey cannot be None must be str or bytes(no child type) of option option keyoption valueschema must be a PyCapsule, int or ArrowS""chemaHandle, not self.array cannot be converted to a Python object for picklingself.connection cannot be converted to a Python object for picklingself.database cannot be converted to a Python object for picklingself.schema cannot be converted to a Python object for picklingself.statement cannot be converted to a Python object for picklingself.stream cannot be converted to a Python object for picklingstr | bytessys.prefix(tree fragment)utf-8value for key ' with open : would need more than ALLALREADY_EXISTSAdbcConnectionAdbcConnection.__reduce_cython__AdbcConnection.__setstate_cython__AdbcConnection.cancelAdbcConnection.closeAdbcConnection.commitAdbcConnection.get_infoAdbcConnection.get_objectsAdbcConnection.get_optionAdbcConnection.get_option_bytesAdbcConnection.get_option_floatAdbcConnection.get_option_intAdbcConnection.get_statistic_namesAdbcConnection.get_statisticsAdbcConnection.get_table_schemaAdbcConnection.get_table_typesAdbcConnection.read_partitionAdbcConnection.rollbackAdbcConnection.set_autocommitAdbcConnection.set_optionsAdbcDatabaseAdbcDatabase.__reduce_cython__AdbcDatabase.__setstate_cython__AdbcDatabase.closeAdbcDatabase.get_optionAdbcDatabase.get_option_bytesAdbcDatabase.get_option_floatAdbcDatabase.get_option_intAdbcDatabase.set_options_AdbcHandle_AdbcHandle.__enter___AdbcHandle.__exit___AdbcHandle.__reduce_cython___AdbcHandle.__setstate_cython__AdbcInfoCodeAdbcStatementAdbcStatement.__reduce_cython__AdbcStatement.__setstate_cython__AdbcStatement.bindAdbcStatement.bind_streamAdbcStatement.cancelAdbcStatement.closeAdbcStatement.execute_partitionsAdbcStatement.execute_queryAdbcStatement.execute_schemaAdbcStatement.execute_updateAdbcStatement.get_optionAdbcStatement.get_option_bytesAdbcStatement.get_option_floatAdbcStatement.get_option_intAdbcStatement.get_parameter_schemaAdbcStatement.prepareAdbcStatement.set_optionsAdbcStatement.set_sql_queryAdbcStatement.set_substrait_planAdbcStatusCodeArrowArrayHandleArrowArrayHandle.__reduce_cython__ArrowArrayHandl""e.__setstate_cython__ArrowArrayHandle.releaseArrowArrayStreamHandleArrowArrayStreamHandle.__arrow_c_schema__ArrowArrayStreamHandle.__arrow_c_stream__ArrowArrayStreamHandle.__reduce_cython__ArrowArrayStreamHandle.__setstate_cython__ArrowArrayStreamHandle.releaseArrowSchemaHandleArrowSchemaHandle.__arrow_c_schema__ArrowSchemaHandle.__reduce_cython__ArrowSchemaHandle.__setstate_cython__ArrowSchemaHandle.releaseCANCELLEDCATALOGSCOLUMNSDB_SCHEMASDRIVER_ARROW_VERSIONDRIVER_NAMEDRIVER_VERSIONDataErrorDatabaseErrorEnumErrorError.__init__GetObjectsDepthINGEST_OPTION_MODEINGEST_OPTION_MODE_APPENDINGEST_OPTION_MODE_CREATEINGEST_OPTION_MODE_CREATE_APPENDINGEST_OPTION_MODE_REPLACEINGEST_OPTION_TARGET_TABLEINTEGRITYINTERNALINVALID_ARGUMENTINVALID_DATAINVALID_STATEIOIntEnumIntegrityErrorInterfaceErrorInternalErrorListLock_MAX_OPTION_SIZENOT_FOUNDNOT_IMPLEMENTEDNoneNotSupportedErrorNotSupportedError.__init__OKOperationalErrorOptionalPathProgrammingError__Pyx_PyDict_NextRefSelfTABLESTIMEOUTTYPE_CHECKINGTupleUNAUTHENTICATEDUNAUTHORIZEDUNKNOWNVENDOR_ARROW_VERSIONVENDOR_NAMEVENDOR_VERSIONWarningadbc_driver_manageradbc_driver_manager._libadditional_manifest_search_path_listadditional_profile_search_path_listallocatedapproximateargs__arrow_c_array____arrow_c_schema____arrow_c_stream__asciiasyncio.coroutinesbase_prefixbindbind_stream_blocking_call_blocking_call_impl_blocking_exc_blocking_lockboolbufbytesc_approximatec_arrayc_bufc_catalogc_column_namec_db_schemac_depthc_doublec_errorc_info_codesc_intc_keyc_lenc_partitionsc_planc_queryc_schemac_streamc_table_namec_valuec_value_lencancelcapsulecatalogchild_type__class__cline_in_tracebackclosecolumn_namecommitconnectionconvert_error.<locals>.genexprcurrent_threaddatadatabasedb_schemadecodedepthdetails__dict___dict__doc__eenabledencodeencoding__enter__enumerrerrorerrorsexcexc_infoexc_tbexc_typeexc_valexecute_partitionsexecute_queryexecute_schemaexecute_update__exit__floatfunc__func__functoolsgenexprget_infoget_objectsget_optionget_option_bytes""get_option_floatget_option_intget_parameter_schemaget_statistic_namesget_statisticsget_table_schemaget_table_types__getstate__handleiinfo_codeinfo_codes__init___init_blocking_callinit_funcint_is_coroutineis_pycapsuleis_validitemsjoinkeykey_byteskwargslengthload_flagslru_cache__main__main_threadmessage__metaclass____module____mro_entries____name__name__new__new_lennextntnum_partitionsobjobjectospartitionpartition_lengthspartitionspathpathlibplanpopprefix__prepare__prepareprivate_data__pyx_capi____pyx_checksum__pyx_result__pyx_state__pyx_type__pyx_unpickle__AdbcHandle__pyx_vtable____qualname__queryquery_datarcread_partition__reduce____reduce_cython____reduce_ex__releaserequested_schemareturnrollbackrows_affectedsschemaselfsendset_autocommit__set_name__set_optionsset_sql_queryset_substrait_plansetdefault__setstate____setstate_cython__sqlstatestatestatusstatus_codestrstreamstrictsupersystable_nametable_types__test___test_errorthreadingthrowtypingupdateuse_setstatevaluevaluesvendor_codewarnwarningswith_tracebackPyObject *(AdbcStatusCode, struct AdbcError *)\000struct AdbcError const *(struct ArrowArrayStream *, AdbcStatusCode *)\000void (AdbcStatusCode, struct AdbcError *)\000convert_error\000PyAdbcErrorFromArrayStream\000check_error\000\000\000\000\000\2501\320\004\"\240)\2501\360\010\000\t#\240+\250Q\340\010-\320-C\3001\360\006\000\t\014\2101\330\014\020\220\r\230Q\330\020\023\220:\230Q\230k\250\021\330\024 \240\n\250!\2501\340\024 \240\n\250!\2509\260A\340\021\022\330\020\031\320\031.\250a\330\024\025\220T\230\021\330\024 \240\005\240Q\330\024 \240\005\240Q\330\024\025\220V\2301\330\024\025\220Q\340\021\022\330\020\031\320\031.\250a\330\024\025\220T\230\021\330\024\025\330\024\025\330\024\025\220V\2301\330\024\025\220Q\340\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004\033\2301\330\010\017\210q\320\0042\260)\2701\340\010\013\320\013\034\230G\2401\330\014\022\320\022%\240Q\240a\340\010\t\330\014!\240\026\240q\250\001\330\010\021\220\033\230A\330\010\022\220-\230q""\330\014\023\220;\320\0364\260A\260Q\340\010\016\210a\210{\230!\2304\230y\250\001\330\010\014\210G\220;\230a\330\010\017\210q\320\0043\2601\340\010\"\240+\250Q\330\010 \240\005\240W\250A\250Q\330\010#\2401\330\r\016\330\014\025\320\025-\250Q\330\020\021\220\024\220\\\240\031\250!\2501\330\010\023\2201\220H\230A\230Q\320\0045\260Q\330\010\014\210F\220!\320\004#\320#5\260_\300A\330\010\r\210R\210y\230\001\330\014\r\330\014\030\230\016\240a\330\014\030\230\001\330\014\025\220Q\330\014\024\220A\320\004)\320)9\3209J\310!\330 2\260)\2701\360\010\000\t#\240+\250Q\340\010-\320-C\3001\330\010\033\230?\250!\2506\260\021\340\010\037\230q\330\010\013\2108\2207\230!\330\014\026\220i\230q\240\t\250\021\330\014\030\230\001\340\010!\240\021\330\010\013\210:\220W\230A\330\014\030\230\t\240\021\240+\250Q\330\014\032\230!\340\010\"\240!\330\010\013\210;\220g\230Q\330\014\031\230\031\240!\240<\250q\330\014\033\2301\340\010#\2401\330\010\013\210<\220w\230a\330\014\032\230)\2401\240M\260\021\330\014\034\230A\340\r\016\330\014\025\320\025-\250Q\330\020\021\220\024\220Q\330\020\021\330\020\021\330\020\021\330\020\021\330\020\021\330\020\021\330\020\021\220\026\220q\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\340\010\017\210q\320\0049\270\021\340\010\"\240+\250Q\330\010%\320%6\260a\330\010\035\230S\240\001\240\021\330\r\016\330\014\025\320\0252\260!\330\020\021\220\024\220\\\240\030\250\030\260\021\260!\330\010\023\2201\220H\230A\230Q\200A\360\016\000\t\014\2104\210v\220Y\230c\240\021\330\014\020\220\006\220h\230a\230q\240\004\240A\330\014\020\220\006\220k\240\021\200A\360\016\000\t\014\2104\210w\220i\230s\240!\330\014\020\220\007\220x\230q\240\001\240\024\240Q\330\014\020\220\007\220{\240!\200A\340\020\021\330\022\023\330\023\024\330\024\025\330\t\n\3606\000\t#\240+\250Q\340\010-\320-C\3001\340\010\037\230q\330\010\013\2108\2207\230!\330\014\026\220i\230q\240\t\250\021\330\014\030\230\001\340\010!\240\021\330\010\013\210:\220W\230A\330\014\030\230\t\240\021\240+\250Q\330\014\032""\230!\340\010\"\240!\330\010\013\210;\220g\230Q\330\014\031\230\031\240!\240<\250q\330\014\033\2301\340\010\"\240%\320'8\270\001\340\r\016\330\014\025\320\0250\260\001\330\020\021\220\024\220Q\330\020\021\330\020\021\330\020\021\330\020\021\330\020\021\220\026\220q\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004$\240A\360\010\000\t(\320'7\260v\270Q\270a\330\010\021\220\033\230A\330\010\022\220-\230q\330\014\023\220;\320\036.\250a\250q\340\010\016\210a\210{\230!\2304\230y\250\001\330\010\014\210G\220;\230a\330\010\017\210q\200A\330\014\r\360 \000\t#\240+\250Q\330\010(\320(9\270\021\330\010;\2701\330\014\017\210v\220V\2306\240\021\330\010%\240Q\340\r\016\330\014\025\320\0253\2601\330\020\021\220\024\220Q\330\020\021\220\026\220q\330\020\021\220\021\330\020\021\220\021\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\340\010\025\220Q\330\010\014\210E\220\025\220a\220|\2401\330\014\025\220\\\320!3\2601\260A\330\014\023\220>\240\034\250[\270\001\270\021\330\014\026\220g\230Q\320\0367\260q\270\006\270a\330\010\024\220H\230A\230Q\230a\340\010\013\2106\220\027\230\t\240\023\240A\330\014\023\220<\230v\240Q\330\010\017\210|\2308\2401\200A\340\r\016\340\021\022\330\017\020\330\t\n\360\036\000\t#\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\034\230A\330\010\033\2301\340\010\016\210i\220q\230\001\330\010\t\330\014\024\220C\220q\230\001\330\014\024\220A\330\021\022\330\020\031\320\0310\260\001\330\024\025\220T\230\035\240g\250W\260A\260W\270A\270Q\330\014\027\220q\230\010\240\001\240\021\330\014\017\210v\220S\230\003\2301\230A\340\020\021\360\006\000\021\033\230#\230Q\230e\2402\240Q\330\020\023\2208\2302\230Q\330\024\032\230,\240a\330\0300\260\001\260\021\330\030/\250q\260\003\2601\260A\330\020\026\220i\230q\240\001\360\006\000\t\014\2106\220\022\2201\330\014\025\220Q\330\010\017\210s\220\"\220F\230'\240\021\240*\250A\200A\340\r\016\340\021\022\330\017\020\330\t\n\360\036\000\t#\240+\250Q\340\010\024\220I\230Q\230e\2401\330""\010\033\2301\330\010\034\230A\330\010\033\2301\340\010\016\210i\220q\230\001\330\010\t\330\014\024\220C\220q\230\001\330\014\024\220A\330\021\022\330\020\031\320\031.\250a\330\024\025\220T\230\033\240G\2507\260!\2607\270!\2701\330\014\027\220q\230\010\240\001\240\021\330\014\017\210v\220S\230\003\2301\230A\340\020\021\360\006\000\021\033\230#\230Q\230e\2402\240Q\330\020\023\2208\2302\230Q\330\024\032\230,\240a\330\0300\260\001\260\021\330\030/\250q\260\003\2601\260A\330\020\026\220i\230q\240\001\360\006\000\t\014\2106\220\022\2201\330\014\025\220Q\330\010\017\210s\220\"\220F\230'\240\021\240*\250A\200A\340\r\016\340\021\022\330\017\020\330\t\n\360\036\000\t#\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\034\230A\330\010\033\2301\340\010\016\210i\220q\230\001\330\010\t\330\014\024\220C\220q\230\001\330\014\024\220A\330\021\022\330\020\031\320\031/\250q\330\024\025\220T\230\034\240W\250G\2601\260G\2701\270A\330\014\027\220q\230\010\240\001\240\021\330\014\017\210v\220S\230\003\2301\230A\340\020\021\360\006\000\021\033\230#\230Q\230e\2402\240Q\330\020\023\2208\2302\230Q\330\024\032\230,\240a\330\0300\260\001\260\021\330\030/\250q\260\003\2601\260A\330\020\026\220i\230q\240\001\360\006\000\t\014\2106\220\022\2201\330\014\025\220Q\330\010\017\210s\220\"\220F\230'\240\021\240*\250A\200A\360\n\000\t\n\330\010\t\330\010\t\340\010\r\210R\210y\230\001\230\021\330\010\014\210O\230>\250\021\250!\330\010\014\210O\2301\330\010\014\210L\230\001\330\010\014\210K\220x\230s\240!\320\000@\300\001\340\004\t\210\033\220A\340\004\016\210i\220q\230\t\240\021\330\004\t\210\033\220A\340\004\007\200q\330\010\r\210_\230A\340\010\r\210_\230A\340\004\007\200q\330\010\023\2208\2307\240!\2401\340\010\023\2201\330\004\010\210\005\210U\220!\2201\330\010\r\210Y\220a\220u\230H\240A\240Q\340\004\013\210;\220a\220~\240Q\240n\260A\260Q\320\004$\240A\340\010\037\230q\340\010\013\2104\210t\2201\330\014\022\220*\230A\230Q\340\010'\320'7\260v\270Q\270a\330\010\021\220\033\230A\330\010""\022\220-\230q\330\014\023\220;\320\036.\250a\250q\340\010\r\210T\220\027\230\013\2401\240A\240T\250\031\260!\330\010\013\2103\210c\220\021\330\014\022\220$\220g\230_\250A\250Q\250d\260!\330\014\017\210t\2203\220a\330\020\026\220l\240!\320#<\270A\270Q\340\020\024\220C\220w\230a\330\020\026\220l\240!\320#<\270A\270U\300!\3001\330\010\017\210q\200A\330\010\017\210t\2202\220X\230Q\320\004B\300!\360\022\000\t#\240+\250Q\340\010(\320(9\270\021\330\010\025\220Y\230a\230|\2501\330\010\"\240!\340\010\037\230q\330\010\013\2108\2207\230!\330\014\026\220i\230q\240\t\250\021\330\014\030\230\001\340\010!\240\021\330\010\013\210:\220W\230A\330\014\030\230\t\240\021\240+\250Q\330\014\032\230!\340\r\016\330\014\025\320\0251\260\021\330\020\021\220\024\220Q\330\020\021\330\020\021\330\020\021\330\020\021\220\026\220q\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\330\010\017\210q\200\001\340\004\014\210E\220\021\320\022'\240s\250'\260\021\260!\330\004\007\200q\330\010\020\220\005\220Q\330\014?\270q\330\014\r\320\004$\240H\250A\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\034\230A\330\010\036\230a\340\010\016\210i\220q\230\001\330\010\t\330\014\024\220C\220q\230\001\330\014\024\220A\330\021\022\330\020\031\320\0313\2601\330\024\025\220T\230\033\240G\2507\260!\2607\270!\2701\330\014\027\220q\230\010\240\001\240\021\330\014\017\210v\220S\230\003\2301\230A\340\020\021\360\006\000\021\033\230#\230Q\230e\2402\240Q\330\020\023\2208\2302\230Q\330\024\032\230,\240a\330\0301\260\021\330\030/\250q\260\003\2601\260A\330\020\026\220i\230q\240\001\340\010\017\210u\220A\220S\230\002\230!\320\004$\240H\250A\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\034\230A\330\010\036\230a\340\010\016\210i\220q\230\001\330\010\t\330\014\024\220C\220q\230\001\330\014\024\220A\330\021\022\330\020\031\320\0314\260A\330\024\025\220T\230\034\240W\250G\2601\260G\2701\270A\330\014\027\220q\230\010\240\001\240\021\330\014\017\210v\220S\230\003\2301""\230A\340\020\021\360\006\000\021\033\230#\230Q\230e\2402\240Q\330\020\023\2208\2302\230Q\330\024\032\230,\240a\330\0301\260\021\330\030/\250q\260\003\2601\260A\330\020\026\220i\230q\240\001\340\010\017\210u\220A\220S\230\002\230!\320\004$\240H\250A\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\034\230A\330\010\036\230a\340\010\016\210i\220q\230\001\330\010\t\330\014\024\220C\220q\230\001\330\014\024\220A\330\021\022\330\020\031\320\0315\260Q\330\024\025\220T\230\035\240g\250W\260A\260W\270A\270Q\330\014\027\220q\230\010\240\001\240\021\330\014\017\210v\220S\230\003\2301\230A\340\020\021\360\006\000\021\033\230#\230Q\230e\2402\240Q\330\020\023\2208\2302\230Q\330\024\032\230,\240a\330\0301\260\021\330\030/\250q\260\003\2601\260A\330\020\026\220i\230q\240\001\340\010\017\210u\220A\220S\230\002\230!\320\004$\240H\250A\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\036\230a\330\r\016\330\014\025\320\0250\260\001\330\020\021\220\024\220[\240\007\240q\250\t\260\021\260!\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004$\240H\250A\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\036\230a\330\r\016\330\014\025\320\0251\260\021\330\020\021\220\024\220\\\240\027\250\001\250\031\260!\2601\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004$\240H\250A\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\036\230a\330\r\016\330\014\025\320\0252\260!\330\020\021\220\024\220]\240'\250\021\250)\2601\260A\330\010\023\2201\220H\230A\230Q\330\010\017\210q\200\001\330\004\n\210+\220Q\320\004\032\230!\340\010\"\240+\250Q\340\r\016\330\014\025\320\025+\2501\250A\250T\260\035\270a\270q\330\010\023\2201\220H\230A\230Q\320\004\031\230\021\340\010\"\240+\250Q\330\r\016\330\014\025\320\025)\250\021\250!\2504\250|\2701\270A\330\010\023\2201\220H\230A\230Q\320\004\030\230\001\340\010\"\240+\250Q\340\r\016\330\014\025\320\025)\250\021\250!\2504\250}\270A\270Q\330\010\023""\2201\220H\230A\230Q\320\004 \240\001\360\022\000\t#\240+\250Q\330\010(\320(9\270\021\330\r\016\330\014\025\320\025/\250q\330\020\021\220\024\220Q\330\020\021\220\026\220q\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\000%\240Q\360\006\000\005\014\320\013\034\230A\230U\240&\250\003\2501\320\004!\240\021\360\010\000\t#\240+\250Q\340\010-\320-C\3001\340\r\016\330\014\025\320\0250\260\001\330\020\021\220\024\220]\240!\2406\250\031\260!\2601\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004:\270!\340\010\"\240+\250Q\340\010-\320-C\3001\330\010#\320#4\260A\330\010\035\230S\240\001\240\021\340\r\016\330\014\025\320\0250\260\001\330\020\021\220\024\220Q\330\020\021\330\020\021\330\020\021\220\026\220q\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004 \240\t\250\021\360\022\000\t#\240+\250Q\360\010\000\t\r\210G\2201\220F\320\032/\250q\330\020\024\220J\230a\230v\240Q\330\014\017\210w\220g\230Q\330\020\026\220j\240\001\330\024\025\360\006\000\r\025\220G\2304\320\0371\260\021\340\010\013\210<\220q\230\006\230a\330\014\026\220o\320%9\270\021\270&\300\001\330\r\027\220q\230\006\230a\330\014\026\220b\320\030+\2505\260\001\330\r\027\220q\230\006\230a\330\014\026\220o\240Q\340\014\022\220)\2301\330\020\021\330GH\310\004\310A\310Q\360\006\000\t\014\210<\220q\230\010\240\001\330\014\027\320\027'\320';\2701\270H\300A\330\r\027\220q\230\010\240\001\330\014\027\220r\320\031-\250W\260A\330\r\027\220q\230\010\240\001\330\014\027\320\027'\240q\340\014\022\220)\2301\230A\330\"#\2404\240q\250\001\340\r\016\330\014\025\320\025&\240a\330\020\021\220\024\220Q\330\020\021\330\020\021\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\320\004\"\240(\250!\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\037\230q\330\r\016\330\014\025\320\025-\250Q\330\020\021\220\024\220[\240\007\240q\250\t\260\021\260!\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004\"\240(\250!\340\010\"\240+\250Q\340\010\024\220I""\230Q\230e\2401\330\010\033\2301\330\010\037\230q\330\r\016\330\014\025\320\025.\250a\330\020\021\220\024\220\\\240\027\250\001\250\031\260!\2601\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004\"\240(\250!\340\010\"\240+\250Q\340\010\024\220I\230Q\230e\2401\330\010\033\2301\330\010\037\230q\330\r\016\330\014\025\320\025/\250q\330\020\021\220\024\220]\240'\250\021\250)\2601\260A\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004%\240Q\360\020\000\t#\240+\250Q\360\006\000\t\n\330\014\023\2201\220H\230A\330\014\020\220\004\220J\230a\230x\240q\340\014\025\220V\320\033.\250a\340\010\013\210<\220q\230\010\240\001\330\014\027\320\027,\320,@\300\001\330\020\030\230\001\340\r\027\220q\230\010\240\001\330\014\027\220r\320\0312\260'\270\021\330\r\027\220q\230\010\240\001\330\014\027\320\027,\250A\340\014\022\220)\2301\230A\330\"#\2404\240q\250\001\340\r\016\330\014\025\320\025,\250A\330\020\021\220\024\220Q\330\020\021\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\320\004%\240Q\360\036\000\t#\240+\250Q\340\010-\320-C\3001\340\r\016\330\014\025\320\0254\260A\330\020\021\220\024\220Q\330\020\021\220\026\220q\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004 \240\001\360\022\000\t#\240+\250Q\330\010%\240Q\330\r\016\330\014\025\320\025.\250a\330\020\021\220\024\220Q\330\020\021\330\020\021\220\021\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\330\010\017\210q\320\004\030\230\001\340\010\"\240+\250Q\340\r\016\330\014\025\320\025(\250\001\250\021\250$\250l\270!\2701\330\010\023\2201\220H\230A\230Q\200\001\360\010\000\005\016\210T\220\036\230t\2408\2504\250q\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\230=\250\007\250u\260C\260t\2707\300'\310\021\330\004\007\200q\330\010\017\320\017,\250D\260\001\260\027\270\013\3007\310!\340\010\017\320\017,\250D\260\001\260\027\270\013\3001\200\001\330\004)\250\021\250&\260\001\320\004.\250a\340\010\"\240+\250Q""\330\010\013\2101\330\014\024\220A\340\014\024\220A\340\r\016\330\014\025\320\025,\250A\330\020\021\220\024\220Q\330\020\021\330\020\021\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\320\004&\240a\360,\000\t#\240+\250Q\340\010(\320(9\270\021\340\r\016\330\014\025\320\0254\260A\330\020\021\220\024\220\\\240\021\240&\250\t\260\021\260!\330\010\023\2201\220H\230A\230Q\330\010\017\210q\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220;\230h\240a\240q\330\004\007\200|\2207\230!\330\010-\250Q\250n\270N\310!\330\004\013\2101\320\004'\240q\360\026\000\t#\240+\250Q\330\010\033\2301\330\010\035\230Q\360\010\000\t\r\210E\220\031\230&\240\006\240a\330\014\022\220)\2301\230E\240\021\330\014\024\220A\340\014\017\210v\220S\230\001\330\020\032\230!\330\025\026\330\024\035\320\0353\2601\330\030\031\230\024\230\\\250\027\260\t\270\021\270!\330\021\033\2301\230G\2401\330\020\030\230\t\240\021\240'\250\021\330\020\032\230!\330\025\026\330\024\035\320\0353\2601\330\030\031\230\024\230\\\250\027\260\t\270\021\270!\330\021\033\2301\230G\2401\330\020\023\2201\330\024\034\230A\340\024\034\230A\330\020\030\230\t\240\021\240'\250\021\330\020\032\230!\330\025\026\330\024\035\320\0353\2601\330\030\031\230\024\230\\\250\027\260\t\270\021\270!\330\021\033\2301\230G\2401\330\020\032\230!\330\020\036\230c\240\021\240!\330\025\026\330\024\035\320\0358\270\001\330\030\031\230\024\230\\\250\027\3200A\300\021\330\030%\240Q\240a\330\021\033\2301\230G\2401\330\020\033\2301\330\025\026\330\024\035\320\0359\270\021\330\030\031\230\024\230\\\250\027\260\n\270!\2701\330\021\033\2301\230G\2401\330\020\030\230\001\330\025\026\330\024\035\320\0356\260a\330\030\031\230\024\230\\\250\027\260\007\260q\270\001\340\020\026\220j\240\001\330\024'\240q\250\004\250A\320-?\270q\300\001\330 !\240\021\340\014\027\220q\230\010\240\001\240\021\320\004'\240q\360\030\000\t#\240+\250Q\330\010\033\2301\330\010\035\230Q\360\010\000\t\r\210E\220\031\230&\240\006\240a\330\014\022\220)\2301\230E\240\021""\330\014\024\220A\340\014\017\210v\220S\230\001\330\020\032\230!\330\025\026\330\024\035\320\0354\260A\330\030\031\230\024\230]\250'\260\031\270!\2701\330\021\033\2301\230G\2401\330\020\030\230\t\240\021\240'\250\021\330\020\032\230!\330\025\026\330\024\035\320\0354\260A\330\030\031\230\024\230]\250'\260\031\270!\2701\330\021\033\2301\230G\2401\330\020\023\2201\330\024\034\230A\340\024\034\230A\330\020\030\230\t\240\021\240'\250\021\330\020\032\230!\330\025\026\330\024\035\320\0354\260A\330\030\031\230\024\230]\250'\260\031\270!\2701\330\021\033\2301\230G\2401\330\020\032\230!\330\020\036\230c\240\021\240!\330\025\026\330\024\035\320\0359\270\021\330\030\031\230\024\230]\250'\3201B\300!\330\030%\240Q\240a\330\021\033\2301\230G\2401\330\020\033\2301\330\025\026\330\024\035\320\035:\270!\330\030\031\230\024\230]\250'\260\032\2701\270A\330\021\033\2301\230G\2401\330\020\030\230\001\330\025\026\330\024\035\320\0357\260q\330\030\031\230\024\230]\250'\260\027\270\001\270\021\340\020\026\220j\240\001\330\024'\240q\250\004\250A\320-?\270q\300\001\330 !\240\021\340\014\027\220q\230\010\240\001\240\021\320\004'\240q\360\030\000\t#\240+\250Q\330\010\033\2301\330\010\035\230Q\330\010\"\240!\360\006\000\t\r\210E\220\031\230&\240\006\240a\330\014\022\220)\2301\230E\240\021\330\014\024\220A\340\014\017\210v\220S\230\001\330\020\032\230!\330\020\031\320\031.\250a\330\024\025\220T\230\033\240G\2509\260A\260Q\330\021\033\2301\230G\2401\330\020\030\230\t\240\021\240'\250\021\330\020\032\230!\330\020\031\320\031.\250a\330\024\025\220T\230\033\240G\2509\260A\260Q\330\021\033\2301\230G\2401\330\020\023\2201\330\024\034\230A\340\024\034\230A\330\020\030\230\t\240\021\240'\250\021\330\020\032\230!\330\020\031\320\031.\250a\330\024\025\220T\230\033\240G\2509\260A\260Q\330\021\033\2301\230G\2401\330\020\032\230!\330\020\036\230c\240\021\240!\330\025\026\330\024\035\320\0357\260q\330\030\031\230\024\230[\250\007\320/@\300\001\330\030%\240Q\240a\330\021\033\2301\230G\2401\330\020\033\2301\330""\025\026\330\024\035\320\0358\270\001\330\030\031\230\024\230[\250\007\250z\270\021\270!\330\021\033\2301\230G\2401\330\020\030\230\001\330\025\026\330\024\035\320\0355\260Q\330\030\031\230\024\230[\250\007\250w\260a\260q\340\020\026\220j\240\001\330\024'\240q\250\004\250A\320-?\270q\300\001\330 !\240\021\340\014\027\220q\230\010\240\001\240\021\320\004\037\230q\360\026\000\t#\240+\250Q\330\010-\320-C\3001\330\010%\240Q\330\r\016\330\014\025\320\025.\250a\330\020\021\220\024\220Q\330\020\021\220\026\220q\330\020\021\220\021\330\020\021\220\021\330\010\023\2201\220H\230A\230Q\330\010\020\220\010\230\001\320\004\027\220q\340\010\"\240+\250Q\340\010\014\210K\220}\240A\330\r\021\220\021\330\014\017\210t\220:\230^\2503\250a\330\020\021\340\021\022\330\020\031\320\031-\250Q\250a\250t\260<\270q\300\001\330\014\027\220q\230\010\240\001\240\021\320\004\027\220q\340\010\"\240+\250Q\340\010\014\320\014!\240\021\330\r\021\220\021\330\014\017\210t\2209\230N\250#\250Q\330\020\021\340\021\022\330\020\031\320\031,\250A\250Q\250d\260+\270Q\270a\330\014\027\220q\230\010\240\001\240\021\320\004\027\220q\340\010\"\240+\250Q\340\010\014\320\014!\240\021\330\r\021\220\021\330\014\017\210t\220;\230n\250C\250q\330\020\021\340\021\022\330\020\031\320\031.\250a\250q\260\004\260M\300\021\300!\330\014\027\220q\230\010\240\001\240\021\330\014\020\220\t\230\035\240a\200\001\360\016\000\005\006\340\004\007\200y\220\017\230s\240'\250\031\260,\270a\330\010\017\210t\2202\220X\230Q\340\004\027\220q\340\t\n\330\010\013\2101\330\014\034\230A\360\010\000\005\r\210E\220\021\330\010\034\230A\230Q\320\0365\260W\270A\330\005\014\210A\210Q\330\004\007\200q\330\010\020\220\005\220Q\330\014-\250Q\330\014\r\360\006\000\005\006\330\010\017\210t\2202\220X\230Q\330\004\013\320\013\034\230A\330\r\016\330\014\017\210q\330\020\026\220a\330\020 \240\001\330\020\026\220g\230S\240\001\240\022\240?\260!\2603\260a\260q\330\010\016\210a\360\006\000\t\021\220\005\220Q\320\026,\250C\250w\260a\260q\330\010\013\2101\330\014""\024\220E\230\021\330\0205\260Q\330\020\021\340\r\016\330\014\017\210q\330\020\026\220a\330\020 \240\001\330\020\026\220c\230\021\230\"\230O\2501\250C\250q\260\t\270\021arrow_arrayarrow_array_streamarrow_schema";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 397; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 67) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 397; i < 460; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 460; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 397;
      for (Py_ssize_t i=0; i<63; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int16_t const cint_constants_2[] = {1024};
    int32_t const cint_constants_4[] = {86393024L,134217728L};
    for (int i = 0; i < 3; i++) {
      numbertab[i] = PyLong_FromLong((i < 1 ? cint_constants_2[i - 0] : cint_constants_4[i - 1]));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<3; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 3;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 3;
    unsigned int nlocals : 4;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 202};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_genexpr, __pyx_mstate->__pyx_kp_b_iso88591_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 4, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 101};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_message, __pyx_mstate->__pyx_n_u_status_code, __pyx_mstate->__pyx_n_u_vendor_code, __pyx_mstate->__pyx_n_u_sqlstate, __pyx_mstate->__pyx_n_u_details};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_Ry_O_O1_L_Kxs, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 3, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 148};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_message, __pyx_mstate->__pyx_n_u_vendor_code, __pyx_mstate->__pyx_n_u_sqlstate, __pyx_mstate->__pyx_n_u_details};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_5_A_Ry_a_Q_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 281};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_status_code, __pyx_mstate->__pyx_n_u_message, __pyx_mstate->__pyx_n_u_vendor_code, __pyx_mstate->__pyx_n_u_sqlstate, __pyx_mstate->__pyx_n_u_error, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_test_error, __pyx_mstate->__pyx_kp_b_iso88591_A_iq_A_q__A__A_q_87_1_1_U_1_Yau, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 317};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_enter, __pyx_mstate->__pyx_kp_b_iso88591_1_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 320};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_exc_type, __pyx_mstate->__pyx_n_u_exc_val, __pyx_mstate->__pyx_n_u_exc_tb};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_exit, __pyx_mstate->__pyx_kp_b_iso88591_5Q_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_t84q_G1F_a_vWE_Q_q_t_uCt7_q_D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591__7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 342};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_obj, __pyx_mstate->__pyx_n_u_name_2};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_is_pycapsule, __pyx_mstate->__pyx_kp_b_iso88591_Q_AU_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 388};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_release, __pyx_mstate->__pyx_kp_b_iso88591_A_4wis_xq_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 399};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_allocated, __pyx_mstate->__pyx_n_u_capsule};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_arrow_c_schema, __pyx_mstate->__pyx_kp_b_iso88591_A_7vQa_A_q_aq_a_4y_G_a_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 430};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_release, __pyx_mstate->__pyx_kp_b_iso88591_A_4vYc_haq_A_k, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 471};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_release, __pyx_mstate->__pyx_kp_b_iso88591_A_4wis_xq_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 482};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_err, __pyx_mstate->__pyx_n_u_allocated, __pyx_mstate->__pyx_n_u_capsule, __pyx_mstate->__pyx_n_u_rc, __pyx_mstate->__pyx_n_u_s};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_arrow_c_schema, __pyx_mstate->__pyx_kp_b_iso88591_A_q_4t1_AQ_7vQa_A_q_aq_T_1AT_3c, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 504};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_requested_schema, __pyx_mstate->__pyx_n_u_allocated, __pyx_mstate->__pyx_n_u_capsule};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_arrow_c_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_1_G1_Qa_q_A_q_4AQ_a_4y_G_a_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 653};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_q_Q_t9N_Q_AQd_Qa_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 2, 12, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 666};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_encoding, __pyx_mstate->__pyx_n_u_errors, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_len, __pyx_mstate->__pyx_n_u_c_buf, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_new_len};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_IQe1_1_A_1_iq_Cq_A_a_T_G7_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 719};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_len, __pyx_mstate->__pyx_n_u_c_buf, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_new_len};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_bytes, __pyx_mstate->__pyx_kp_b_iso88591_HA_Q_IQe1_1_A_a_iq_Cq_A_31_T_G7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 750};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_float, __pyx_mstate->__pyx_kp_b_iso88591_HA_Q_IQe1_1_a_0_q_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 763};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_int, __pyx_mstate->__pyx_kp_b_iso88591_Q_IQe1_1_q_Q_q_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 11, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 776};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value, __pyx_mstate->__pyx_n_u_c_value_len, __pyx_mstate->__pyx_n_u_c_double, __pyx_mstate->__pyx_n_u_c_int, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_set_options, __pyx_mstate->__pyx_kp_b_iso88591_q_Q_1_Q_E_a_1E_A_vS_a_T_G9AQ_1G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 938};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_cancel, __pyx_mstate->__pyx_kp_b_iso88591_Q_4_AQ_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 946};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_commit, __pyx_mstate->__pyx_kp_b_iso88591_Q_4_AQ_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 954};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_info_codes, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_stream, __pyx_mstate->__pyx_n_u_c_info_codes, __pyx_mstate->__pyx_n_u_info_code};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_info, __pyx_mstate->__pyx_kp_b_iso88591_1_Q_C1_1_Q_Qk_1_9A_a_T_Q_Q_V1_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {7, 0, 0, 15, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 989};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_depth, __pyx_mstate->__pyx_n_u_catalog, __pyx_mstate->__pyx_n_u_db_schema, __pyx_mstate->__pyx_n_u_table_name, __pyx_mstate->__pyx_n_u_table_types, __pyx_mstate->__pyx_n_u_column_name, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_stream, __pyx_mstate->__pyx_n_u_c_depth, __pyx_mstate->__pyx_n_u_c_catalog, __pyx_mstate->__pyx_n_u_c_db_schema, __pyx_mstate->__pyx_n_u_c_table_name, __pyx_mstate->__pyx_n_u_c_column_name};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_objects, __pyx_mstate->__pyx_kp_b_iso88591_99J_2_1_Q_C1_6_q_87_iq_WA_Q_gQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 2, 12, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1034};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_encoding, __pyx_mstate->__pyx_n_u_errors, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_len, __pyx_mstate->__pyx_n_u_c_buf, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_new_len};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_IQe1_1_A_1_iq_Cq_A_0_T_gWAWA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1087};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_len, __pyx_mstate->__pyx_n_u_c_buf, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_new_len};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_bytes, __pyx_mstate->__pyx_kp_b_iso88591_HA_Q_IQe1_1_A_a_iq_Cq_A_5Q_T_gW, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1118};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_float, __pyx_mstate->__pyx_kp_b_iso88591_HA_Q_IQe1_1_a_2_1A_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1131};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_int, __pyx_mstate->__pyx_kp_b_iso88591_Q_IQe1_1_q_q_1A_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 12, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1144};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_catalog, __pyx_mstate->__pyx_n_u_db_schema, __pyx_mstate->__pyx_n_u_table_name, __pyx_mstate->__pyx_n_u_approximate, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_stream, __pyx_mstate->__pyx_n_u_c_catalog, __pyx_mstate->__pyx_n_u_c_db_schema, __pyx_mstate->__pyx_n_u_c_table_name, __pyx_mstate->__pyx_n_u_c_approximate};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_statistics, __pyx_mstate->__pyx_kp_b_iso88591_A_6_Q_C1_q_87_iq_WA_Q_gQ_q_1_8_0, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1210};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_statistic_names, __pyx_mstate->__pyx_kp_b_iso88591_Q_Q_C1_4A_Q_q_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1237};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_catalog, __pyx_mstate->__pyx_n_u_db_schema, __pyx_mstate->__pyx_n_u_table_name, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_handle, __pyx_mstate->__pyx_n_u_c_table_name, __pyx_mstate->__pyx_n_u_c_catalog, __pyx_mstate->__pyx_n_u_c_db_schema};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_table_schema, __pyx_mstate->__pyx_kp_b_iso88591_B_Q_9_Ya_1_q_87_iq_WA_Q_1_Q_q_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1273};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_table_types, __pyx_mstate->__pyx_kp_b_iso88591_Q_C1_0_6_1_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1287};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_partition, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_stream, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_length};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_read_partition, __pyx_mstate->__pyx_kp_b_iso88591_Q_C1_4A_S_0_Q_q_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1305};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_rollback, __pyx_mstate->__pyx_kp_b_iso88591_Q_1AT_aq_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1313};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_enabled, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_set_autocommit, __pyx_mstate->__pyx_kp_b_iso88591_a_Q_1_A_A_A_Q_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 11, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 1329};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value, __pyx_mstate->__pyx_n_u_c_value_len, __pyx_mstate->__pyx_n_u_c_double, __pyx_mstate->__pyx_n_u_c_int, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_set_options, __pyx_mstate->__pyx_kp_b_iso88591_q_Q_1_Q_E_a_1E_A_vS_4A_1_1G1_4A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1396};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_q_Q_t_nCq_aq_M_q_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1442};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_schema, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_c_array, __pyx_mstate->__pyx_n_u_c_schema, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_bind, __pyx_mstate->__pyx_kp_b_iso88591_Q_G1F_q_JavQ_wgQ_j_G4_1_q_a_o_9, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1494};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_stream, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_c_stream, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_bind_stream, __pyx_mstate->__pyx_kp_b_iso88591_Q_Q_1HA_Jaxq_V_a_q_q_r_2_q_A_1A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1530};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_cancel, __pyx_mstate->__pyx_kp_b_iso88591_Q_l_1_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1538};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_q_Q_K_A_t_3a_Qat_q_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1551};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_stream, __pyx_mstate->__pyx_n_u_rows_affected, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[52] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_execute_query, __pyx_mstate->__pyx_kp_b_iso88591_q_Q_C1_Q_a_Q_q_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[52])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1574};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_schema, __pyx_mstate->__pyx_n_u_c_partitions, __pyx_mstate->__pyx_n_u_rows_affected, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_partitions, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_data};
    __pyx_mstate_global->__pyx_codeobj_tab[53] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_execute_partitions, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_9_1_vV6_Q_31_Q_q_1HAQ_Q_E_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[53])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1617};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_schema, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[54] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_execute_schema, __pyx_mstate->__pyx_kp_b_iso88591_Q_9_q_Q_q_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[54])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1636};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_rows_affected, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[55] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_execute_update, __pyx_mstate->__pyx_kp_b_iso88591_Q_Q_a_Q_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[55])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 2, 12, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1656};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_encoding, __pyx_mstate->__pyx_n_u_errors, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_len, __pyx_mstate->__pyx_n_u_c_buf, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_new_len};
    __pyx_mstate_global->__pyx_codeobj_tab[56] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_IQe1_1_A_1_iq_Cq_A_q_T_WG1G1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[56])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1709};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_len, __pyx_mstate->__pyx_n_u_c_buf, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_new_len};
    __pyx_mstate_global->__pyx_codeobj_tab[57] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_bytes, __pyx_mstate->__pyx_kp_b_iso88591_HA_Q_IQe1_1_A_a_iq_Cq_A_4A_T_WG, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[57])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1740};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value};
    __pyx_mstate_global->__pyx_codeobj_tab[58] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_float, __pyx_mstate->__pyx_kp_b_iso88591_HA_Q_IQe1_1_a_1_1_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[58])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1753};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_key_bytes, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value};
    __pyx_mstate_global->__pyx_codeobj_tab[59] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_option_int, __pyx_mstate->__pyx_kp_b_iso88591_Q_IQe1_1_q_a_1_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[59])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1766};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_handle};
    __pyx_mstate_global->__pyx_codeobj_tab[60] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_get_parameter_schema, __pyx_mstate->__pyx_kp_b_iso88591_a_Q_9_4A_1HAQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[60])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1798};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[61] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_prepare_2, __pyx_mstate->__pyx_kp_b_iso88591_Q_4_1A_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[61])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 11, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 1805};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_c_key, __pyx_mstate->__pyx_n_u_c_value, __pyx_mstate->__pyx_n_u_c_value_len, __pyx_mstate->__pyx_n_u_c_double, __pyx_mstate->__pyx_n_u_c_int, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[62] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_set_options, __pyx_mstate->__pyx_kp_b_iso88591_q_Q_1_Q_E_a_1E_A_vS_31_1G1_31_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[62])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1871};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_query, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_query_data, __pyx_mstate->__pyx_n_u_c_query, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[63] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_set_sql_query, __pyx_mstate->__pyx_kp_b_iso88591_31_Q_WAQ_1_Q_1_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[63])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1881};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_plan, __pyx_mstate->__pyx_n_u_c_error, __pyx_mstate->__pyx_n_u_c_plan, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[64] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_set_substrait_plan, __pyx_mstate->__pyx_kp_b_iso88591_9_Q_6a_S_2_1HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[64])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[65] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[65])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[66] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[66])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1905};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_error};
    __pyx_mstate_global->__pyx_codeobj_tab[67] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_init_blocking_call, __pyx_mstate->__pyx_kp_b_iso88591_E_s_q_Q_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[67])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1919};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_func, __pyx_mstate->__pyx_n_u_args, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_cancel, __pyx_mstate->__pyx_n_u_error, __pyx_mstate->__pyx_n_u_e, __pyx_mstate->__pyx_n_u_exc};
    __pyx_mstate_global->__pyx_codeobj_tab[68] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_blocking_call_impl, __pyx_mstate->__pyx_kp_b_iso88591_y_s_a_t2XQ_q_1_A_E_AQ_5WA_AQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[68])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1976};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_func, __pyx_mstate->__pyx_n_u_args, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_cancel};
    __pyx_mstate_global->__pyx_codeobj_tab[69] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_adbc_driver_manager__lib_pyx, __pyx_mstate->__pyx_n_u_blocking_call, __pyx_mstate->__pyx_kp_b_iso88591_A_t2XQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[69])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[70] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle__AdbcHandle, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_haq_7_QnN_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[70])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* Generator.init */
  if (likely(__pyx_Generator_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseKeywordRequired */
static void __Pyx_RaiseKeywordRequired(const char* func_name, PyObject* kw_name) {
    PyErr_Format(PyExc_TypeError,
        "%s() needs keyword-only argument %U", func_name, kw_name);
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyObjectVectorCallKwBuilder (used by PyObjectVectorCallMethodKwBuilder) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* GetException (used by pep479) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc, *new_exc;
    __Pyx_PyThreadState_declare
    int is_async_stopiteration = 0;
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    __Pyx_PyThreadState_assign
    cur_exc = __Pyx_PyErr_CurrentExceptionType();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else {
            return;
        }
    }
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(tb);
    new_exc = PyObject_CallFunction(PyExc_RuntimeError, "s",
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        "generator raised StopIteration");
    if (!new_exc) {
        Py_XDECREF(val);
        return;
    }
    PyException_SetCause(new_exc, val); // steals ref to val
    PyErr_SetObject(PyExc_RuntimeError, new_exc);
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* decode_c_bytes (used by decode_bytes) */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_mstate_global->__pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON
static int
__Pyx_unicode_modifiable(PyObject *unicode, int unsafe_shared)
{
    if (!__Pyx_IS_UNIQUELY_REFERENCED(unicode, unsafe_shared))
        return 0;
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030F0000
    if (PyUnstable_Unicode_GET_CACHED_HASH(unicode) != -1)
        return 0;
#endif
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left, unsafe_shared)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* PyObjectLookupSpecial */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* CIntToDigits (used by CIntToPyUnicode) */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp = PySequence_Repeat(padding, uoffset - prepend_sign);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyUnicode_Concat(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyUnicode_Concat(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* COrdinalToPyUnicode (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value) {
    return value <= 1114111;
}
static PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t ulength, char padding_char) {
    Py_ssize_t padding_length = ulength - 1;
    if (likely((padding_length <= 250) && (value < 0xD800 || value > 0xDFFF))) {
        char chars[256];
        if (value <= 255) {
            memset(chars, padding_char, (size_t) padding_length);
            chars[ulength-1] = (char) value;
            return PyUnicode_DecodeLatin1(chars, ulength, NULL);
        }
        char *cpos = chars + sizeof(chars);
        if (value < 0x800) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xc0 | (value & 0x1f));
        } else if (value < 0x10000) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xe0 | (value & 0x0f));
        } else {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xf0 | (value & 0x07));
        }
        cpos -= padding_length;
        memset(cpos, padding_char, (size_t) padding_length);
        return PyUnicode_DecodeUTF8(cpos, chars + sizeof(chars) - cpos, NULL);
    }
    if (value <= 127 && CYTHON_USE_UNICODE_INTERNALS) {
        const char chars[1] = {(char) value};
        return __Pyx_PyUnicode_BuildFromAscii(ulength, chars, 1, 0, padding_char);
    }
    {
        PyObject *uchar, *padding_uchar, *padding, *result;
        padding_uchar = PyUnicode_FromOrdinal(padding_char);
        if (unlikely(!padding_uchar)) return NULL;
        padding = PySequence_Repeat(padding_uchar, padding_length);
        Py_DECREF(padding_uchar);
        if (unlikely(!padding)) return NULL;
        uchar = PyUnicode_FromOrdinal(value);
        if (unlikely(!uchar)) {
            Py_DECREF(padding);
            return NULL;
        }
        result = PyUnicode_Concat(padding, uchar);
        Py_DECREF(padding);
        Py_DECREF(uchar);
        return result;
    }
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (int) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    const char* function_name,
    PyObject *kw)
{
#if CYTHON_COMPILING_IN_PYPY && !defined(PyArg_ValidateKeywordArguments)
    CYTHON_UNUSED_VAR(function_name);
    CYTHON_UNUSED_VAR(kw);
    return 0;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
#if PY_VERSION_HEX >= 0x03090000
        CYTHON_UNUSED_VAR(function_name);
#else
        Py_ssize_t kwsize;
        #if CYTHON_ASSUME_SAFE_SIZE
        kwsize = PyTuple_GET_SIZE(kw);
        #else
        kwsize = PyTuple_Size(kw);
        if (unlikely(kwsize < 0)) return -1;
        #endif
        for (Py_ssize_t pos = 0; pos < kwsize; pos++) {
            PyObject* key = NULL;
            #if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
            #else
            key = PyTuple_GetItem(kw, pos);
            if (unlikely(!key)) return -1;
            #endif
            if (unlikely(!PyUnicode_Check(key))) {
                PyErr_Format(PyExc_TypeError,
                    "%.200s() keywords must be strings", function_name);
                return -1;
            }
        }
#endif
    } else {
        if (unlikely(!PyArg_ValidateKeywordArguments(kw))) return -1;
    }
    return 0;
#endif
}

/* IterFinish (used by dict_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* RaiseNeedMoreValuesToUnpack (used by UnpackTuple2) */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack (used by UnpackItemEndCheck) */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck (used by UnpackTuple2) */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError (used by UnpackTupleError) */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else {
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(t);
 #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(size < 0)) return;
 #endif
      if (size < index) {
        __Pyx_RaiseNeedMoreValuesError(size);
      } else {
        __Pyx_RaiseTooManyValuesError(index);
      }
    }
}

/* UnpackTuple2 (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
        PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple) {
    if (likely(is_tuple || PyTuple_Check(tuple))) {
        Py_ssize_t size;
        if (has_known_size) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        size = __Pyx_PyTuple_GET_SIZE(tuple);
        if (likely(size == 2)) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        if (size >= 0) {
            __Pyx_UnpackTupleError(tuple, 2);
        }
        return -1;
    } else {
        return __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple);
    }
}
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
    value1 = __Pyx_PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = __Pyx_PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#else
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
#if !CYTHON_AVOID_BORROWED_REFS
static CYTHON_INLINE int __Pyx_dict_iter_next_source_is_dict(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem) {
    PyObject *key, *value;
    if (unlikely(orig_length != PyDict_Size(iter_obj))) {
        PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
        return -1;
    }
    if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
        return 0;
    }
    if (pitem) {
        PyObject* tuple = PyTuple_New(2);
        if (unlikely(!tuple)) {
            return -1;
        }
        Py_INCREF(key);
        Py_INCREF(value);
        #if CYTHON_ASSUME_SAFE_MACROS
        PyTuple_SET_ITEM(tuple, 0, key);
        PyTuple_SET_ITEM(tuple, 1, value);
        #else
        if (unlikely(PyTuple_SetItem(tuple, 0, key) < 0)) {
            Py_DECREF(value);
            Py_DECREF(tuple);
            return -1;
        }
        if (unlikely(PyTuple_SetItem(tuple, 1, value) < 0)) {
            Py_DECREF(tuple);
            return -1;
        }
        #endif
        *pitem = tuple;
    } else {
        if (pkey) {
            Py_INCREF(key);
            *pkey = key;
        }
        if (pvalue) {
            Py_INCREF(value);
            *pvalue = value;
        }
    }
    return 1;
}
#endif
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_AVOID_BORROWED_REFS
    if (source_is_dict) {
        int result;
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_BEGIN_CRITICAL_SECTION(iter_obj);
#endif
        result = __Pyx_dict_iter_next_source_is_dict(iter_obj, orig_length, ppos, pkey, pvalue, pitem);
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_END_CRITICAL_SECTION();
#endif
        return result;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(tuple_size < 0)) return -1;
        #endif
        if (unlikely(pos >= tuple_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_ASSUME_SAFE_MACROS
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        #else
        next_item = PyTuple_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #endif
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t list_size = __Pyx_PyList_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(list_size < 0)) return -1;
        #endif
        if (unlikely(pos >= list_size)) return 0;
        *ppos = pos + 1;
        next_item = __Pyx_PyList_GetItemRef(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (Py_ssize_t) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* pybuiltin_invalid (used by pyint_simplify) */
static void __Pyx_PyBuiltin_Invalid(PyObject *obj, const char *type_name, const char *argname) {
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    if (argname) {
        PyErr_Format(PyExc_TypeError,
            "Argument '%.200s' has incorrect type (expected %.200s, got " __Pyx_FMT_TYPENAME ")",
            argname, type_name, obj_type_name
        );
    } else {
        PyErr_Format(PyExc_TypeError,
            "Expected %.200s, got " __Pyx_FMT_TYPENAME,
            type_name, obj_type_name
        );
    }
    __Pyx_DECREF_TypeName(obj_type_name);
}

/* pyint_simplify */
static CYTHON_INLINE int __Pyx_PyInt_FromNumber(PyObject **number_var, const char *argname, int accept_none) {
    PyObject *number = *number_var;
    if (likely((accept_none && number == Py_None) || PyLong_CheckExact(number))) {
        return 0;
    }
    PyObject *int_object;
    if (likely(PyNumber_Check(number))) {
        int_object = PyNumber_Long(number);
        if (unlikely(!int_object)) goto bad;
    } else {
        __Pyx_PyBuiltin_Invalid(number, "int", argname);
        goto bad;
    }
    *number_var = int_object;
    Py_DECREF(number);
    return 0;
bad:
    *number_var = NULL;
    Py_DECREF(number);
    return -1;
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* HasAttr */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* FunctionExport */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig) {
    PyObject *cobj;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(api_dict, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(cobj);
    return -1;
}

/* GetApiDict */
static PyObject *__Pyx_ApiExport_GetApiDict(void) {
    PyObject *d;
    if (__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_capi, &d) == -1)
        return NULL;
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_pyx_capi, d) < 0)
            goto bad;
    }
    return d;
bad:
    Py_XDECREF(d);
    return NULL;
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__5);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base = NULL, *meth, *new_base, *result, *new_bases = NULL;
#if CYTHON_ASSUME_SAFE_SIZE
    size_bases = PyTuple_GET_SIZE(bases);
#else
    size_bases = PyTuple_Size(bases);
    if (size_bases < 0) return NULL;
#endif
    for (i = 0; i < size_bases; i++) {
#if CYTHON_AVOID_BORROWED_REFS
        Py_CLEAR(base);
#endif
#if CYTHON_ASSUME_SAFE_MACROS
        base = PyTuple_GET_ITEM(bases, i);
#else
        base = PyTuple_GetItem(bases, i);
        if (!base) goto error;
#endif
#if CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(base);
#endif
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_mstate_global->__pyx_n_u_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                PyObject *base_from_list;
#if CYTHON_ASSUME_SAFE_MACROS
                base_from_list = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base_from_list);
                Py_INCREF(base_from_list);
#else
                base_from_list = PyTuple_GetItem(bases, j);
                if (!base_from_list) goto error;
                Py_INCREF(base_from_list);
                if (PyList_SetItem(new_bases, j, base_from_list) < 0) goto error;
#endif
            }
        }
#if CYTHON_ASSUME_SAFE_SIZE
        j = PyList_GET_SIZE(new_bases);
#else
        j = PyList_Size(new_bases);
        if (j < 0) goto error;
#endif
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return result;
error:
    Py_XDECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_SIZE
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
        metaclass = &PyType_Type;
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectCall2Args (used by Py3ClassCreate) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_mstate_global->__pyx_n_u_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_mstate_global->__pyx_n_u_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
    Py_XDECREF(owned_metaclass);
    return result;
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func_2);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CyFunctionClassCell */
static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj) {
    Py_ssize_t i, count = __Pyx_PyList_GET_SIZE(cyfunctions);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(count < 0)) return -1;
    #endif
    for (i = 0; i < count; i++) {
        __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *)
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && !CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
            PyList_GET_ITEM(cyfunctions, i);
#else
            __Pyx_PySequence_ITEM(cyfunctions, i);
        if (unlikely(!m))
            return -1;
#endif
        __Pyx_CyFunction_SetClassObj(m, classobj);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && !CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
        Py_DECREF((PyObject*)m);
#endif
    }
    return 0;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CheckUnpickleChecksum */
static void __Pyx_RaiseUnpickleChecksumError(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    PyObject *pickle_module = PyImport_ImportModule("pickle");
    if (unlikely(!pickle_module)) return;
    PyObject *pickle_error = PyObject_GetAttrString(pickle_module, "PickleError");
    Py_DECREF(pickle_module);
    if (unlikely(!pickle_error)) return;
    if (checksum2 == checksum1) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x) = (%s))",
            checksum, checksum1, members);
    } else if (checksum3 == checksum2) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, members);
    } else {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, checksum3, members);
    }
    Py_DECREF(pickle_error);
}
static int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    int found = 0;
    found |= checksum1 == checksum;
    found |= checksum2 == checksum;
    found |= checksum3 == checksum;
    if (likely(found))
        return 0;
    __Pyx_RaiseUnpickleChecksumError(checksum, checksum1, checksum2, checksum3, members);
    return -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint8_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int32_t(int32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_char(char value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(char) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(char) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(char),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(char));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int32_t __Pyx_PyLong_As_int32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyLong_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 2 * PyLong_SHIFT)) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 3 * PyLong_SHIFT)) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 4 * PyLong_SHIFT)) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int32_t) 1) << (sizeof(int32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE char __Pyx_PyLong_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        char val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyLong_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) >= 2 * PyLong_SHIFT)) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) >= 3 * PyLong_SHIFT)) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) >= 4 * PyLong_SHIFT)) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (char) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(char) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(char) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(char) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(char) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
        } else if ((sizeof(char) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        char val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (char) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (char) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (char) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (char) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(char) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((char) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(char) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((char) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((char) 1) << (sizeof(char) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (char) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyLong_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint8_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint8_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint8_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint8_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint8_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint8_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint8_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint8_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint8_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyLong_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyLong_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* UpdateUnpickledDict */
static int __Pyx__UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    PyObject *state_dict = __Pyx_PySequence_ITEM(state, index);
    if (unlikely(!state_dict)) {
        return -1;
    }
    int non_empty = PyObject_IsTrue(state_dict);
    if (non_empty == 0) {
        Py_DECREF(state_dict);
        return 0;
    } else if (unlikely(non_empty == -1)) {
        return -1;
    }
    PyObject *dict;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = PyObject_GetAttrString(obj, "__dict__");
    #else
    dict = PyObject_GenericGetDict(obj, NULL);
    #endif
    if (unlikely(!dict)) {
        Py_DECREF(state_dict);
        return -1;
    }
    int result;
    if (likely(PyDict_CheckExact(dict))) {
        result = PyDict_Update(dict, state_dict);
    } else {
        PyObject *obj_result = __Pyx_PyObject_CallMethod1(dict, __pyx_mstate_global->__pyx_n_u_update, state_dict);
        if (likely(obj_result)) {
            Py_DECREF(obj_result);
            result = 0;
        } else {
            result = -1;
        }
    }
    Py_DECREF(state_dict);
    Py_DECREF(dict);
    return result;
}
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    Py_ssize_t state_size = __Pyx_PyTuple_GET_SIZE(state);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(state_size == -1)) return -1;
    #endif
    if (state_size <= index) {
        return 0;
    }
    return __Pyx__UpdateUnpickledDict(obj, state, index);
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__6);
    }
    goto done;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* IterNextPlain (used by CoroutineBase) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void) {
    if (unlikely(!__pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache))
        __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache = __Pyx_GetBuiltinName(__pyx_mstate_global->__pyx_n_u_next);
    return __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator) {
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    PyObject *result;
    PyObject *next = __Pyx_GetBuiltinNext_LimitedAPI();
    if (unlikely(!next)) return NULL;
    result = PyObject_CallFunctionObjArgs(next, iterator, NULL);
    return result;
#else
    (void)__Pyx_GetBuiltinName; // only for early limited API
    iternextfunc iternext = __Pyx_PyObject_GetIterNextFunc(iterator);
    assert(iternext);
    return iternext(iterator);
#endif
}

/* ReturnWithStopIteration (used by CoroutineBase) */
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async);
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext) {
    if (value == Py_None) {
        if (async || !iternext)
            PyErr_SetNone(async ? PyExc_StopAsyncIteration : PyExc_StopIteration);
        return;
    }
    __Pyx__ReturnWithStopIteration(value, async);
}
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async) {
#if CYTHON_COMPILING_IN_CPYTHON
    __Pyx_PyThreadState_declare
#endif
    PyObject *exc;
    PyObject *exc_type = async ? PyExc_StopAsyncIteration : PyExc_StopIteration;
#if CYTHON_COMPILING_IN_CPYTHON
    if ((PY_VERSION_HEX >= (0x030C00A6)) || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        if (PY_VERSION_HEX >= (0x030e00A1)) {
            exc = __Pyx_PyObject_CallOneArg(exc_type, value);
        } else {
            PyObject *args_tuple = PyTuple_New(1);
            if (unlikely(!args_tuple)) return;
            Py_INCREF(value);
            PyTuple_SET_ITEM(args_tuple, 0, value);
            exc = PyObject_Call(exc_type, args_tuple, NULL);
            Py_DECREF(args_tuple);
        }
        if (unlikely(!exc)) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_value)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(exc_type);
        __Pyx_ErrRestore(exc_type, exc, NULL);
        return;
    }
    #endif
#else
    exc = __Pyx_PyObject_CallOneArg(exc_type, value);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(exc_type, exc);
    Py_DECREF(exc);
}

/* CoroutineBase (used by Generator) */
#if !CYTHON_COMPILING_IN_LIMITED_API
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6 && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#endif // CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void
__Pyx_Coroutine_Undelegate(__pyx_CoroutineObject *gen) {
#if CYTHON_USE_AM_SEND
    gen->yieldfrom_am_send = NULL;
#endif
    Py_CLEAR(gen->yieldfrom);
}
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
            value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
            if (unlikely(!value)) goto limited_api_failure;
            #else
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            #endif
            Py_DECREF(ev);
        }
        else if (unlikely(PyTuple_Check(ev))) {
            Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(ev);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely(tuple_size < 0)) {
                Py_XDECREF(tb);
                Py_DECREF(ev);
                Py_DECREF(et);
                return -1;
            }
            #endif
            if (tuple_size >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#elif CYTHON_ASSUME_SAFE_MACROS
                value = PySequence_ITEM(ev, 0);
#else
                value = PySequence_GetItem(ev, 0);
                if (!value) goto limited_api_failure;
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
#else
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
#endif
    Py_DECREF(ev);
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!value)) return -1;
#endif
    *pvalue = value;
    return 0;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL || !CYTHON_ASSUME_SAFE_MACROS
  limited_api_failure:
    Py_XDECREF(et);
    Py_XDECREF(tb);
    Py_XDECREF(ev);
    return -1;
#endif
}
static CYTHON_INLINE
__Pyx_PySendResult __Pyx_Coroutine_status_from_result(PyObject **retval) {
    if (*retval) {
        return PYGEN_NEXT;
    } else if (likely(__Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, retval) == 0)) {
        return PYGEN_RETURN;
    } else {
        return PYGEN_ERROR;
    }
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
static void __Pyx_Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
__Pyx_PySendResult __Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, PyObject **result, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(__Pyx_Coroutine_get_is_running(self));  // Callers should ensure is_running
    if (unlikely(self->resume_label == -1)) {
        __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
        return PYGEN_ERROR;
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    retval = self->body(self, tstate, value);
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    *result = retval;
    if (self->resume_label == -1) {
        return likely(retval) ? PYGEN_RETURN : PYGEN_ERROR;
    }
    return PYGEN_NEXT;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
#define __Pyx_Coroutine_MethodReturnFromResult(gen, result, retval, iternext)\
    ((result) == PYGEN_NEXT ? (retval) : __Pyx__Coroutine_MethodReturnFromResult(gen, result, retval, iternext))
static PyObject *
__Pyx__Coroutine_MethodReturnFromResult(PyObject* gen, __Pyx_PySendResult result, PyObject *retval, int iternext) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (likely(result == PYGEN_RETURN)) {
        int is_async = 0;
        #ifdef __Pyx_AsyncGen_USED
        is_async = __Pyx_AsyncGen_CheckExact(gen);
        #endif
        __Pyx_ReturnWithStopIteration(retval, is_async, iternext);
        Py_XDECREF(retval);
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE __Pyx_PySendResult
__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen, PyObject** retval) {
    __Pyx_PySendResult result;
    PyObject *val = NULL;
    assert(__Pyx_Coroutine_get_is_running(gen));
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    result = __Pyx_Coroutine_SendEx(gen, val, retval, 0);
    Py_XDECREF(val);
    return result;
}
#if CYTHON_USE_AM_SEND
static __Pyx_PySendResult
__Pyx_Coroutine_SendToDelegate(__pyx_CoroutineObject *gen, __Pyx_pyiter_sendfunc gen_am_send, PyObject *value, PyObject **retval) {
    PyObject *ret = NULL;
    __Pyx_PySendResult delegate_result, result;
    assert(__Pyx_Coroutine_get_is_running(gen));
    delegate_result = gen_am_send(gen->yieldfrom, value, &ret);
    if (delegate_result == PYGEN_NEXT) {
        assert (ret != NULL);
        *retval = ret;
        return PYGEN_NEXT;
    }
    assert (delegate_result != PYGEN_ERROR || ret == NULL);
    __Pyx_Coroutine_Undelegate(gen);
    result = __Pyx_Coroutine_SendEx(gen, ret, retval, 0);
    Py_XDECREF(ret);
    return result;
}
#endif
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result = __Pyx_Coroutine_AmSend(self, value, &retval);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
}
static __Pyx_PySendResult
__Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval) {
    __Pyx_PySendResult result;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, value, retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
      #if !CYTHON_USE_AM_SEND
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
      #endif
        {
            #if !CYTHON_COMPILING_IN_LIMITED_API || __PYX_LIMITED_VERSION_HEX >= 0x03080000
            if (value == Py_None && PyIter_Check(yf))
                ret = __Pyx_PyIter_Next_Plain(yf);
            else
            #endif
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_mstate_global->__pyx_n_u_send, value);
        }
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            *retval = ret;
            return PYGEN_NEXT;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, value, retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return result;
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    CYTHON_UNUSED_VAR(gen);
    assert(__Pyx_Coroutine_get_is_running(gen));
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        result = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf);
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    #endif
    {
        PyObject *meth;
        result = PYGEN_RETURN;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval)) {
                result = PYGEN_ERROR;
            }
        }
    }
    Py_XDECREF(retval);
    return result == PYGEN_ERROR ? -1 : 0;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, Py_None, &retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && (PY_VERSION_HEX < 0x030A00A3 || !CYTHON_USE_AM_SEND)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
            ret = __Pyx_PyIter_Next_Plain(yf);
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 1);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result;
    CYTHON_UNUSED_VAR(arg);
    result = __Pyx_Coroutine_Close(self, &retval);
    if (unlikely(result == PYGEN_ERROR))
        return NULL;
    Py_XDECREF(retval);
    Py_RETURN_NONE;
}
static __Pyx_PySendResult
__Pyx_Coroutine_Close(PyObject *self, PyObject **retval) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PySendResult result;
    PyObject *yf;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    yf = gen->yieldfrom;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    result = __Pyx_Coroutine_SendEx(gen, NULL, retval, 1);
    if (result == PYGEN_ERROR) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_Coroutine_unset_is_running(gen);
        if (!__Pyx_PyErr_Occurred()) {
            return PYGEN_RETURN;
        } else if (likely(__Pyx_PyErr_ExceptionMatches2(PyExc_GeneratorExit, PyExc_StopIteration))) {
            __Pyx_PyErr_Clear();
            return PYGEN_RETURN;
        }
        return PYGEN_ERROR;
    } else if (likely(result == PYGEN_RETURN && *retval == Py_None)) {
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_RETURN;
    } else {
        const char *msg;
        Py_DECREF(*retval);
        *retval = NULL;
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
            msg = "async generator ignored GeneratorExit";
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_ERROR;
    }
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen)))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    yf = gen->yieldfrom;
    if (yf) {
        __Pyx_PySendResult result;
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                goto propagate_exception;
            goto throw_here;
        }
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    __Pyx_Coroutine_unset_is_running(gen);
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        Py_DECREF(yf);
        if (ret) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &ret);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, ret, 0);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
propagate_exception:
    {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, NULL, &retval, 0);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
    }
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, "throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    {
        int e = __Pyx_call_type_traverse((PyObject*)gen, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    if (gen->gi_weakreflist != NULL)
    #endif
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        {
            destructor del = __Pyx_PyObject_GetSlot(gen, tp_del, destructor);
            if (del) del(self);
        }
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
#if CYTHON_USE_TP_FINALIZE
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_Close(self, &retval);
        if (result == PYGEN_ERROR) {
            PyErr_WriteUnraisable(self);
        } else {
            Py_XDECREF(retval);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
}
#endif
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx__Coroutine_get_frame(__pyx_CoroutineObject *self)
{
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyObject *frame;
    #if PY_VERSION_HEX >= 0x030d0000
    Py_BEGIN_CRITICAL_SECTION(self);
    #endif
    frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        PyObject *globals = PyDict_New();
        if (unlikely(!globals)) return NULL;
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            globals,                        /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        Py_DECREF(globals);
        if (unlikely(!frame))
            return NULL;
        if (unlikely(self->gi_frame)) {
            Py_DECREF(frame);
            frame = self->gi_frame;
        } else {
            self->gi_frame = frame;
        }
    }
    Py_INCREF(frame);
    #if PY_VERSION_HEX >= 0x030d0000
    Py_END_CRITICAL_SECTION();
    #endif
    return frame;
#else
    CYTHON_UNUSED_VAR(self);
    Py_RETURN_NONE;
#endif
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context) {
    CYTHON_UNUSED_VAR(context);
    PyObject *frame = self->gi_frame;
    if (frame)
        return __Pyx_NewRef(frame);
    return __Pyx__Coroutine_get_frame(self);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->yieldfrom_am_send = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_weakreflist = NULL;
#endif
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    gen->is_running = 1;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen) {
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    assert(gen->is_running);
    gen->is_running = 0;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
}
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure) {
    CYTHON_UNUSED_VAR(closure);
    char result = __Pyx_Coroutine_get_is_running((__pyx_CoroutineObject*)gen);
    if (result) Py_RETURN_TRUE;
    else Py_RETURN_FALSE;
}
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send) {
    Py_ssize_t ptr_offset = (char*)(type->tp_as_async) - (char*)type;
    if (ptr_offset < 0 || ptr_offset > type->tp_basicsize) {
        return;
    }
    memcpy((void*)static_amsend_methods, (void*)(type->tp_as_async), sizeof(*type->tp_as_async));
    static_amsend_methods->am_send = am_send;
    type->tp_as_async = __Pyx_SlotTpAsAsync(static_amsend_methods);
}
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    __Pyx_TypeName self_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE((PyObject*)self));
    PyErr_Format(PyExc_TypeError, "cannot pickle '" __Pyx_FMT_TYPENAME "' object",
                         self_type_name);
    __Pyx_DECREF_TypeName(self_type_name);
    return NULL;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {"gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     PyDoc_STR("object being iterated by 'yield from', or None")},
    {"gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {"__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {"__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     PyDoc_STR("name of the generator"), 0},
    {"__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     PyDoc_STR("qualified name of the generator"), 0},
    {"gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     PyDoc_STR("Frame of the generator"), 0},
    {"gi_running", __Pyx_Coroutine_get_is_running_getter, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) PyObject_GenericGetAttr},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    {Py_am_send, (void *)__Pyx_Coroutine_AmSend},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_WEAKREF |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | __Pyx_TPFLAGS_HAVE_AM_SEND,
    __pyx_GeneratorType_slots
};
#if __PYX_HAS_PY_AM_SEND == 2
static __Pyx_PyAsyncMethodsStruct __pyx_Generator_as_async;
#endif
static int __pyx_Generator_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_GeneratorType_spec, NULL);
    if (unlikely(!mstate->__pyx_GeneratorType)) {
        return -1;
    }
#if __PYX_HAS_PY_AM_SEND == 2
    __Pyx_SetBackportTypeAmSend(mstate->__pyx_GeneratorType, &__pyx_Generator_as_async, &__Pyx_Coroutine_AmSend);
#endif
    return 0;
}
static PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *retval = NULL;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    __Pyx_Coroutine_unset_is_running(gen);
    (void) result;
    assert (result == PYGEN_RETURN || result == PYGEN_ERROR);
    assert ((result == PYGEN_RETURN && retval != NULL) || (result == PYGEN_ERROR && retval == NULL));
    return retval;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
